/*
 * HCS API
 *
 * No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)
 *
 * API version: 2.1
 * Generated by: Swagger Codegen (https://github.com/swagger-api/swagger-codegen.git)
 */

package hcsschema

type ProcessParameters struct {
	ApplicationName string `json:"ApplicationName,omitempty"`

	CommandLine string `json:"CommandLine,omitempty"`

	//  optional alternative to CommandLine, currently only supported by Linux GCS
	CommandArgs []string `json:"CommandArgs,omitempty"`

	User string `json:"User,omitempty"`

	WorkingDirectory string `json:"WorkingDirectory,omitempty"`

	Environment map[string]string `json:"Environment,omitempty"`

	//  if set, will run as low-privilege process
	RestrictedToken bool `json:"RestrictedToken,omitempty"`

	//  if set, ignore StdErrPipe
	EmulateConsole bool `json:"EmulateConsole,omitempty"`

	CreateStdInPipe bool `json:"CreateStdInPipe,omitempty"`

	CreateStdOutPipe bool `json:"CreateStdOutPipe,omitempty"`

	CreateStdErrPipe bool `json:"CreateStdErrPipe,omitempty"`

	//  height then width
	ConsoleSize []int32 `json:"ConsoleSize,omitempty"`

	//  if set, find an existing session for the user and create the process in it
	UseExistingLogin bool `json:"UseExistingLogin,omitempty"`

	//  if set, use the legacy console instead of conhost
	UseLegacyConsole bool `json:"UseLegacyConsole,omitempty"`
}
