package dockerversion // import "github.com/docker/docker/dockerversion"

// Default build-time variable for library-import.
// These variables are overridden on build with build-time information.
var (
	GitCommit             = "library-import"
	Version               = "library-import"
	BuildTime             = "library-import"
	PlatformName          = ""
	ProductName           = ""
	DefaultProductLicense = ""
)
