syntax = "proto3";

package fsutil.types;

option go_package = "github.com/tonistiigi/fsutil/types";

import "github.com/planetscale/vtprotobuf/vtproto/ext.proto";

message Stat {
  string path = 1;
  uint32 mode = 2;
  uint32 uid = 3;
  uint32 gid = 4;
  int64 size = 5;
  int64 modTime = 6;
  // int32 typeflag = 7;
  string linkname = 7;
  int64 devmajor = 8;
  int64 devminor = 9;
  map<string, bytes> xattrs = 10;
}
