// This file is automatically generated
// Run "go run cmd/gen/gen.go" to update the list.

package publicsuffix

const defaultListVersion = "PSL version 598c63 (Thu May  6 04:03:10 2021)"

func DefaultRules() [9169]Rule {
	return r
}

var r = [9169]Rule{
	{1, "ac", 1, false},
	{1, "com.ac", 2, false},
	{1, "edu.ac", 2, false},
	{1, "gov.ac", 2, false},
	{1, "net.ac", 2, false},
	{1, "mil.ac", 2, false},
	{1, "org.ac", 2, false},
	{1, "ad", 1, false},
	{1, "nom.ad", 2, false},
	{1, "ae", 1, false},
	{1, "co.ae", 2, false},
	{1, "net.ae", 2, false},
	{1, "org.ae", 2, false},
	{1, "sch.ae", 2, false},
	{1, "ac.ae", 2, false},
	{1, "gov.ae", 2, false},
	{1, "mil.ae", 2, false},
	{1, "aero", 1, false},
	{1, "accident-investigation.aero", 2, false},
	{1, "accident-prevention.aero", 2, false},
	{1, "aerobatic.aero", 2, false},
	{1, "aeroclub.aero", 2, false},
	{1, "aerodrome.aero", 2, false},
	{1, "agents.aero", 2, false},
	{1, "aircraft.aero", 2, false},
	{1, "airline.aero", 2, false},
	{1, "airport.aero", 2, false},
	{1, "air-surveillance.aero", 2, false},
	{1, "airtraffic.aero", 2, false},
	{1, "air-traffic-control.aero", 2, false},
	{1, "ambulance.aero", 2, false},
	{1, "amusement.aero", 2, false},
	{1, "association.aero", 2, false},
	{1, "author.aero", 2, false},
	{1, "ballooning.aero", 2, false},
	{1, "broker.aero", 2, false},
	{1, "caa.aero", 2, false},
	{1, "cargo.aero", 2, false},
	{1, "catering.aero", 2, false},
	{1, "certification.aero", 2, false},
	{1, "championship.aero", 2, false},
	{1, "charter.aero", 2, false},
	{1, "civilaviation.aero", 2, false},
	{1, "club.aero", 2, false},
	{1, "conference.aero", 2, false},
	{1, "consultant.aero", 2, false},
	{1, "consulting.aero", 2, false},
	{1, "control.aero", 2, false},
	{1, "council.aero", 2, false},
	{1, "crew.aero", 2, false},
	{1, "design.aero", 2, false},
	{1, "dgca.aero", 2, false},
	{1, "educator.aero", 2, false},
	{1, "emergency.aero", 2, false},
	{1, "engine.aero", 2, false},
	{1, "engineer.aero", 2, false},
	{1, "entertainment.aero", 2, false},
	{1, "equipment.aero", 2, false},
	{1, "exchange.aero", 2, false},
	{1, "express.aero", 2, false},
	{1, "federation.aero", 2, false},
	{1, "flight.aero", 2, false},
	{1, "fuel.aero", 2, false},
	{1, "gliding.aero", 2, false},
	{1, "government.aero", 2, false},
	{1, "groundhandling.aero", 2, false},
	{1, "group.aero", 2, false},
	{1, "hanggliding.aero", 2, false},
	{1, "homebuilt.aero", 2, false},
	{1, "insurance.aero", 2, false},
	{1, "journal.aero", 2, false},
	{1, "journalist.aero", 2, false},
	{1, "leasing.aero", 2, false},
	{1, "logistics.aero", 2, false},
	{1, "magazine.aero", 2, false},
	{1, "maintenance.aero", 2, false},
	{1, "media.aero", 2, false},
	{1, "microlight.aero", 2, false},
	{1, "modelling.aero", 2, false},
	{1, "navigation.aero", 2, false},
	{1, "parachuting.aero", 2, false},
	{1, "paragliding.aero", 2, false},
	{1, "passenger-association.aero", 2, false},
	{1, "pilot.aero", 2, false},
	{1, "press.aero", 2, false},
	{1, "production.aero", 2, false},
	{1, "recreation.aero", 2, false},
	{1, "repbody.aero", 2, false},
	{1, "res.aero", 2, false},
	{1, "research.aero", 2, false},
	{1, "rotorcraft.aero", 2, false},
	{1, "safety.aero", 2, false},
	{1, "scientist.aero", 2, false},
	{1, "services.aero", 2, false},
	{1, "show.aero", 2, false},
	{1, "skydiving.aero", 2, false},
	{1, "software.aero", 2, false},
	{1, "student.aero", 2, false},
	{1, "trader.aero", 2, false},
	{1, "trading.aero", 2, false},
	{1, "trainer.aero", 2, false},
	{1, "union.aero", 2, false},
	{1, "workinggroup.aero", 2, false},
	{1, "works.aero", 2, false},
	{1, "af", 1, false},
	{1, "gov.af", 2, false},
	{1, "com.af", 2, false},
	{1, "org.af", 2, false},
	{1, "net.af", 2, false},
	{1, "edu.af", 2, false},
	{1, "ag", 1, false},
	{1, "com.ag", 2, false},
	{1, "org.ag", 2, false},
	{1, "net.ag", 2, false},
	{1, "co.ag", 2, false},
	{1, "nom.ag", 2, false},
	{1, "ai", 1, false},
	{1, "off.ai", 2, false},
	{1, "com.ai", 2, false},
	{1, "net.ai", 2, false},
	{1, "org.ai", 2, false},
	{1, "al", 1, false},
	{1, "com.al", 2, false},
	{1, "edu.al", 2, false},
	{1, "gov.al", 2, false},
	{1, "mil.al", 2, false},
	{1, "net.al", 2, false},
	{1, "org.al", 2, false},
	{1, "am", 1, false},
	{1, "co.am", 2, false},
	{1, "com.am", 2, false},
	{1, "commune.am", 2, false},
	{1, "net.am", 2, false},
	{1, "org.am", 2, false},
	{1, "ao", 1, false},
	{1, "ed.ao", 2, false},
	{1, "gv.ao", 2, false},
	{1, "og.ao", 2, false},
	{1, "co.ao", 2, false},
	{1, "pb.ao", 2, false},
	{1, "it.ao", 2, false},
	{1, "aq", 1, false},
	{1, "ar", 1, false},
	{1, "com.ar", 2, false},
	{1, "edu.ar", 2, false},
	{1, "gob.ar", 2, false},
	{1, "gov.ar", 2, false},
	{1, "int.ar", 2, false},
	{1, "mil.ar", 2, false},
	{1, "musica.ar", 2, false},
	{1, "net.ar", 2, false},
	{1, "org.ar", 2, false},
	{1, "tur.ar", 2, false},
	{1, "arpa", 1, false},
	{1, "e164.arpa", 2, false},
	{1, "in-addr.arpa", 2, false},
	{1, "ip6.arpa", 2, false},
	{1, "iris.arpa", 2, false},
	{1, "uri.arpa", 2, false},
	{1, "urn.arpa", 2, false},
	{1, "as", 1, false},
	{1, "gov.as", 2, false},
	{1, "asia", 1, false},
	{1, "at", 1, false},
	{1, "ac.at", 2, false},
	{1, "co.at", 2, false},
	{1, "gv.at", 2, false},
	{1, "or.at", 2, false},
	{1, "sth.ac.at", 3, false},
	{1, "au", 1, false},
	{1, "com.au", 2, false},
	{1, "net.au", 2, false},
	{1, "org.au", 2, false},
	{1, "edu.au", 2, false},
	{1, "gov.au", 2, false},
	{1, "asn.au", 2, false},
	{1, "id.au", 2, false},
	{1, "info.au", 2, false},
	{1, "conf.au", 2, false},
	{1, "oz.au", 2, false},
	{1, "act.au", 2, false},
	{1, "nsw.au", 2, false},
	{1, "nt.au", 2, false},
	{1, "qld.au", 2, false},
	{1, "sa.au", 2, false},
	{1, "tas.au", 2, false},
	{1, "vic.au", 2, false},
	{1, "wa.au", 2, false},
	{1, "act.edu.au", 3, false},
	{1, "catholic.edu.au", 3, false},
	{1, "nsw.edu.au", 3, false},
	{1, "nt.edu.au", 3, false},
	{1, "qld.edu.au", 3, false},
	{1, "sa.edu.au", 3, false},
	{1, "tas.edu.au", 3, false},
	{1, "vic.edu.au", 3, false},
	{1, "wa.edu.au", 3, false},
	{1, "qld.gov.au", 3, false},
	{1, "sa.gov.au", 3, false},
	{1, "tas.gov.au", 3, false},
	{1, "vic.gov.au", 3, false},
	{1, "wa.gov.au", 3, false},
	{1, "schools.nsw.edu.au", 4, false},
	{1, "aw", 1, false},
	{1, "com.aw", 2, false},
	{1, "ax", 1, false},
	{1, "az", 1, false},
	{1, "com.az", 2, false},
	{1, "net.az", 2, false},
	{1, "int.az", 2, false},
	{1, "gov.az", 2, false},
	{1, "org.az", 2, false},
	{1, "edu.az", 2, false},
	{1, "info.az", 2, false},
	{1, "pp.az", 2, false},
	{1, "mil.az", 2, false},
	{1, "name.az", 2, false},
	{1, "pro.az", 2, false},
	{1, "biz.az", 2, false},
	{1, "ba", 1, false},
	{1, "com.ba", 2, false},
	{1, "edu.ba", 2, false},
	{1, "gov.ba", 2, false},
	{1, "mil.ba", 2, false},
	{1, "net.ba", 2, false},
	{1, "org.ba", 2, false},
	{1, "bb", 1, false},
	{1, "biz.bb", 2, false},
	{1, "co.bb", 2, false},
	{1, "com.bb", 2, false},
	{1, "edu.bb", 2, false},
	{1, "gov.bb", 2, false},
	{1, "info.bb", 2, false},
	{1, "net.bb", 2, false},
	{1, "org.bb", 2, false},
	{1, "store.bb", 2, false},
	{1, "tv.bb", 2, false},
	{2, "bd", 2, false},
	{1, "be", 1, false},
	{1, "ac.be", 2, false},
	{1, "bf", 1, false},
	{1, "gov.bf", 2, false},
	{1, "bg", 1, false},
	{1, "a.bg", 2, false},
	{1, "b.bg", 2, false},
	{1, "c.bg", 2, false},
	{1, "d.bg", 2, false},
	{1, "e.bg", 2, false},
	{1, "f.bg", 2, false},
	{1, "g.bg", 2, false},
	{1, "h.bg", 2, false},
	{1, "i.bg", 2, false},
	{1, "j.bg", 2, false},
	{1, "k.bg", 2, false},
	{1, "l.bg", 2, false},
	{1, "m.bg", 2, false},
	{1, "n.bg", 2, false},
	{1, "o.bg", 2, false},
	{1, "p.bg", 2, false},
	{1, "q.bg", 2, false},
	{1, "r.bg", 2, false},
	{1, "s.bg", 2, false},
	{1, "t.bg", 2, false},
	{1, "u.bg", 2, false},
	{1, "v.bg", 2, false},
	{1, "w.bg", 2, false},
	{1, "x.bg", 2, false},
	{1, "y.bg", 2, false},
	{1, "z.bg", 2, false},
	{1, "0.bg", 2, false},
	{1, "1.bg", 2, false},
	{1, "2.bg", 2, false},
	{1, "3.bg", 2, false},
	{1, "4.bg", 2, false},
	{1, "5.bg", 2, false},
	{1, "6.bg", 2, false},
	{1, "7.bg", 2, false},
	{1, "8.bg", 2, false},
	{1, "9.bg", 2, false},
	{1, "bh", 1, false},
	{1, "com.bh", 2, false},
	{1, "edu.bh", 2, false},
	{1, "net.bh", 2, false},
	{1, "org.bh", 2, false},
	{1, "gov.bh", 2, false},
	{1, "bi", 1, false},
	{1, "co.bi", 2, false},
	{1, "com.bi", 2, false},
	{1, "edu.bi", 2, false},
	{1, "or.bi", 2, false},
	{1, "org.bi", 2, false},
	{1, "biz", 1, false},
	{1, "bj", 1, false},
	{1, "asso.bj", 2, false},
	{1, "barreau.bj", 2, false},
	{1, "gouv.bj", 2, false},
	{1, "bm", 1, false},
	{1, "com.bm", 2, false},
	{1, "edu.bm", 2, false},
	{1, "gov.bm", 2, false},
	{1, "net.bm", 2, false},
	{1, "org.bm", 2, false},
	{1, "bn", 1, false},
	{1, "com.bn", 2, false},
	{1, "edu.bn", 2, false},
	{1, "gov.bn", 2, false},
	{1, "net.bn", 2, false},
	{1, "org.bn", 2, false},
	{1, "bo", 1, false},
	{1, "com.bo", 2, false},
	{1, "edu.bo", 2, false},
	{1, "gob.bo", 2, false},
	{1, "int.bo", 2, false},
	{1, "org.bo", 2, false},
	{1, "net.bo", 2, false},
	{1, "mil.bo", 2, false},
	{1, "tv.bo", 2, false},
	{1, "web.bo", 2, false},
	{1, "academia.bo", 2, false},
	{1, "agro.bo", 2, false},
	{1, "arte.bo", 2, false},
	{1, "blog.bo", 2, false},
	{1, "bolivia.bo", 2, false},
	{1, "ciencia.bo", 2, false},
	{1, "cooperativa.bo", 2, false},
	{1, "democracia.bo", 2, false},
	{1, "deporte.bo", 2, false},
	{1, "ecologia.bo", 2, false},
	{1, "economia.bo", 2, false},
	{1, "empresa.bo", 2, false},
	{1, "indigena.bo", 2, false},
	{1, "industria.bo", 2, false},
	{1, "info.bo", 2, false},
	{1, "medicina.bo", 2, false},
	{1, "movimiento.bo", 2, false},
	{1, "musica.bo", 2, false},
	{1, "natural.bo", 2, false},
	{1, "nombre.bo", 2, false},
	{1, "noticias.bo", 2, false},
	{1, "patria.bo", 2, false},
	{1, "politica.bo", 2, false},
	{1, "profesional.bo", 2, false},
	{1, "plurinacional.bo", 2, false},
	{1, "pueblo.bo", 2, false},
	{1, "revista.bo", 2, false},
	{1, "salud.bo", 2, false},
	{1, "tecnologia.bo", 2, false},
	{1, "tksat.bo", 2, false},
	{1, "transporte.bo", 2, false},
	{1, "wiki.bo", 2, false},
	{1, "br", 1, false},
	{1, "9guacu.br", 2, false},
	{1, "abc.br", 2, false},
	{1, "adm.br", 2, false},
	{1, "adv.br", 2, false},
	{1, "agr.br", 2, false},
	{1, "aju.br", 2, false},
	{1, "am.br", 2, false},
	{1, "anani.br", 2, false},
	{1, "aparecida.br", 2, false},
	{1, "app.br", 2, false},
	{1, "arq.br", 2, false},
	{1, "art.br", 2, false},
	{1, "ato.br", 2, false},
	{1, "b.br", 2, false},
	{1, "barueri.br", 2, false},
	{1, "belem.br", 2, false},
	{1, "bhz.br", 2, false},
	{1, "bib.br", 2, false},
	{1, "bio.br", 2, false},
	{1, "blog.br", 2, false},
	{1, "bmd.br", 2, false},
	{1, "boavista.br", 2, false},
	{1, "bsb.br", 2, false},
	{1, "campinagrande.br", 2, false},
	{1, "campinas.br", 2, false},
	{1, "caxias.br", 2, false},
	{1, "cim.br", 2, false},
	{1, "cng.br", 2, false},
	{1, "cnt.br", 2, false},
	{1, "com.br", 2, false},
	{1, "contagem.br", 2, false},
	{1, "coop.br", 2, false},
	{1, "coz.br", 2, false},
	{1, "cri.br", 2, false},
	{1, "cuiaba.br", 2, false},
	{1, "curitiba.br", 2, false},
	{1, "def.br", 2, false},
	{1, "des.br", 2, false},
	{1, "det.br", 2, false},
	{1, "dev.br", 2, false},
	{1, "ecn.br", 2, false},
	{1, "eco.br", 2, false},
	{1, "edu.br", 2, false},
	{1, "emp.br", 2, false},
	{1, "enf.br", 2, false},
	{1, "eng.br", 2, false},
	{1, "esp.br", 2, false},
	{1, "etc.br", 2, false},
	{1, "eti.br", 2, false},
	{1, "far.br", 2, false},
	{1, "feira.br", 2, false},
	{1, "flog.br", 2, false},
	{1, "floripa.br", 2, false},
	{1, "fm.br", 2, false},
	{1, "fnd.br", 2, false},
	{1, "fortal.br", 2, false},
	{1, "fot.br", 2, false},
	{1, "foz.br", 2, false},
	{1, "fst.br", 2, false},
	{1, "g12.br", 2, false},
	{1, "geo.br", 2, false},
	{1, "ggf.br", 2, false},
	{1, "goiania.br", 2, false},
	{1, "gov.br", 2, false},
	{1, "ac.gov.br", 3, false},
	{1, "al.gov.br", 3, false},
	{1, "am.gov.br", 3, false},
	{1, "ap.gov.br", 3, false},
	{1, "ba.gov.br", 3, false},
	{1, "ce.gov.br", 3, false},
	{1, "df.gov.br", 3, false},
	{1, "es.gov.br", 3, false},
	{1, "go.gov.br", 3, false},
	{1, "ma.gov.br", 3, false},
	{1, "mg.gov.br", 3, false},
	{1, "ms.gov.br", 3, false},
	{1, "mt.gov.br", 3, false},
	{1, "pa.gov.br", 3, false},
	{1, "pb.gov.br", 3, false},
	{1, "pe.gov.br", 3, false},
	{1, "pi.gov.br", 3, false},
	{1, "pr.gov.br", 3, false},
	{1, "rj.gov.br", 3, false},
	{1, "rn.gov.br", 3, false},
	{1, "ro.gov.br", 3, false},
	{1, "rr.gov.br", 3, false},
	{1, "rs.gov.br", 3, false},
	{1, "sc.gov.br", 3, false},
	{1, "se.gov.br", 3, false},
	{1, "sp.gov.br", 3, false},
	{1, "to.gov.br", 3, false},
	{1, "gru.br", 2, false},
	{1, "imb.br", 2, false},
	{1, "ind.br", 2, false},
	{1, "inf.br", 2, false},
	{1, "jab.br", 2, false},
	{1, "jampa.br", 2, false},
	{1, "jdf.br", 2, false},
	{1, "joinville.br", 2, false},
	{1, "jor.br", 2, false},
	{1, "jus.br", 2, false},
	{1, "leg.br", 2, false},
	{1, "lel.br", 2, false},
	{1, "log.br", 2, false},
	{1, "londrina.br", 2, false},
	{1, "macapa.br", 2, false},
	{1, "maceio.br", 2, false},
	{1, "manaus.br", 2, false},
	{1, "maringa.br", 2, false},
	{1, "mat.br", 2, false},
	{1, "med.br", 2, false},
	{1, "mil.br", 2, false},
	{1, "morena.br", 2, false},
	{1, "mp.br", 2, false},
	{1, "mus.br", 2, false},
	{1, "natal.br", 2, false},
	{1, "net.br", 2, false},
	{1, "niteroi.br", 2, false},
	{2, "nom.br", 3, false},
	{1, "not.br", 2, false},
	{1, "ntr.br", 2, false},
	{1, "odo.br", 2, false},
	{1, "ong.br", 2, false},
	{1, "org.br", 2, false},
	{1, "osasco.br", 2, false},
	{1, "palmas.br", 2, false},
	{1, "poa.br", 2, false},
	{1, "ppg.br", 2, false},
	{1, "pro.br", 2, false},
	{1, "psc.br", 2, false},
	{1, "psi.br", 2, false},
	{1, "pvh.br", 2, false},
	{1, "qsl.br", 2, false},
	{1, "radio.br", 2, false},
	{1, "rec.br", 2, false},
	{1, "recife.br", 2, false},
	{1, "rep.br", 2, false},
	{1, "ribeirao.br", 2, false},
	{1, "rio.br", 2, false},
	{1, "riobranco.br", 2, false},
	{1, "riopreto.br", 2, false},
	{1, "salvador.br", 2, false},
	{1, "sampa.br", 2, false},
	{1, "santamaria.br", 2, false},
	{1, "santoandre.br", 2, false},
	{1, "saobernardo.br", 2, false},
	{1, "saogonca.br", 2, false},
	{1, "seg.br", 2, false},
	{1, "sjc.br", 2, false},
	{1, "slg.br", 2, false},
	{1, "slz.br", 2, false},
	{1, "sorocaba.br", 2, false},
	{1, "srv.br", 2, false},
	{1, "taxi.br", 2, false},
	{1, "tc.br", 2, false},
	{1, "tec.br", 2, false},
	{1, "teo.br", 2, false},
	{1, "the.br", 2, false},
	{1, "tmp.br", 2, false},
	{1, "trd.br", 2, false},
	{1, "tur.br", 2, false},
	{1, "tv.br", 2, false},
	{1, "udi.br", 2, false},
	{1, "vet.br", 2, false},
	{1, "vix.br", 2, false},
	{1, "vlog.br", 2, false},
	{1, "wiki.br", 2, false},
	{1, "zlg.br", 2, false},
	{1, "bs", 1, false},
	{1, "com.bs", 2, false},
	{1, "net.bs", 2, false},
	{1, "org.bs", 2, false},
	{1, "edu.bs", 2, false},
	{1, "gov.bs", 2, false},
	{1, "bt", 1, false},
	{1, "com.bt", 2, false},
	{1, "edu.bt", 2, false},
	{1, "gov.bt", 2, false},
	{1, "net.bt", 2, false},
	{1, "org.bt", 2, false},
	{1, "bv", 1, false},
	{1, "bw", 1, false},
	{1, "co.bw", 2, false},
	{1, "org.bw", 2, false},
	{1, "by", 1, false},
	{1, "gov.by", 2, false},
	{1, "mil.by", 2, false},
	{1, "com.by", 2, false},
	{1, "of.by", 2, false},
	{1, "bz", 1, false},
	{1, "com.bz", 2, false},
	{1, "net.bz", 2, false},
	{1, "org.bz", 2, false},
	{1, "edu.bz", 2, false},
	{1, "gov.bz", 2, false},
	{1, "ca", 1, false},
	{1, "ab.ca", 2, false},
	{1, "bc.ca", 2, false},
	{1, "mb.ca", 2, false},
	{1, "nb.ca", 2, false},
	{1, "nf.ca", 2, false},
	{1, "nl.ca", 2, false},
	{1, "ns.ca", 2, false},
	{1, "nt.ca", 2, false},
	{1, "nu.ca", 2, false},
	{1, "on.ca", 2, false},
	{1, "pe.ca", 2, false},
	{1, "qc.ca", 2, false},
	{1, "sk.ca", 2, false},
	{1, "yk.ca", 2, false},
	{1, "gc.ca", 2, false},
	{1, "cat", 1, false},
	{1, "cc", 1, false},
	{1, "cd", 1, false},
	{1, "gov.cd", 2, false},
	{1, "cf", 1, false},
	{1, "cg", 1, false},
	{1, "ch", 1, false},
	{1, "ci", 1, false},
	{1, "org.ci", 2, false},
	{1, "or.ci", 2, false},
	{1, "com.ci", 2, false},
	{1, "co.ci", 2, false},
	{1, "edu.ci", 2, false},
	{1, "ed.ci", 2, false},
	{1, "ac.ci", 2, false},
	{1, "net.ci", 2, false},
	{1, "go.ci", 2, false},
	{1, "asso.ci", 2, false},
	{1, "xn--aroport-bya.ci", 2, false},
	{1, "int.ci", 2, false},
	{1, "presse.ci", 2, false},
	{1, "md.ci", 2, false},
	{1, "gouv.ci", 2, false},
	{2, "ck", 2, false},
	{3, "www.ck", 2, false},
	{1, "cl", 1, false},
	{1, "aprendemas.cl", 2, false},
	{1, "co.cl", 2, false},
	{1, "gob.cl", 2, false},
	{1, "gov.cl", 2, false},
	{1, "mil.cl", 2, false},
	{1, "cm", 1, false},
	{1, "co.cm", 2, false},
	{1, "com.cm", 2, false},
	{1, "gov.cm", 2, false},
	{1, "net.cm", 2, false},
	{1, "cn", 1, false},
	{1, "ac.cn", 2, false},
	{1, "com.cn", 2, false},
	{1, "edu.cn", 2, false},
	{1, "gov.cn", 2, false},
	{1, "net.cn", 2, false},
	{1, "org.cn", 2, false},
	{1, "mil.cn", 2, false},
	{1, "xn--55qx5d.cn", 2, false},
	{1, "xn--io0a7i.cn", 2, false},
	{1, "xn--od0alg.cn", 2, false},
	{1, "ah.cn", 2, false},
	{1, "bj.cn", 2, false},
	{1, "cq.cn", 2, false},
	{1, "fj.cn", 2, false},
	{1, "gd.cn", 2, false},
	{1, "gs.cn", 2, false},
	{1, "gz.cn", 2, false},
	{1, "gx.cn", 2, false},
	{1, "ha.cn", 2, false},
	{1, "hb.cn", 2, false},
	{1, "he.cn", 2, false},
	{1, "hi.cn", 2, false},
	{1, "hl.cn", 2, false},
	{1, "hn.cn", 2, false},
	{1, "jl.cn", 2, false},
	{1, "js.cn", 2, false},
	{1, "jx.cn", 2, false},
	{1, "ln.cn", 2, false},
	{1, "nm.cn", 2, false},
	{1, "nx.cn", 2, false},
	{1, "qh.cn", 2, false},
	{1, "sc.cn", 2, false},
	{1, "sd.cn", 2, false},
	{1, "sh.cn", 2, false},
	{1, "sn.cn", 2, false},
	{1, "sx.cn", 2, false},
	{1, "tj.cn", 2, false},
	{1, "xj.cn", 2, false},
	{1, "xz.cn", 2, false},
	{1, "yn.cn", 2, false},
	{1, "zj.cn", 2, false},
	{1, "hk.cn", 2, false},
	{1, "mo.cn", 2, false},
	{1, "tw.cn", 2, false},
	{1, "co", 1, false},
	{1, "arts.co", 2, false},
	{1, "com.co", 2, false},
	{1, "edu.co", 2, false},
	{1, "firm.co", 2, false},
	{1, "gov.co", 2, false},
	{1, "info.co", 2, false},
	{1, "int.co", 2, false},
	{1, "mil.co", 2, false},
	{1, "net.co", 2, false},
	{1, "nom.co", 2, false},
	{1, "org.co", 2, false},
	{1, "rec.co", 2, false},
	{1, "web.co", 2, false},
	{1, "com", 1, false},
	{1, "coop", 1, false},
	{1, "cr", 1, false},
	{1, "ac.cr", 2, false},
	{1, "co.cr", 2, false},
	{1, "ed.cr", 2, false},
	{1, "fi.cr", 2, false},
	{1, "go.cr", 2, false},
	{1, "or.cr", 2, false},
	{1, "sa.cr", 2, false},
	{1, "cu", 1, false},
	{1, "com.cu", 2, false},
	{1, "edu.cu", 2, false},
	{1, "org.cu", 2, false},
	{1, "net.cu", 2, false},
	{1, "gov.cu", 2, false},
	{1, "inf.cu", 2, false},
	{1, "cv", 1, false},
	{1, "cw", 1, false},
	{1, "com.cw", 2, false},
	{1, "edu.cw", 2, false},
	{1, "net.cw", 2, false},
	{1, "org.cw", 2, false},
	{1, "cx", 1, false},
	{1, "gov.cx", 2, false},
	{1, "cy", 1, false},
	{1, "ac.cy", 2, false},
	{1, "biz.cy", 2, false},
	{1, "com.cy", 2, false},
	{1, "ekloges.cy", 2, false},
	{1, "gov.cy", 2, false},
	{1, "ltd.cy", 2, false},
	{1, "name.cy", 2, false},
	{1, "net.cy", 2, false},
	{1, "org.cy", 2, false},
	{1, "parliament.cy", 2, false},
	{1, "press.cy", 2, false},
	{1, "pro.cy", 2, false},
	{1, "tm.cy", 2, false},
	{1, "cz", 1, false},
	{1, "de", 1, false},
	{1, "dj", 1, false},
	{1, "dk", 1, false},
	{1, "dm", 1, false},
	{1, "com.dm", 2, false},
	{1, "net.dm", 2, false},
	{1, "org.dm", 2, false},
	{1, "edu.dm", 2, false},
	{1, "gov.dm", 2, false},
	{1, "do", 1, false},
	{1, "art.do", 2, false},
	{1, "com.do", 2, false},
	{1, "edu.do", 2, false},
	{1, "gob.do", 2, false},
	{1, "gov.do", 2, false},
	{1, "mil.do", 2, false},
	{1, "net.do", 2, false},
	{1, "org.do", 2, false},
	{1, "sld.do", 2, false},
	{1, "web.do", 2, false},
	{1, "dz", 1, false},
	{1, "art.dz", 2, false},
	{1, "asso.dz", 2, false},
	{1, "com.dz", 2, false},
	{1, "edu.dz", 2, false},
	{1, "gov.dz", 2, false},
	{1, "org.dz", 2, false},
	{1, "net.dz", 2, false},
	{1, "pol.dz", 2, false},
	{1, "soc.dz", 2, false},
	{1, "tm.dz", 2, false},
	{1, "ec", 1, false},
	{1, "com.ec", 2, false},
	{1, "info.ec", 2, false},
	{1, "net.ec", 2, false},
	{1, "fin.ec", 2, false},
	{1, "k12.ec", 2, false},
	{1, "med.ec", 2, false},
	{1, "pro.ec", 2, false},
	{1, "org.ec", 2, false},
	{1, "edu.ec", 2, false},
	{1, "gov.ec", 2, false},
	{1, "gob.ec", 2, false},
	{1, "mil.ec", 2, false},
	{1, "edu", 1, false},
	{1, "ee", 1, false},
	{1, "edu.ee", 2, false},
	{1, "gov.ee", 2, false},
	{1, "riik.ee", 2, false},
	{1, "lib.ee", 2, false},
	{1, "med.ee", 2, false},
	{1, "com.ee", 2, false},
	{1, "pri.ee", 2, false},
	{1, "aip.ee", 2, false},
	{1, "org.ee", 2, false},
	{1, "fie.ee", 2, false},
	{1, "eg", 1, false},
	{1, "com.eg", 2, false},
	{1, "edu.eg", 2, false},
	{1, "eun.eg", 2, false},
	{1, "gov.eg", 2, false},
	{1, "mil.eg", 2, false},
	{1, "name.eg", 2, false},
	{1, "net.eg", 2, false},
	{1, "org.eg", 2, false},
	{1, "sci.eg", 2, false},
	{2, "er", 2, false},
	{1, "es", 1, false},
	{1, "com.es", 2, false},
	{1, "nom.es", 2, false},
	{1, "org.es", 2, false},
	{1, "gob.es", 2, false},
	{1, "edu.es", 2, false},
	{1, "et", 1, false},
	{1, "com.et", 2, false},
	{1, "gov.et", 2, false},
	{1, "org.et", 2, false},
	{1, "edu.et", 2, false},
	{1, "biz.et", 2, false},
	{1, "name.et", 2, false},
	{1, "info.et", 2, false},
	{1, "net.et", 2, false},
	{1, "eu", 1, false},
	{1, "fi", 1, false},
	{1, "aland.fi", 2, false},
	{1, "fj", 1, false},
	{1, "ac.fj", 2, false},
	{1, "biz.fj", 2, false},
	{1, "com.fj", 2, false},
	{1, "gov.fj", 2, false},
	{1, "info.fj", 2, false},
	{1, "mil.fj", 2, false},
	{1, "name.fj", 2, false},
	{1, "net.fj", 2, false},
	{1, "org.fj", 2, false},
	{1, "pro.fj", 2, false},
	{2, "fk", 2, false},
	{1, "com.fm", 2, false},
	{1, "edu.fm", 2, false},
	{1, "net.fm", 2, false},
	{1, "org.fm", 2, false},
	{1, "fm", 1, false},
	{1, "fo", 1, false},
	{1, "fr", 1, false},
	{1, "asso.fr", 2, false},
	{1, "com.fr", 2, false},
	{1, "gouv.fr", 2, false},
	{1, "nom.fr", 2, false},
	{1, "prd.fr", 2, false},
	{1, "tm.fr", 2, false},
	{1, "aeroport.fr", 2, false},
	{1, "avocat.fr", 2, false},
	{1, "avoues.fr", 2, false},
	{1, "cci.fr", 2, false},
	{1, "chambagri.fr", 2, false},
	{1, "chirurgiens-dentistes.fr", 2, false},
	{1, "experts-comptables.fr", 2, false},
	{1, "geometre-expert.fr", 2, false},
	{1, "greta.fr", 2, false},
	{1, "huissier-justice.fr", 2, false},
	{1, "medecin.fr", 2, false},
	{1, "notaires.fr", 2, false},
	{1, "pharmacien.fr", 2, false},
	{1, "port.fr", 2, false},
	{1, "veterinaire.fr", 2, false},
	{1, "ga", 1, false},
	{1, "gb", 1, false},
	{1, "edu.gd", 2, false},
	{1, "gov.gd", 2, false},
	{1, "gd", 1, false},
	{1, "ge", 1, false},
	{1, "com.ge", 2, false},
	{1, "edu.ge", 2, false},
	{1, "gov.ge", 2, false},
	{1, "org.ge", 2, false},
	{1, "mil.ge", 2, false},
	{1, "net.ge", 2, false},
	{1, "pvt.ge", 2, false},
	{1, "gf", 1, false},
	{1, "gg", 1, false},
	{1, "co.gg", 2, false},
	{1, "net.gg", 2, false},
	{1, "org.gg", 2, false},
	{1, "gh", 1, false},
	{1, "com.gh", 2, false},
	{1, "edu.gh", 2, false},
	{1, "gov.gh", 2, false},
	{1, "org.gh", 2, false},
	{1, "mil.gh", 2, false},
	{1, "gi", 1, false},
	{1, "com.gi", 2, false},
	{1, "ltd.gi", 2, false},
	{1, "gov.gi", 2, false},
	{1, "mod.gi", 2, false},
	{1, "edu.gi", 2, false},
	{1, "org.gi", 2, false},
	{1, "gl", 1, false},
	{1, "co.gl", 2, false},
	{1, "com.gl", 2, false},
	{1, "edu.gl", 2, false},
	{1, "net.gl", 2, false},
	{1, "org.gl", 2, false},
	{1, "gm", 1, false},
	{1, "gn", 1, false},
	{1, "ac.gn", 2, false},
	{1, "com.gn", 2, false},
	{1, "edu.gn", 2, false},
	{1, "gov.gn", 2, false},
	{1, "org.gn", 2, false},
	{1, "net.gn", 2, false},
	{1, "gov", 1, false},
	{1, "gp", 1, false},
	{1, "com.gp", 2, false},
	{1, "net.gp", 2, false},
	{1, "mobi.gp", 2, false},
	{1, "edu.gp", 2, false},
	{1, "org.gp", 2, false},
	{1, "asso.gp", 2, false},
	{1, "gq", 1, false},
	{1, "gr", 1, false},
	{1, "com.gr", 2, false},
	{1, "edu.gr", 2, false},
	{1, "net.gr", 2, false},
	{1, "org.gr", 2, false},
	{1, "gov.gr", 2, false},
	{1, "gs", 1, false},
	{1, "gt", 1, false},
	{1, "com.gt", 2, false},
	{1, "edu.gt", 2, false},
	{1, "gob.gt", 2, false},
	{1, "ind.gt", 2, false},
	{1, "mil.gt", 2, false},
	{1, "net.gt", 2, false},
	{1, "org.gt", 2, false},
	{1, "gu", 1, false},
	{1, "com.gu", 2, false},
	{1, "edu.gu", 2, false},
	{1, "gov.gu", 2, false},
	{1, "guam.gu", 2, false},
	{1, "info.gu", 2, false},
	{1, "net.gu", 2, false},
	{1, "org.gu", 2, false},
	{1, "web.gu", 2, false},
	{1, "gw", 1, false},
	{1, "gy", 1, false},
	{1, "co.gy", 2, false},
	{1, "com.gy", 2, false},
	{1, "edu.gy", 2, false},
	{1, "gov.gy", 2, false},
	{1, "net.gy", 2, false},
	{1, "org.gy", 2, false},
	{1, "hk", 1, false},
	{1, "com.hk", 2, false},
	{1, "edu.hk", 2, false},
	{1, "gov.hk", 2, false},
	{1, "idv.hk", 2, false},
	{1, "net.hk", 2, false},
	{1, "org.hk", 2, false},
	{1, "xn--55qx5d.hk", 2, false},
	{1, "xn--wcvs22d.hk", 2, false},
	{1, "xn--lcvr32d.hk", 2, false},
	{1, "xn--mxtq1m.hk", 2, false},
	{1, "xn--gmqw5a.hk", 2, false},
	{1, "xn--ciqpn.hk", 2, false},
	{1, "xn--gmq050i.hk", 2, false},
	{1, "xn--zf0avx.hk", 2, false},
	{1, "xn--io0a7i.hk", 2, false},
	{1, "xn--mk0axi.hk", 2, false},
	{1, "xn--od0alg.hk", 2, false},
	{1, "xn--od0aq3b.hk", 2, false},
	{1, "xn--tn0ag.hk", 2, false},
	{1, "xn--uc0atv.hk", 2, false},
	{1, "xn--uc0ay4a.hk", 2, false},
	{1, "hm", 1, false},
	{1, "hn", 1, false},
	{1, "com.hn", 2, false},
	{1, "edu.hn", 2, false},
	{1, "org.hn", 2, false},
	{1, "net.hn", 2, false},
	{1, "mil.hn", 2, false},
	{1, "gob.hn", 2, false},
	{1, "hr", 1, false},
	{1, "iz.hr", 2, false},
	{1, "from.hr", 2, false},
	{1, "name.hr", 2, false},
	{1, "com.hr", 2, false},
	{1, "ht", 1, false},
	{1, "com.ht", 2, false},
	{1, "shop.ht", 2, false},
	{1, "firm.ht", 2, false},
	{1, "info.ht", 2, false},
	{1, "adult.ht", 2, false},
	{1, "net.ht", 2, false},
	{1, "pro.ht", 2, false},
	{1, "org.ht", 2, false},
	{1, "med.ht", 2, false},
	{1, "art.ht", 2, false},
	{1, "coop.ht", 2, false},
	{1, "pol.ht", 2, false},
	{1, "asso.ht", 2, false},
	{1, "edu.ht", 2, false},
	{1, "rel.ht", 2, false},
	{1, "gouv.ht", 2, false},
	{1, "perso.ht", 2, false},
	{1, "hu", 1, false},
	{1, "co.hu", 2, false},
	{1, "info.hu", 2, false},
	{1, "org.hu", 2, false},
	{1, "priv.hu", 2, false},
	{1, "sport.hu", 2, false},
	{1, "tm.hu", 2, false},
	{1, "2000.hu", 2, false},
	{1, "agrar.hu", 2, false},
	{1, "bolt.hu", 2, false},
	{1, "casino.hu", 2, false},
	{1, "city.hu", 2, false},
	{1, "erotica.hu", 2, false},
	{1, "erotika.hu", 2, false},
	{1, "film.hu", 2, false},
	{1, "forum.hu", 2, false},
	{1, "games.hu", 2, false},
	{1, "hotel.hu", 2, false},
	{1, "ingatlan.hu", 2, false},
	{1, "jogasz.hu", 2, false},
	{1, "konyvelo.hu", 2, false},
	{1, "lakas.hu", 2, false},
	{1, "media.hu", 2, false},
	{1, "news.hu", 2, false},
	{1, "reklam.hu", 2, false},
	{1, "sex.hu", 2, false},
	{1, "shop.hu", 2, false},
	{1, "suli.hu", 2, false},
	{1, "szex.hu", 2, false},
	{1, "tozsde.hu", 2, false},
	{1, "utazas.hu", 2, false},
	{1, "video.hu", 2, false},
	{1, "id", 1, false},
	{1, "ac.id", 2, false},
	{1, "biz.id", 2, false},
	{1, "co.id", 2, false},
	{1, "desa.id", 2, false},
	{1, "go.id", 2, false},
	{1, "mil.id", 2, false},
	{1, "my.id", 2, false},
	{1, "net.id", 2, false},
	{1, "or.id", 2, false},
	{1, "ponpes.id", 2, false},
	{1, "sch.id", 2, false},
	{1, "web.id", 2, false},
	{1, "ie", 1, false},
	{1, "gov.ie", 2, false},
	{1, "il", 1, false},
	{1, "ac.il", 2, false},
	{1, "co.il", 2, false},
	{1, "gov.il", 2, false},
	{1, "idf.il", 2, false},
	{1, "k12.il", 2, false},
	{1, "muni.il", 2, false},
	{1, "net.il", 2, false},
	{1, "org.il", 2, false},
	{1, "im", 1, false},
	{1, "ac.im", 2, false},
	{1, "co.im", 2, false},
	{1, "com.im", 2, false},
	{1, "ltd.co.im", 3, false},
	{1, "net.im", 2, false},
	{1, "org.im", 2, false},
	{1, "plc.co.im", 3, false},
	{1, "tt.im", 2, false},
	{1, "tv.im", 2, false},
	{1, "in", 1, false},
	{1, "co.in", 2, false},
	{1, "firm.in", 2, false},
	{1, "net.in", 2, false},
	{1, "org.in", 2, false},
	{1, "gen.in", 2, false},
	{1, "ind.in", 2, false},
	{1, "nic.in", 2, false},
	{1, "ac.in", 2, false},
	{1, "edu.in", 2, false},
	{1, "res.in", 2, false},
	{1, "gov.in", 2, false},
	{1, "mil.in", 2, false},
	{1, "info", 1, false},
	{1, "int", 1, false},
	{1, "eu.int", 2, false},
	{1, "io", 1, false},
	{1, "com.io", 2, false},
	{1, "iq", 1, false},
	{1, "gov.iq", 2, false},
	{1, "edu.iq", 2, false},
	{1, "mil.iq", 2, false},
	{1, "com.iq", 2, false},
	{1, "org.iq", 2, false},
	{1, "net.iq", 2, false},
	{1, "ir", 1, false},
	{1, "ac.ir", 2, false},
	{1, "co.ir", 2, false},
	{1, "gov.ir", 2, false},
	{1, "id.ir", 2, false},
	{1, "net.ir", 2, false},
	{1, "org.ir", 2, false},
	{1, "sch.ir", 2, false},
	{1, "xn--mgba3a4f16a.ir", 2, false},
	{1, "xn--mgba3a4fra.ir", 2, false},
	{1, "is", 1, false},
	{1, "net.is", 2, false},
	{1, "com.is", 2, false},
	{1, "edu.is", 2, false},
	{1, "gov.is", 2, false},
	{1, "org.is", 2, false},
	{1, "int.is", 2, false},
	{1, "it", 1, false},
	{1, "gov.it", 2, false},
	{1, "edu.it", 2, false},
	{1, "abr.it", 2, false},
	{1, "abruzzo.it", 2, false},
	{1, "aosta-valley.it", 2, false},
	{1, "aostavalley.it", 2, false},
	{1, "bas.it", 2, false},
	{1, "basilicata.it", 2, false},
	{1, "cal.it", 2, false},
	{1, "calabria.it", 2, false},
	{1, "cam.it", 2, false},
	{1, "campania.it", 2, false},
	{1, "emilia-romagna.it", 2, false},
	{1, "emiliaromagna.it", 2, false},
	{1, "emr.it", 2, false},
	{1, "friuli-v-giulia.it", 2, false},
	{1, "friuli-ve-giulia.it", 2, false},
	{1, "friuli-vegiulia.it", 2, false},
	{1, "friuli-venezia-giulia.it", 2, false},
	{1, "friuli-veneziagiulia.it", 2, false},
	{1, "friuli-vgiulia.it", 2, false},
	{1, "friuliv-giulia.it", 2, false},
	{1, "friulive-giulia.it", 2, false},
	{1, "friulivegiulia.it", 2, false},
	{1, "friulivenezia-giulia.it", 2, false},
	{1, "friuliveneziagiulia.it", 2, false},
	{1, "friulivgiulia.it", 2, false},
	{1, "fvg.it", 2, false},
	{1, "laz.it", 2, false},
	{1, "lazio.it", 2, false},
	{1, "lig.it", 2, false},
	{1, "liguria.it", 2, false},
	{1, "lom.it", 2, false},
	{1, "lombardia.it", 2, false},
	{1, "lombardy.it", 2, false},
	{1, "lucania.it", 2, false},
	{1, "mar.it", 2, false},
	{1, "marche.it", 2, false},
	{1, "mol.it", 2, false},
	{1, "molise.it", 2, false},
	{1, "piedmont.it", 2, false},
	{1, "piemonte.it", 2, false},
	{1, "pmn.it", 2, false},
	{1, "pug.it", 2, false},
	{1, "puglia.it", 2, false},
	{1, "sar.it", 2, false},
	{1, "sardegna.it", 2, false},
	{1, "sardinia.it", 2, false},
	{1, "sic.it", 2, false},
	{1, "sicilia.it", 2, false},
	{1, "sicily.it", 2, false},
	{1, "taa.it", 2, false},
	{1, "tos.it", 2, false},
	{1, "toscana.it", 2, false},
	{1, "trentin-sud-tirol.it", 2, false},
	{1, "xn--trentin-sd-tirol-rzb.it", 2, false},
	{1, "trentin-sudtirol.it", 2, false},
	{1, "xn--trentin-sdtirol-7vb.it", 2, false},
	{1, "trentin-sued-tirol.it", 2, false},
	{1, "trentin-suedtirol.it", 2, false},
	{1, "trentino-a-adige.it", 2, false},
	{1, "trentino-aadige.it", 2, false},
	{1, "trentino-alto-adige.it", 2, false},
	{1, "trentino-altoadige.it", 2, false},
	{1, "trentino-s-tirol.it", 2, false},
	{1, "trentino-stirol.it", 2, false},
	{1, "trentino-sud-tirol.it", 2, false},
	{1, "xn--trentino-sd-tirol-c3b.it", 2, false},
	{1, "trentino-sudtirol.it", 2, false},
	{1, "xn--trentino-sdtirol-szb.it", 2, false},
	{1, "trentino-sued-tirol.it", 2, false},
	{1, "trentino-suedtirol.it", 2, false},
	{1, "trentino.it", 2, false},
	{1, "trentinoa-adige.it", 2, false},
	{1, "trentinoaadige.it", 2, false},
	{1, "trentinoalto-adige.it", 2, false},
	{1, "trentinoaltoadige.it", 2, false},
	{1, "trentinos-tirol.it", 2, false},
	{1, "trentinostirol.it", 2, false},
	{1, "trentinosud-tirol.it", 2, false},
	{1, "xn--trentinosd-tirol-rzb.it", 2, false},
	{1, "trentinosudtirol.it", 2, false},
	{1, "xn--trentinosdtirol-7vb.it", 2, false},
	{1, "trentinosued-tirol.it", 2, false},
	{1, "trentinosuedtirol.it", 2, false},
	{1, "trentinsud-tirol.it", 2, false},
	{1, "xn--trentinsd-tirol-6vb.it", 2, false},
	{1, "trentinsudtirol.it", 2, false},
	{1, "xn--trentinsdtirol-nsb.it", 2, false},
	{1, "trentinsued-tirol.it", 2, false},
	{1, "trentinsuedtirol.it", 2, false},
	{1, "tuscany.it", 2, false},
	{1, "umb.it", 2, false},
	{1, "umbria.it", 2, false},
	{1, "val-d-aosta.it", 2, false},
	{1, "val-daosta.it", 2, false},
	{1, "vald-aosta.it", 2, false},
	{1, "valdaosta.it", 2, false},
	{1, "valle-aosta.it", 2, false},
	{1, "valle-d-aosta.it", 2, false},
	{1, "valle-daosta.it", 2, false},
	{1, "valleaosta.it", 2, false},
	{1, "valled-aosta.it", 2, false},
	{1, "valledaosta.it", 2, false},
	{1, "vallee-aoste.it", 2, false},
	{1, "xn--valle-aoste-ebb.it", 2, false},
	{1, "vallee-d-aoste.it", 2, false},
	{1, "xn--valle-d-aoste-ehb.it", 2, false},
	{1, "valleeaoste.it", 2, false},
	{1, "xn--valleaoste-e7a.it", 2, false},
	{1, "valleedaoste.it", 2, false},
	{1, "xn--valledaoste-ebb.it", 2, false},
	{1, "vao.it", 2, false},
	{1, "vda.it", 2, false},
	{1, "ven.it", 2, false},
	{1, "veneto.it", 2, false},
	{1, "ag.it", 2, false},
	{1, "agrigento.it", 2, false},
	{1, "al.it", 2, false},
	{1, "alessandria.it", 2, false},
	{1, "alto-adige.it", 2, false},
	{1, "altoadige.it", 2, false},
	{1, "an.it", 2, false},
	{1, "ancona.it", 2, false},
	{1, "andria-barletta-trani.it", 2, false},
	{1, "andria-trani-barletta.it", 2, false},
	{1, "andriabarlettatrani.it", 2, false},
	{1, "andriatranibarletta.it", 2, false},
	{1, "ao.it", 2, false},
	{1, "aosta.it", 2, false},
	{1, "aoste.it", 2, false},
	{1, "ap.it", 2, false},
	{1, "aq.it", 2, false},
	{1, "aquila.it", 2, false},
	{1, "ar.it", 2, false},
	{1, "arezzo.it", 2, false},
	{1, "ascoli-piceno.it", 2, false},
	{1, "ascolipiceno.it", 2, false},
	{1, "asti.it", 2, false},
	{1, "at.it", 2, false},
	{1, "av.it", 2, false},
	{1, "avellino.it", 2, false},
	{1, "ba.it", 2, false},
	{1, "balsan-sudtirol.it", 2, false},
	{1, "xn--balsan-sdtirol-nsb.it", 2, false},
	{1, "balsan-suedtirol.it", 2, false},
	{1, "balsan.it", 2, false},
	{1, "bari.it", 2, false},
	{1, "barletta-trani-andria.it", 2, false},
	{1, "barlettatraniandria.it", 2, false},
	{1, "belluno.it", 2, false},
	{1, "benevento.it", 2, false},
	{1, "bergamo.it", 2, false},
	{1, "bg.it", 2, false},
	{1, "bi.it", 2, false},
	{1, "biella.it", 2, false},
	{1, "bl.it", 2, false},
	{1, "bn.it", 2, false},
	{1, "bo.it", 2, false},
	{1, "bologna.it", 2, false},
	{1, "bolzano-altoadige.it", 2, false},
	{1, "bolzano.it", 2, false},
	{1, "bozen-sudtirol.it", 2, false},
	{1, "xn--bozen-sdtirol-2ob.it", 2, false},
	{1, "bozen-suedtirol.it", 2, false},
	{1, "bozen.it", 2, false},
	{1, "br.it", 2, false},
	{1, "brescia.it", 2, false},
	{1, "brindisi.it", 2, false},
	{1, "bs.it", 2, false},
	{1, "bt.it", 2, false},
	{1, "bulsan-sudtirol.it", 2, false},
	{1, "xn--bulsan-sdtirol-nsb.it", 2, false},
	{1, "bulsan-suedtirol.it", 2, false},
	{1, "bulsan.it", 2, false},
	{1, "bz.it", 2, false},
	{1, "ca.it", 2, false},
	{1, "cagliari.it", 2, false},
	{1, "caltanissetta.it", 2, false},
	{1, "campidano-medio.it", 2, false},
	{1, "campidanomedio.it", 2, false},
	{1, "campobasso.it", 2, false},
	{1, "carbonia-iglesias.it", 2, false},
	{1, "carboniaiglesias.it", 2, false},
	{1, "carrara-massa.it", 2, false},
	{1, "carraramassa.it", 2, false},
	{1, "caserta.it", 2, false},
	{1, "catania.it", 2, false},
	{1, "catanzaro.it", 2, false},
	{1, "cb.it", 2, false},
	{1, "ce.it", 2, false},
	{1, "cesena-forli.it", 2, false},
	{1, "xn--cesena-forl-mcb.it", 2, false},
	{1, "cesenaforli.it", 2, false},
	{1, "xn--cesenaforl-i8a.it", 2, false},
	{1, "ch.it", 2, false},
	{1, "chieti.it", 2, false},
	{1, "ci.it", 2, false},
	{1, "cl.it", 2, false},
	{1, "cn.it", 2, false},
	{1, "co.it", 2, false},
	{1, "como.it", 2, false},
	{1, "cosenza.it", 2, false},
	{1, "cr.it", 2, false},
	{1, "cremona.it", 2, false},
	{1, "crotone.it", 2, false},
	{1, "cs.it", 2, false},
	{1, "ct.it", 2, false},
	{1, "cuneo.it", 2, false},
	{1, "cz.it", 2, false},
	{1, "dell-ogliastra.it", 2, false},
	{1, "dellogliastra.it", 2, false},
	{1, "en.it", 2, false},
	{1, "enna.it", 2, false},
	{1, "fc.it", 2, false},
	{1, "fe.it", 2, false},
	{1, "fermo.it", 2, false},
	{1, "ferrara.it", 2, false},
	{1, "fg.it", 2, false},
	{1, "fi.it", 2, false},
	{1, "firenze.it", 2, false},
	{1, "florence.it", 2, false},
	{1, "fm.it", 2, false},
	{1, "foggia.it", 2, false},
	{1, "forli-cesena.it", 2, false},
	{1, "xn--forl-cesena-fcb.it", 2, false},
	{1, "forlicesena.it", 2, false},
	{1, "xn--forlcesena-c8a.it", 2, false},
	{1, "fr.it", 2, false},
	{1, "frosinone.it", 2, false},
	{1, "ge.it", 2, false},
	{1, "genoa.it", 2, false},
	{1, "genova.it", 2, false},
	{1, "go.it", 2, false},
	{1, "gorizia.it", 2, false},
	{1, "gr.it", 2, false},
	{1, "grosseto.it", 2, false},
	{1, "iglesias-carbonia.it", 2, false},
	{1, "iglesiascarbonia.it", 2, false},
	{1, "im.it", 2, false},
	{1, "imperia.it", 2, false},
	{1, "is.it", 2, false},
	{1, "isernia.it", 2, false},
	{1, "kr.it", 2, false},
	{1, "la-spezia.it", 2, false},
	{1, "laquila.it", 2, false},
	{1, "laspezia.it", 2, false},
	{1, "latina.it", 2, false},
	{1, "lc.it", 2, false},
	{1, "le.it", 2, false},
	{1, "lecce.it", 2, false},
	{1, "lecco.it", 2, false},
	{1, "li.it", 2, false},
	{1, "livorno.it", 2, false},
	{1, "lo.it", 2, false},
	{1, "lodi.it", 2, false},
	{1, "lt.it", 2, false},
	{1, "lu.it", 2, false},
	{1, "lucca.it", 2, false},
	{1, "macerata.it", 2, false},
	{1, "mantova.it", 2, false},
	{1, "massa-carrara.it", 2, false},
	{1, "massacarrara.it", 2, false},
	{1, "matera.it", 2, false},
	{1, "mb.it", 2, false},
	{1, "mc.it", 2, false},
	{1, "me.it", 2, false},
	{1, "medio-campidano.it", 2, false},
	{1, "mediocampidano.it", 2, false},
	{1, "messina.it", 2, false},
	{1, "mi.it", 2, false},
	{1, "milan.it", 2, false},
	{1, "milano.it", 2, false},
	{1, "mn.it", 2, false},
	{1, "mo.it", 2, false},
	{1, "modena.it", 2, false},
	{1, "monza-brianza.it", 2, false},
	{1, "monza-e-della-brianza.it", 2, false},
	{1, "monza.it", 2, false},
	{1, "monzabrianza.it", 2, false},
	{1, "monzaebrianza.it", 2, false},
	{1, "monzaedellabrianza.it", 2, false},
	{1, "ms.it", 2, false},
	{1, "mt.it", 2, false},
	{1, "na.it", 2, false},
	{1, "naples.it", 2, false},
	{1, "napoli.it", 2, false},
	{1, "no.it", 2, false},
	{1, "novara.it", 2, false},
	{1, "nu.it", 2, false},
	{1, "nuoro.it", 2, false},
	{1, "og.it", 2, false},
	{1, "ogliastra.it", 2, false},
	{1, "olbia-tempio.it", 2, false},
	{1, "olbiatempio.it", 2, false},
	{1, "or.it", 2, false},
	{1, "oristano.it", 2, false},
	{1, "ot.it", 2, false},
	{1, "pa.it", 2, false},
	{1, "padova.it", 2, false},
	{1, "padua.it", 2, false},
	{1, "palermo.it", 2, false},
	{1, "parma.it", 2, false},
	{1, "pavia.it", 2, false},
	{1, "pc.it", 2, false},
	{1, "pd.it", 2, false},
	{1, "pe.it", 2, false},
	{1, "perugia.it", 2, false},
	{1, "pesaro-urbino.it", 2, false},
	{1, "pesarourbino.it", 2, false},
	{1, "pescara.it", 2, false},
	{1, "pg.it", 2, false},
	{1, "pi.it", 2, false},
	{1, "piacenza.it", 2, false},
	{1, "pisa.it", 2, false},
	{1, "pistoia.it", 2, false},
	{1, "pn.it", 2, false},
	{1, "po.it", 2, false},
	{1, "pordenone.it", 2, false},
	{1, "potenza.it", 2, false},
	{1, "pr.it", 2, false},
	{1, "prato.it", 2, false},
	{1, "pt.it", 2, false},
	{1, "pu.it", 2, false},
	{1, "pv.it", 2, false},
	{1, "pz.it", 2, false},
	{1, "ra.it", 2, false},
	{1, "ragusa.it", 2, false},
	{1, "ravenna.it", 2, false},
	{1, "rc.it", 2, false},
	{1, "re.it", 2, false},
	{1, "reggio-calabria.it", 2, false},
	{1, "reggio-emilia.it", 2, false},
	{1, "reggiocalabria.it", 2, false},
	{1, "reggioemilia.it", 2, false},
	{1, "rg.it", 2, false},
	{1, "ri.it", 2, false},
	{1, "rieti.it", 2, false},
	{1, "rimini.it", 2, false},
	{1, "rm.it", 2, false},
	{1, "rn.it", 2, false},
	{1, "ro.it", 2, false},
	{1, "roma.it", 2, false},
	{1, "rome.it", 2, false},
	{1, "rovigo.it", 2, false},
	{1, "sa.it", 2, false},
	{1, "salerno.it", 2, false},
	{1, "sassari.it", 2, false},
	{1, "savona.it", 2, false},
	{1, "si.it", 2, false},
	{1, "siena.it", 2, false},
	{1, "siracusa.it", 2, false},
	{1, "so.it", 2, false},
	{1, "sondrio.it", 2, false},
	{1, "sp.it", 2, false},
	{1, "sr.it", 2, false},
	{1, "ss.it", 2, false},
	{1, "suedtirol.it", 2, false},
	{1, "xn--sdtirol-n2a.it", 2, false},
	{1, "sv.it", 2, false},
	{1, "ta.it", 2, false},
	{1, "taranto.it", 2, false},
	{1, "te.it", 2, false},
	{1, "tempio-olbia.it", 2, false},
	{1, "tempioolbia.it", 2, false},
	{1, "teramo.it", 2, false},
	{1, "terni.it", 2, false},
	{1, "tn.it", 2, false},
	{1, "to.it", 2, false},
	{1, "torino.it", 2, false},
	{1, "tp.it", 2, false},
	{1, "tr.it", 2, false},
	{1, "trani-andria-barletta.it", 2, false},
	{1, "trani-barletta-andria.it", 2, false},
	{1, "traniandriabarletta.it", 2, false},
	{1, "tranibarlettaandria.it", 2, false},
	{1, "trapani.it", 2, false},
	{1, "trento.it", 2, false},
	{1, "treviso.it", 2, false},
	{1, "trieste.it", 2, false},
	{1, "ts.it", 2, false},
	{1, "turin.it", 2, false},
	{1, "tv.it", 2, false},
	{1, "ud.it", 2, false},
	{1, "udine.it", 2, false},
	{1, "urbino-pesaro.it", 2, false},
	{1, "urbinopesaro.it", 2, false},
	{1, "va.it", 2, false},
	{1, "varese.it", 2, false},
	{1, "vb.it", 2, false},
	{1, "vc.it", 2, false},
	{1, "ve.it", 2, false},
	{1, "venezia.it", 2, false},
	{1, "venice.it", 2, false},
	{1, "verbania.it", 2, false},
	{1, "vercelli.it", 2, false},
	{1, "verona.it", 2, false},
	{1, "vi.it", 2, false},
	{1, "vibo-valentia.it", 2, false},
	{1, "vibovalentia.it", 2, false},
	{1, "vicenza.it", 2, false},
	{1, "viterbo.it", 2, false},
	{1, "vr.it", 2, false},
	{1, "vs.it", 2, false},
	{1, "vt.it", 2, false},
	{1, "vv.it", 2, false},
	{1, "je", 1, false},
	{1, "co.je", 2, false},
	{1, "net.je", 2, false},
	{1, "org.je", 2, false},
	{2, "jm", 2, false},
	{1, "jo", 1, false},
	{1, "com.jo", 2, false},
	{1, "org.jo", 2, false},
	{1, "net.jo", 2, false},
	{1, "edu.jo", 2, false},
	{1, "sch.jo", 2, false},
	{1, "gov.jo", 2, false},
	{1, "mil.jo", 2, false},
	{1, "name.jo", 2, false},
	{1, "jobs", 1, false},
	{1, "jp", 1, false},
	{1, "ac.jp", 2, false},
	{1, "ad.jp", 2, false},
	{1, "co.jp", 2, false},
	{1, "ed.jp", 2, false},
	{1, "go.jp", 2, false},
	{1, "gr.jp", 2, false},
	{1, "lg.jp", 2, false},
	{1, "ne.jp", 2, false},
	{1, "or.jp", 2, false},
	{1, "aichi.jp", 2, false},
	{1, "akita.jp", 2, false},
	{1, "aomori.jp", 2, false},
	{1, "chiba.jp", 2, false},
	{1, "ehime.jp", 2, false},
	{1, "fukui.jp", 2, false},
	{1, "fukuoka.jp", 2, false},
	{1, "fukushima.jp", 2, false},
	{1, "gifu.jp", 2, false},
	{1, "gunma.jp", 2, false},
	{1, "hiroshima.jp", 2, false},
	{1, "hokkaido.jp", 2, false},
	{1, "hyogo.jp", 2, false},
	{1, "ibaraki.jp", 2, false},
	{1, "ishikawa.jp", 2, false},
	{1, "iwate.jp", 2, false},
	{1, "kagawa.jp", 2, false},
	{1, "kagoshima.jp", 2, false},
	{1, "kanagawa.jp", 2, false},
	{1, "kochi.jp", 2, false},
	{1, "kumamoto.jp", 2, false},
	{1, "kyoto.jp", 2, false},
	{1, "mie.jp", 2, false},
	{1, "miyagi.jp", 2, false},
	{1, "miyazaki.jp", 2, false},
	{1, "nagano.jp", 2, false},
	{1, "nagasaki.jp", 2, false},
	{1, "nara.jp", 2, false},
	{1, "niigata.jp", 2, false},
	{1, "oita.jp", 2, false},
	{1, "okayama.jp", 2, false},
	{1, "okinawa.jp", 2, false},
	{1, "osaka.jp", 2, false},
	{1, "saga.jp", 2, false},
	{1, "saitama.jp", 2, false},
	{1, "shiga.jp", 2, false},
	{1, "shimane.jp", 2, false},
	{1, "shizuoka.jp", 2, false},
	{1, "tochigi.jp", 2, false},
	{1, "tokushima.jp", 2, false},
	{1, "tokyo.jp", 2, false},
	{1, "tottori.jp", 2, false},
	{1, "toyama.jp", 2, false},
	{1, "wakayama.jp", 2, false},
	{1, "yamagata.jp", 2, false},
	{1, "yamaguchi.jp", 2, false},
	{1, "yamanashi.jp", 2, false},
	{1, "xn--4pvxs.jp", 2, false},
	{1, "xn--vgu402c.jp", 2, false},
	{1, "xn--c3s14m.jp", 2, false},
	{1, "xn--f6qx53a.jp", 2, false},
	{1, "xn--8pvr4u.jp", 2, false},
	{1, "xn--uist22h.jp", 2, false},
	{1, "xn--djrs72d6uy.jp", 2, false},
	{1, "xn--mkru45i.jp", 2, false},
	{1, "xn--0trq7p7nn.jp", 2, false},
	{1, "xn--8ltr62k.jp", 2, false},
	{1, "xn--2m4a15e.jp", 2, false},
	{1, "xn--efvn9s.jp", 2, false},
	{1, "xn--32vp30h.jp", 2, false},
	{1, "xn--4it797k.jp", 2, false},
	{1, "xn--1lqs71d.jp", 2, false},
	{1, "xn--5rtp49c.jp", 2, false},
	{1, "xn--5js045d.jp", 2, false},
	{1, "xn--ehqz56n.jp", 2, false},
	{1, "xn--1lqs03n.jp", 2, false},
	{1, "xn--qqqt11m.jp", 2, false},
	{1, "xn--kbrq7o.jp", 2, false},
	{1, "xn--pssu33l.jp", 2, false},
	{1, "xn--ntsq17g.jp", 2, false},
	{1, "xn--uisz3g.jp", 2, false},
	{1, "xn--6btw5a.jp", 2, false},
	{1, "xn--1ctwo.jp", 2, false},
	{1, "xn--6orx2r.jp", 2, false},
	{1, "xn--rht61e.jp", 2, false},
	{1, "xn--rht27z.jp", 2, false},
	{1, "xn--djty4k.jp", 2, false},
	{1, "xn--nit225k.jp", 2, false},
	{1, "xn--rht3d.jp", 2, false},
	{1, "xn--klty5x.jp", 2, false},
	{1, "xn--kltx9a.jp", 2, false},
	{1, "xn--kltp7d.jp", 2, false},
	{1, "xn--uuwu58a.jp", 2, false},
	{1, "xn--zbx025d.jp", 2, false},
	{1, "xn--ntso0iqx3a.jp", 2, false},
	{1, "xn--elqq16h.jp", 2, false},
	{1, "xn--4it168d.jp", 2, false},
	{1, "xn--klt787d.jp", 2, false},
	{1, "xn--rny31h.jp", 2, false},
	{1, "xn--7t0a264c.jp", 2, false},
	{1, "xn--5rtq34k.jp", 2, false},
	{1, "xn--k7yn95e.jp", 2, false},
	{1, "xn--tor131o.jp", 2, false},
	{1, "xn--d5qv7z876c.jp", 2, false},
	{2, "kawasaki.jp", 3, false},
	{2, "kitakyushu.jp", 3, false},
	{2, "kobe.jp", 3, false},
	{2, "nagoya.jp", 3, false},
	{2, "sapporo.jp", 3, false},
	{2, "sendai.jp", 3, false},
	{2, "yokohama.jp", 3, false},
	{3, "city.kawasaki.jp", 3, false},
	{3, "city.kitakyushu.jp", 3, false},
	{3, "city.kobe.jp", 3, false},
	{3, "city.nagoya.jp", 3, false},
	{3, "city.sapporo.jp", 3, false},
	{3, "city.sendai.jp", 3, false},
	{3, "city.yokohama.jp", 3, false},
	{1, "aisai.aichi.jp", 3, false},
	{1, "ama.aichi.jp", 3, false},
	{1, "anjo.aichi.jp", 3, false},
	{1, "asuke.aichi.jp", 3, false},
	{1, "chiryu.aichi.jp", 3, false},
	{1, "chita.aichi.jp", 3, false},
	{1, "fuso.aichi.jp", 3, false},
	{1, "gamagori.aichi.jp", 3, false},
	{1, "handa.aichi.jp", 3, false},
	{1, "hazu.aichi.jp", 3, false},
	{1, "hekinan.aichi.jp", 3, false},
	{1, "higashiura.aichi.jp", 3, false},
	{1, "ichinomiya.aichi.jp", 3, false},
	{1, "inazawa.aichi.jp", 3, false},
	{1, "inuyama.aichi.jp", 3, false},
	{1, "isshiki.aichi.jp", 3, false},
	{1, "iwakura.aichi.jp", 3, false},
	{1, "kanie.aichi.jp", 3, false},
	{1, "kariya.aichi.jp", 3, false},
	{1, "kasugai.aichi.jp", 3, false},
	{1, "kira.aichi.jp", 3, false},
	{1, "kiyosu.aichi.jp", 3, false},
	{1, "komaki.aichi.jp", 3, false},
	{1, "konan.aichi.jp", 3, false},
	{1, "kota.aichi.jp", 3, false},
	{1, "mihama.aichi.jp", 3, false},
	{1, "miyoshi.aichi.jp", 3, false},
	{1, "nishio.aichi.jp", 3, false},
	{1, "nisshin.aichi.jp", 3, false},
	{1, "obu.aichi.jp", 3, false},
	{1, "oguchi.aichi.jp", 3, false},
	{1, "oharu.aichi.jp", 3, false},
	{1, "okazaki.aichi.jp", 3, false},
	{1, "owariasahi.aichi.jp", 3, false},
	{1, "seto.aichi.jp", 3, false},
	{1, "shikatsu.aichi.jp", 3, false},
	{1, "shinshiro.aichi.jp", 3, false},
	{1, "shitara.aichi.jp", 3, false},
	{1, "tahara.aichi.jp", 3, false},
	{1, "takahama.aichi.jp", 3, false},
	{1, "tobishima.aichi.jp", 3, false},
	{1, "toei.aichi.jp", 3, false},
	{1, "togo.aichi.jp", 3, false},
	{1, "tokai.aichi.jp", 3, false},
	{1, "tokoname.aichi.jp", 3, false},
	{1, "toyoake.aichi.jp", 3, false},
	{1, "toyohashi.aichi.jp", 3, false},
	{1, "toyokawa.aichi.jp", 3, false},
	{1, "toyone.aichi.jp", 3, false},
	{1, "toyota.aichi.jp", 3, false},
	{1, "tsushima.aichi.jp", 3, false},
	{1, "yatomi.aichi.jp", 3, false},
	{1, "akita.akita.jp", 3, false},
	{1, "daisen.akita.jp", 3, false},
	{1, "fujisato.akita.jp", 3, false},
	{1, "gojome.akita.jp", 3, false},
	{1, "hachirogata.akita.jp", 3, false},
	{1, "happou.akita.jp", 3, false},
	{1, "higashinaruse.akita.jp", 3, false},
	{1, "honjo.akita.jp", 3, false},
	{1, "honjyo.akita.jp", 3, false},
	{1, "ikawa.akita.jp", 3, false},
	{1, "kamikoani.akita.jp", 3, false},
	{1, "kamioka.akita.jp", 3, false},
	{1, "katagami.akita.jp", 3, false},
	{1, "kazuno.akita.jp", 3, false},
	{1, "kitaakita.akita.jp", 3, false},
	{1, "kosaka.akita.jp", 3, false},
	{1, "kyowa.akita.jp", 3, false},
	{1, "misato.akita.jp", 3, false},
	{1, "mitane.akita.jp", 3, false},
	{1, "moriyoshi.akita.jp", 3, false},
	{1, "nikaho.akita.jp", 3, false},
	{1, "noshiro.akita.jp", 3, false},
	{1, "odate.akita.jp", 3, false},
	{1, "oga.akita.jp", 3, false},
	{1, "ogata.akita.jp", 3, false},
	{1, "semboku.akita.jp", 3, false},
	{1, "yokote.akita.jp", 3, false},
	{1, "yurihonjo.akita.jp", 3, false},
	{1, "aomori.aomori.jp", 3, false},
	{1, "gonohe.aomori.jp", 3, false},
	{1, "hachinohe.aomori.jp", 3, false},
	{1, "hashikami.aomori.jp", 3, false},
	{1, "hiranai.aomori.jp", 3, false},
	{1, "hirosaki.aomori.jp", 3, false},
	{1, "itayanagi.aomori.jp", 3, false},
	{1, "kuroishi.aomori.jp", 3, false},
	{1, "misawa.aomori.jp", 3, false},
	{1, "mutsu.aomori.jp", 3, false},
	{1, "nakadomari.aomori.jp", 3, false},
	{1, "noheji.aomori.jp", 3, false},
	{1, "oirase.aomori.jp", 3, false},
	{1, "owani.aomori.jp", 3, false},
	{1, "rokunohe.aomori.jp", 3, false},
	{1, "sannohe.aomori.jp", 3, false},
	{1, "shichinohe.aomori.jp", 3, false},
	{1, "shingo.aomori.jp", 3, false},
	{1, "takko.aomori.jp", 3, false},
	{1, "towada.aomori.jp", 3, false},
	{1, "tsugaru.aomori.jp", 3, false},
	{1, "tsuruta.aomori.jp", 3, false},
	{1, "abiko.chiba.jp", 3, false},
	{1, "asahi.chiba.jp", 3, false},
	{1, "chonan.chiba.jp", 3, false},
	{1, "chosei.chiba.jp", 3, false},
	{1, "choshi.chiba.jp", 3, false},
	{1, "chuo.chiba.jp", 3, false},
	{1, "funabashi.chiba.jp", 3, false},
	{1, "futtsu.chiba.jp", 3, false},
	{1, "hanamigawa.chiba.jp", 3, false},
	{1, "ichihara.chiba.jp", 3, false},
	{1, "ichikawa.chiba.jp", 3, false},
	{1, "ichinomiya.chiba.jp", 3, false},
	{1, "inzai.chiba.jp", 3, false},
	{1, "isumi.chiba.jp", 3, false},
	{1, "kamagaya.chiba.jp", 3, false},
	{1, "kamogawa.chiba.jp", 3, false},
	{1, "kashiwa.chiba.jp", 3, false},
	{1, "katori.chiba.jp", 3, false},
	{1, "katsuura.chiba.jp", 3, false},
	{1, "kimitsu.chiba.jp", 3, false},
	{1, "kisarazu.chiba.jp", 3, false},
	{1, "kozaki.chiba.jp", 3, false},
	{1, "kujukuri.chiba.jp", 3, false},
	{1, "kyonan.chiba.jp", 3, false},
	{1, "matsudo.chiba.jp", 3, false},
	{1, "midori.chiba.jp", 3, false},
	{1, "mihama.chiba.jp", 3, false},
	{1, "minamiboso.chiba.jp", 3, false},
	{1, "mobara.chiba.jp", 3, false},
	{1, "mutsuzawa.chiba.jp", 3, false},
	{1, "nagara.chiba.jp", 3, false},
	{1, "nagareyama.chiba.jp", 3, false},
	{1, "narashino.chiba.jp", 3, false},
	{1, "narita.chiba.jp", 3, false},
	{1, "noda.chiba.jp", 3, false},
	{1, "oamishirasato.chiba.jp", 3, false},
	{1, "omigawa.chiba.jp", 3, false},
	{1, "onjuku.chiba.jp", 3, false},
	{1, "otaki.chiba.jp", 3, false},
	{1, "sakae.chiba.jp", 3, false},
	{1, "sakura.chiba.jp", 3, false},
	{1, "shimofusa.chiba.jp", 3, false},
	{1, "shirako.chiba.jp", 3, false},
	{1, "shiroi.chiba.jp", 3, false},
	{1, "shisui.chiba.jp", 3, false},
	{1, "sodegaura.chiba.jp", 3, false},
	{1, "sosa.chiba.jp", 3, false},
	{1, "tako.chiba.jp", 3, false},
	{1, "tateyama.chiba.jp", 3, false},
	{1, "togane.chiba.jp", 3, false},
	{1, "tohnosho.chiba.jp", 3, false},
	{1, "tomisato.chiba.jp", 3, false},
	{1, "urayasu.chiba.jp", 3, false},
	{1, "yachimata.chiba.jp", 3, false},
	{1, "yachiyo.chiba.jp", 3, false},
	{1, "yokaichiba.chiba.jp", 3, false},
	{1, "yokoshibahikari.chiba.jp", 3, false},
	{1, "yotsukaido.chiba.jp", 3, false},
	{1, "ainan.ehime.jp", 3, false},
	{1, "honai.ehime.jp", 3, false},
	{1, "ikata.ehime.jp", 3, false},
	{1, "imabari.ehime.jp", 3, false},
	{1, "iyo.ehime.jp", 3, false},
	{1, "kamijima.ehime.jp", 3, false},
	{1, "kihoku.ehime.jp", 3, false},
	{1, "kumakogen.ehime.jp", 3, false},
	{1, "masaki.ehime.jp", 3, false},
	{1, "matsuno.ehime.jp", 3, false},
	{1, "matsuyama.ehime.jp", 3, false},
	{1, "namikata.ehime.jp", 3, false},
	{1, "niihama.ehime.jp", 3, false},
	{1, "ozu.ehime.jp", 3, false},
	{1, "saijo.ehime.jp", 3, false},
	{1, "seiyo.ehime.jp", 3, false},
	{1, "shikokuchuo.ehime.jp", 3, false},
	{1, "tobe.ehime.jp", 3, false},
	{1, "toon.ehime.jp", 3, false},
	{1, "uchiko.ehime.jp", 3, false},
	{1, "uwajima.ehime.jp", 3, false},
	{1, "yawatahama.ehime.jp", 3, false},
	{1, "echizen.fukui.jp", 3, false},
	{1, "eiheiji.fukui.jp", 3, false},
	{1, "fukui.fukui.jp", 3, false},
	{1, "ikeda.fukui.jp", 3, false},
	{1, "katsuyama.fukui.jp", 3, false},
	{1, "mihama.fukui.jp", 3, false},
	{1, "minamiechizen.fukui.jp", 3, false},
	{1, "obama.fukui.jp", 3, false},
	{1, "ohi.fukui.jp", 3, false},
	{1, "ono.fukui.jp", 3, false},
	{1, "sabae.fukui.jp", 3, false},
	{1, "sakai.fukui.jp", 3, false},
	{1, "takahama.fukui.jp", 3, false},
	{1, "tsuruga.fukui.jp", 3, false},
	{1, "wakasa.fukui.jp", 3, false},
	{1, "ashiya.fukuoka.jp", 3, false},
	{1, "buzen.fukuoka.jp", 3, false},
	{1, "chikugo.fukuoka.jp", 3, false},
	{1, "chikuho.fukuoka.jp", 3, false},
	{1, "chikujo.fukuoka.jp", 3, false},
	{1, "chikushino.fukuoka.jp", 3, false},
	{1, "chikuzen.fukuoka.jp", 3, false},
	{1, "chuo.fukuoka.jp", 3, false},
	{1, "dazaifu.fukuoka.jp", 3, false},
	{1, "fukuchi.fukuoka.jp", 3, false},
	{1, "hakata.fukuoka.jp", 3, false},
	{1, "higashi.fukuoka.jp", 3, false},
	{1, "hirokawa.fukuoka.jp", 3, false},
	{1, "hisayama.fukuoka.jp", 3, false},
	{1, "iizuka.fukuoka.jp", 3, false},
	{1, "inatsuki.fukuoka.jp", 3, false},
	{1, "kaho.fukuoka.jp", 3, false},
	{1, "kasuga.fukuoka.jp", 3, false},
	{1, "kasuya.fukuoka.jp", 3, false},
	{1, "kawara.fukuoka.jp", 3, false},
	{1, "keisen.fukuoka.jp", 3, false},
	{1, "koga.fukuoka.jp", 3, false},
	{1, "kurate.fukuoka.jp", 3, false},
	{1, "kurogi.fukuoka.jp", 3, false},
	{1, "kurume.fukuoka.jp", 3, false},
	{1, "minami.fukuoka.jp", 3, false},
	{1, "miyako.fukuoka.jp", 3, false},
	{1, "miyama.fukuoka.jp", 3, false},
	{1, "miyawaka.fukuoka.jp", 3, false},
	{1, "mizumaki.fukuoka.jp", 3, false},
	{1, "munakata.fukuoka.jp", 3, false},
	{1, "nakagawa.fukuoka.jp", 3, false},
	{1, "nakama.fukuoka.jp", 3, false},
	{1, "nishi.fukuoka.jp", 3, false},
	{1, "nogata.fukuoka.jp", 3, false},
	{1, "ogori.fukuoka.jp", 3, false},
	{1, "okagaki.fukuoka.jp", 3, false},
	{1, "okawa.fukuoka.jp", 3, false},
	{1, "oki.fukuoka.jp", 3, false},
	{1, "omuta.fukuoka.jp", 3, false},
	{1, "onga.fukuoka.jp", 3, false},
	{1, "onojo.fukuoka.jp", 3, false},
	{1, "oto.fukuoka.jp", 3, false},
	{1, "saigawa.fukuoka.jp", 3, false},
	{1, "sasaguri.fukuoka.jp", 3, false},
	{1, "shingu.fukuoka.jp", 3, false},
	{1, "shinyoshitomi.fukuoka.jp", 3, false},
	{1, "shonai.fukuoka.jp", 3, false},
	{1, "soeda.fukuoka.jp", 3, false},
	{1, "sue.fukuoka.jp", 3, false},
	{1, "tachiarai.fukuoka.jp", 3, false},
	{1, "tagawa.fukuoka.jp", 3, false},
	{1, "takata.fukuoka.jp", 3, false},
	{1, "toho.fukuoka.jp", 3, false},
	{1, "toyotsu.fukuoka.jp", 3, false},
	{1, "tsuiki.fukuoka.jp", 3, false},
	{1, "ukiha.fukuoka.jp", 3, false},
	{1, "umi.fukuoka.jp", 3, false},
	{1, "usui.fukuoka.jp", 3, false},
	{1, "yamada.fukuoka.jp", 3, false},
	{1, "yame.fukuoka.jp", 3, false},
	{1, "yanagawa.fukuoka.jp", 3, false},
	{1, "yukuhashi.fukuoka.jp", 3, false},
	{1, "aizubange.fukushima.jp", 3, false},
	{1, "aizumisato.fukushima.jp", 3, false},
	{1, "aizuwakamatsu.fukushima.jp", 3, false},
	{1, "asakawa.fukushima.jp", 3, false},
	{1, "bandai.fukushima.jp", 3, false},
	{1, "date.fukushima.jp", 3, false},
	{1, "fukushima.fukushima.jp", 3, false},
	{1, "furudono.fukushima.jp", 3, false},
	{1, "futaba.fukushima.jp", 3, false},
	{1, "hanawa.fukushima.jp", 3, false},
	{1, "higashi.fukushima.jp", 3, false},
	{1, "hirata.fukushima.jp", 3, false},
	{1, "hirono.fukushima.jp", 3, false},
	{1, "iitate.fukushima.jp", 3, false},
	{1, "inawashiro.fukushima.jp", 3, false},
	{1, "ishikawa.fukushima.jp", 3, false},
	{1, "iwaki.fukushima.jp", 3, false},
	{1, "izumizaki.fukushima.jp", 3, false},
	{1, "kagamiishi.fukushima.jp", 3, false},
	{1, "kaneyama.fukushima.jp", 3, false},
	{1, "kawamata.fukushima.jp", 3, false},
	{1, "kitakata.fukushima.jp", 3, false},
	{1, "kitashiobara.fukushima.jp", 3, false},
	{1, "koori.fukushima.jp", 3, false},
	{1, "koriyama.fukushima.jp", 3, false},
	{1, "kunimi.fukushima.jp", 3, false},
	{1, "miharu.fukushima.jp", 3, false},
	{1, "mishima.fukushima.jp", 3, false},
	{1, "namie.fukushima.jp", 3, false},
	{1, "nango.fukushima.jp", 3, false},
	{1, "nishiaizu.fukushima.jp", 3, false},
	{1, "nishigo.fukushima.jp", 3, false},
	{1, "okuma.fukushima.jp", 3, false},
	{1, "omotego.fukushima.jp", 3, false},
	{1, "ono.fukushima.jp", 3, false},
	{1, "otama.fukushima.jp", 3, false},
	{1, "samegawa.fukushima.jp", 3, false},
	{1, "shimogo.fukushima.jp", 3, false},
	{1, "shirakawa.fukushima.jp", 3, false},
	{1, "showa.fukushima.jp", 3, false},
	{1, "soma.fukushima.jp", 3, false},
	{1, "sukagawa.fukushima.jp", 3, false},
	{1, "taishin.fukushima.jp", 3, false},
	{1, "tamakawa.fukushima.jp", 3, false},
	{1, "tanagura.fukushima.jp", 3, false},
	{1, "tenei.fukushima.jp", 3, false},
	{1, "yabuki.fukushima.jp", 3, false},
	{1, "yamato.fukushima.jp", 3, false},
	{1, "yamatsuri.fukushima.jp", 3, false},
	{1, "yanaizu.fukushima.jp", 3, false},
	{1, "yugawa.fukushima.jp", 3, false},
	{1, "anpachi.gifu.jp", 3, false},
	{1, "ena.gifu.jp", 3, false},
	{1, "gifu.gifu.jp", 3, false},
	{1, "ginan.gifu.jp", 3, false},
	{1, "godo.gifu.jp", 3, false},
	{1, "gujo.gifu.jp", 3, false},
	{1, "hashima.gifu.jp", 3, false},
	{1, "hichiso.gifu.jp", 3, false},
	{1, "hida.gifu.jp", 3, false},
	{1, "higashishirakawa.gifu.jp", 3, false},
	{1, "ibigawa.gifu.jp", 3, false},
	{1, "ikeda.gifu.jp", 3, false},
	{1, "kakamigahara.gifu.jp", 3, false},
	{1, "kani.gifu.jp", 3, false},
	{1, "kasahara.gifu.jp", 3, false},
	{1, "kasamatsu.gifu.jp", 3, false},
	{1, "kawaue.gifu.jp", 3, false},
	{1, "kitagata.gifu.jp", 3, false},
	{1, "mino.gifu.jp", 3, false},
	{1, "minokamo.gifu.jp", 3, false},
	{1, "mitake.gifu.jp", 3, false},
	{1, "mizunami.gifu.jp", 3, false},
	{1, "motosu.gifu.jp", 3, false},
	{1, "nakatsugawa.gifu.jp", 3, false},
	{1, "ogaki.gifu.jp", 3, false},
	{1, "sakahogi.gifu.jp", 3, false},
	{1, "seki.gifu.jp", 3, false},
	{1, "sekigahara.gifu.jp", 3, false},
	{1, "shirakawa.gifu.jp", 3, false},
	{1, "tajimi.gifu.jp", 3, false},
	{1, "takayama.gifu.jp", 3, false},
	{1, "tarui.gifu.jp", 3, false},
	{1, "toki.gifu.jp", 3, false},
	{1, "tomika.gifu.jp", 3, false},
	{1, "wanouchi.gifu.jp", 3, false},
	{1, "yamagata.gifu.jp", 3, false},
	{1, "yaotsu.gifu.jp", 3, false},
	{1, "yoro.gifu.jp", 3, false},
	{1, "annaka.gunma.jp", 3, false},
	{1, "chiyoda.gunma.jp", 3, false},
	{1, "fujioka.gunma.jp", 3, false},
	{1, "higashiagatsuma.gunma.jp", 3, false},
	{1, "isesaki.gunma.jp", 3, false},
	{1, "itakura.gunma.jp", 3, false},
	{1, "kanna.gunma.jp", 3, false},
	{1, "kanra.gunma.jp", 3, false},
	{1, "katashina.gunma.jp", 3, false},
	{1, "kawaba.gunma.jp", 3, false},
	{1, "kiryu.gunma.jp", 3, false},
	{1, "kusatsu.gunma.jp", 3, false},
	{1, "maebashi.gunma.jp", 3, false},
	{1, "meiwa.gunma.jp", 3, false},
	{1, "midori.gunma.jp", 3, false},
	{1, "minakami.gunma.jp", 3, false},
	{1, "naganohara.gunma.jp", 3, false},
	{1, "nakanojo.gunma.jp", 3, false},
	{1, "nanmoku.gunma.jp", 3, false},
	{1, "numata.gunma.jp", 3, false},
	{1, "oizumi.gunma.jp", 3, false},
	{1, "ora.gunma.jp", 3, false},
	{1, "ota.gunma.jp", 3, false},
	{1, "shibukawa.gunma.jp", 3, false},
	{1, "shimonita.gunma.jp", 3, false},
	{1, "shinto.gunma.jp", 3, false},
	{1, "showa.gunma.jp", 3, false},
	{1, "takasaki.gunma.jp", 3, false},
	{1, "takayama.gunma.jp", 3, false},
	{1, "tamamura.gunma.jp", 3, false},
	{1, "tatebayashi.gunma.jp", 3, false},
	{1, "tomioka.gunma.jp", 3, false},
	{1, "tsukiyono.gunma.jp", 3, false},
	{1, "tsumagoi.gunma.jp", 3, false},
	{1, "ueno.gunma.jp", 3, false},
	{1, "yoshioka.gunma.jp", 3, false},
	{1, "asaminami.hiroshima.jp", 3, false},
	{1, "daiwa.hiroshima.jp", 3, false},
	{1, "etajima.hiroshima.jp", 3, false},
	{1, "fuchu.hiroshima.jp", 3, false},
	{1, "fukuyama.hiroshima.jp", 3, false},
	{1, "hatsukaichi.hiroshima.jp", 3, false},
	{1, "higashihiroshima.hiroshima.jp", 3, false},
	{1, "hongo.hiroshima.jp", 3, false},
	{1, "jinsekikogen.hiroshima.jp", 3, false},
	{1, "kaita.hiroshima.jp", 3, false},
	{1, "kui.hiroshima.jp", 3, false},
	{1, "kumano.hiroshima.jp", 3, false},
	{1, "kure.hiroshima.jp", 3, false},
	{1, "mihara.hiroshima.jp", 3, false},
	{1, "miyoshi.hiroshima.jp", 3, false},
	{1, "naka.hiroshima.jp", 3, false},
	{1, "onomichi.hiroshima.jp", 3, false},
	{1, "osakikamijima.hiroshima.jp", 3, false},
	{1, "otake.hiroshima.jp", 3, false},
	{1, "saka.hiroshima.jp", 3, false},
	{1, "sera.hiroshima.jp", 3, false},
	{1, "seranishi.hiroshima.jp", 3, false},
	{1, "shinichi.hiroshima.jp", 3, false},
	{1, "shobara.hiroshima.jp", 3, false},
	{1, "takehara.hiroshima.jp", 3, false},
	{1, "abashiri.hokkaido.jp", 3, false},
	{1, "abira.hokkaido.jp", 3, false},
	{1, "aibetsu.hokkaido.jp", 3, false},
	{1, "akabira.hokkaido.jp", 3, false},
	{1, "akkeshi.hokkaido.jp", 3, false},
	{1, "asahikawa.hokkaido.jp", 3, false},
	{1, "ashibetsu.hokkaido.jp", 3, false},
	{1, "ashoro.hokkaido.jp", 3, false},
	{1, "assabu.hokkaido.jp", 3, false},
	{1, "atsuma.hokkaido.jp", 3, false},
	{1, "bibai.hokkaido.jp", 3, false},
	{1, "biei.hokkaido.jp", 3, false},
	{1, "bifuka.hokkaido.jp", 3, false},
	{1, "bihoro.hokkaido.jp", 3, false},
	{1, "biratori.hokkaido.jp", 3, false},
	{1, "chippubetsu.hokkaido.jp", 3, false},
	{1, "chitose.hokkaido.jp", 3, false},
	{1, "date.hokkaido.jp", 3, false},
	{1, "ebetsu.hokkaido.jp", 3, false},
	{1, "embetsu.hokkaido.jp", 3, false},
	{1, "eniwa.hokkaido.jp", 3, false},
	{1, "erimo.hokkaido.jp", 3, false},
	{1, "esan.hokkaido.jp", 3, false},
	{1, "esashi.hokkaido.jp", 3, false},
	{1, "fukagawa.hokkaido.jp", 3, false},
	{1, "fukushima.hokkaido.jp", 3, false},
	{1, "furano.hokkaido.jp", 3, false},
	{1, "furubira.hokkaido.jp", 3, false},
	{1, "haboro.hokkaido.jp", 3, false},
	{1, "hakodate.hokkaido.jp", 3, false},
	{1, "hamatonbetsu.hokkaido.jp", 3, false},
	{1, "hidaka.hokkaido.jp", 3, false},
	{1, "higashikagura.hokkaido.jp", 3, false},
	{1, "higashikawa.hokkaido.jp", 3, false},
	{1, "hiroo.hokkaido.jp", 3, false},
	{1, "hokuryu.hokkaido.jp", 3, false},
	{1, "hokuto.hokkaido.jp", 3, false},
	{1, "honbetsu.hokkaido.jp", 3, false},
	{1, "horokanai.hokkaido.jp", 3, false},
	{1, "horonobe.hokkaido.jp", 3, false},
	{1, "ikeda.hokkaido.jp", 3, false},
	{1, "imakane.hokkaido.jp", 3, false},
	{1, "ishikari.hokkaido.jp", 3, false},
	{1, "iwamizawa.hokkaido.jp", 3, false},
	{1, "iwanai.hokkaido.jp", 3, false},
	{1, "kamifurano.hokkaido.jp", 3, false},
	{1, "kamikawa.hokkaido.jp", 3, false},
	{1, "kamishihoro.hokkaido.jp", 3, false},
	{1, "kamisunagawa.hokkaido.jp", 3, false},
	{1, "kamoenai.hokkaido.jp", 3, false},
	{1, "kayabe.hokkaido.jp", 3, false},
	{1, "kembuchi.hokkaido.jp", 3, false},
	{1, "kikonai.hokkaido.jp", 3, false},
	{1, "kimobetsu.hokkaido.jp", 3, false},
	{1, "kitahiroshima.hokkaido.jp", 3, false},
	{1, "kitami.hokkaido.jp", 3, false},
	{1, "kiyosato.hokkaido.jp", 3, false},
	{1, "koshimizu.hokkaido.jp", 3, false},
	{1, "kunneppu.hokkaido.jp", 3, false},
	{1, "kuriyama.hokkaido.jp", 3, false},
	{1, "kuromatsunai.hokkaido.jp", 3, false},
	{1, "kushiro.hokkaido.jp", 3, false},
	{1, "kutchan.hokkaido.jp", 3, false},
	{1, "kyowa.hokkaido.jp", 3, false},
	{1, "mashike.hokkaido.jp", 3, false},
	{1, "matsumae.hokkaido.jp", 3, false},
	{1, "mikasa.hokkaido.jp", 3, false},
	{1, "minamifurano.hokkaido.jp", 3, false},
	{1, "mombetsu.hokkaido.jp", 3, false},
	{1, "moseushi.hokkaido.jp", 3, false},
	{1, "mukawa.hokkaido.jp", 3, false},
	{1, "muroran.hokkaido.jp", 3, false},
	{1, "naie.hokkaido.jp", 3, false},
	{1, "nakagawa.hokkaido.jp", 3, false},
	{1, "nakasatsunai.hokkaido.jp", 3, false},
	{1, "nakatombetsu.hokkaido.jp", 3, false},
	{1, "nanae.hokkaido.jp", 3, false},
	{1, "nanporo.hokkaido.jp", 3, false},
	{1, "nayoro.hokkaido.jp", 3, false},
	{1, "nemuro.hokkaido.jp", 3, false},
	{1, "niikappu.hokkaido.jp", 3, false},
	{1, "niki.hokkaido.jp", 3, false},
	{1, "nishiokoppe.hokkaido.jp", 3, false},
	{1, "noboribetsu.hokkaido.jp", 3, false},
	{1, "numata.hokkaido.jp", 3, false},
	{1, "obihiro.hokkaido.jp", 3, false},
	{1, "obira.hokkaido.jp", 3, false},
	{1, "oketo.hokkaido.jp", 3, false},
	{1, "okoppe.hokkaido.jp", 3, false},
	{1, "otaru.hokkaido.jp", 3, false},
	{1, "otobe.hokkaido.jp", 3, false},
	{1, "otofuke.hokkaido.jp", 3, false},
	{1, "otoineppu.hokkaido.jp", 3, false},
	{1, "oumu.hokkaido.jp", 3, false},
	{1, "ozora.hokkaido.jp", 3, false},
	{1, "pippu.hokkaido.jp", 3, false},
	{1, "rankoshi.hokkaido.jp", 3, false},
	{1, "rebun.hokkaido.jp", 3, false},
	{1, "rikubetsu.hokkaido.jp", 3, false},
	{1, "rishiri.hokkaido.jp", 3, false},
	{1, "rishirifuji.hokkaido.jp", 3, false},
	{1, "saroma.hokkaido.jp", 3, false},
	{1, "sarufutsu.hokkaido.jp", 3, false},
	{1, "shakotan.hokkaido.jp", 3, false},
	{1, "shari.hokkaido.jp", 3, false},
	{1, "shibecha.hokkaido.jp", 3, false},
	{1, "shibetsu.hokkaido.jp", 3, false},
	{1, "shikabe.hokkaido.jp", 3, false},
	{1, "shikaoi.hokkaido.jp", 3, false},
	{1, "shimamaki.hokkaido.jp", 3, false},
	{1, "shimizu.hokkaido.jp", 3, false},
	{1, "shimokawa.hokkaido.jp", 3, false},
	{1, "shinshinotsu.hokkaido.jp", 3, false},
	{1, "shintoku.hokkaido.jp", 3, false},
	{1, "shiranuka.hokkaido.jp", 3, false},
	{1, "shiraoi.hokkaido.jp", 3, false},
	{1, "shiriuchi.hokkaido.jp", 3, false},
	{1, "sobetsu.hokkaido.jp", 3, false},
	{1, "sunagawa.hokkaido.jp", 3, false},
	{1, "taiki.hokkaido.jp", 3, false},
	{1, "takasu.hokkaido.jp", 3, false},
	{1, "takikawa.hokkaido.jp", 3, false},
	{1, "takinoue.hokkaido.jp", 3, false},
	{1, "teshikaga.hokkaido.jp", 3, false},
	{1, "tobetsu.hokkaido.jp", 3, false},
	{1, "tohma.hokkaido.jp", 3, false},
	{1, "tomakomai.hokkaido.jp", 3, false},
	{1, "tomari.hokkaido.jp", 3, false},
	{1, "toya.hokkaido.jp", 3, false},
	{1, "toyako.hokkaido.jp", 3, false},
	{1, "toyotomi.hokkaido.jp", 3, false},
	{1, "toyoura.hokkaido.jp", 3, false},
	{1, "tsubetsu.hokkaido.jp", 3, false},
	{1, "tsukigata.hokkaido.jp", 3, false},
	{1, "urakawa.hokkaido.jp", 3, false},
	{1, "urausu.hokkaido.jp", 3, false},
	{1, "uryu.hokkaido.jp", 3, false},
	{1, "utashinai.hokkaido.jp", 3, false},
	{1, "wakkanai.hokkaido.jp", 3, false},
	{1, "wassamu.hokkaido.jp", 3, false},
	{1, "yakumo.hokkaido.jp", 3, false},
	{1, "yoichi.hokkaido.jp", 3, false},
	{1, "aioi.hyogo.jp", 3, false},
	{1, "akashi.hyogo.jp", 3, false},
	{1, "ako.hyogo.jp", 3, false},
	{1, "amagasaki.hyogo.jp", 3, false},
	{1, "aogaki.hyogo.jp", 3, false},
	{1, "asago.hyogo.jp", 3, false},
	{1, "ashiya.hyogo.jp", 3, false},
	{1, "awaji.hyogo.jp", 3, false},
	{1, "fukusaki.hyogo.jp", 3, false},
	{1, "goshiki.hyogo.jp", 3, false},
	{1, "harima.hyogo.jp", 3, false},
	{1, "himeji.hyogo.jp", 3, false},
	{1, "ichikawa.hyogo.jp", 3, false},
	{1, "inagawa.hyogo.jp", 3, false},
	{1, "itami.hyogo.jp", 3, false},
	{1, "kakogawa.hyogo.jp", 3, false},
	{1, "kamigori.hyogo.jp", 3, false},
	{1, "kamikawa.hyogo.jp", 3, false},
	{1, "kasai.hyogo.jp", 3, false},
	{1, "kasuga.hyogo.jp", 3, false},
	{1, "kawanishi.hyogo.jp", 3, false},
	{1, "miki.hyogo.jp", 3, false},
	{1, "minamiawaji.hyogo.jp", 3, false},
	{1, "nishinomiya.hyogo.jp", 3, false},
	{1, "nishiwaki.hyogo.jp", 3, false},
	{1, "ono.hyogo.jp", 3, false},
	{1, "sanda.hyogo.jp", 3, false},
	{1, "sannan.hyogo.jp", 3, false},
	{1, "sasayama.hyogo.jp", 3, false},
	{1, "sayo.hyogo.jp", 3, false},
	{1, "shingu.hyogo.jp", 3, false},
	{1, "shinonsen.hyogo.jp", 3, false},
	{1, "shiso.hyogo.jp", 3, false},
	{1, "sumoto.hyogo.jp", 3, false},
	{1, "taishi.hyogo.jp", 3, false},
	{1, "taka.hyogo.jp", 3, false},
	{1, "takarazuka.hyogo.jp", 3, false},
	{1, "takasago.hyogo.jp", 3, false},
	{1, "takino.hyogo.jp", 3, false},
	{1, "tamba.hyogo.jp", 3, false},
	{1, "tatsuno.hyogo.jp", 3, false},
	{1, "toyooka.hyogo.jp", 3, false},
	{1, "yabu.hyogo.jp", 3, false},
	{1, "yashiro.hyogo.jp", 3, false},
	{1, "yoka.hyogo.jp", 3, false},
	{1, "yokawa.hyogo.jp", 3, false},
	{1, "ami.ibaraki.jp", 3, false},
	{1, "asahi.ibaraki.jp", 3, false},
	{1, "bando.ibaraki.jp", 3, false},
	{1, "chikusei.ibaraki.jp", 3, false},
	{1, "daigo.ibaraki.jp", 3, false},
	{1, "fujishiro.ibaraki.jp", 3, false},
	{1, "hitachi.ibaraki.jp", 3, false},
	{1, "hitachinaka.ibaraki.jp", 3, false},
	{1, "hitachiomiya.ibaraki.jp", 3, false},
	{1, "hitachiota.ibaraki.jp", 3, false},
	{1, "ibaraki.ibaraki.jp", 3, false},
	{1, "ina.ibaraki.jp", 3, false},
	{1, "inashiki.ibaraki.jp", 3, false},
	{1, "itako.ibaraki.jp", 3, false},
	{1, "iwama.ibaraki.jp", 3, false},
	{1, "joso.ibaraki.jp", 3, false},
	{1, "kamisu.ibaraki.jp", 3, false},
	{1, "kasama.ibaraki.jp", 3, false},
	{1, "kashima.ibaraki.jp", 3, false},
	{1, "kasumigaura.ibaraki.jp", 3, false},
	{1, "koga.ibaraki.jp", 3, false},
	{1, "miho.ibaraki.jp", 3, false},
	{1, "mito.ibaraki.jp", 3, false},
	{1, "moriya.ibaraki.jp", 3, false},
	{1, "naka.ibaraki.jp", 3, false},
	{1, "namegata.ibaraki.jp", 3, false},
	{1, "oarai.ibaraki.jp", 3, false},
	{1, "ogawa.ibaraki.jp", 3, false},
	{1, "omitama.ibaraki.jp", 3, false},
	{1, "ryugasaki.ibaraki.jp", 3, false},
	{1, "sakai.ibaraki.jp", 3, false},
	{1, "sakuragawa.ibaraki.jp", 3, false},
	{1, "shimodate.ibaraki.jp", 3, false},
	{1, "shimotsuma.ibaraki.jp", 3, false},
	{1, "shirosato.ibaraki.jp", 3, false},
	{1, "sowa.ibaraki.jp", 3, false},
	{1, "suifu.ibaraki.jp", 3, false},
	{1, "takahagi.ibaraki.jp", 3, false},
	{1, "tamatsukuri.ibaraki.jp", 3, false},
	{1, "tokai.ibaraki.jp", 3, false},
	{1, "tomobe.ibaraki.jp", 3, false},
	{1, "tone.ibaraki.jp", 3, false},
	{1, "toride.ibaraki.jp", 3, false},
	{1, "tsuchiura.ibaraki.jp", 3, false},
	{1, "tsukuba.ibaraki.jp", 3, false},
	{1, "uchihara.ibaraki.jp", 3, false},
	{1, "ushiku.ibaraki.jp", 3, false},
	{1, "yachiyo.ibaraki.jp", 3, false},
	{1, "yamagata.ibaraki.jp", 3, false},
	{1, "yawara.ibaraki.jp", 3, false},
	{1, "yuki.ibaraki.jp", 3, false},
	{1, "anamizu.ishikawa.jp", 3, false},
	{1, "hakui.ishikawa.jp", 3, false},
	{1, "hakusan.ishikawa.jp", 3, false},
	{1, "kaga.ishikawa.jp", 3, false},
	{1, "kahoku.ishikawa.jp", 3, false},
	{1, "kanazawa.ishikawa.jp", 3, false},
	{1, "kawakita.ishikawa.jp", 3, false},
	{1, "komatsu.ishikawa.jp", 3, false},
	{1, "nakanoto.ishikawa.jp", 3, false},
	{1, "nanao.ishikawa.jp", 3, false},
	{1, "nomi.ishikawa.jp", 3, false},
	{1, "nonoichi.ishikawa.jp", 3, false},
	{1, "noto.ishikawa.jp", 3, false},
	{1, "shika.ishikawa.jp", 3, false},
	{1, "suzu.ishikawa.jp", 3, false},
	{1, "tsubata.ishikawa.jp", 3, false},
	{1, "tsurugi.ishikawa.jp", 3, false},
	{1, "uchinada.ishikawa.jp", 3, false},
	{1, "wajima.ishikawa.jp", 3, false},
	{1, "fudai.iwate.jp", 3, false},
	{1, "fujisawa.iwate.jp", 3, false},
	{1, "hanamaki.iwate.jp", 3, false},
	{1, "hiraizumi.iwate.jp", 3, false},
	{1, "hirono.iwate.jp", 3, false},
	{1, "ichinohe.iwate.jp", 3, false},
	{1, "ichinoseki.iwate.jp", 3, false},
	{1, "iwaizumi.iwate.jp", 3, false},
	{1, "iwate.iwate.jp", 3, false},
	{1, "joboji.iwate.jp", 3, false},
	{1, "kamaishi.iwate.jp", 3, false},
	{1, "kanegasaki.iwate.jp", 3, false},
	{1, "karumai.iwate.jp", 3, false},
	{1, "kawai.iwate.jp", 3, false},
	{1, "kitakami.iwate.jp", 3, false},
	{1, "kuji.iwate.jp", 3, false},
	{1, "kunohe.iwate.jp", 3, false},
	{1, "kuzumaki.iwate.jp", 3, false},
	{1, "miyako.iwate.jp", 3, false},
	{1, "mizusawa.iwate.jp", 3, false},
	{1, "morioka.iwate.jp", 3, false},
	{1, "ninohe.iwate.jp", 3, false},
	{1, "noda.iwate.jp", 3, false},
	{1, "ofunato.iwate.jp", 3, false},
	{1, "oshu.iwate.jp", 3, false},
	{1, "otsuchi.iwate.jp", 3, false},
	{1, "rikuzentakata.iwate.jp", 3, false},
	{1, "shiwa.iwate.jp", 3, false},
	{1, "shizukuishi.iwate.jp", 3, false},
	{1, "sumita.iwate.jp", 3, false},
	{1, "tanohata.iwate.jp", 3, false},
	{1, "tono.iwate.jp", 3, false},
	{1, "yahaba.iwate.jp", 3, false},
	{1, "yamada.iwate.jp", 3, false},
	{1, "ayagawa.kagawa.jp", 3, false},
	{1, "higashikagawa.kagawa.jp", 3, false},
	{1, "kanonji.kagawa.jp", 3, false},
	{1, "kotohira.kagawa.jp", 3, false},
	{1, "manno.kagawa.jp", 3, false},
	{1, "marugame.kagawa.jp", 3, false},
	{1, "mitoyo.kagawa.jp", 3, false},
	{1, "naoshima.kagawa.jp", 3, false},
	{1, "sanuki.kagawa.jp", 3, false},
	{1, "tadotsu.kagawa.jp", 3, false},
	{1, "takamatsu.kagawa.jp", 3, false},
	{1, "tonosho.kagawa.jp", 3, false},
	{1, "uchinomi.kagawa.jp", 3, false},
	{1, "utazu.kagawa.jp", 3, false},
	{1, "zentsuji.kagawa.jp", 3, false},
	{1, "akune.kagoshima.jp", 3, false},
	{1, "amami.kagoshima.jp", 3, false},
	{1, "hioki.kagoshima.jp", 3, false},
	{1, "isa.kagoshima.jp", 3, false},
	{1, "isen.kagoshima.jp", 3, false},
	{1, "izumi.kagoshima.jp", 3, false},
	{1, "kagoshima.kagoshima.jp", 3, false},
	{1, "kanoya.kagoshima.jp", 3, false},
	{1, "kawanabe.kagoshima.jp", 3, false},
	{1, "kinko.kagoshima.jp", 3, false},
	{1, "kouyama.kagoshima.jp", 3, false},
	{1, "makurazaki.kagoshima.jp", 3, false},
	{1, "matsumoto.kagoshima.jp", 3, false},
	{1, "minamitane.kagoshima.jp", 3, false},
	{1, "nakatane.kagoshima.jp", 3, false},
	{1, "nishinoomote.kagoshima.jp", 3, false},
	{1, "satsumasendai.kagoshima.jp", 3, false},
	{1, "soo.kagoshima.jp", 3, false},
	{1, "tarumizu.kagoshima.jp", 3, false},
	{1, "yusui.kagoshima.jp", 3, false},
	{1, "aikawa.kanagawa.jp", 3, false},
	{1, "atsugi.kanagawa.jp", 3, false},
	{1, "ayase.kanagawa.jp", 3, false},
	{1, "chigasaki.kanagawa.jp", 3, false},
	{1, "ebina.kanagawa.jp", 3, false},
	{1, "fujisawa.kanagawa.jp", 3, false},
	{1, "hadano.kanagawa.jp", 3, false},
	{1, "hakone.kanagawa.jp", 3, false},
	{1, "hiratsuka.kanagawa.jp", 3, false},
	{1, "isehara.kanagawa.jp", 3, false},
	{1, "kaisei.kanagawa.jp", 3, false},
	{1, "kamakura.kanagawa.jp", 3, false},
	{1, "kiyokawa.kanagawa.jp", 3, false},
	{1, "matsuda.kanagawa.jp", 3, false},
	{1, "minamiashigara.kanagawa.jp", 3, false},
	{1, "miura.kanagawa.jp", 3, false},
	{1, "nakai.kanagawa.jp", 3, false},
	{1, "ninomiya.kanagawa.jp", 3, false},
	{1, "odawara.kanagawa.jp", 3, false},
	{1, "oi.kanagawa.jp", 3, false},
	{1, "oiso.kanagawa.jp", 3, false},
	{1, "sagamihara.kanagawa.jp", 3, false},
	{1, "samukawa.kanagawa.jp", 3, false},
	{1, "tsukui.kanagawa.jp", 3, false},
	{1, "yamakita.kanagawa.jp", 3, false},
	{1, "yamato.kanagawa.jp", 3, false},
	{1, "yokosuka.kanagawa.jp", 3, false},
	{1, "yugawara.kanagawa.jp", 3, false},
	{1, "zama.kanagawa.jp", 3, false},
	{1, "zushi.kanagawa.jp", 3, false},
	{1, "aki.kochi.jp", 3, false},
	{1, "geisei.kochi.jp", 3, false},
	{1, "hidaka.kochi.jp", 3, false},
	{1, "higashitsuno.kochi.jp", 3, false},
	{1, "ino.kochi.jp", 3, false},
	{1, "kagami.kochi.jp", 3, false},
	{1, "kami.kochi.jp", 3, false},
	{1, "kitagawa.kochi.jp", 3, false},
	{1, "kochi.kochi.jp", 3, false},
	{1, "mihara.kochi.jp", 3, false},
	{1, "motoyama.kochi.jp", 3, false},
	{1, "muroto.kochi.jp", 3, false},
	{1, "nahari.kochi.jp", 3, false},
	{1, "nakamura.kochi.jp", 3, false},
	{1, "nankoku.kochi.jp", 3, false},
	{1, "nishitosa.kochi.jp", 3, false},
	{1, "niyodogawa.kochi.jp", 3, false},
	{1, "ochi.kochi.jp", 3, false},
	{1, "okawa.kochi.jp", 3, false},
	{1, "otoyo.kochi.jp", 3, false},
	{1, "otsuki.kochi.jp", 3, false},
	{1, "sakawa.kochi.jp", 3, false},
	{1, "sukumo.kochi.jp", 3, false},
	{1, "susaki.kochi.jp", 3, false},
	{1, "tosa.kochi.jp", 3, false},
	{1, "tosashimizu.kochi.jp", 3, false},
	{1, "toyo.kochi.jp", 3, false},
	{1, "tsuno.kochi.jp", 3, false},
	{1, "umaji.kochi.jp", 3, false},
	{1, "yasuda.kochi.jp", 3, false},
	{1, "yusuhara.kochi.jp", 3, false},
	{1, "amakusa.kumamoto.jp", 3, false},
	{1, "arao.kumamoto.jp", 3, false},
	{1, "aso.kumamoto.jp", 3, false},
	{1, "choyo.kumamoto.jp", 3, false},
	{1, "gyokuto.kumamoto.jp", 3, false},
	{1, "kamiamakusa.kumamoto.jp", 3, false},
	{1, "kikuchi.kumamoto.jp", 3, false},
	{1, "kumamoto.kumamoto.jp", 3, false},
	{1, "mashiki.kumamoto.jp", 3, false},
	{1, "mifune.kumamoto.jp", 3, false},
	{1, "minamata.kumamoto.jp", 3, false},
	{1, "minamioguni.kumamoto.jp", 3, false},
	{1, "nagasu.kumamoto.jp", 3, false},
	{1, "nishihara.kumamoto.jp", 3, false},
	{1, "oguni.kumamoto.jp", 3, false},
	{1, "ozu.kumamoto.jp", 3, false},
	{1, "sumoto.kumamoto.jp", 3, false},
	{1, "takamori.kumamoto.jp", 3, false},
	{1, "uki.kumamoto.jp", 3, false},
	{1, "uto.kumamoto.jp", 3, false},
	{1, "yamaga.kumamoto.jp", 3, false},
	{1, "yamato.kumamoto.jp", 3, false},
	{1, "yatsushiro.kumamoto.jp", 3, false},
	{1, "ayabe.kyoto.jp", 3, false},
	{1, "fukuchiyama.kyoto.jp", 3, false},
	{1, "higashiyama.kyoto.jp", 3, false},
	{1, "ide.kyoto.jp", 3, false},
	{1, "ine.kyoto.jp", 3, false},
	{1, "joyo.kyoto.jp", 3, false},
	{1, "kameoka.kyoto.jp", 3, false},
	{1, "kamo.kyoto.jp", 3, false},
	{1, "kita.kyoto.jp", 3, false},
	{1, "kizu.kyoto.jp", 3, false},
	{1, "kumiyama.kyoto.jp", 3, false},
	{1, "kyotamba.kyoto.jp", 3, false},
	{1, "kyotanabe.kyoto.jp", 3, false},
	{1, "kyotango.kyoto.jp", 3, false},
	{1, "maizuru.kyoto.jp", 3, false},
	{1, "minami.kyoto.jp", 3, false},
	{1, "minamiyamashiro.kyoto.jp", 3, false},
	{1, "miyazu.kyoto.jp", 3, false},
	{1, "muko.kyoto.jp", 3, false},
	{1, "nagaokakyo.kyoto.jp", 3, false},
	{1, "nakagyo.kyoto.jp", 3, false},
	{1, "nantan.kyoto.jp", 3, false},
	{1, "oyamazaki.kyoto.jp", 3, false},
	{1, "sakyo.kyoto.jp", 3, false},
	{1, "seika.kyoto.jp", 3, false},
	{1, "tanabe.kyoto.jp", 3, false},
	{1, "uji.kyoto.jp", 3, false},
	{1, "ujitawara.kyoto.jp", 3, false},
	{1, "wazuka.kyoto.jp", 3, false},
	{1, "yamashina.kyoto.jp", 3, false},
	{1, "yawata.kyoto.jp", 3, false},
	{1, "asahi.mie.jp", 3, false},
	{1, "inabe.mie.jp", 3, false},
	{1, "ise.mie.jp", 3, false},
	{1, "kameyama.mie.jp", 3, false},
	{1, "kawagoe.mie.jp", 3, false},
	{1, "kiho.mie.jp", 3, false},
	{1, "kisosaki.mie.jp", 3, false},
	{1, "kiwa.mie.jp", 3, false},
	{1, "komono.mie.jp", 3, false},
	{1, "kumano.mie.jp", 3, false},
	{1, "kuwana.mie.jp", 3, false},
	{1, "matsusaka.mie.jp", 3, false},
	{1, "meiwa.mie.jp", 3, false},
	{1, "mihama.mie.jp", 3, false},
	{1, "minamiise.mie.jp", 3, false},
	{1, "misugi.mie.jp", 3, false},
	{1, "miyama.mie.jp", 3, false},
	{1, "nabari.mie.jp", 3, false},
	{1, "shima.mie.jp", 3, false},
	{1, "suzuka.mie.jp", 3, false},
	{1, "tado.mie.jp", 3, false},
	{1, "taiki.mie.jp", 3, false},
	{1, "taki.mie.jp", 3, false},
	{1, "tamaki.mie.jp", 3, false},
	{1, "toba.mie.jp", 3, false},
	{1, "tsu.mie.jp", 3, false},
	{1, "udono.mie.jp", 3, false},
	{1, "ureshino.mie.jp", 3, false},
	{1, "watarai.mie.jp", 3, false},
	{1, "yokkaichi.mie.jp", 3, false},
	{1, "furukawa.miyagi.jp", 3, false},
	{1, "higashimatsushima.miyagi.jp", 3, false},
	{1, "ishinomaki.miyagi.jp", 3, false},
	{1, "iwanuma.miyagi.jp", 3, false},
	{1, "kakuda.miyagi.jp", 3, false},
	{1, "kami.miyagi.jp", 3, false},
	{1, "kawasaki.miyagi.jp", 3, false},
	{1, "marumori.miyagi.jp", 3, false},
	{1, "matsushima.miyagi.jp", 3, false},
	{1, "minamisanriku.miyagi.jp", 3, false},
	{1, "misato.miyagi.jp", 3, false},
	{1, "murata.miyagi.jp", 3, false},
	{1, "natori.miyagi.jp", 3, false},
	{1, "ogawara.miyagi.jp", 3, false},
	{1, "ohira.miyagi.jp", 3, false},
	{1, "onagawa.miyagi.jp", 3, false},
	{1, "osaki.miyagi.jp", 3, false},
	{1, "rifu.miyagi.jp", 3, false},
	{1, "semine.miyagi.jp", 3, false},
	{1, "shibata.miyagi.jp", 3, false},
	{1, "shichikashuku.miyagi.jp", 3, false},
	{1, "shikama.miyagi.jp", 3, false},
	{1, "shiogama.miyagi.jp", 3, false},
	{1, "shiroishi.miyagi.jp", 3, false},
	{1, "tagajo.miyagi.jp", 3, false},
	{1, "taiwa.miyagi.jp", 3, false},
	{1, "tome.miyagi.jp", 3, false},
	{1, "tomiya.miyagi.jp", 3, false},
	{1, "wakuya.miyagi.jp", 3, false},
	{1, "watari.miyagi.jp", 3, false},
	{1, "yamamoto.miyagi.jp", 3, false},
	{1, "zao.miyagi.jp", 3, false},
	{1, "aya.miyazaki.jp", 3, false},
	{1, "ebino.miyazaki.jp", 3, false},
	{1, "gokase.miyazaki.jp", 3, false},
	{1, "hyuga.miyazaki.jp", 3, false},
	{1, "kadogawa.miyazaki.jp", 3, false},
	{1, "kawaminami.miyazaki.jp", 3, false},
	{1, "kijo.miyazaki.jp", 3, false},
	{1, "kitagawa.miyazaki.jp", 3, false},
	{1, "kitakata.miyazaki.jp", 3, false},
	{1, "kitaura.miyazaki.jp", 3, false},
	{1, "kobayashi.miyazaki.jp", 3, false},
	{1, "kunitomi.miyazaki.jp", 3, false},
	{1, "kushima.miyazaki.jp", 3, false},
	{1, "mimata.miyazaki.jp", 3, false},
	{1, "miyakonojo.miyazaki.jp", 3, false},
	{1, "miyazaki.miyazaki.jp", 3, false},
	{1, "morotsuka.miyazaki.jp", 3, false},
	{1, "nichinan.miyazaki.jp", 3, false},
	{1, "nishimera.miyazaki.jp", 3, false},
	{1, "nobeoka.miyazaki.jp", 3, false},
	{1, "saito.miyazaki.jp", 3, false},
	{1, "shiiba.miyazaki.jp", 3, false},
	{1, "shintomi.miyazaki.jp", 3, false},
	{1, "takaharu.miyazaki.jp", 3, false},
	{1, "takanabe.miyazaki.jp", 3, false},
	{1, "takazaki.miyazaki.jp", 3, false},
	{1, "tsuno.miyazaki.jp", 3, false},
	{1, "achi.nagano.jp", 3, false},
	{1, "agematsu.nagano.jp", 3, false},
	{1, "anan.nagano.jp", 3, false},
	{1, "aoki.nagano.jp", 3, false},
	{1, "asahi.nagano.jp", 3, false},
	{1, "azumino.nagano.jp", 3, false},
	{1, "chikuhoku.nagano.jp", 3, false},
	{1, "chikuma.nagano.jp", 3, false},
	{1, "chino.nagano.jp", 3, false},
	{1, "fujimi.nagano.jp", 3, false},
	{1, "hakuba.nagano.jp", 3, false},
	{1, "hara.nagano.jp", 3, false},
	{1, "hiraya.nagano.jp", 3, false},
	{1, "iida.nagano.jp", 3, false},
	{1, "iijima.nagano.jp", 3, false},
	{1, "iiyama.nagano.jp", 3, false},
	{1, "iizuna.nagano.jp", 3, false},
	{1, "ikeda.nagano.jp", 3, false},
	{1, "ikusaka.nagano.jp", 3, false},
	{1, "ina.nagano.jp", 3, false},
	{1, "karuizawa.nagano.jp", 3, false},
	{1, "kawakami.nagano.jp", 3, false},
	{1, "kiso.nagano.jp", 3, false},
	{1, "kisofukushima.nagano.jp", 3, false},
	{1, "kitaaiki.nagano.jp", 3, false},
	{1, "komagane.nagano.jp", 3, false},
	{1, "komoro.nagano.jp", 3, false},
	{1, "matsukawa.nagano.jp", 3, false},
	{1, "matsumoto.nagano.jp", 3, false},
	{1, "miasa.nagano.jp", 3, false},
	{1, "minamiaiki.nagano.jp", 3, false},
	{1, "minamimaki.nagano.jp", 3, false},
	{1, "minamiminowa.nagano.jp", 3, false},
	{1, "minowa.nagano.jp", 3, false},
	{1, "miyada.nagano.jp", 3, false},
	{1, "miyota.nagano.jp", 3, false},
	{1, "mochizuki.nagano.jp", 3, false},
	{1, "nagano.nagano.jp", 3, false},
	{1, "nagawa.nagano.jp", 3, false},
	{1, "nagiso.nagano.jp", 3, false},
	{1, "nakagawa.nagano.jp", 3, false},
	{1, "nakano.nagano.jp", 3, false},
	{1, "nozawaonsen.nagano.jp", 3, false},
	{1, "obuse.nagano.jp", 3, false},
	{1, "ogawa.nagano.jp", 3, false},
	{1, "okaya.nagano.jp", 3, false},
	{1, "omachi.nagano.jp", 3, false},
	{1, "omi.nagano.jp", 3, false},
	{1, "ookuwa.nagano.jp", 3, false},
	{1, "ooshika.nagano.jp", 3, false},
	{1, "otaki.nagano.jp", 3, false},
	{1, "otari.nagano.jp", 3, false},
	{1, "sakae.nagano.jp", 3, false},
	{1, "sakaki.nagano.jp", 3, false},
	{1, "saku.nagano.jp", 3, false},
	{1, "sakuho.nagano.jp", 3, false},
	{1, "shimosuwa.nagano.jp", 3, false},
	{1, "shinanomachi.nagano.jp", 3, false},
	{1, "shiojiri.nagano.jp", 3, false},
	{1, "suwa.nagano.jp", 3, false},
	{1, "suzaka.nagano.jp", 3, false},
	{1, "takagi.nagano.jp", 3, false},
	{1, "takamori.nagano.jp", 3, false},
	{1, "takayama.nagano.jp", 3, false},
	{1, "tateshina.nagano.jp", 3, false},
	{1, "tatsuno.nagano.jp", 3, false},
	{1, "togakushi.nagano.jp", 3, false},
	{1, "togura.nagano.jp", 3, false},
	{1, "tomi.nagano.jp", 3, false},
	{1, "ueda.nagano.jp", 3, false},
	{1, "wada.nagano.jp", 3, false},
	{1, "yamagata.nagano.jp", 3, false},
	{1, "yamanouchi.nagano.jp", 3, false},
	{1, "yasaka.nagano.jp", 3, false},
	{1, "yasuoka.nagano.jp", 3, false},
	{1, "chijiwa.nagasaki.jp", 3, false},
	{1, "futsu.nagasaki.jp", 3, false},
	{1, "goto.nagasaki.jp", 3, false},
	{1, "hasami.nagasaki.jp", 3, false},
	{1, "hirado.nagasaki.jp", 3, false},
	{1, "iki.nagasaki.jp", 3, false},
	{1, "isahaya.nagasaki.jp", 3, false},
	{1, "kawatana.nagasaki.jp", 3, false},
	{1, "kuchinotsu.nagasaki.jp", 3, false},
	{1, "matsuura.nagasaki.jp", 3, false},
	{1, "nagasaki.nagasaki.jp", 3, false},
	{1, "obama.nagasaki.jp", 3, false},
	{1, "omura.nagasaki.jp", 3, false},
	{1, "oseto.nagasaki.jp", 3, false},
	{1, "saikai.nagasaki.jp", 3, false},
	{1, "sasebo.nagasaki.jp", 3, false},
	{1, "seihi.nagasaki.jp", 3, false},
	{1, "shimabara.nagasaki.jp", 3, false},
	{1, "shinkamigoto.nagasaki.jp", 3, false},
	{1, "togitsu.nagasaki.jp", 3, false},
	{1, "tsushima.nagasaki.jp", 3, false},
	{1, "unzen.nagasaki.jp", 3, false},
	{1, "ando.nara.jp", 3, false},
	{1, "gose.nara.jp", 3, false},
	{1, "heguri.nara.jp", 3, false},
	{1, "higashiyoshino.nara.jp", 3, false},
	{1, "ikaruga.nara.jp", 3, false},
	{1, "ikoma.nara.jp", 3, false},
	{1, "kamikitayama.nara.jp", 3, false},
	{1, "kanmaki.nara.jp", 3, false},
	{1, "kashiba.nara.jp", 3, false},
	{1, "kashihara.nara.jp", 3, false},
	{1, "katsuragi.nara.jp", 3, false},
	{1, "kawai.nara.jp", 3, false},
	{1, "kawakami.nara.jp", 3, false},
	{1, "kawanishi.nara.jp", 3, false},
	{1, "koryo.nara.jp", 3, false},
	{1, "kurotaki.nara.jp", 3, false},
	{1, "mitsue.nara.jp", 3, false},
	{1, "miyake.nara.jp", 3, false},
	{1, "nara.nara.jp", 3, false},
	{1, "nosegawa.nara.jp", 3, false},
	{1, "oji.nara.jp", 3, false},
	{1, "ouda.nara.jp", 3, false},
	{1, "oyodo.nara.jp", 3, false},
	{1, "sakurai.nara.jp", 3, false},
	{1, "sango.nara.jp", 3, false},
	{1, "shimoichi.nara.jp", 3, false},
	{1, "shimokitayama.nara.jp", 3, false},
	{1, "shinjo.nara.jp", 3, false},
	{1, "soni.nara.jp", 3, false},
	{1, "takatori.nara.jp", 3, false},
	{1, "tawaramoto.nara.jp", 3, false},
	{1, "tenkawa.nara.jp", 3, false},
	{1, "tenri.nara.jp", 3, false},
	{1, "uda.nara.jp", 3, false},
	{1, "yamatokoriyama.nara.jp", 3, false},
	{1, "yamatotakada.nara.jp", 3, false},
	{1, "yamazoe.nara.jp", 3, false},
	{1, "yoshino.nara.jp", 3, false},
	{1, "aga.niigata.jp", 3, false},
	{1, "agano.niigata.jp", 3, false},
	{1, "gosen.niigata.jp", 3, false},
	{1, "itoigawa.niigata.jp", 3, false},
	{1, "izumozaki.niigata.jp", 3, false},
	{1, "joetsu.niigata.jp", 3, false},
	{1, "kamo.niigata.jp", 3, false},
	{1, "kariwa.niigata.jp", 3, false},
	{1, "kashiwazaki.niigata.jp", 3, false},
	{1, "minamiuonuma.niigata.jp", 3, false},
	{1, "mitsuke.niigata.jp", 3, false},
	{1, "muika.niigata.jp", 3, false},
	{1, "murakami.niigata.jp", 3, false},
	{1, "myoko.niigata.jp", 3, false},
	{1, "nagaoka.niigata.jp", 3, false},
	{1, "niigata.niigata.jp", 3, false},
	{1, "ojiya.niigata.jp", 3, false},
	{1, "omi.niigata.jp", 3, false},
	{1, "sado.niigata.jp", 3, false},
	{1, "sanjo.niigata.jp", 3, false},
	{1, "seiro.niigata.jp", 3, false},
	{1, "seirou.niigata.jp", 3, false},
	{1, "sekikawa.niigata.jp", 3, false},
	{1, "shibata.niigata.jp", 3, false},
	{1, "tagami.niigata.jp", 3, false},
	{1, "tainai.niigata.jp", 3, false},
	{1, "tochio.niigata.jp", 3, false},
	{1, "tokamachi.niigata.jp", 3, false},
	{1, "tsubame.niigata.jp", 3, false},
	{1, "tsunan.niigata.jp", 3, false},
	{1, "uonuma.niigata.jp", 3, false},
	{1, "yahiko.niigata.jp", 3, false},
	{1, "yoita.niigata.jp", 3, false},
	{1, "yuzawa.niigata.jp", 3, false},
	{1, "beppu.oita.jp", 3, false},
	{1, "bungoono.oita.jp", 3, false},
	{1, "bungotakada.oita.jp", 3, false},
	{1, "hasama.oita.jp", 3, false},
	{1, "hiji.oita.jp", 3, false},
	{1, "himeshima.oita.jp", 3, false},
	{1, "hita.oita.jp", 3, false},
	{1, "kamitsue.oita.jp", 3, false},
	{1, "kokonoe.oita.jp", 3, false},
	{1, "kuju.oita.jp", 3, false},
	{1, "kunisaki.oita.jp", 3, false},
	{1, "kusu.oita.jp", 3, false},
	{1, "oita.oita.jp", 3, false},
	{1, "saiki.oita.jp", 3, false},
	{1, "taketa.oita.jp", 3, false},
	{1, "tsukumi.oita.jp", 3, false},
	{1, "usa.oita.jp", 3, false},
	{1, "usuki.oita.jp", 3, false},
	{1, "yufu.oita.jp", 3, false},
	{1, "akaiwa.okayama.jp", 3, false},
	{1, "asakuchi.okayama.jp", 3, false},
	{1, "bizen.okayama.jp", 3, false},
	{1, "hayashima.okayama.jp", 3, false},
	{1, "ibara.okayama.jp", 3, false},
	{1, "kagamino.okayama.jp", 3, false},
	{1, "kasaoka.okayama.jp", 3, false},
	{1, "kibichuo.okayama.jp", 3, false},
	{1, "kumenan.okayama.jp", 3, false},
	{1, "kurashiki.okayama.jp", 3, false},
	{1, "maniwa.okayama.jp", 3, false},
	{1, "misaki.okayama.jp", 3, false},
	{1, "nagi.okayama.jp", 3, false},
	{1, "niimi.okayama.jp", 3, false},
	{1, "nishiawakura.okayama.jp", 3, false},
	{1, "okayama.okayama.jp", 3, false},
	{1, "satosho.okayama.jp", 3, false},
	{1, "setouchi.okayama.jp", 3, false},
	{1, "shinjo.okayama.jp", 3, false},
	{1, "shoo.okayama.jp", 3, false},
	{1, "soja.okayama.jp", 3, false},
	{1, "takahashi.okayama.jp", 3, false},
	{1, "tamano.okayama.jp", 3, false},
	{1, "tsuyama.okayama.jp", 3, false},
	{1, "wake.okayama.jp", 3, false},
	{1, "yakage.okayama.jp", 3, false},
	{1, "aguni.okinawa.jp", 3, false},
	{1, "ginowan.okinawa.jp", 3, false},
	{1, "ginoza.okinawa.jp", 3, false},
	{1, "gushikami.okinawa.jp", 3, false},
	{1, "haebaru.okinawa.jp", 3, false},
	{1, "higashi.okinawa.jp", 3, false},
	{1, "hirara.okinawa.jp", 3, false},
	{1, "iheya.okinawa.jp", 3, false},
	{1, "ishigaki.okinawa.jp", 3, false},
	{1, "ishikawa.okinawa.jp", 3, false},
	{1, "itoman.okinawa.jp", 3, false},
	{1, "izena.okinawa.jp", 3, false},
	{1, "kadena.okinawa.jp", 3, false},
	{1, "kin.okinawa.jp", 3, false},
	{1, "kitadaito.okinawa.jp", 3, false},
	{1, "kitanakagusuku.okinawa.jp", 3, false},
	{1, "kumejima.okinawa.jp", 3, false},
	{1, "kunigami.okinawa.jp", 3, false},
	{1, "minamidaito.okinawa.jp", 3, false},
	{1, "motobu.okinawa.jp", 3, false},
	{1, "nago.okinawa.jp", 3, false},
	{1, "naha.okinawa.jp", 3, false},
	{1, "nakagusuku.okinawa.jp", 3, false},
	{1, "nakijin.okinawa.jp", 3, false},
	{1, "nanjo.okinawa.jp", 3, false},
	{1, "nishihara.okinawa.jp", 3, false},
	{1, "ogimi.okinawa.jp", 3, false},
	{1, "okinawa.okinawa.jp", 3, false},
	{1, "onna.okinawa.jp", 3, false},
	{1, "shimoji.okinawa.jp", 3, false},
	{1, "taketomi.okinawa.jp", 3, false},
	{1, "tarama.okinawa.jp", 3, false},
	{1, "tokashiki.okinawa.jp", 3, false},
	{1, "tomigusuku.okinawa.jp", 3, false},
	{1, "tonaki.okinawa.jp", 3, false},
	{1, "urasoe.okinawa.jp", 3, false},
	{1, "uruma.okinawa.jp", 3, false},
	{1, "yaese.okinawa.jp", 3, false},
	{1, "yomitan.okinawa.jp", 3, false},
	{1, "yonabaru.okinawa.jp", 3, false},
	{1, "yonaguni.okinawa.jp", 3, false},
	{1, "zamami.okinawa.jp", 3, false},
	{1, "abeno.osaka.jp", 3, false},
	{1, "chihayaakasaka.osaka.jp", 3, false},
	{1, "chuo.osaka.jp", 3, false},
	{1, "daito.osaka.jp", 3, false},
	{1, "fujiidera.osaka.jp", 3, false},
	{1, "habikino.osaka.jp", 3, false},
	{1, "hannan.osaka.jp", 3, false},
	{1, "higashiosaka.osaka.jp", 3, false},
	{1, "higashisumiyoshi.osaka.jp", 3, false},
	{1, "higashiyodogawa.osaka.jp", 3, false},
	{1, "hirakata.osaka.jp", 3, false},
	{1, "ibaraki.osaka.jp", 3, false},
	{1, "ikeda.osaka.jp", 3, false},
	{1, "izumi.osaka.jp", 3, false},
	{1, "izumiotsu.osaka.jp", 3, false},
	{1, "izumisano.osaka.jp", 3, false},
	{1, "kadoma.osaka.jp", 3, false},
	{1, "kaizuka.osaka.jp", 3, false},
	{1, "kanan.osaka.jp", 3, false},
	{1, "kashiwara.osaka.jp", 3, false},
	{1, "katano.osaka.jp", 3, false},
	{1, "kawachinagano.osaka.jp", 3, false},
	{1, "kishiwada.osaka.jp", 3, false},
	{1, "kita.osaka.jp", 3, false},
	{1, "kumatori.osaka.jp", 3, false},
	{1, "matsubara.osaka.jp", 3, false},
	{1, "minato.osaka.jp", 3, false},
	{1, "minoh.osaka.jp", 3, false},
	{1, "misaki.osaka.jp", 3, false},
	{1, "moriguchi.osaka.jp", 3, false},
	{1, "neyagawa.osaka.jp", 3, false},
	{1, "nishi.osaka.jp", 3, false},
	{1, "nose.osaka.jp", 3, false},
	{1, "osakasayama.osaka.jp", 3, false},
	{1, "sakai.osaka.jp", 3, false},
	{1, "sayama.osaka.jp", 3, false},
	{1, "sennan.osaka.jp", 3, false},
	{1, "settsu.osaka.jp", 3, false},
	{1, "shijonawate.osaka.jp", 3, false},
	{1, "shimamoto.osaka.jp", 3, false},
	{1, "suita.osaka.jp", 3, false},
	{1, "tadaoka.osaka.jp", 3, false},
	{1, "taishi.osaka.jp", 3, false},
	{1, "tajiri.osaka.jp", 3, false},
	{1, "takaishi.osaka.jp", 3, false},
	{1, "takatsuki.osaka.jp", 3, false},
	{1, "tondabayashi.osaka.jp", 3, false},
	{1, "toyonaka.osaka.jp", 3, false},
	{1, "toyono.osaka.jp", 3, false},
	{1, "yao.osaka.jp", 3, false},
	{1, "ariake.saga.jp", 3, false},
	{1, "arita.saga.jp", 3, false},
	{1, "fukudomi.saga.jp", 3, false},
	{1, "genkai.saga.jp", 3, false},
	{1, "hamatama.saga.jp", 3, false},
	{1, "hizen.saga.jp", 3, false},
	{1, "imari.saga.jp", 3, false},
	{1, "kamimine.saga.jp", 3, false},
	{1, "kanzaki.saga.jp", 3, false},
	{1, "karatsu.saga.jp", 3, false},
	{1, "kashima.saga.jp", 3, false},
	{1, "kitagata.saga.jp", 3, false},
	{1, "kitahata.saga.jp", 3, false},
	{1, "kiyama.saga.jp", 3, false},
	{1, "kouhoku.saga.jp", 3, false},
	{1, "kyuragi.saga.jp", 3, false},
	{1, "nishiarita.saga.jp", 3, false},
	{1, "ogi.saga.jp", 3, false},
	{1, "omachi.saga.jp", 3, false},
	{1, "ouchi.saga.jp", 3, false},
	{1, "saga.saga.jp", 3, false},
	{1, "shiroishi.saga.jp", 3, false},
	{1, "taku.saga.jp", 3, false},
	{1, "tara.saga.jp", 3, false},
	{1, "tosu.saga.jp", 3, false},
	{1, "yoshinogari.saga.jp", 3, false},
	{1, "arakawa.saitama.jp", 3, false},
	{1, "asaka.saitama.jp", 3, false},
	{1, "chichibu.saitama.jp", 3, false},
	{1, "fujimi.saitama.jp", 3, false},
	{1, "fujimino.saitama.jp", 3, false},
	{1, "fukaya.saitama.jp", 3, false},
	{1, "hanno.saitama.jp", 3, false},
	{1, "hanyu.saitama.jp", 3, false},
	{1, "hasuda.saitama.jp", 3, false},
	{1, "hatogaya.saitama.jp", 3, false},
	{1, "hatoyama.saitama.jp", 3, false},
	{1, "hidaka.saitama.jp", 3, false},
	{1, "higashichichibu.saitama.jp", 3, false},
	{1, "higashimatsuyama.saitama.jp", 3, false},
	{1, "honjo.saitama.jp", 3, false},
	{1, "ina.saitama.jp", 3, false},
	{1, "iruma.saitama.jp", 3, false},
	{1, "iwatsuki.saitama.jp", 3, false},
	{1, "kamiizumi.saitama.jp", 3, false},
	{1, "kamikawa.saitama.jp", 3, false},
	{1, "kamisato.saitama.jp", 3, false},
	{1, "kasukabe.saitama.jp", 3, false},
	{1, "kawagoe.saitama.jp", 3, false},
	{1, "kawaguchi.saitama.jp", 3, false},
	{1, "kawajima.saitama.jp", 3, false},
	{1, "kazo.saitama.jp", 3, false},
	{1, "kitamoto.saitama.jp", 3, false},
	{1, "koshigaya.saitama.jp", 3, false},
	{1, "kounosu.saitama.jp", 3, false},
	{1, "kuki.saitama.jp", 3, false},
	{1, "kumagaya.saitama.jp", 3, false},
	{1, "matsubushi.saitama.jp", 3, false},
	{1, "minano.saitama.jp", 3, false},
	{1, "misato.saitama.jp", 3, false},
	{1, "miyashiro.saitama.jp", 3, false},
	{1, "miyoshi.saitama.jp", 3, false},
	{1, "moroyama.saitama.jp", 3, false},
	{1, "nagatoro.saitama.jp", 3, false},
	{1, "namegawa.saitama.jp", 3, false},
	{1, "niiza.saitama.jp", 3, false},
	{1, "ogano.saitama.jp", 3, false},
	{1, "ogawa.saitama.jp", 3, false},
	{1, "ogose.saitama.jp", 3, false},
	{1, "okegawa.saitama.jp", 3, false},
	{1, "omiya.saitama.jp", 3, false},
	{1, "otaki.saitama.jp", 3, false},
	{1, "ranzan.saitama.jp", 3, false},
	{1, "ryokami.saitama.jp", 3, false},
	{1, "saitama.saitama.jp", 3, false},
	{1, "sakado.saitama.jp", 3, false},
	{1, "satte.saitama.jp", 3, false},
	{1, "sayama.saitama.jp", 3, false},
	{1, "shiki.saitama.jp", 3, false},
	{1, "shiraoka.saitama.jp", 3, false},
	{1, "soka.saitama.jp", 3, false},
	{1, "sugito.saitama.jp", 3, false},
	{1, "toda.saitama.jp", 3, false},
	{1, "tokigawa.saitama.jp", 3, false},
	{1, "tokorozawa.saitama.jp", 3, false},
	{1, "tsurugashima.saitama.jp", 3, false},
	{1, "urawa.saitama.jp", 3, false},
	{1, "warabi.saitama.jp", 3, false},
	{1, "yashio.saitama.jp", 3, false},
	{1, "yokoze.saitama.jp", 3, false},
	{1, "yono.saitama.jp", 3, false},
	{1, "yorii.saitama.jp", 3, false},
	{1, "yoshida.saitama.jp", 3, false},
	{1, "yoshikawa.saitama.jp", 3, false},
	{1, "yoshimi.saitama.jp", 3, false},
	{1, "aisho.shiga.jp", 3, false},
	{1, "gamo.shiga.jp", 3, false},
	{1, "higashiomi.shiga.jp", 3, false},
	{1, "hikone.shiga.jp", 3, false},
	{1, "koka.shiga.jp", 3, false},
	{1, "konan.shiga.jp", 3, false},
	{1, "kosei.shiga.jp", 3, false},
	{1, "koto.shiga.jp", 3, false},
	{1, "kusatsu.shiga.jp", 3, false},
	{1, "maibara.shiga.jp", 3, false},
	{1, "moriyama.shiga.jp", 3, false},
	{1, "nagahama.shiga.jp", 3, false},
	{1, "nishiazai.shiga.jp", 3, false},
	{1, "notogawa.shiga.jp", 3, false},
	{1, "omihachiman.shiga.jp", 3, false},
	{1, "otsu.shiga.jp", 3, false},
	{1, "ritto.shiga.jp", 3, false},
	{1, "ryuoh.shiga.jp", 3, false},
	{1, "takashima.shiga.jp", 3, false},
	{1, "takatsuki.shiga.jp", 3, false},
	{1, "torahime.shiga.jp", 3, false},
	{1, "toyosato.shiga.jp", 3, false},
	{1, "yasu.shiga.jp", 3, false},
	{1, "akagi.shimane.jp", 3, false},
	{1, "ama.shimane.jp", 3, false},
	{1, "gotsu.shimane.jp", 3, false},
	{1, "hamada.shimane.jp", 3, false},
	{1, "higashiizumo.shimane.jp", 3, false},
	{1, "hikawa.shimane.jp", 3, false},
	{1, "hikimi.shimane.jp", 3, false},
	{1, "izumo.shimane.jp", 3, false},
	{1, "kakinoki.shimane.jp", 3, false},
	{1, "masuda.shimane.jp", 3, false},
	{1, "matsue.shimane.jp", 3, false},
	{1, "misato.shimane.jp", 3, false},
	{1, "nishinoshima.shimane.jp", 3, false},
	{1, "ohda.shimane.jp", 3, false},
	{1, "okinoshima.shimane.jp", 3, false},
	{1, "okuizumo.shimane.jp", 3, false},
	{1, "shimane.shimane.jp", 3, false},
	{1, "tamayu.shimane.jp", 3, false},
	{1, "tsuwano.shimane.jp", 3, false},
	{1, "unnan.shimane.jp", 3, false},
	{1, "yakumo.shimane.jp", 3, false},
	{1, "yasugi.shimane.jp", 3, false},
	{1, "yatsuka.shimane.jp", 3, false},
	{1, "arai.shizuoka.jp", 3, false},
	{1, "atami.shizuoka.jp", 3, false},
	{1, "fuji.shizuoka.jp", 3, false},
	{1, "fujieda.shizuoka.jp", 3, false},
	{1, "fujikawa.shizuoka.jp", 3, false},
	{1, "fujinomiya.shizuoka.jp", 3, false},
	{1, "fukuroi.shizuoka.jp", 3, false},
	{1, "gotemba.shizuoka.jp", 3, false},
	{1, "haibara.shizuoka.jp", 3, false},
	{1, "hamamatsu.shizuoka.jp", 3, false},
	{1, "higashiizu.shizuoka.jp", 3, false},
	{1, "ito.shizuoka.jp", 3, false},
	{1, "iwata.shizuoka.jp", 3, false},
	{1, "izu.shizuoka.jp", 3, false},
	{1, "izunokuni.shizuoka.jp", 3, false},
	{1, "kakegawa.shizuoka.jp", 3, false},
	{1, "kannami.shizuoka.jp", 3, false},
	{1, "kawanehon.shizuoka.jp", 3, false},
	{1, "kawazu.shizuoka.jp", 3, false},
	{1, "kikugawa.shizuoka.jp", 3, false},
	{1, "kosai.shizuoka.jp", 3, false},
	{1, "makinohara.shizuoka.jp", 3, false},
	{1, "matsuzaki.shizuoka.jp", 3, false},
	{1, "minamiizu.shizuoka.jp", 3, false},
	{1, "mishima.shizuoka.jp", 3, false},
	{1, "morimachi.shizuoka.jp", 3, false},
	{1, "nishiizu.shizuoka.jp", 3, false},
	{1, "numazu.shizuoka.jp", 3, false},
	{1, "omaezaki.shizuoka.jp", 3, false},
	{1, "shimada.shizuoka.jp", 3, false},
	{1, "shimizu.shizuoka.jp", 3, false},
	{1, "shimoda.shizuoka.jp", 3, false},
	{1, "shizuoka.shizuoka.jp", 3, false},
	{1, "susono.shizuoka.jp", 3, false},
	{1, "yaizu.shizuoka.jp", 3, false},
	{1, "yoshida.shizuoka.jp", 3, false},
	{1, "ashikaga.tochigi.jp", 3, false},
	{1, "bato.tochigi.jp", 3, false},
	{1, "haga.tochigi.jp", 3, false},
	{1, "ichikai.tochigi.jp", 3, false},
	{1, "iwafune.tochigi.jp", 3, false},
	{1, "kaminokawa.tochigi.jp", 3, false},
	{1, "kanuma.tochigi.jp", 3, false},
	{1, "karasuyama.tochigi.jp", 3, false},
	{1, "kuroiso.tochigi.jp", 3, false},
	{1, "mashiko.tochigi.jp", 3, false},
	{1, "mibu.tochigi.jp", 3, false},
	{1, "moka.tochigi.jp", 3, false},
	{1, "motegi.tochigi.jp", 3, false},
	{1, "nasu.tochigi.jp", 3, false},
	{1, "nasushiobara.tochigi.jp", 3, false},
	{1, "nikko.tochigi.jp", 3, false},
	{1, "nishikata.tochigi.jp", 3, false},
	{1, "nogi.tochigi.jp", 3, false},
	{1, "ohira.tochigi.jp", 3, false},
	{1, "ohtawara.tochigi.jp", 3, false},
	{1, "oyama.tochigi.jp", 3, false},
	{1, "sakura.tochigi.jp", 3, false},
	{1, "sano.tochigi.jp", 3, false},
	{1, "shimotsuke.tochigi.jp", 3, false},
	{1, "shioya.tochigi.jp", 3, false},
	{1, "takanezawa.tochigi.jp", 3, false},
	{1, "tochigi.tochigi.jp", 3, false},
	{1, "tsuga.tochigi.jp", 3, false},
	{1, "ujiie.tochigi.jp", 3, false},
	{1, "utsunomiya.tochigi.jp", 3, false},
	{1, "yaita.tochigi.jp", 3, false},
	{1, "aizumi.tokushima.jp", 3, false},
	{1, "anan.tokushima.jp", 3, false},
	{1, "ichiba.tokushima.jp", 3, false},
	{1, "itano.tokushima.jp", 3, false},
	{1, "kainan.tokushima.jp", 3, false},
	{1, "komatsushima.tokushima.jp", 3, false},
	{1, "matsushige.tokushima.jp", 3, false},
	{1, "mima.tokushima.jp", 3, false},
	{1, "minami.tokushima.jp", 3, false},
	{1, "miyoshi.tokushima.jp", 3, false},
	{1, "mugi.tokushima.jp", 3, false},
	{1, "nakagawa.tokushima.jp", 3, false},
	{1, "naruto.tokushima.jp", 3, false},
	{1, "sanagochi.tokushima.jp", 3, false},
	{1, "shishikui.tokushima.jp", 3, false},
	{1, "tokushima.tokushima.jp", 3, false},
	{1, "wajiki.tokushima.jp", 3, false},
	{1, "adachi.tokyo.jp", 3, false},
	{1, "akiruno.tokyo.jp", 3, false},
	{1, "akishima.tokyo.jp", 3, false},
	{1, "aogashima.tokyo.jp", 3, false},
	{1, "arakawa.tokyo.jp", 3, false},
	{1, "bunkyo.tokyo.jp", 3, false},
	{1, "chiyoda.tokyo.jp", 3, false},
	{1, "chofu.tokyo.jp", 3, false},
	{1, "chuo.tokyo.jp", 3, false},
	{1, "edogawa.tokyo.jp", 3, false},
	{1, "fuchu.tokyo.jp", 3, false},
	{1, "fussa.tokyo.jp", 3, false},
	{1, "hachijo.tokyo.jp", 3, false},
	{1, "hachioji.tokyo.jp", 3, false},
	{1, "hamura.tokyo.jp", 3, false},
	{1, "higashikurume.tokyo.jp", 3, false},
	{1, "higashimurayama.tokyo.jp", 3, false},
	{1, "higashiyamato.tokyo.jp", 3, false},
	{1, "hino.tokyo.jp", 3, false},
	{1, "hinode.tokyo.jp", 3, false},
	{1, "hinohara.tokyo.jp", 3, false},
	{1, "inagi.tokyo.jp", 3, false},
	{1, "itabashi.tokyo.jp", 3, false},
	{1, "katsushika.tokyo.jp", 3, false},
	{1, "kita.tokyo.jp", 3, false},
	{1, "kiyose.tokyo.jp", 3, false},
	{1, "kodaira.tokyo.jp", 3, false},
	{1, "koganei.tokyo.jp", 3, false},
	{1, "kokubunji.tokyo.jp", 3, false},
	{1, "komae.tokyo.jp", 3, false},
	{1, "koto.tokyo.jp", 3, false},
	{1, "kouzushima.tokyo.jp", 3, false},
	{1, "kunitachi.tokyo.jp", 3, false},
	{1, "machida.tokyo.jp", 3, false},
	{1, "meguro.tokyo.jp", 3, false},
	{1, "minato.tokyo.jp", 3, false},
	{1, "mitaka.tokyo.jp", 3, false},
	{1, "mizuho.tokyo.jp", 3, false},
	{1, "musashimurayama.tokyo.jp", 3, false},
	{1, "musashino.tokyo.jp", 3, false},
	{1, "nakano.tokyo.jp", 3, false},
	{1, "nerima.tokyo.jp", 3, false},
	{1, "ogasawara.tokyo.jp", 3, false},
	{1, "okutama.tokyo.jp", 3, false},
	{1, "ome.tokyo.jp", 3, false},
	{1, "oshima.tokyo.jp", 3, false},
	{1, "ota.tokyo.jp", 3, false},
	{1, "setagaya.tokyo.jp", 3, false},
	{1, "shibuya.tokyo.jp", 3, false},
	{1, "shinagawa.tokyo.jp", 3, false},
	{1, "shinjuku.tokyo.jp", 3, false},
	{1, "suginami.tokyo.jp", 3, false},
	{1, "sumida.tokyo.jp", 3, false},
	{1, "tachikawa.tokyo.jp", 3, false},
	{1, "taito.tokyo.jp", 3, false},
	{1, "tama.tokyo.jp", 3, false},
	{1, "toshima.tokyo.jp", 3, false},
	{1, "chizu.tottori.jp", 3, false},
	{1, "hino.tottori.jp", 3, false},
	{1, "kawahara.tottori.jp", 3, false},
	{1, "koge.tottori.jp", 3, false},
	{1, "kotoura.tottori.jp", 3, false},
	{1, "misasa.tottori.jp", 3, false},
	{1, "nanbu.tottori.jp", 3, false},
	{1, "nichinan.tottori.jp", 3, false},
	{1, "sakaiminato.tottori.jp", 3, false},
	{1, "tottori.tottori.jp", 3, false},
	{1, "wakasa.tottori.jp", 3, false},
	{1, "yazu.tottori.jp", 3, false},
	{1, "yonago.tottori.jp", 3, false},
	{1, "asahi.toyama.jp", 3, false},
	{1, "fuchu.toyama.jp", 3, false},
	{1, "fukumitsu.toyama.jp", 3, false},
	{1, "funahashi.toyama.jp", 3, false},
	{1, "himi.toyama.jp", 3, false},
	{1, "imizu.toyama.jp", 3, false},
	{1, "inami.toyama.jp", 3, false},
	{1, "johana.toyama.jp", 3, false},
	{1, "kamiichi.toyama.jp", 3, false},
	{1, "kurobe.toyama.jp", 3, false},
	{1, "nakaniikawa.toyama.jp", 3, false},
	{1, "namerikawa.toyama.jp", 3, false},
	{1, "nanto.toyama.jp", 3, false},
	{1, "nyuzen.toyama.jp", 3, false},
	{1, "oyabe.toyama.jp", 3, false},
	{1, "taira.toyama.jp", 3, false},
	{1, "takaoka.toyama.jp", 3, false},
	{1, "tateyama.toyama.jp", 3, false},
	{1, "toga.toyama.jp", 3, false},
	{1, "tonami.toyama.jp", 3, false},
	{1, "toyama.toyama.jp", 3, false},
	{1, "unazuki.toyama.jp", 3, false},
	{1, "uozu.toyama.jp", 3, false},
	{1, "yamada.toyama.jp", 3, false},
	{1, "arida.wakayama.jp", 3, false},
	{1, "aridagawa.wakayama.jp", 3, false},
	{1, "gobo.wakayama.jp", 3, false},
	{1, "hashimoto.wakayama.jp", 3, false},
	{1, "hidaka.wakayama.jp", 3, false},
	{1, "hirogawa.wakayama.jp", 3, false},
	{1, "inami.wakayama.jp", 3, false},
	{1, "iwade.wakayama.jp", 3, false},
	{1, "kainan.wakayama.jp", 3, false},
	{1, "kamitonda.wakayama.jp", 3, false},
	{1, "katsuragi.wakayama.jp", 3, false},
	{1, "kimino.wakayama.jp", 3, false},
	{1, "kinokawa.wakayama.jp", 3, false},
	{1, "kitayama.wakayama.jp", 3, false},
	{1, "koya.wakayama.jp", 3, false},
	{1, "koza.wakayama.jp", 3, false},
	{1, "kozagawa.wakayama.jp", 3, false},
	{1, "kudoyama.wakayama.jp", 3, false},
	{1, "kushimoto.wakayama.jp", 3, false},
	{1, "mihama.wakayama.jp", 3, false},
	{1, "misato.wakayama.jp", 3, false},
	{1, "nachikatsuura.wakayama.jp", 3, false},
	{1, "shingu.wakayama.jp", 3, false},
	{1, "shirahama.wakayama.jp", 3, false},
	{1, "taiji.wakayama.jp", 3, false},
	{1, "tanabe.wakayama.jp", 3, false},
	{1, "wakayama.wakayama.jp", 3, false},
	{1, "yuasa.wakayama.jp", 3, false},
	{1, "yura.wakayama.jp", 3, false},
	{1, "asahi.yamagata.jp", 3, false},
	{1, "funagata.yamagata.jp", 3, false},
	{1, "higashine.yamagata.jp", 3, false},
	{1, "iide.yamagata.jp", 3, false},
	{1, "kahoku.yamagata.jp", 3, false},
	{1, "kaminoyama.yamagata.jp", 3, false},
	{1, "kaneyama.yamagata.jp", 3, false},
	{1, "kawanishi.yamagata.jp", 3, false},
	{1, "mamurogawa.yamagata.jp", 3, false},
	{1, "mikawa.yamagata.jp", 3, false},
	{1, "murayama.yamagata.jp", 3, false},
	{1, "nagai.yamagata.jp", 3, false},
	{1, "nakayama.yamagata.jp", 3, false},
	{1, "nanyo.yamagata.jp", 3, false},
	{1, "nishikawa.yamagata.jp", 3, false},
	{1, "obanazawa.yamagata.jp", 3, false},
	{1, "oe.yamagata.jp", 3, false},
	{1, "oguni.yamagata.jp", 3, false},
	{1, "ohkura.yamagata.jp", 3, false},
	{1, "oishida.yamagata.jp", 3, false},
	{1, "sagae.yamagata.jp", 3, false},
	{1, "sakata.yamagata.jp", 3, false},
	{1, "sakegawa.yamagata.jp", 3, false},
	{1, "shinjo.yamagata.jp", 3, false},
	{1, "shirataka.yamagata.jp", 3, false},
	{1, "shonai.yamagata.jp", 3, false},
	{1, "takahata.yamagata.jp", 3, false},
	{1, "tendo.yamagata.jp", 3, false},
	{1, "tozawa.yamagata.jp", 3, false},
	{1, "tsuruoka.yamagata.jp", 3, false},
	{1, "yamagata.yamagata.jp", 3, false},
	{1, "yamanobe.yamagata.jp", 3, false},
	{1, "yonezawa.yamagata.jp", 3, false},
	{1, "yuza.yamagata.jp", 3, false},
	{1, "abu.yamaguchi.jp", 3, false},
	{1, "hagi.yamaguchi.jp", 3, false},
	{1, "hikari.yamaguchi.jp", 3, false},
	{1, "hofu.yamaguchi.jp", 3, false},
	{1, "iwakuni.yamaguchi.jp", 3, false},
	{1, "kudamatsu.yamaguchi.jp", 3, false},
	{1, "mitou.yamaguchi.jp", 3, false},
	{1, "nagato.yamaguchi.jp", 3, false},
	{1, "oshima.yamaguchi.jp", 3, false},
	{1, "shimonoseki.yamaguchi.jp", 3, false},
	{1, "shunan.yamaguchi.jp", 3, false},
	{1, "tabuse.yamaguchi.jp", 3, false},
	{1, "tokuyama.yamaguchi.jp", 3, false},
	{1, "toyota.yamaguchi.jp", 3, false},
	{1, "ube.yamaguchi.jp", 3, false},
	{1, "yuu.yamaguchi.jp", 3, false},
	{1, "chuo.yamanashi.jp", 3, false},
	{1, "doshi.yamanashi.jp", 3, false},
	{1, "fuefuki.yamanashi.jp", 3, false},
	{1, "fujikawa.yamanashi.jp", 3, false},
	{1, "fujikawaguchiko.yamanashi.jp", 3, false},
	{1, "fujiyoshida.yamanashi.jp", 3, false},
	{1, "hayakawa.yamanashi.jp", 3, false},
	{1, "hokuto.yamanashi.jp", 3, false},
	{1, "ichikawamisato.yamanashi.jp", 3, false},
	{1, "kai.yamanashi.jp", 3, false},
	{1, "kofu.yamanashi.jp", 3, false},
	{1, "koshu.yamanashi.jp", 3, false},
	{1, "kosuge.yamanashi.jp", 3, false},
	{1, "minami-alps.yamanashi.jp", 3, false},
	{1, "minobu.yamanashi.jp", 3, false},
	{1, "nakamichi.yamanashi.jp", 3, false},
	{1, "nanbu.yamanashi.jp", 3, false},
	{1, "narusawa.yamanashi.jp", 3, false},
	{1, "nirasaki.yamanashi.jp", 3, false},
	{1, "nishikatsura.yamanashi.jp", 3, false},
	{1, "oshino.yamanashi.jp", 3, false},
	{1, "otsuki.yamanashi.jp", 3, false},
	{1, "showa.yamanashi.jp", 3, false},
	{1, "tabayama.yamanashi.jp", 3, false},
	{1, "tsuru.yamanashi.jp", 3, false},
	{1, "uenohara.yamanashi.jp", 3, false},
	{1, "yamanakako.yamanashi.jp", 3, false},
	{1, "yamanashi.yamanashi.jp", 3, false},
	{1, "ke", 1, false},
	{1, "ac.ke", 2, false},
	{1, "co.ke", 2, false},
	{1, "go.ke", 2, false},
	{1, "info.ke", 2, false},
	{1, "me.ke", 2, false},
	{1, "mobi.ke", 2, false},
	{1, "ne.ke", 2, false},
	{1, "or.ke", 2, false},
	{1, "sc.ke", 2, false},
	{1, "kg", 1, false},
	{1, "org.kg", 2, false},
	{1, "net.kg", 2, false},
	{1, "com.kg", 2, false},
	{1, "edu.kg", 2, false},
	{1, "gov.kg", 2, false},
	{1, "mil.kg", 2, false},
	{2, "kh", 2, false},
	{1, "ki", 1, false},
	{1, "edu.ki", 2, false},
	{1, "biz.ki", 2, false},
	{1, "net.ki", 2, false},
	{1, "org.ki", 2, false},
	{1, "gov.ki", 2, false},
	{1, "info.ki", 2, false},
	{1, "com.ki", 2, false},
	{1, "km", 1, false},
	{1, "org.km", 2, false},
	{1, "nom.km", 2, false},
	{1, "gov.km", 2, false},
	{1, "prd.km", 2, false},
	{1, "tm.km", 2, false},
	{1, "edu.km", 2, false},
	{1, "mil.km", 2, false},
	{1, "ass.km", 2, false},
	{1, "com.km", 2, false},
	{1, "coop.km", 2, false},
	{1, "asso.km", 2, false},
	{1, "presse.km", 2, false},
	{1, "medecin.km", 2, false},
	{1, "notaires.km", 2, false},
	{1, "pharmaciens.km", 2, false},
	{1, "veterinaire.km", 2, false},
	{1, "gouv.km", 2, false},
	{1, "kn", 1, false},
	{1, "net.kn", 2, false},
	{1, "org.kn", 2, false},
	{1, "edu.kn", 2, false},
	{1, "gov.kn", 2, false},
	{1, "kp", 1, false},
	{1, "com.kp", 2, false},
	{1, "edu.kp", 2, false},
	{1, "gov.kp", 2, false},
	{1, "org.kp", 2, false},
	{1, "rep.kp", 2, false},
	{1, "tra.kp", 2, false},
	{1, "kr", 1, false},
	{1, "ac.kr", 2, false},
	{1, "co.kr", 2, false},
	{1, "es.kr", 2, false},
	{1, "go.kr", 2, false},
	{1, "hs.kr", 2, false},
	{1, "kg.kr", 2, false},
	{1, "mil.kr", 2, false},
	{1, "ms.kr", 2, false},
	{1, "ne.kr", 2, false},
	{1, "or.kr", 2, false},
	{1, "pe.kr", 2, false},
	{1, "re.kr", 2, false},
	{1, "sc.kr", 2, false},
	{1, "busan.kr", 2, false},
	{1, "chungbuk.kr", 2, false},
	{1, "chungnam.kr", 2, false},
	{1, "daegu.kr", 2, false},
	{1, "daejeon.kr", 2, false},
	{1, "gangwon.kr", 2, false},
	{1, "gwangju.kr", 2, false},
	{1, "gyeongbuk.kr", 2, false},
	{1, "gyeonggi.kr", 2, false},
	{1, "gyeongnam.kr", 2, false},
	{1, "incheon.kr", 2, false},
	{1, "jeju.kr", 2, false},
	{1, "jeonbuk.kr", 2, false},
	{1, "jeonnam.kr", 2, false},
	{1, "seoul.kr", 2, false},
	{1, "ulsan.kr", 2, false},
	{1, "kw", 1, false},
	{1, "com.kw", 2, false},
	{1, "edu.kw", 2, false},
	{1, "emb.kw", 2, false},
	{1, "gov.kw", 2, false},
	{1, "ind.kw", 2, false},
	{1, "net.kw", 2, false},
	{1, "org.kw", 2, false},
	{1, "ky", 1, false},
	{1, "edu.ky", 2, false},
	{1, "gov.ky", 2, false},
	{1, "com.ky", 2, false},
	{1, "org.ky", 2, false},
	{1, "net.ky", 2, false},
	{1, "kz", 1, false},
	{1, "org.kz", 2, false},
	{1, "edu.kz", 2, false},
	{1, "net.kz", 2, false},
	{1, "gov.kz", 2, false},
	{1, "mil.kz", 2, false},
	{1, "com.kz", 2, false},
	{1, "la", 1, false},
	{1, "int.la", 2, false},
	{1, "net.la", 2, false},
	{1, "info.la", 2, false},
	{1, "edu.la", 2, false},
	{1, "gov.la", 2, false},
	{1, "per.la", 2, false},
	{1, "com.la", 2, false},
	{1, "org.la", 2, false},
	{1, "lb", 1, false},
	{1, "com.lb", 2, false},
	{1, "edu.lb", 2, false},
	{1, "gov.lb", 2, false},
	{1, "net.lb", 2, false},
	{1, "org.lb", 2, false},
	{1, "lc", 1, false},
	{1, "com.lc", 2, false},
	{1, "net.lc", 2, false},
	{1, "co.lc", 2, false},
	{1, "org.lc", 2, false},
	{1, "edu.lc", 2, false},
	{1, "gov.lc", 2, false},
	{1, "li", 1, false},
	{1, "lk", 1, false},
	{1, "gov.lk", 2, false},
	{1, "sch.lk", 2, false},
	{1, "net.lk", 2, false},
	{1, "int.lk", 2, false},
	{1, "com.lk", 2, false},
	{1, "org.lk", 2, false},
	{1, "edu.lk", 2, false},
	{1, "ngo.lk", 2, false},
	{1, "soc.lk", 2, false},
	{1, "web.lk", 2, false},
	{1, "ltd.lk", 2, false},
	{1, "assn.lk", 2, false},
	{1, "grp.lk", 2, false},
	{1, "hotel.lk", 2, false},
	{1, "ac.lk", 2, false},
	{1, "lr", 1, false},
	{1, "com.lr", 2, false},
	{1, "edu.lr", 2, false},
	{1, "gov.lr", 2, false},
	{1, "org.lr", 2, false},
	{1, "net.lr", 2, false},
	{1, "ls", 1, false},
	{1, "ac.ls", 2, false},
	{1, "biz.ls", 2, false},
	{1, "co.ls", 2, false},
	{1, "edu.ls", 2, false},
	{1, "gov.ls", 2, false},
	{1, "info.ls", 2, false},
	{1, "net.ls", 2, false},
	{1, "org.ls", 2, false},
	{1, "sc.ls", 2, false},
	{1, "lt", 1, false},
	{1, "gov.lt", 2, false},
	{1, "lu", 1, false},
	{1, "lv", 1, false},
	{1, "com.lv", 2, false},
	{1, "edu.lv", 2, false},
	{1, "gov.lv", 2, false},
	{1, "org.lv", 2, false},
	{1, "mil.lv", 2, false},
	{1, "id.lv", 2, false},
	{1, "net.lv", 2, false},
	{1, "asn.lv", 2, false},
	{1, "conf.lv", 2, false},
	{1, "ly", 1, false},
	{1, "com.ly", 2, false},
	{1, "net.ly", 2, false},
	{1, "gov.ly", 2, false},
	{1, "plc.ly", 2, false},
	{1, "edu.ly", 2, false},
	{1, "sch.ly", 2, false},
	{1, "med.ly", 2, false},
	{1, "org.ly", 2, false},
	{1, "id.ly", 2, false},
	{1, "ma", 1, false},
	{1, "co.ma", 2, false},
	{1, "net.ma", 2, false},
	{1, "gov.ma", 2, false},
	{1, "org.ma", 2, false},
	{1, "ac.ma", 2, false},
	{1, "press.ma", 2, false},
	{1, "mc", 1, false},
	{1, "tm.mc", 2, false},
	{1, "asso.mc", 2, false},
	{1, "md", 1, false},
	{1, "me", 1, false},
	{1, "co.me", 2, false},
	{1, "net.me", 2, false},
	{1, "org.me", 2, false},
	{1, "edu.me", 2, false},
	{1, "ac.me", 2, false},
	{1, "gov.me", 2, false},
	{1, "its.me", 2, false},
	{1, "priv.me", 2, false},
	{1, "mg", 1, false},
	{1, "org.mg", 2, false},
	{1, "nom.mg", 2, false},
	{1, "gov.mg", 2, false},
	{1, "prd.mg", 2, false},
	{1, "tm.mg", 2, false},
	{1, "edu.mg", 2, false},
	{1, "mil.mg", 2, false},
	{1, "com.mg", 2, false},
	{1, "co.mg", 2, false},
	{1, "mh", 1, false},
	{1, "mil", 1, false},
	{1, "mk", 1, false},
	{1, "com.mk", 2, false},
	{1, "org.mk", 2, false},
	{1, "net.mk", 2, false},
	{1, "edu.mk", 2, false},
	{1, "gov.mk", 2, false},
	{1, "inf.mk", 2, false},
	{1, "name.mk", 2, false},
	{1, "ml", 1, false},
	{1, "com.ml", 2, false},
	{1, "edu.ml", 2, false},
	{1, "gouv.ml", 2, false},
	{1, "gov.ml", 2, false},
	{1, "net.ml", 2, false},
	{1, "org.ml", 2, false},
	{1, "presse.ml", 2, false},
	{2, "mm", 2, false},
	{1, "mn", 1, false},
	{1, "gov.mn", 2, false},
	{1, "edu.mn", 2, false},
	{1, "org.mn", 2, false},
	{1, "mo", 1, false},
	{1, "com.mo", 2, false},
	{1, "net.mo", 2, false},
	{1, "org.mo", 2, false},
	{1, "edu.mo", 2, false},
	{1, "gov.mo", 2, false},
	{1, "mobi", 1, false},
	{1, "mp", 1, false},
	{1, "mq", 1, false},
	{1, "mr", 1, false},
	{1, "gov.mr", 2, false},
	{1, "ms", 1, false},
	{1, "com.ms", 2, false},
	{1, "edu.ms", 2, false},
	{1, "gov.ms", 2, false},
	{1, "net.ms", 2, false},
	{1, "org.ms", 2, false},
	{1, "mt", 1, false},
	{1, "com.mt", 2, false},
	{1, "edu.mt", 2, false},
	{1, "net.mt", 2, false},
	{1, "org.mt", 2, false},
	{1, "mu", 1, false},
	{1, "com.mu", 2, false},
	{1, "net.mu", 2, false},
	{1, "org.mu", 2, false},
	{1, "gov.mu", 2, false},
	{1, "ac.mu", 2, false},
	{1, "co.mu", 2, false},
	{1, "or.mu", 2, false},
	{1, "museum", 1, false},
	{1, "academy.museum", 2, false},
	{1, "agriculture.museum", 2, false},
	{1, "air.museum", 2, false},
	{1, "airguard.museum", 2, false},
	{1, "alabama.museum", 2, false},
	{1, "alaska.museum", 2, false},
	{1, "amber.museum", 2, false},
	{1, "ambulance.museum", 2, false},
	{1, "american.museum", 2, false},
	{1, "americana.museum", 2, false},
	{1, "americanantiques.museum", 2, false},
	{1, "americanart.museum", 2, false},
	{1, "amsterdam.museum", 2, false},
	{1, "and.museum", 2, false},
	{1, "annefrank.museum", 2, false},
	{1, "anthro.museum", 2, false},
	{1, "anthropology.museum", 2, false},
	{1, "antiques.museum", 2, false},
	{1, "aquarium.museum", 2, false},
	{1, "arboretum.museum", 2, false},
	{1, "archaeological.museum", 2, false},
	{1, "archaeology.museum", 2, false},
	{1, "architecture.museum", 2, false},
	{1, "art.museum", 2, false},
	{1, "artanddesign.museum", 2, false},
	{1, "artcenter.museum", 2, false},
	{1, "artdeco.museum", 2, false},
	{1, "arteducation.museum", 2, false},
	{1, "artgallery.museum", 2, false},
	{1, "arts.museum", 2, false},
	{1, "artsandcrafts.museum", 2, false},
	{1, "asmatart.museum", 2, false},
	{1, "assassination.museum", 2, false},
	{1, "assisi.museum", 2, false},
	{1, "association.museum", 2, false},
	{1, "astronomy.museum", 2, false},
	{1, "atlanta.museum", 2, false},
	{1, "austin.museum", 2, false},
	{1, "australia.museum", 2, false},
	{1, "automotive.museum", 2, false},
	{1, "aviation.museum", 2, false},
	{1, "axis.museum", 2, false},
	{1, "badajoz.museum", 2, false},
	{1, "baghdad.museum", 2, false},
	{1, "bahn.museum", 2, false},
	{1, "bale.museum", 2, false},
	{1, "baltimore.museum", 2, false},
	{1, "barcelona.museum", 2, false},
	{1, "baseball.museum", 2, false},
	{1, "basel.museum", 2, false},
	{1, "baths.museum", 2, false},
	{1, "bauern.museum", 2, false},
	{1, "beauxarts.museum", 2, false},
	{1, "beeldengeluid.museum", 2, false},
	{1, "bellevue.museum", 2, false},
	{1, "bergbau.museum", 2, false},
	{1, "berkeley.museum", 2, false},
	{1, "berlin.museum", 2, false},
	{1, "bern.museum", 2, false},
	{1, "bible.museum", 2, false},
	{1, "bilbao.museum", 2, false},
	{1, "bill.museum", 2, false},
	{1, "birdart.museum", 2, false},
	{1, "birthplace.museum", 2, false},
	{1, "bonn.museum", 2, false},
	{1, "boston.museum", 2, false},
	{1, "botanical.museum", 2, false},
	{1, "botanicalgarden.museum", 2, false},
	{1, "botanicgarden.museum", 2, false},
	{1, "botany.museum", 2, false},
	{1, "brandywinevalley.museum", 2, false},
	{1, "brasil.museum", 2, false},
	{1, "bristol.museum", 2, false},
	{1, "british.museum", 2, false},
	{1, "britishcolumbia.museum", 2, false},
	{1, "broadcast.museum", 2, false},
	{1, "brunel.museum", 2, false},
	{1, "brussel.museum", 2, false},
	{1, "brussels.museum", 2, false},
	{1, "bruxelles.museum", 2, false},
	{1, "building.museum", 2, false},
	{1, "burghof.museum", 2, false},
	{1, "bus.museum", 2, false},
	{1, "bushey.museum", 2, false},
	{1, "cadaques.museum", 2, false},
	{1, "california.museum", 2, false},
	{1, "cambridge.museum", 2, false},
	{1, "can.museum", 2, false},
	{1, "canada.museum", 2, false},
	{1, "capebreton.museum", 2, false},
	{1, "carrier.museum", 2, false},
	{1, "cartoonart.museum", 2, false},
	{1, "casadelamoneda.museum", 2, false},
	{1, "castle.museum", 2, false},
	{1, "castres.museum", 2, false},
	{1, "celtic.museum", 2, false},
	{1, "center.museum", 2, false},
	{1, "chattanooga.museum", 2, false},
	{1, "cheltenham.museum", 2, false},
	{1, "chesapeakebay.museum", 2, false},
	{1, "chicago.museum", 2, false},
	{1, "children.museum", 2, false},
	{1, "childrens.museum", 2, false},
	{1, "childrensgarden.museum", 2, false},
	{1, "chiropractic.museum", 2, false},
	{1, "chocolate.museum", 2, false},
	{1, "christiansburg.museum", 2, false},
	{1, "cincinnati.museum", 2, false},
	{1, "cinema.museum", 2, false},
	{1, "circus.museum", 2, false},
	{1, "civilisation.museum", 2, false},
	{1, "civilization.museum", 2, false},
	{1, "civilwar.museum", 2, false},
	{1, "clinton.museum", 2, false},
	{1, "clock.museum", 2, false},
	{1, "coal.museum", 2, false},
	{1, "coastaldefence.museum", 2, false},
	{1, "cody.museum", 2, false},
	{1, "coldwar.museum", 2, false},
	{1, "collection.museum", 2, false},
	{1, "colonialwilliamsburg.museum", 2, false},
	{1, "coloradoplateau.museum", 2, false},
	{1, "columbia.museum", 2, false},
	{1, "columbus.museum", 2, false},
	{1, "communication.museum", 2, false},
	{1, "communications.museum", 2, false},
	{1, "community.museum", 2, false},
	{1, "computer.museum", 2, false},
	{1, "computerhistory.museum", 2, false},
	{1, "xn--comunicaes-v6a2o.museum", 2, false},
	{1, "contemporary.museum", 2, false},
	{1, "contemporaryart.museum", 2, false},
	{1, "convent.museum", 2, false},
	{1, "copenhagen.museum", 2, false},
	{1, "corporation.museum", 2, false},
	{1, "xn--correios-e-telecomunicaes-ghc29a.museum", 2, false},
	{1, "corvette.museum", 2, false},
	{1, "costume.museum", 2, false},
	{1, "countryestate.museum", 2, false},
	{1, "county.museum", 2, false},
	{1, "crafts.museum", 2, false},
	{1, "cranbrook.museum", 2, false},
	{1, "creation.museum", 2, false},
	{1, "cultural.museum", 2, false},
	{1, "culturalcenter.museum", 2, false},
	{1, "culture.museum", 2, false},
	{1, "cyber.museum", 2, false},
	{1, "cymru.museum", 2, false},
	{1, "dali.museum", 2, false},
	{1, "dallas.museum", 2, false},
	{1, "database.museum", 2, false},
	{1, "ddr.museum", 2, false},
	{1, "decorativearts.museum", 2, false},
	{1, "delaware.museum", 2, false},
	{1, "delmenhorst.museum", 2, false},
	{1, "denmark.museum", 2, false},
	{1, "depot.museum", 2, false},
	{1, "design.museum", 2, false},
	{1, "detroit.museum", 2, false},
	{1, "dinosaur.museum", 2, false},
	{1, "discovery.museum", 2, false},
	{1, "dolls.museum", 2, false},
	{1, "donostia.museum", 2, false},
	{1, "durham.museum", 2, false},
	{1, "eastafrica.museum", 2, false},
	{1, "eastcoast.museum", 2, false},
	{1, "education.museum", 2, false},
	{1, "educational.museum", 2, false},
	{1, "egyptian.museum", 2, false},
	{1, "eisenbahn.museum", 2, false},
	{1, "elburg.museum", 2, false},
	{1, "elvendrell.museum", 2, false},
	{1, "embroidery.museum", 2, false},
	{1, "encyclopedic.museum", 2, false},
	{1, "england.museum", 2, false},
	{1, "entomology.museum", 2, false},
	{1, "environment.museum", 2, false},
	{1, "environmentalconservation.museum", 2, false},
	{1, "epilepsy.museum", 2, false},
	{1, "essex.museum", 2, false},
	{1, "estate.museum", 2, false},
	{1, "ethnology.museum", 2, false},
	{1, "exeter.museum", 2, false},
	{1, "exhibition.museum", 2, false},
	{1, "family.museum", 2, false},
	{1, "farm.museum", 2, false},
	{1, "farmequipment.museum", 2, false},
	{1, "farmers.museum", 2, false},
	{1, "farmstead.museum", 2, false},
	{1, "field.museum", 2, false},
	{1, "figueres.museum", 2, false},
	{1, "filatelia.museum", 2, false},
	{1, "film.museum", 2, false},
	{1, "fineart.museum", 2, false},
	{1, "finearts.museum", 2, false},
	{1, "finland.museum", 2, false},
	{1, "flanders.museum", 2, false},
	{1, "florida.museum", 2, false},
	{1, "force.museum", 2, false},
	{1, "fortmissoula.museum", 2, false},
	{1, "fortworth.museum", 2, false},
	{1, "foundation.museum", 2, false},
	{1, "francaise.museum", 2, false},
	{1, "frankfurt.museum", 2, false},
	{1, "franziskaner.museum", 2, false},
	{1, "freemasonry.museum", 2, false},
	{1, "freiburg.museum", 2, false},
	{1, "fribourg.museum", 2, false},
	{1, "frog.museum", 2, false},
	{1, "fundacio.museum", 2, false},
	{1, "furniture.museum", 2, false},
	{1, "gallery.museum", 2, false},
	{1, "garden.museum", 2, false},
	{1, "gateway.museum", 2, false},
	{1, "geelvinck.museum", 2, false},
	{1, "gemological.museum", 2, false},
	{1, "geology.museum", 2, false},
	{1, "georgia.museum", 2, false},
	{1, "giessen.museum", 2, false},
	{1, "glas.museum", 2, false},
	{1, "glass.museum", 2, false},
	{1, "gorge.museum", 2, false},
	{1, "grandrapids.museum", 2, false},
	{1, "graz.museum", 2, false},
	{1, "guernsey.museum", 2, false},
	{1, "halloffame.museum", 2, false},
	{1, "hamburg.museum", 2, false},
	{1, "handson.museum", 2, false},
	{1, "harvestcelebration.museum", 2, false},
	{1, "hawaii.museum", 2, false},
	{1, "health.museum", 2, false},
	{1, "heimatunduhren.museum", 2, false},
	{1, "hellas.museum", 2, false},
	{1, "helsinki.museum", 2, false},
	{1, "hembygdsforbund.museum", 2, false},
	{1, "heritage.museum", 2, false},
	{1, "histoire.museum", 2, false},
	{1, "historical.museum", 2, false},
	{1, "historicalsociety.museum", 2, false},
	{1, "historichouses.museum", 2, false},
	{1, "historisch.museum", 2, false},
	{1, "historisches.museum", 2, false},
	{1, "history.museum", 2, false},
	{1, "historyofscience.museum", 2, false},
	{1, "horology.museum", 2, false},
	{1, "house.museum", 2, false},
	{1, "humanities.museum", 2, false},
	{1, "illustration.museum", 2, false},
	{1, "imageandsound.museum", 2, false},
	{1, "indian.museum", 2, false},
	{1, "indiana.museum", 2, false},
	{1, "indianapolis.museum", 2, false},
	{1, "indianmarket.museum", 2, false},
	{1, "intelligence.museum", 2, false},
	{1, "interactive.museum", 2, false},
	{1, "iraq.museum", 2, false},
	{1, "iron.museum", 2, false},
	{1, "isleofman.museum", 2, false},
	{1, "jamison.museum", 2, false},
	{1, "jefferson.museum", 2, false},
	{1, "jerusalem.museum", 2, false},
	{1, "jewelry.museum", 2, false},
	{1, "jewish.museum", 2, false},
	{1, "jewishart.museum", 2, false},
	{1, "jfk.museum", 2, false},
	{1, "journalism.museum", 2, false},
	{1, "judaica.museum", 2, false},
	{1, "judygarland.museum", 2, false},
	{1, "juedisches.museum", 2, false},
	{1, "juif.museum", 2, false},
	{1, "karate.museum", 2, false},
	{1, "karikatur.museum", 2, false},
	{1, "kids.museum", 2, false},
	{1, "koebenhavn.museum", 2, false},
	{1, "koeln.museum", 2, false},
	{1, "kunst.museum", 2, false},
	{1, "kunstsammlung.museum", 2, false},
	{1, "kunstunddesign.museum", 2, false},
	{1, "labor.museum", 2, false},
	{1, "labour.museum", 2, false},
	{1, "lajolla.museum", 2, false},
	{1, "lancashire.museum", 2, false},
	{1, "landes.museum", 2, false},
	{1, "lans.museum", 2, false},
	{1, "xn--lns-qla.museum", 2, false},
	{1, "larsson.museum", 2, false},
	{1, "lewismiller.museum", 2, false},
	{1, "lincoln.museum", 2, false},
	{1, "linz.museum", 2, false},
	{1, "living.museum", 2, false},
	{1, "livinghistory.museum", 2, false},
	{1, "localhistory.museum", 2, false},
	{1, "london.museum", 2, false},
	{1, "losangeles.museum", 2, false},
	{1, "louvre.museum", 2, false},
	{1, "loyalist.museum", 2, false},
	{1, "lucerne.museum", 2, false},
	{1, "luxembourg.museum", 2, false},
	{1, "luzern.museum", 2, false},
	{1, "mad.museum", 2, false},
	{1, "madrid.museum", 2, false},
	{1, "mallorca.museum", 2, false},
	{1, "manchester.museum", 2, false},
	{1, "mansion.museum", 2, false},
	{1, "mansions.museum", 2, false},
	{1, "manx.museum", 2, false},
	{1, "marburg.museum", 2, false},
	{1, "maritime.museum", 2, false},
	{1, "maritimo.museum", 2, false},
	{1, "maryland.museum", 2, false},
	{1, "marylhurst.museum", 2, false},
	{1, "media.museum", 2, false},
	{1, "medical.museum", 2, false},
	{1, "medizinhistorisches.museum", 2, false},
	{1, "meeres.museum", 2, false},
	{1, "memorial.museum", 2, false},
	{1, "mesaverde.museum", 2, false},
	{1, "michigan.museum", 2, false},
	{1, "midatlantic.museum", 2, false},
	{1, "military.museum", 2, false},
	{1, "mill.museum", 2, false},
	{1, "miners.museum", 2, false},
	{1, "mining.museum", 2, false},
	{1, "minnesota.museum", 2, false},
	{1, "missile.museum", 2, false},
	{1, "missoula.museum", 2, false},
	{1, "modern.museum", 2, false},
	{1, "moma.museum", 2, false},
	{1, "money.museum", 2, false},
	{1, "monmouth.museum", 2, false},
	{1, "monticello.museum", 2, false},
	{1, "montreal.museum", 2, false},
	{1, "moscow.museum", 2, false},
	{1, "motorcycle.museum", 2, false},
	{1, "muenchen.museum", 2, false},
	{1, "muenster.museum", 2, false},
	{1, "mulhouse.museum", 2, false},
	{1, "muncie.museum", 2, false},
	{1, "museet.museum", 2, false},
	{1, "museumcenter.museum", 2, false},
	{1, "museumvereniging.museum", 2, false},
	{1, "music.museum", 2, false},
	{1, "national.museum", 2, false},
	{1, "nationalfirearms.museum", 2, false},
	{1, "nationalheritage.museum", 2, false},
	{1, "nativeamerican.museum", 2, false},
	{1, "naturalhistory.museum", 2, false},
	{1, "naturalhistorymuseum.museum", 2, false},
	{1, "naturalsciences.museum", 2, false},
	{1, "nature.museum", 2, false},
	{1, "naturhistorisches.museum", 2, false},
	{1, "natuurwetenschappen.museum", 2, false},
	{1, "naumburg.museum", 2, false},
	{1, "naval.museum", 2, false},
	{1, "nebraska.museum", 2, false},
	{1, "neues.museum", 2, false},
	{1, "newhampshire.museum", 2, false},
	{1, "newjersey.museum", 2, false},
	{1, "newmexico.museum", 2, false},
	{1, "newport.museum", 2, false},
	{1, "newspaper.museum", 2, false},
	{1, "newyork.museum", 2, false},
	{1, "niepce.museum", 2, false},
	{1, "norfolk.museum", 2, false},
	{1, "north.museum", 2, false},
	{1, "nrw.museum", 2, false},
	{1, "nyc.museum", 2, false},
	{1, "nyny.museum", 2, false},
	{1, "oceanographic.museum", 2, false},
	{1, "oceanographique.museum", 2, false},
	{1, "omaha.museum", 2, false},
	{1, "online.museum", 2, false},
	{1, "ontario.museum", 2, false},
	{1, "openair.museum", 2, false},
	{1, "oregon.museum", 2, false},
	{1, "oregontrail.museum", 2, false},
	{1, "otago.museum", 2, false},
	{1, "oxford.museum", 2, false},
	{1, "pacific.museum", 2, false},
	{1, "paderborn.museum", 2, false},
	{1, "palace.museum", 2, false},
	{1, "paleo.museum", 2, false},
	{1, "palmsprings.museum", 2, false},
	{1, "panama.museum", 2, false},
	{1, "paris.museum", 2, false},
	{1, "pasadena.museum", 2, false},
	{1, "pharmacy.museum", 2, false},
	{1, "philadelphia.museum", 2, false},
	{1, "philadelphiaarea.museum", 2, false},
	{1, "philately.museum", 2, false},
	{1, "phoenix.museum", 2, false},
	{1, "photography.museum", 2, false},
	{1, "pilots.museum", 2, false},
	{1, "pittsburgh.museum", 2, false},
	{1, "planetarium.museum", 2, false},
	{1, "plantation.museum", 2, false},
	{1, "plants.museum", 2, false},
	{1, "plaza.museum", 2, false},
	{1, "portal.museum", 2, false},
	{1, "portland.museum", 2, false},
	{1, "portlligat.museum", 2, false},
	{1, "posts-and-telecommunications.museum", 2, false},
	{1, "preservation.museum", 2, false},
	{1, "presidio.museum", 2, false},
	{1, "press.museum", 2, false},
	{1, "project.museum", 2, false},
	{1, "public.museum", 2, false},
	{1, "pubol.museum", 2, false},
	{1, "quebec.museum", 2, false},
	{1, "railroad.museum", 2, false},
	{1, "railway.museum", 2, false},
	{1, "research.museum", 2, false},
	{1, "resistance.museum", 2, false},
	{1, "riodejaneiro.museum", 2, false},
	{1, "rochester.museum", 2, false},
	{1, "rockart.museum", 2, false},
	{1, "roma.museum", 2, false},
	{1, "russia.museum", 2, false},
	{1, "saintlouis.museum", 2, false},
	{1, "salem.museum", 2, false},
	{1, "salvadordali.museum", 2, false},
	{1, "salzburg.museum", 2, false},
	{1, "sandiego.museum", 2, false},
	{1, "sanfrancisco.museum", 2, false},
	{1, "santabarbara.museum", 2, false},
	{1, "santacruz.museum", 2, false},
	{1, "santafe.museum", 2, false},
	{1, "saskatchewan.museum", 2, false},
	{1, "satx.museum", 2, false},
	{1, "savannahga.museum", 2, false},
	{1, "schlesisches.museum", 2, false},
	{1, "schoenbrunn.museum", 2, false},
	{1, "schokoladen.museum", 2, false},
	{1, "school.museum", 2, false},
	{1, "schweiz.museum", 2, false},
	{1, "science.museum", 2, false},
	{1, "scienceandhistory.museum", 2, false},
	{1, "scienceandindustry.museum", 2, false},
	{1, "sciencecenter.museum", 2, false},
	{1, "sciencecenters.museum", 2, false},
	{1, "science-fiction.museum", 2, false},
	{1, "sciencehistory.museum", 2, false},
	{1, "sciences.museum", 2, false},
	{1, "sciencesnaturelles.museum", 2, false},
	{1, "scotland.museum", 2, false},
	{1, "seaport.museum", 2, false},
	{1, "settlement.museum", 2, false},
	{1, "settlers.museum", 2, false},
	{1, "shell.museum", 2, false},
	{1, "sherbrooke.museum", 2, false},
	{1, "sibenik.museum", 2, false},
	{1, "silk.museum", 2, false},
	{1, "ski.museum", 2, false},
	{1, "skole.museum", 2, false},
	{1, "society.museum", 2, false},
	{1, "sologne.museum", 2, false},
	{1, "soundandvision.museum", 2, false},
	{1, "southcarolina.museum", 2, false},
	{1, "southwest.museum", 2, false},
	{1, "space.museum", 2, false},
	{1, "spy.museum", 2, false},
	{1, "square.museum", 2, false},
	{1, "stadt.museum", 2, false},
	{1, "stalbans.museum", 2, false},
	{1, "starnberg.museum", 2, false},
	{1, "state.museum", 2, false},
	{1, "stateofdelaware.museum", 2, false},
	{1, "station.museum", 2, false},
	{1, "steam.museum", 2, false},
	{1, "steiermark.museum", 2, false},
	{1, "stjohn.museum", 2, false},
	{1, "stockholm.museum", 2, false},
	{1, "stpetersburg.museum", 2, false},
	{1, "stuttgart.museum", 2, false},
	{1, "suisse.museum", 2, false},
	{1, "surgeonshall.museum", 2, false},
	{1, "surrey.museum", 2, false},
	{1, "svizzera.museum", 2, false},
	{1, "sweden.museum", 2, false},
	{1, "sydney.museum", 2, false},
	{1, "tank.museum", 2, false},
	{1, "tcm.museum", 2, false},
	{1, "technology.museum", 2, false},
	{1, "telekommunikation.museum", 2, false},
	{1, "television.museum", 2, false},
	{1, "texas.museum", 2, false},
	{1, "textile.museum", 2, false},
	{1, "theater.museum", 2, false},
	{1, "time.museum", 2, false},
	{1, "timekeeping.museum", 2, false},
	{1, "topology.museum", 2, false},
	{1, "torino.museum", 2, false},
	{1, "touch.museum", 2, false},
	{1, "town.museum", 2, false},
	{1, "transport.museum", 2, false},
	{1, "tree.museum", 2, false},
	{1, "trolley.museum", 2, false},
	{1, "trust.museum", 2, false},
	{1, "trustee.museum", 2, false},
	{1, "uhren.museum", 2, false},
	{1, "ulm.museum", 2, false},
	{1, "undersea.museum", 2, false},
	{1, "university.museum", 2, false},
	{1, "usa.museum", 2, false},
	{1, "usantiques.museum", 2, false},
	{1, "usarts.museum", 2, false},
	{1, "uscountryestate.museum", 2, false},
	{1, "usculture.museum", 2, false},
	{1, "usdecorativearts.museum", 2, false},
	{1, "usgarden.museum", 2, false},
	{1, "ushistory.museum", 2, false},
	{1, "ushuaia.museum", 2, false},
	{1, "uslivinghistory.museum", 2, false},
	{1, "utah.museum", 2, false},
	{1, "uvic.museum", 2, false},
	{1, "valley.museum", 2, false},
	{1, "vantaa.museum", 2, false},
	{1, "versailles.museum", 2, false},
	{1, "viking.museum", 2, false},
	{1, "village.museum", 2, false},
	{1, "virginia.museum", 2, false},
	{1, "virtual.museum", 2, false},
	{1, "virtuel.museum", 2, false},
	{1, "vlaanderen.museum", 2, false},
	{1, "volkenkunde.museum", 2, false},
	{1, "wales.museum", 2, false},
	{1, "wallonie.museum", 2, false},
	{1, "war.museum", 2, false},
	{1, "washingtondc.museum", 2, false},
	{1, "watchandclock.museum", 2, false},
	{1, "watch-and-clock.museum", 2, false},
	{1, "western.museum", 2, false},
	{1, "westfalen.museum", 2, false},
	{1, "whaling.museum", 2, false},
	{1, "wildlife.museum", 2, false},
	{1, "williamsburg.museum", 2, false},
	{1, "windmill.museum", 2, false},
	{1, "workshop.museum", 2, false},
	{1, "york.museum", 2, false},
	{1, "yorkshire.museum", 2, false},
	{1, "yosemite.museum", 2, false},
	{1, "youth.museum", 2, false},
	{1, "zoological.museum", 2, false},
	{1, "zoology.museum", 2, false},
	{1, "xn--9dbhblg6di.museum", 2, false},
	{1, "xn--h1aegh.museum", 2, false},
	{1, "mv", 1, false},
	{1, "aero.mv", 2, false},
	{1, "biz.mv", 2, false},
	{1, "com.mv", 2, false},
	{1, "coop.mv", 2, false},
	{1, "edu.mv", 2, false},
	{1, "gov.mv", 2, false},
	{1, "info.mv", 2, false},
	{1, "int.mv", 2, false},
	{1, "mil.mv", 2, false},
	{1, "museum.mv", 2, false},
	{1, "name.mv", 2, false},
	{1, "net.mv", 2, false},
	{1, "org.mv", 2, false},
	{1, "pro.mv", 2, false},
	{1, "mw", 1, false},
	{1, "ac.mw", 2, false},
	{1, "biz.mw", 2, false},
	{1, "co.mw", 2, false},
	{1, "com.mw", 2, false},
	{1, "coop.mw", 2, false},
	{1, "edu.mw", 2, false},
	{1, "gov.mw", 2, false},
	{1, "int.mw", 2, false},
	{1, "museum.mw", 2, false},
	{1, "net.mw", 2, false},
	{1, "org.mw", 2, false},
	{1, "mx", 1, false},
	{1, "com.mx", 2, false},
	{1, "org.mx", 2, false},
	{1, "gob.mx", 2, false},
	{1, "edu.mx", 2, false},
	{1, "net.mx", 2, false},
	{1, "my", 1, false},
	{1, "biz.my", 2, false},
	{1, "com.my", 2, false},
	{1, "edu.my", 2, false},
	{1, "gov.my", 2, false},
	{1, "mil.my", 2, false},
	{1, "name.my", 2, false},
	{1, "net.my", 2, false},
	{1, "org.my", 2, false},
	{1, "mz", 1, false},
	{1, "ac.mz", 2, false},
	{1, "adv.mz", 2, false},
	{1, "co.mz", 2, false},
	{1, "edu.mz", 2, false},
	{1, "gov.mz", 2, false},
	{1, "mil.mz", 2, false},
	{1, "net.mz", 2, false},
	{1, "org.mz", 2, false},
	{1, "na", 1, false},
	{1, "info.na", 2, false},
	{1, "pro.na", 2, false},
	{1, "name.na", 2, false},
	{1, "school.na", 2, false},
	{1, "or.na", 2, false},
	{1, "dr.na", 2, false},
	{1, "us.na", 2, false},
	{1, "mx.na", 2, false},
	{1, "ca.na", 2, false},
	{1, "in.na", 2, false},
	{1, "cc.na", 2, false},
	{1, "tv.na", 2, false},
	{1, "ws.na", 2, false},
	{1, "mobi.na", 2, false},
	{1, "co.na", 2, false},
	{1, "com.na", 2, false},
	{1, "org.na", 2, false},
	{1, "name", 1, false},
	{1, "nc", 1, false},
	{1, "asso.nc", 2, false},
	{1, "nom.nc", 2, false},
	{1, "ne", 1, false},
	{1, "net", 1, false},
	{1, "nf", 1, false},
	{1, "com.nf", 2, false},
	{1, "net.nf", 2, false},
	{1, "per.nf", 2, false},
	{1, "rec.nf", 2, false},
	{1, "web.nf", 2, false},
	{1, "arts.nf", 2, false},
	{1, "firm.nf", 2, false},
	{1, "info.nf", 2, false},
	{1, "other.nf", 2, false},
	{1, "store.nf", 2, false},
	{1, "ng", 1, false},
	{1, "com.ng", 2, false},
	{1, "edu.ng", 2, false},
	{1, "gov.ng", 2, false},
	{1, "i.ng", 2, false},
	{1, "mil.ng", 2, false},
	{1, "mobi.ng", 2, false},
	{1, "name.ng", 2, false},
	{1, "net.ng", 2, false},
	{1, "org.ng", 2, false},
	{1, "sch.ng", 2, false},
	{1, "ni", 1, false},
	{1, "ac.ni", 2, false},
	{1, "biz.ni", 2, false},
	{1, "co.ni", 2, false},
	{1, "com.ni", 2, false},
	{1, "edu.ni", 2, false},
	{1, "gob.ni", 2, false},
	{1, "in.ni", 2, false},
	{1, "info.ni", 2, false},
	{1, "int.ni", 2, false},
	{1, "mil.ni", 2, false},
	{1, "net.ni", 2, false},
	{1, "nom.ni", 2, false},
	{1, "org.ni", 2, false},
	{1, "web.ni", 2, false},
	{1, "nl", 1, false},
	{1, "no", 1, false},
	{1, "fhs.no", 2, false},
	{1, "vgs.no", 2, false},
	{1, "fylkesbibl.no", 2, false},
	{1, "folkebibl.no", 2, false},
	{1, "museum.no", 2, false},
	{1, "idrett.no", 2, false},
	{1, "priv.no", 2, false},
	{1, "mil.no", 2, false},
	{1, "stat.no", 2, false},
	{1, "dep.no", 2, false},
	{1, "kommune.no", 2, false},
	{1, "herad.no", 2, false},
	{1, "aa.no", 2, false},
	{1, "ah.no", 2, false},
	{1, "bu.no", 2, false},
	{1, "fm.no", 2, false},
	{1, "hl.no", 2, false},
	{1, "hm.no", 2, false},
	{1, "jan-mayen.no", 2, false},
	{1, "mr.no", 2, false},
	{1, "nl.no", 2, false},
	{1, "nt.no", 2, false},
	{1, "of.no", 2, false},
	{1, "ol.no", 2, false},
	{1, "oslo.no", 2, false},
	{1, "rl.no", 2, false},
	{1, "sf.no", 2, false},
	{1, "st.no", 2, false},
	{1, "svalbard.no", 2, false},
	{1, "tm.no", 2, false},
	{1, "tr.no", 2, false},
	{1, "va.no", 2, false},
	{1, "vf.no", 2, false},
	{1, "gs.aa.no", 3, false},
	{1, "gs.ah.no", 3, false},
	{1, "gs.bu.no", 3, false},
	{1, "gs.fm.no", 3, false},
	{1, "gs.hl.no", 3, false},
	{1, "gs.hm.no", 3, false},
	{1, "gs.jan-mayen.no", 3, false},
	{1, "gs.mr.no", 3, false},
	{1, "gs.nl.no", 3, false},
	{1, "gs.nt.no", 3, false},
	{1, "gs.of.no", 3, false},
	{1, "gs.ol.no", 3, false},
	{1, "gs.oslo.no", 3, false},
	{1, "gs.rl.no", 3, false},
	{1, "gs.sf.no", 3, false},
	{1, "gs.st.no", 3, false},
	{1, "gs.svalbard.no", 3, false},
	{1, "gs.tm.no", 3, false},
	{1, "gs.tr.no", 3, false},
	{1, "gs.va.no", 3, false},
	{1, "gs.vf.no", 3, false},
	{1, "akrehamn.no", 2, false},
	{1, "xn--krehamn-dxa.no", 2, false},
	{1, "algard.no", 2, false},
	{1, "xn--lgrd-poac.no", 2, false},
	{1, "arna.no", 2, false},
	{1, "brumunddal.no", 2, false},
	{1, "bryne.no", 2, false},
	{1, "bronnoysund.no", 2, false},
	{1, "xn--brnnysund-m8ac.no", 2, false},
	{1, "drobak.no", 2, false},
	{1, "xn--drbak-wua.no", 2, false},
	{1, "egersund.no", 2, false},
	{1, "fetsund.no", 2, false},
	{1, "floro.no", 2, false},
	{1, "xn--flor-jra.no", 2, false},
	{1, "fredrikstad.no", 2, false},
	{1, "hokksund.no", 2, false},
	{1, "honefoss.no", 2, false},
	{1, "xn--hnefoss-q1a.no", 2, false},
	{1, "jessheim.no", 2, false},
	{1, "jorpeland.no", 2, false},
	{1, "xn--jrpeland-54a.no", 2, false},
	{1, "kirkenes.no", 2, false},
	{1, "kopervik.no", 2, false},
	{1, "krokstadelva.no", 2, false},
	{1, "langevag.no", 2, false},
	{1, "xn--langevg-jxa.no", 2, false},
	{1, "leirvik.no", 2, false},
	{1, "mjondalen.no", 2, false},
	{1, "xn--mjndalen-64a.no", 2, false},
	{1, "mo-i-rana.no", 2, false},
	{1, "mosjoen.no", 2, false},
	{1, "xn--mosjen-eya.no", 2, false},
	{1, "nesoddtangen.no", 2, false},
	{1, "orkanger.no", 2, false},
	{1, "osoyro.no", 2, false},
	{1, "xn--osyro-wua.no", 2, false},
	{1, "raholt.no", 2, false},
	{1, "xn--rholt-mra.no", 2, false},
	{1, "sandnessjoen.no", 2, false},
	{1, "xn--sandnessjen-ogb.no", 2, false},
	{1, "skedsmokorset.no", 2, false},
	{1, "slattum.no", 2, false},
	{1, "spjelkavik.no", 2, false},
	{1, "stathelle.no", 2, false},
	{1, "stavern.no", 2, false},
	{1, "stjordalshalsen.no", 2, false},
	{1, "xn--stjrdalshalsen-sqb.no", 2, false},
	{1, "tananger.no", 2, false},
	{1, "tranby.no", 2, false},
	{1, "vossevangen.no", 2, false},
	{1, "afjord.no", 2, false},
	{1, "xn--fjord-lra.no", 2, false},
	{1, "agdenes.no", 2, false},
	{1, "al.no", 2, false},
	{1, "xn--l-1fa.no", 2, false},
	{1, "alesund.no", 2, false},
	{1, "xn--lesund-hua.no", 2, false},
	{1, "alstahaug.no", 2, false},
	{1, "alta.no", 2, false},
	{1, "xn--lt-liac.no", 2, false},
	{1, "alaheadju.no", 2, false},
	{1, "xn--laheadju-7ya.no", 2, false},
	{1, "alvdal.no", 2, false},
	{1, "amli.no", 2, false},
	{1, "xn--mli-tla.no", 2, false},
	{1, "amot.no", 2, false},
	{1, "xn--mot-tla.no", 2, false},
	{1, "andebu.no", 2, false},
	{1, "andoy.no", 2, false},
	{1, "xn--andy-ira.no", 2, false},
	{1, "andasuolo.no", 2, false},
	{1, "ardal.no", 2, false},
	{1, "xn--rdal-poa.no", 2, false},
	{1, "aremark.no", 2, false},
	{1, "arendal.no", 2, false},
	{1, "xn--s-1fa.no", 2, false},
	{1, "aseral.no", 2, false},
	{1, "xn--seral-lra.no", 2, false},
	{1, "asker.no", 2, false},
	{1, "askim.no", 2, false},
	{1, "askvoll.no", 2, false},
	{1, "askoy.no", 2, false},
	{1, "xn--asky-ira.no", 2, false},
	{1, "asnes.no", 2, false},
	{1, "xn--snes-poa.no", 2, false},
	{1, "audnedaln.no", 2, false},
	{1, "aukra.no", 2, false},
	{1, "aure.no", 2, false},
	{1, "aurland.no", 2, false},
	{1, "aurskog-holand.no", 2, false},
	{1, "xn--aurskog-hland-jnb.no", 2, false},
	{1, "austevoll.no", 2, false},
	{1, "austrheim.no", 2, false},
	{1, "averoy.no", 2, false},
	{1, "xn--avery-yua.no", 2, false},
	{1, "balestrand.no", 2, false},
	{1, "ballangen.no", 2, false},
	{1, "balat.no", 2, false},
	{1, "xn--blt-elab.no", 2, false},
	{1, "balsfjord.no", 2, false},
	{1, "bahccavuotna.no", 2, false},
	{1, "xn--bhccavuotna-k7a.no", 2, false},
	{1, "bamble.no", 2, false},
	{1, "bardu.no", 2, false},
	{1, "beardu.no", 2, false},
	{1, "beiarn.no", 2, false},
	{1, "bajddar.no", 2, false},
	{1, "xn--bjddar-pta.no", 2, false},
	{1, "baidar.no", 2, false},
	{1, "xn--bidr-5nac.no", 2, false},
	{1, "berg.no", 2, false},
	{1, "bergen.no", 2, false},
	{1, "berlevag.no", 2, false},
	{1, "xn--berlevg-jxa.no", 2, false},
	{1, "bearalvahki.no", 2, false},
	{1, "xn--bearalvhki-y4a.no", 2, false},
	{1, "bindal.no", 2, false},
	{1, "birkenes.no", 2, false},
	{1, "bjarkoy.no", 2, false},
	{1, "xn--bjarky-fya.no", 2, false},
	{1, "bjerkreim.no", 2, false},
	{1, "bjugn.no", 2, false},
	{1, "bodo.no", 2, false},
	{1, "xn--bod-2na.no", 2, false},
	{1, "badaddja.no", 2, false},
	{1, "xn--bdddj-mrabd.no", 2, false},
	{1, "budejju.no", 2, false},
	{1, "bokn.no", 2, false},
	{1, "bremanger.no", 2, false},
	{1, "bronnoy.no", 2, false},
	{1, "xn--brnny-wuac.no", 2, false},
	{1, "bygland.no", 2, false},
	{1, "bykle.no", 2, false},
	{1, "barum.no", 2, false},
	{1, "xn--brum-voa.no", 2, false},
	{1, "bo.telemark.no", 3, false},
	{1, "xn--b-5ga.telemark.no", 3, false},
	{1, "bo.nordland.no", 3, false},
	{1, "xn--b-5ga.nordland.no", 3, false},
	{1, "bievat.no", 2, false},
	{1, "xn--bievt-0qa.no", 2, false},
	{1, "bomlo.no", 2, false},
	{1, "xn--bmlo-gra.no", 2, false},
	{1, "batsfjord.no", 2, false},
	{1, "xn--btsfjord-9za.no", 2, false},
	{1, "bahcavuotna.no", 2, false},
	{1, "xn--bhcavuotna-s4a.no", 2, false},
	{1, "dovre.no", 2, false},
	{1, "drammen.no", 2, false},
	{1, "drangedal.no", 2, false},
	{1, "dyroy.no", 2, false},
	{1, "xn--dyry-ira.no", 2, false},
	{1, "donna.no", 2, false},
	{1, "xn--dnna-gra.no", 2, false},
	{1, "eid.no", 2, false},
	{1, "eidfjord.no", 2, false},
	{1, "eidsberg.no", 2, false},
	{1, "eidskog.no", 2, false},
	{1, "eidsvoll.no", 2, false},
	{1, "eigersund.no", 2, false},
	{1, "elverum.no", 2, false},
	{1, "enebakk.no", 2, false},
	{1, "engerdal.no", 2, false},
	{1, "etne.no", 2, false},
	{1, "etnedal.no", 2, false},
	{1, "evenes.no", 2, false},
	{1, "evenassi.no", 2, false},
	{1, "xn--eveni-0qa01ga.no", 2, false},
	{1, "evje-og-hornnes.no", 2, false},
	{1, "farsund.no", 2, false},
	{1, "fauske.no", 2, false},
	{1, "fuossko.no", 2, false},
	{1, "fuoisku.no", 2, false},
	{1, "fedje.no", 2, false},
	{1, "fet.no", 2, false},
	{1, "finnoy.no", 2, false},
	{1, "xn--finny-yua.no", 2, false},
	{1, "fitjar.no", 2, false},
	{1, "fjaler.no", 2, false},
	{1, "fjell.no", 2, false},
	{1, "flakstad.no", 2, false},
	{1, "flatanger.no", 2, false},
	{1, "flekkefjord.no", 2, false},
	{1, "flesberg.no", 2, false},
	{1, "flora.no", 2, false},
	{1, "fla.no", 2, false},
	{1, "xn--fl-zia.no", 2, false},
	{1, "folldal.no", 2, false},
	{1, "forsand.no", 2, false},
	{1, "fosnes.no", 2, false},
	{1, "frei.no", 2, false},
	{1, "frogn.no", 2, false},
	{1, "froland.no", 2, false},
	{1, "frosta.no", 2, false},
	{1, "frana.no", 2, false},
	{1, "xn--frna-woa.no", 2, false},
	{1, "froya.no", 2, false},
	{1, "xn--frya-hra.no", 2, false},
	{1, "fusa.no", 2, false},
	{1, "fyresdal.no", 2, false},
	{1, "forde.no", 2, false},
	{1, "xn--frde-gra.no", 2, false},
	{1, "gamvik.no", 2, false},
	{1, "gangaviika.no", 2, false},
	{1, "xn--ggaviika-8ya47h.no", 2, false},
	{1, "gaular.no", 2, false},
	{1, "gausdal.no", 2, false},
	{1, "gildeskal.no", 2, false},
	{1, "xn--gildeskl-g0a.no", 2, false},
	{1, "giske.no", 2, false},
	{1, "gjemnes.no", 2, false},
	{1, "gjerdrum.no", 2, false},
	{1, "gjerstad.no", 2, false},
	{1, "gjesdal.no", 2, false},
	{1, "gjovik.no", 2, false},
	{1, "xn--gjvik-wua.no", 2, false},
	{1, "gloppen.no", 2, false},
	{1, "gol.no", 2, false},
	{1, "gran.no", 2, false},
	{1, "grane.no", 2, false},
	{1, "granvin.no", 2, false},
	{1, "gratangen.no", 2, false},
	{1, "grimstad.no", 2, false},
	{1, "grong.no", 2, false},
	{1, "kraanghke.no", 2, false},
	{1, "xn--kranghke-b0a.no", 2, false},
	{1, "grue.no", 2, false},
	{1, "gulen.no", 2, false},
	{1, "hadsel.no", 2, false},
	{1, "halden.no", 2, false},
	{1, "halsa.no", 2, false},
	{1, "hamar.no", 2, false},
	{1, "hamaroy.no", 2, false},
	{1, "habmer.no", 2, false},
	{1, "xn--hbmer-xqa.no", 2, false},
	{1, "hapmir.no", 2, false},
	{1, "xn--hpmir-xqa.no", 2, false},
	{1, "hammerfest.no", 2, false},
	{1, "hammarfeasta.no", 2, false},
	{1, "xn--hmmrfeasta-s4ac.no", 2, false},
	{1, "haram.no", 2, false},
	{1, "hareid.no", 2, false},
	{1, "harstad.no", 2, false},
	{1, "hasvik.no", 2, false},
	{1, "aknoluokta.no", 2, false},
	{1, "xn--koluokta-7ya57h.no", 2, false},
	{1, "hattfjelldal.no", 2, false},
	{1, "aarborte.no", 2, false},
	{1, "haugesund.no", 2, false},
	{1, "hemne.no", 2, false},
	{1, "hemnes.no", 2, false},
	{1, "hemsedal.no", 2, false},
	{1, "heroy.more-og-romsdal.no", 3, false},
	{1, "xn--hery-ira.xn--mre-og-romsdal-qqb.no", 3, false},
	{1, "heroy.nordland.no", 3, false},
	{1, "xn--hery-ira.nordland.no", 3, false},
	{1, "hitra.no", 2, false},
	{1, "hjartdal.no", 2, false},
	{1, "hjelmeland.no", 2, false},
	{1, "hobol.no", 2, false},
	{1, "xn--hobl-ira.no", 2, false},
	{1, "hof.no", 2, false},
	{1, "hol.no", 2, false},
	{1, "hole.no", 2, false},
	{1, "holmestrand.no", 2, false},
	{1, "holtalen.no", 2, false},
	{1, "xn--holtlen-hxa.no", 2, false},
	{1, "hornindal.no", 2, false},
	{1, "horten.no", 2, false},
	{1, "hurdal.no", 2, false},
	{1, "hurum.no", 2, false},
	{1, "hvaler.no", 2, false},
	{1, "hyllestad.no", 2, false},
	{1, "hagebostad.no", 2, false},
	{1, "xn--hgebostad-g3a.no", 2, false},
	{1, "hoyanger.no", 2, false},
	{1, "xn--hyanger-q1a.no", 2, false},
	{1, "hoylandet.no", 2, false},
	{1, "xn--hylandet-54a.no", 2, false},
	{1, "ha.no", 2, false},
	{1, "xn--h-2fa.no", 2, false},
	{1, "ibestad.no", 2, false},
	{1, "inderoy.no", 2, false},
	{1, "xn--indery-fya.no", 2, false},
	{1, "iveland.no", 2, false},
	{1, "jevnaker.no", 2, false},
	{1, "jondal.no", 2, false},
	{1, "jolster.no", 2, false},
	{1, "xn--jlster-bya.no", 2, false},
	{1, "karasjok.no", 2, false},
	{1, "karasjohka.no", 2, false},
	{1, "xn--krjohka-hwab49j.no", 2, false},
	{1, "karlsoy.no", 2, false},
	{1, "galsa.no", 2, false},
	{1, "xn--gls-elac.no", 2, false},
	{1, "karmoy.no", 2, false},
	{1, "xn--karmy-yua.no", 2, false},
	{1, "kautokeino.no", 2, false},
	{1, "guovdageaidnu.no", 2, false},
	{1, "klepp.no", 2, false},
	{1, "klabu.no", 2, false},
	{1, "xn--klbu-woa.no", 2, false},
	{1, "kongsberg.no", 2, false},
	{1, "kongsvinger.no", 2, false},
	{1, "kragero.no", 2, false},
	{1, "xn--krager-gya.no", 2, false},
	{1, "kristiansand.no", 2, false},
	{1, "kristiansund.no", 2, false},
	{1, "krodsherad.no", 2, false},
	{1, "xn--krdsherad-m8a.no", 2, false},
	{1, "kvalsund.no", 2, false},
	{1, "rahkkeravju.no", 2, false},
	{1, "xn--rhkkervju-01af.no", 2, false},
	{1, "kvam.no", 2, false},
	{1, "kvinesdal.no", 2, false},
	{1, "kvinnherad.no", 2, false},
	{1, "kviteseid.no", 2, false},
	{1, "kvitsoy.no", 2, false},
	{1, "xn--kvitsy-fya.no", 2, false},
	{1, "kvafjord.no", 2, false},
	{1, "xn--kvfjord-nxa.no", 2, false},
	{1, "giehtavuoatna.no", 2, false},
	{1, "kvanangen.no", 2, false},
	{1, "xn--kvnangen-k0a.no", 2, false},
	{1, "navuotna.no", 2, false},
	{1, "xn--nvuotna-hwa.no", 2, false},
	{1, "kafjord.no", 2, false},
	{1, "xn--kfjord-iua.no", 2, false},
	{1, "gaivuotna.no", 2, false},
	{1, "xn--givuotna-8ya.no", 2, false},
	{1, "larvik.no", 2, false},
	{1, "lavangen.no", 2, false},
	{1, "lavagis.no", 2, false},
	{1, "loabat.no", 2, false},
	{1, "xn--loabt-0qa.no", 2, false},
	{1, "lebesby.no", 2, false},
	{1, "davvesiida.no", 2, false},
	{1, "leikanger.no", 2, false},
	{1, "leirfjord.no", 2, false},
	{1, "leka.no", 2, false},
	{1, "leksvik.no", 2, false},
	{1, "lenvik.no", 2, false},
	{1, "leangaviika.no", 2, false},
	{1, "xn--leagaviika-52b.no", 2, false},
	{1, "lesja.no", 2, false},
	{1, "levanger.no", 2, false},
	{1, "lier.no", 2, false},
	{1, "lierne.no", 2, false},
	{1, "lillehammer.no", 2, false},
	{1, "lillesand.no", 2, false},
	{1, "lindesnes.no", 2, false},
	{1, "lindas.no", 2, false},
	{1, "xn--linds-pra.no", 2, false},
	{1, "lom.no", 2, false},
	{1, "loppa.no", 2, false},
	{1, "lahppi.no", 2, false},
	{1, "xn--lhppi-xqa.no", 2, false},
	{1, "lund.no", 2, false},
	{1, "lunner.no", 2, false},
	{1, "luroy.no", 2, false},
	{1, "xn--lury-ira.no", 2, false},
	{1, "luster.no", 2, false},
	{1, "lyngdal.no", 2, false},
	{1, "lyngen.no", 2, false},
	{1, "ivgu.no", 2, false},
	{1, "lardal.no", 2, false},
	{1, "lerdal.no", 2, false},
	{1, "xn--lrdal-sra.no", 2, false},
	{1, "lodingen.no", 2, false},
	{1, "xn--ldingen-q1a.no", 2, false},
	{1, "lorenskog.no", 2, false},
	{1, "xn--lrenskog-54a.no", 2, false},
	{1, "loten.no", 2, false},
	{1, "xn--lten-gra.no", 2, false},
	{1, "malvik.no", 2, false},
	{1, "masoy.no", 2, false},
	{1, "xn--msy-ula0h.no", 2, false},
	{1, "muosat.no", 2, false},
	{1, "xn--muost-0qa.no", 2, false},
	{1, "mandal.no", 2, false},
	{1, "marker.no", 2, false},
	{1, "marnardal.no", 2, false},
	{1, "masfjorden.no", 2, false},
	{1, "meland.no", 2, false},
	{1, "meldal.no", 2, false},
	{1, "melhus.no", 2, false},
	{1, "meloy.no", 2, false},
	{1, "xn--mely-ira.no", 2, false},
	{1, "meraker.no", 2, false},
	{1, "xn--merker-kua.no", 2, false},
	{1, "moareke.no", 2, false},
	{1, "xn--moreke-jua.no", 2, false},
	{1, "midsund.no", 2, false},
	{1, "midtre-gauldal.no", 2, false},
	{1, "modalen.no", 2, false},
	{1, "modum.no", 2, false},
	{1, "molde.no", 2, false},
	{1, "moskenes.no", 2, false},
	{1, "moss.no", 2, false},
	{1, "mosvik.no", 2, false},
	{1, "malselv.no", 2, false},
	{1, "xn--mlselv-iua.no", 2, false},
	{1, "malatvuopmi.no", 2, false},
	{1, "xn--mlatvuopmi-s4a.no", 2, false},
	{1, "namdalseid.no", 2, false},
	{1, "aejrie.no", 2, false},
	{1, "namsos.no", 2, false},
	{1, "namsskogan.no", 2, false},
	{1, "naamesjevuemie.no", 2, false},
	{1, "xn--nmesjevuemie-tcba.no", 2, false},
	{1, "laakesvuemie.no", 2, false},
	{1, "nannestad.no", 2, false},
	{1, "narvik.no", 2, false},
	{1, "narviika.no", 2, false},
	{1, "naustdal.no", 2, false},
	{1, "nedre-eiker.no", 2, false},
	{1, "nes.akershus.no", 3, false},
	{1, "nes.buskerud.no", 3, false},
	{1, "nesna.no", 2, false},
	{1, "nesodden.no", 2, false},
	{1, "nesseby.no", 2, false},
	{1, "unjarga.no", 2, false},
	{1, "xn--unjrga-rta.no", 2, false},
	{1, "nesset.no", 2, false},
	{1, "nissedal.no", 2, false},
	{1, "nittedal.no", 2, false},
	{1, "nord-aurdal.no", 2, false},
	{1, "nord-fron.no", 2, false},
	{1, "nord-odal.no", 2, false},
	{1, "norddal.no", 2, false},
	{1, "nordkapp.no", 2, false},
	{1, "davvenjarga.no", 2, false},
	{1, "xn--davvenjrga-y4a.no", 2, false},
	{1, "nordre-land.no", 2, false},
	{1, "nordreisa.no", 2, false},
	{1, "raisa.no", 2, false},
	{1, "xn--risa-5na.no", 2, false},
	{1, "nore-og-uvdal.no", 2, false},
	{1, "notodden.no", 2, false},
	{1, "naroy.no", 2, false},
	{1, "xn--nry-yla5g.no", 2, false},
	{1, "notteroy.no", 2, false},
	{1, "xn--nttery-byae.no", 2, false},
	{1, "odda.no", 2, false},
	{1, "oksnes.no", 2, false},
	{1, "xn--ksnes-uua.no", 2, false},
	{1, "oppdal.no", 2, false},
	{1, "oppegard.no", 2, false},
	{1, "xn--oppegrd-ixa.no", 2, false},
	{1, "orkdal.no", 2, false},
	{1, "orland.no", 2, false},
	{1, "xn--rland-uua.no", 2, false},
	{1, "orskog.no", 2, false},
	{1, "xn--rskog-uua.no", 2, false},
	{1, "orsta.no", 2, false},
	{1, "xn--rsta-fra.no", 2, false},
	{1, "os.hedmark.no", 3, false},
	{1, "os.hordaland.no", 3, false},
	{1, "osen.no", 2, false},
	{1, "osteroy.no", 2, false},
	{1, "xn--ostery-fya.no", 2, false},
	{1, "ostre-toten.no", 2, false},
	{1, "xn--stre-toten-zcb.no", 2, false},
	{1, "overhalla.no", 2, false},
	{1, "ovre-eiker.no", 2, false},
	{1, "xn--vre-eiker-k8a.no", 2, false},
	{1, "oyer.no", 2, false},
	{1, "xn--yer-zna.no", 2, false},
	{1, "oygarden.no", 2, false},
	{1, "xn--ygarden-p1a.no", 2, false},
	{1, "oystre-slidre.no", 2, false},
	{1, "xn--ystre-slidre-ujb.no", 2, false},
	{1, "porsanger.no", 2, false},
	{1, "porsangu.no", 2, false},
	{1, "xn--porsgu-sta26f.no", 2, false},
	{1, "porsgrunn.no", 2, false},
	{1, "radoy.no", 2, false},
	{1, "xn--rady-ira.no", 2, false},
	{1, "rakkestad.no", 2, false},
	{1, "rana.no", 2, false},
	{1, "ruovat.no", 2, false},
	{1, "randaberg.no", 2, false},
	{1, "rauma.no", 2, false},
	{1, "rendalen.no", 2, false},
	{1, "rennebu.no", 2, false},
	{1, "rennesoy.no", 2, false},
	{1, "xn--rennesy-v1a.no", 2, false},
	{1, "rindal.no", 2, false},
	{1, "ringebu.no", 2, false},
	{1, "ringerike.no", 2, false},
	{1, "ringsaker.no", 2, false},
	{1, "rissa.no", 2, false},
	{1, "risor.no", 2, false},
	{1, "xn--risr-ira.no", 2, false},
	{1, "roan.no", 2, false},
	{1, "rollag.no", 2, false},
	{1, "rygge.no", 2, false},
	{1, "ralingen.no", 2, false},
	{1, "xn--rlingen-mxa.no", 2, false},
	{1, "rodoy.no", 2, false},
	{1, "xn--rdy-0nab.no", 2, false},
	{1, "romskog.no", 2, false},
	{1, "xn--rmskog-bya.no", 2, false},
	{1, "roros.no", 2, false},
	{1, "xn--rros-gra.no", 2, false},
	{1, "rost.no", 2, false},
	{1, "xn--rst-0na.no", 2, false},
	{1, "royken.no", 2, false},
	{1, "xn--ryken-vua.no", 2, false},
	{1, "royrvik.no", 2, false},
	{1, "xn--ryrvik-bya.no", 2, false},
	{1, "rade.no", 2, false},
	{1, "xn--rde-ula.no", 2, false},
	{1, "salangen.no", 2, false},
	{1, "siellak.no", 2, false},
	{1, "saltdal.no", 2, false},
	{1, "salat.no", 2, false},
	{1, "xn--slt-elab.no", 2, false},
	{1, "xn--slat-5na.no", 2, false},
	{1, "samnanger.no", 2, false},
	{1, "sande.more-og-romsdal.no", 3, false},
	{1, "sande.xn--mre-og-romsdal-qqb.no", 3, false},
	{1, "sande.vestfold.no", 3, false},
	{1, "sandefjord.no", 2, false},
	{1, "sandnes.no", 2, false},
	{1, "sandoy.no", 2, false},
	{1, "xn--sandy-yua.no", 2, false},
	{1, "sarpsborg.no", 2, false},
	{1, "sauda.no", 2, false},
	{1, "sauherad.no", 2, false},
	{1, "sel.no", 2, false},
	{1, "selbu.no", 2, false},
	{1, "selje.no", 2, false},
	{1, "seljord.no", 2, false},
	{1, "sigdal.no", 2, false},
	{1, "siljan.no", 2, false},
	{1, "sirdal.no", 2, false},
	{1, "skaun.no", 2, false},
	{1, "skedsmo.no", 2, false},
	{1, "ski.no", 2, false},
	{1, "skien.no", 2, false},
	{1, "skiptvet.no", 2, false},
	{1, "skjervoy.no", 2, false},
	{1, "xn--skjervy-v1a.no", 2, false},
	{1, "skierva.no", 2, false},
	{1, "xn--skierv-uta.no", 2, false},
	{1, "skjak.no", 2, false},
	{1, "xn--skjk-soa.no", 2, false},
	{1, "skodje.no", 2, false},
	{1, "skanland.no", 2, false},
	{1, "xn--sknland-fxa.no", 2, false},
	{1, "skanit.no", 2, false},
	{1, "xn--sknit-yqa.no", 2, false},
	{1, "smola.no", 2, false},
	{1, "xn--smla-hra.no", 2, false},
	{1, "snillfjord.no", 2, false},
	{1, "snasa.no", 2, false},
	{1, "xn--snsa-roa.no", 2, false},
	{1, "snoasa.no", 2, false},
	{1, "snaase.no", 2, false},
	{1, "xn--snase-nra.no", 2, false},
	{1, "sogndal.no", 2, false},
	{1, "sokndal.no", 2, false},
	{1, "sola.no", 2, false},
	{1, "solund.no", 2, false},
	{1, "songdalen.no", 2, false},
	{1, "sortland.no", 2, false},
	{1, "spydeberg.no", 2, false},
	{1, "stange.no", 2, false},
	{1, "stavanger.no", 2, false},
	{1, "steigen.no", 2, false},
	{1, "steinkjer.no", 2, false},
	{1, "stjordal.no", 2, false},
	{1, "xn--stjrdal-s1a.no", 2, false},
	{1, "stokke.no", 2, false},
	{1, "stor-elvdal.no", 2, false},
	{1, "stord.no", 2, false},
	{1, "stordal.no", 2, false},
	{1, "storfjord.no", 2, false},
	{1, "omasvuotna.no", 2, false},
	{1, "strand.no", 2, false},
	{1, "stranda.no", 2, false},
	{1, "stryn.no", 2, false},
	{1, "sula.no", 2, false},
	{1, "suldal.no", 2, false},
	{1, "sund.no", 2, false},
	{1, "sunndal.no", 2, false},
	{1, "surnadal.no", 2, false},
	{1, "sveio.no", 2, false},
	{1, "svelvik.no", 2, false},
	{1, "sykkylven.no", 2, false},
	{1, "sogne.no", 2, false},
	{1, "xn--sgne-gra.no", 2, false},
	{1, "somna.no", 2, false},
	{1, "xn--smna-gra.no", 2, false},
	{1, "sondre-land.no", 2, false},
	{1, "xn--sndre-land-0cb.no", 2, false},
	{1, "sor-aurdal.no", 2, false},
	{1, "xn--sr-aurdal-l8a.no", 2, false},
	{1, "sor-fron.no", 2, false},
	{1, "xn--sr-fron-q1a.no", 2, false},
	{1, "sor-odal.no", 2, false},
	{1, "xn--sr-odal-q1a.no", 2, false},
	{1, "sor-varanger.no", 2, false},
	{1, "xn--sr-varanger-ggb.no", 2, false},
	{1, "matta-varjjat.no", 2, false},
	{1, "xn--mtta-vrjjat-k7af.no", 2, false},
	{1, "sorfold.no", 2, false},
	{1, "xn--srfold-bya.no", 2, false},
	{1, "sorreisa.no", 2, false},
	{1, "xn--srreisa-q1a.no", 2, false},
	{1, "sorum.no", 2, false},
	{1, "xn--srum-gra.no", 2, false},
	{1, "tana.no", 2, false},
	{1, "deatnu.no", 2, false},
	{1, "time.no", 2, false},
	{1, "tingvoll.no", 2, false},
	{1, "tinn.no", 2, false},
	{1, "tjeldsund.no", 2, false},
	{1, "dielddanuorri.no", 2, false},
	{1, "tjome.no", 2, false},
	{1, "xn--tjme-hra.no", 2, false},
	{1, "tokke.no", 2, false},
	{1, "tolga.no", 2, false},
	{1, "torsken.no", 2, false},
	{1, "tranoy.no", 2, false},
	{1, "xn--trany-yua.no", 2, false},
	{1, "tromso.no", 2, false},
	{1, "xn--troms-zua.no", 2, false},
	{1, "tromsa.no", 2, false},
	{1, "romsa.no", 2, false},
	{1, "trondheim.no", 2, false},
	{1, "troandin.no", 2, false},
	{1, "trysil.no", 2, false},
	{1, "trana.no", 2, false},
	{1, "xn--trna-woa.no", 2, false},
	{1, "trogstad.no", 2, false},
	{1, "xn--trgstad-r1a.no", 2, false},
	{1, "tvedestrand.no", 2, false},
	{1, "tydal.no", 2, false},
	{1, "tynset.no", 2, false},
	{1, "tysfjord.no", 2, false},
	{1, "divtasvuodna.no", 2, false},
	{1, "divttasvuotna.no", 2, false},
	{1, "tysnes.no", 2, false},
	{1, "tysvar.no", 2, false},
	{1, "xn--tysvr-vra.no", 2, false},
	{1, "tonsberg.no", 2, false},
	{1, "xn--tnsberg-q1a.no", 2, false},
	{1, "ullensaker.no", 2, false},
	{1, "ullensvang.no", 2, false},
	{1, "ulvik.no", 2, false},
	{1, "utsira.no", 2, false},
	{1, "vadso.no", 2, false},
	{1, "xn--vads-jra.no", 2, false},
	{1, "cahcesuolo.no", 2, false},
	{1, "xn--hcesuolo-7ya35b.no", 2, false},
	{1, "vaksdal.no", 2, false},
	{1, "valle.no", 2, false},
	{1, "vang.no", 2, false},
	{1, "vanylven.no", 2, false},
	{1, "vardo.no", 2, false},
	{1, "xn--vard-jra.no", 2, false},
	{1, "varggat.no", 2, false},
	{1, "xn--vrggt-xqad.no", 2, false},
	{1, "vefsn.no", 2, false},
	{1, "vaapste.no", 2, false},
	{1, "vega.no", 2, false},
	{1, "vegarshei.no", 2, false},
	{1, "xn--vegrshei-c0a.no", 2, false},
	{1, "vennesla.no", 2, false},
	{1, "verdal.no", 2, false},
	{1, "verran.no", 2, false},
	{1, "vestby.no", 2, false},
	{1, "vestnes.no", 2, false},
	{1, "vestre-slidre.no", 2, false},
	{1, "vestre-toten.no", 2, false},
	{1, "vestvagoy.no", 2, false},
	{1, "xn--vestvgy-ixa6o.no", 2, false},
	{1, "vevelstad.no", 2, false},
	{1, "vik.no", 2, false},
	{1, "vikna.no", 2, false},
	{1, "vindafjord.no", 2, false},
	{1, "volda.no", 2, false},
	{1, "voss.no", 2, false},
	{1, "varoy.no", 2, false},
	{1, "xn--vry-yla5g.no", 2, false},
	{1, "vagan.no", 2, false},
	{1, "xn--vgan-qoa.no", 2, false},
	{1, "voagat.no", 2, false},
	{1, "vagsoy.no", 2, false},
	{1, "xn--vgsy-qoa0j.no", 2, false},
	{1, "vaga.no", 2, false},
	{1, "xn--vg-yiab.no", 2, false},
	{1, "valer.ostfold.no", 3, false},
	{1, "xn--vler-qoa.xn--stfold-9xa.no", 3, false},
	{1, "valer.hedmark.no", 3, false},
	{1, "xn--vler-qoa.hedmark.no", 3, false},
	{2, "np", 2, false},
	{1, "nr", 1, false},
	{1, "biz.nr", 2, false},
	{1, "info.nr", 2, false},
	{1, "gov.nr", 2, false},
	{1, "edu.nr", 2, false},
	{1, "org.nr", 2, false},
	{1, "net.nr", 2, false},
	{1, "com.nr", 2, false},
	{1, "nu", 1, false},
	{1, "nz", 1, false},
	{1, "ac.nz", 2, false},
	{1, "co.nz", 2, false},
	{1, "cri.nz", 2, false},
	{1, "geek.nz", 2, false},
	{1, "gen.nz", 2, false},
	{1, "govt.nz", 2, false},
	{1, "health.nz", 2, false},
	{1, "iwi.nz", 2, false},
	{1, "kiwi.nz", 2, false},
	{1, "maori.nz", 2, false},
	{1, "mil.nz", 2, false},
	{1, "xn--mori-qsa.nz", 2, false},
	{1, "net.nz", 2, false},
	{1, "org.nz", 2, false},
	{1, "parliament.nz", 2, false},
	{1, "school.nz", 2, false},
	{1, "om", 1, false},
	{1, "co.om", 2, false},
	{1, "com.om", 2, false},
	{1, "edu.om", 2, false},
	{1, "gov.om", 2, false},
	{1, "med.om", 2, false},
	{1, "museum.om", 2, false},
	{1, "net.om", 2, false},
	{1, "org.om", 2, false},
	{1, "pro.om", 2, false},
	{1, "onion", 1, false},
	{1, "org", 1, false},
	{1, "pa", 1, false},
	{1, "ac.pa", 2, false},
	{1, "gob.pa", 2, false},
	{1, "com.pa", 2, false},
	{1, "org.pa", 2, false},
	{1, "sld.pa", 2, false},
	{1, "edu.pa", 2, false},
	{1, "net.pa", 2, false},
	{1, "ing.pa", 2, false},
	{1, "abo.pa", 2, false},
	{1, "med.pa", 2, false},
	{1, "nom.pa", 2, false},
	{1, "pe", 1, false},
	{1, "edu.pe", 2, false},
	{1, "gob.pe", 2, false},
	{1, "nom.pe", 2, false},
	{1, "mil.pe", 2, false},
	{1, "org.pe", 2, false},
	{1, "com.pe", 2, false},
	{1, "net.pe", 2, false},
	{1, "pf", 1, false},
	{1, "com.pf", 2, false},
	{1, "org.pf", 2, false},
	{1, "edu.pf", 2, false},
	{2, "pg", 2, false},
	{1, "ph", 1, false},
	{1, "com.ph", 2, false},
	{1, "net.ph", 2, false},
	{1, "org.ph", 2, false},
	{1, "gov.ph", 2, false},
	{1, "edu.ph", 2, false},
	{1, "ngo.ph", 2, false},
	{1, "mil.ph", 2, false},
	{1, "i.ph", 2, false},
	{1, "pk", 1, false},
	{1, "com.pk", 2, false},
	{1, "net.pk", 2, false},
	{1, "edu.pk", 2, false},
	{1, "org.pk", 2, false},
	{1, "fam.pk", 2, false},
	{1, "biz.pk", 2, false},
	{1, "web.pk", 2, false},
	{1, "gov.pk", 2, false},
	{1, "gob.pk", 2, false},
	{1, "gok.pk", 2, false},
	{1, "gon.pk", 2, false},
	{1, "gop.pk", 2, false},
	{1, "gos.pk", 2, false},
	{1, "info.pk", 2, false},
	{1, "pl", 1, false},
	{1, "com.pl", 2, false},
	{1, "net.pl", 2, false},
	{1, "org.pl", 2, false},
	{1, "aid.pl", 2, false},
	{1, "agro.pl", 2, false},
	{1, "atm.pl", 2, false},
	{1, "auto.pl", 2, false},
	{1, "biz.pl", 2, false},
	{1, "edu.pl", 2, false},
	{1, "gmina.pl", 2, false},
	{1, "gsm.pl", 2, false},
	{1, "info.pl", 2, false},
	{1, "mail.pl", 2, false},
	{1, "miasta.pl", 2, false},
	{1, "media.pl", 2, false},
	{1, "mil.pl", 2, false},
	{1, "nieruchomosci.pl", 2, false},
	{1, "nom.pl", 2, false},
	{1, "pc.pl", 2, false},
	{1, "powiat.pl", 2, false},
	{1, "priv.pl", 2, false},
	{1, "realestate.pl", 2, false},
	{1, "rel.pl", 2, false},
	{1, "sex.pl", 2, false},
	{1, "shop.pl", 2, false},
	{1, "sklep.pl", 2, false},
	{1, "sos.pl", 2, false},
	{1, "szkola.pl", 2, false},
	{1, "targi.pl", 2, false},
	{1, "tm.pl", 2, false},
	{1, "tourism.pl", 2, false},
	{1, "travel.pl", 2, false},
	{1, "turystyka.pl", 2, false},
	{1, "gov.pl", 2, false},
	{1, "ap.gov.pl", 3, false},
	{1, "ic.gov.pl", 3, false},
	{1, "is.gov.pl", 3, false},
	{1, "us.gov.pl", 3, false},
	{1, "kmpsp.gov.pl", 3, false},
	{1, "kppsp.gov.pl", 3, false},
	{1, "kwpsp.gov.pl", 3, false},
	{1, "psp.gov.pl", 3, false},
	{1, "wskr.gov.pl", 3, false},
	{1, "kwp.gov.pl", 3, false},
	{1, "mw.gov.pl", 3, false},
	{1, "ug.gov.pl", 3, false},
	{1, "um.gov.pl", 3, false},
	{1, "umig.gov.pl", 3, false},
	{1, "ugim.gov.pl", 3, false},
	{1, "upow.gov.pl", 3, false},
	{1, "uw.gov.pl", 3, false},
	{1, "starostwo.gov.pl", 3, false},
	{1, "pa.gov.pl", 3, false},
	{1, "po.gov.pl", 3, false},
	{1, "psse.gov.pl", 3, false},
	{1, "pup.gov.pl", 3, false},
	{1, "rzgw.gov.pl", 3, false},
	{1, "sa.gov.pl", 3, false},
	{1, "so.gov.pl", 3, false},
	{1, "sr.gov.pl", 3, false},
	{1, "wsa.gov.pl", 3, false},
	{1, "sko.gov.pl", 3, false},
	{1, "uzs.gov.pl", 3, false},
	{1, "wiih.gov.pl", 3, false},
	{1, "winb.gov.pl", 3, false},
	{1, "pinb.gov.pl", 3, false},
	{1, "wios.gov.pl", 3, false},
	{1, "witd.gov.pl", 3, false},
	{1, "wzmiuw.gov.pl", 3, false},
	{1, "piw.gov.pl", 3, false},
	{1, "wiw.gov.pl", 3, false},
	{1, "griw.gov.pl", 3, false},
	{1, "wif.gov.pl", 3, false},
	{1, "oum.gov.pl", 3, false},
	{1, "sdn.gov.pl", 3, false},
	{1, "zp.gov.pl", 3, false},
	{1, "uppo.gov.pl", 3, false},
	{1, "mup.gov.pl", 3, false},
	{1, "wuoz.gov.pl", 3, false},
	{1, "konsulat.gov.pl", 3, false},
	{1, "oirm.gov.pl", 3, false},
	{1, "augustow.pl", 2, false},
	{1, "babia-gora.pl", 2, false},
	{1, "bedzin.pl", 2, false},
	{1, "beskidy.pl", 2, false},
	{1, "bialowieza.pl", 2, false},
	{1, "bialystok.pl", 2, false},
	{1, "bielawa.pl", 2, false},
	{1, "bieszczady.pl", 2, false},
	{1, "boleslawiec.pl", 2, false},
	{1, "bydgoszcz.pl", 2, false},
	{1, "bytom.pl", 2, false},
	{1, "cieszyn.pl", 2, false},
	{1, "czeladz.pl", 2, false},
	{1, "czest.pl", 2, false},
	{1, "dlugoleka.pl", 2, false},
	{1, "elblag.pl", 2, false},
	{1, "elk.pl", 2, false},
	{1, "glogow.pl", 2, false},
	{1, "gniezno.pl", 2, false},
	{1, "gorlice.pl", 2, false},
	{1, "grajewo.pl", 2, false},
	{1, "ilawa.pl", 2, false},
	{1, "jaworzno.pl", 2, false},
	{1, "jelenia-gora.pl", 2, false},
	{1, "jgora.pl", 2, false},
	{1, "kalisz.pl", 2, false},
	{1, "kazimierz-dolny.pl", 2, false},
	{1, "karpacz.pl", 2, false},
	{1, "kartuzy.pl", 2, false},
	{1, "kaszuby.pl", 2, false},
	{1, "katowice.pl", 2, false},
	{1, "kepno.pl", 2, false},
	{1, "ketrzyn.pl", 2, false},
	{1, "klodzko.pl", 2, false},
	{1, "kobierzyce.pl", 2, false},
	{1, "kolobrzeg.pl", 2, false},
	{1, "konin.pl", 2, false},
	{1, "konskowola.pl", 2, false},
	{1, "kutno.pl", 2, false},
	{1, "lapy.pl", 2, false},
	{1, "lebork.pl", 2, false},
	{1, "legnica.pl", 2, false},
	{1, "lezajsk.pl", 2, false},
	{1, "limanowa.pl", 2, false},
	{1, "lomza.pl", 2, false},
	{1, "lowicz.pl", 2, false},
	{1, "lubin.pl", 2, false},
	{1, "lukow.pl", 2, false},
	{1, "malbork.pl", 2, false},
	{1, "malopolska.pl", 2, false},
	{1, "mazowsze.pl", 2, false},
	{1, "mazury.pl", 2, false},
	{1, "mielec.pl", 2, false},
	{1, "mielno.pl", 2, false},
	{1, "mragowo.pl", 2, false},
	{1, "naklo.pl", 2, false},
	{1, "nowaruda.pl", 2, false},
	{1, "nysa.pl", 2, false},
	{1, "olawa.pl", 2, false},
	{1, "olecko.pl", 2, false},
	{1, "olkusz.pl", 2, false},
	{1, "olsztyn.pl", 2, false},
	{1, "opoczno.pl", 2, false},
	{1, "opole.pl", 2, false},
	{1, "ostroda.pl", 2, false},
	{1, "ostroleka.pl", 2, false},
	{1, "ostrowiec.pl", 2, false},
	{1, "ostrowwlkp.pl", 2, false},
	{1, "pila.pl", 2, false},
	{1, "pisz.pl", 2, false},
	{1, "podhale.pl", 2, false},
	{1, "podlasie.pl", 2, false},
	{1, "polkowice.pl", 2, false},
	{1, "pomorze.pl", 2, false},
	{1, "pomorskie.pl", 2, false},
	{1, "prochowice.pl", 2, false},
	{1, "pruszkow.pl", 2, false},
	{1, "przeworsk.pl", 2, false},
	{1, "pulawy.pl", 2, false},
	{1, "radom.pl", 2, false},
	{1, "rawa-maz.pl", 2, false},
	{1, "rybnik.pl", 2, false},
	{1, "rzeszow.pl", 2, false},
	{1, "sanok.pl", 2, false},
	{1, "sejny.pl", 2, false},
	{1, "slask.pl", 2, false},
	{1, "slupsk.pl", 2, false},
	{1, "sosnowiec.pl", 2, false},
	{1, "stalowa-wola.pl", 2, false},
	{1, "skoczow.pl", 2, false},
	{1, "starachowice.pl", 2, false},
	{1, "stargard.pl", 2, false},
	{1, "suwalki.pl", 2, false},
	{1, "swidnica.pl", 2, false},
	{1, "swiebodzin.pl", 2, false},
	{1, "swinoujscie.pl", 2, false},
	{1, "szczecin.pl", 2, false},
	{1, "szczytno.pl", 2, false},
	{1, "tarnobrzeg.pl", 2, false},
	{1, "tgory.pl", 2, false},
	{1, "turek.pl", 2, false},
	{1, "tychy.pl", 2, false},
	{1, "ustka.pl", 2, false},
	{1, "walbrzych.pl", 2, false},
	{1, "warmia.pl", 2, false},
	{1, "warszawa.pl", 2, false},
	{1, "waw.pl", 2, false},
	{1, "wegrow.pl", 2, false},
	{1, "wielun.pl", 2, false},
	{1, "wlocl.pl", 2, false},
	{1, "wloclawek.pl", 2, false},
	{1, "wodzislaw.pl", 2, false},
	{1, "wolomin.pl", 2, false},
	{1, "wroclaw.pl", 2, false},
	{1, "zachpomor.pl", 2, false},
	{1, "zagan.pl", 2, false},
	{1, "zarow.pl", 2, false},
	{1, "zgora.pl", 2, false},
	{1, "zgorzelec.pl", 2, false},
	{1, "pm", 1, false},
	{1, "pn", 1, false},
	{1, "gov.pn", 2, false},
	{1, "co.pn", 2, false},
	{1, "org.pn", 2, false},
	{1, "edu.pn", 2, false},
	{1, "net.pn", 2, false},
	{1, "post", 1, false},
	{1, "pr", 1, false},
	{1, "com.pr", 2, false},
	{1, "net.pr", 2, false},
	{1, "org.pr", 2, false},
	{1, "gov.pr", 2, false},
	{1, "edu.pr", 2, false},
	{1, "isla.pr", 2, false},
	{1, "pro.pr", 2, false},
	{1, "biz.pr", 2, false},
	{1, "info.pr", 2, false},
	{1, "name.pr", 2, false},
	{1, "est.pr", 2, false},
	{1, "prof.pr", 2, false},
	{1, "ac.pr", 2, false},
	{1, "pro", 1, false},
	{1, "aaa.pro", 2, false},
	{1, "aca.pro", 2, false},
	{1, "acct.pro", 2, false},
	{1, "avocat.pro", 2, false},
	{1, "bar.pro", 2, false},
	{1, "cpa.pro", 2, false},
	{1, "eng.pro", 2, false},
	{1, "jur.pro", 2, false},
	{1, "law.pro", 2, false},
	{1, "med.pro", 2, false},
	{1, "recht.pro", 2, false},
	{1, "ps", 1, false},
	{1, "edu.ps", 2, false},
	{1, "gov.ps", 2, false},
	{1, "sec.ps", 2, false},
	{1, "plo.ps", 2, false},
	{1, "com.ps", 2, false},
	{1, "org.ps", 2, false},
	{1, "net.ps", 2, false},
	{1, "pt", 1, false},
	{1, "net.pt", 2, false},
	{1, "gov.pt", 2, false},
	{1, "org.pt", 2, false},
	{1, "edu.pt", 2, false},
	{1, "int.pt", 2, false},
	{1, "publ.pt", 2, false},
	{1, "com.pt", 2, false},
	{1, "nome.pt", 2, false},
	{1, "pw", 1, false},
	{1, "co.pw", 2, false},
	{1, "ne.pw", 2, false},
	{1, "or.pw", 2, false},
	{1, "ed.pw", 2, false},
	{1, "go.pw", 2, false},
	{1, "belau.pw", 2, false},
	{1, "py", 1, false},
	{1, "com.py", 2, false},
	{1, "coop.py", 2, false},
	{1, "edu.py", 2, false},
	{1, "gov.py", 2, false},
	{1, "mil.py", 2, false},
	{1, "net.py", 2, false},
	{1, "org.py", 2, false},
	{1, "qa", 1, false},
	{1, "com.qa", 2, false},
	{1, "edu.qa", 2, false},
	{1, "gov.qa", 2, false},
	{1, "mil.qa", 2, false},
	{1, "name.qa", 2, false},
	{1, "net.qa", 2, false},
	{1, "org.qa", 2, false},
	{1, "sch.qa", 2, false},
	{1, "re", 1, false},
	{1, "asso.re", 2, false},
	{1, "com.re", 2, false},
	{1, "nom.re", 2, false},
	{1, "ro", 1, false},
	{1, "arts.ro", 2, false},
	{1, "com.ro", 2, false},
	{1, "firm.ro", 2, false},
	{1, "info.ro", 2, false},
	{1, "nom.ro", 2, false},
	{1, "nt.ro", 2, false},
	{1, "org.ro", 2, false},
	{1, "rec.ro", 2, false},
	{1, "store.ro", 2, false},
	{1, "tm.ro", 2, false},
	{1, "www.ro", 2, false},
	{1, "rs", 1, false},
	{1, "ac.rs", 2, false},
	{1, "co.rs", 2, false},
	{1, "edu.rs", 2, false},
	{1, "gov.rs", 2, false},
	{1, "in.rs", 2, false},
	{1, "org.rs", 2, false},
	{1, "ru", 1, false},
	{1, "rw", 1, false},
	{1, "ac.rw", 2, false},
	{1, "co.rw", 2, false},
	{1, "coop.rw", 2, false},
	{1, "gov.rw", 2, false},
	{1, "mil.rw", 2, false},
	{1, "net.rw", 2, false},
	{1, "org.rw", 2, false},
	{1, "sa", 1, false},
	{1, "com.sa", 2, false},
	{1, "net.sa", 2, false},
	{1, "org.sa", 2, false},
	{1, "gov.sa", 2, false},
	{1, "med.sa", 2, false},
	{1, "pub.sa", 2, false},
	{1, "edu.sa", 2, false},
	{1, "sch.sa", 2, false},
	{1, "sb", 1, false},
	{1, "com.sb", 2, false},
	{1, "edu.sb", 2, false},
	{1, "gov.sb", 2, false},
	{1, "net.sb", 2, false},
	{1, "org.sb", 2, false},
	{1, "sc", 1, false},
	{1, "com.sc", 2, false},
	{1, "gov.sc", 2, false},
	{1, "net.sc", 2, false},
	{1, "org.sc", 2, false},
	{1, "edu.sc", 2, false},
	{1, "sd", 1, false},
	{1, "com.sd", 2, false},
	{1, "net.sd", 2, false},
	{1, "org.sd", 2, false},
	{1, "edu.sd", 2, false},
	{1, "med.sd", 2, false},
	{1, "tv.sd", 2, false},
	{1, "gov.sd", 2, false},
	{1, "info.sd", 2, false},
	{1, "se", 1, false},
	{1, "a.se", 2, false},
	{1, "ac.se", 2, false},
	{1, "b.se", 2, false},
	{1, "bd.se", 2, false},
	{1, "brand.se", 2, false},
	{1, "c.se", 2, false},
	{1, "d.se", 2, false},
	{1, "e.se", 2, false},
	{1, "f.se", 2, false},
	{1, "fh.se", 2, false},
	{1, "fhsk.se", 2, false},
	{1, "fhv.se", 2, false},
	{1, "g.se", 2, false},
	{1, "h.se", 2, false},
	{1, "i.se", 2, false},
	{1, "k.se", 2, false},
	{1, "komforb.se", 2, false},
	{1, "kommunalforbund.se", 2, false},
	{1, "komvux.se", 2, false},
	{1, "l.se", 2, false},
	{1, "lanbib.se", 2, false},
	{1, "m.se", 2, false},
	{1, "n.se", 2, false},
	{1, "naturbruksgymn.se", 2, false},
	{1, "o.se", 2, false},
	{1, "org.se", 2, false},
	{1, "p.se", 2, false},
	{1, "parti.se", 2, false},
	{1, "pp.se", 2, false},
	{1, "press.se", 2, false},
	{1, "r.se", 2, false},
	{1, "s.se", 2, false},
	{1, "t.se", 2, false},
	{1, "tm.se", 2, false},
	{1, "u.se", 2, false},
	{1, "w.se", 2, false},
	{1, "x.se", 2, false},
	{1, "y.se", 2, false},
	{1, "z.se", 2, false},
	{1, "sg", 1, false},
	{1, "com.sg", 2, false},
	{1, "net.sg", 2, false},
	{1, "org.sg", 2, false},
	{1, "gov.sg", 2, false},
	{1, "edu.sg", 2, false},
	{1, "per.sg", 2, false},
	{1, "sh", 1, false},
	{1, "com.sh", 2, false},
	{1, "net.sh", 2, false},
	{1, "gov.sh", 2, false},
	{1, "org.sh", 2, false},
	{1, "mil.sh", 2, false},
	{1, "si", 1, false},
	{1, "sj", 1, false},
	{1, "sk", 1, false},
	{1, "sl", 1, false},
	{1, "com.sl", 2, false},
	{1, "net.sl", 2, false},
	{1, "edu.sl", 2, false},
	{1, "gov.sl", 2, false},
	{1, "org.sl", 2, false},
	{1, "sm", 1, false},
	{1, "sn", 1, false},
	{1, "art.sn", 2, false},
	{1, "com.sn", 2, false},
	{1, "edu.sn", 2, false},
	{1, "gouv.sn", 2, false},
	{1, "org.sn", 2, false},
	{1, "perso.sn", 2, false},
	{1, "univ.sn", 2, false},
	{1, "so", 1, false},
	{1, "com.so", 2, false},
	{1, "edu.so", 2, false},
	{1, "gov.so", 2, false},
	{1, "me.so", 2, false},
	{1, "net.so", 2, false},
	{1, "org.so", 2, false},
	{1, "sr", 1, false},
	{1, "ss", 1, false},
	{1, "biz.ss", 2, false},
	{1, "com.ss", 2, false},
	{1, "edu.ss", 2, false},
	{1, "gov.ss", 2, false},
	{1, "me.ss", 2, false},
	{1, "net.ss", 2, false},
	{1, "org.ss", 2, false},
	{1, "sch.ss", 2, false},
	{1, "st", 1, false},
	{1, "co.st", 2, false},
	{1, "com.st", 2, false},
	{1, "consulado.st", 2, false},
	{1, "edu.st", 2, false},
	{1, "embaixada.st", 2, false},
	{1, "mil.st", 2, false},
	{1, "net.st", 2, false},
	{1, "org.st", 2, false},
	{1, "principe.st", 2, false},
	{1, "saotome.st", 2, false},
	{1, "store.st", 2, false},
	{1, "su", 1, false},
	{1, "sv", 1, false},
	{1, "com.sv", 2, false},
	{1, "edu.sv", 2, false},
	{1, "gob.sv", 2, false},
	{1, "org.sv", 2, false},
	{1, "red.sv", 2, false},
	{1, "sx", 1, false},
	{1, "gov.sx", 2, false},
	{1, "sy", 1, false},
	{1, "edu.sy", 2, false},
	{1, "gov.sy", 2, false},
	{1, "net.sy", 2, false},
	{1, "mil.sy", 2, false},
	{1, "com.sy", 2, false},
	{1, "org.sy", 2, false},
	{1, "sz", 1, false},
	{1, "co.sz", 2, false},
	{1, "ac.sz", 2, false},
	{1, "org.sz", 2, false},
	{1, "tc", 1, false},
	{1, "td", 1, false},
	{1, "tel", 1, false},
	{1, "tf", 1, false},
	{1, "tg", 1, false},
	{1, "th", 1, false},
	{1, "ac.th", 2, false},
	{1, "co.th", 2, false},
	{1, "go.th", 2, false},
	{1, "in.th", 2, false},
	{1, "mi.th", 2, false},
	{1, "net.th", 2, false},
	{1, "or.th", 2, false},
	{1, "tj", 1, false},
	{1, "ac.tj", 2, false},
	{1, "biz.tj", 2, false},
	{1, "co.tj", 2, false},
	{1, "com.tj", 2, false},
	{1, "edu.tj", 2, false},
	{1, "go.tj", 2, false},
	{1, "gov.tj", 2, false},
	{1, "int.tj", 2, false},
	{1, "mil.tj", 2, false},
	{1, "name.tj", 2, false},
	{1, "net.tj", 2, false},
	{1, "nic.tj", 2, false},
	{1, "org.tj", 2, false},
	{1, "test.tj", 2, false},
	{1, "web.tj", 2, false},
	{1, "tk", 1, false},
	{1, "tl", 1, false},
	{1, "gov.tl", 2, false},
	{1, "tm", 1, false},
	{1, "com.tm", 2, false},
	{1, "co.tm", 2, false},
	{1, "org.tm", 2, false},
	{1, "net.tm", 2, false},
	{1, "nom.tm", 2, false},
	{1, "gov.tm", 2, false},
	{1, "mil.tm", 2, false},
	{1, "edu.tm", 2, false},
	{1, "tn", 1, false},
	{1, "com.tn", 2, false},
	{1, "ens.tn", 2, false},
	{1, "fin.tn", 2, false},
	{1, "gov.tn", 2, false},
	{1, "ind.tn", 2, false},
	{1, "intl.tn", 2, false},
	{1, "nat.tn", 2, false},
	{1, "net.tn", 2, false},
	{1, "org.tn", 2, false},
	{1, "info.tn", 2, false},
	{1, "perso.tn", 2, false},
	{1, "tourism.tn", 2, false},
	{1, "edunet.tn", 2, false},
	{1, "rnrt.tn", 2, false},
	{1, "rns.tn", 2, false},
	{1, "rnu.tn", 2, false},
	{1, "mincom.tn", 2, false},
	{1, "agrinet.tn", 2, false},
	{1, "defense.tn", 2, false},
	{1, "turen.tn", 2, false},
	{1, "to", 1, false},
	{1, "com.to", 2, false},
	{1, "gov.to", 2, false},
	{1, "net.to", 2, false},
	{1, "org.to", 2, false},
	{1, "edu.to", 2, false},
	{1, "mil.to", 2, false},
	{1, "tr", 1, false},
	{1, "av.tr", 2, false},
	{1, "bbs.tr", 2, false},
	{1, "bel.tr", 2, false},
	{1, "biz.tr", 2, false},
	{1, "com.tr", 2, false},
	{1, "dr.tr", 2, false},
	{1, "edu.tr", 2, false},
	{1, "gen.tr", 2, false},
	{1, "gov.tr", 2, false},
	{1, "info.tr", 2, false},
	{1, "mil.tr", 2, false},
	{1, "k12.tr", 2, false},
	{1, "kep.tr", 2, false},
	{1, "name.tr", 2, false},
	{1, "net.tr", 2, false},
	{1, "org.tr", 2, false},
	{1, "pol.tr", 2, false},
	{1, "tel.tr", 2, false},
	{1, "tsk.tr", 2, false},
	{1, "tv.tr", 2, false},
	{1, "web.tr", 2, false},
	{1, "nc.tr", 2, false},
	{1, "gov.nc.tr", 3, false},
	{1, "tt", 1, false},
	{1, "co.tt", 2, false},
	{1, "com.tt", 2, false},
	{1, "org.tt", 2, false},
	{1, "net.tt", 2, false},
	{1, "biz.tt", 2, false},
	{1, "info.tt", 2, false},
	{1, "pro.tt", 2, false},
	{1, "int.tt", 2, false},
	{1, "coop.tt", 2, false},
	{1, "jobs.tt", 2, false},
	{1, "mobi.tt", 2, false},
	{1, "travel.tt", 2, false},
	{1, "museum.tt", 2, false},
	{1, "aero.tt", 2, false},
	{1, "name.tt", 2, false},
	{1, "gov.tt", 2, false},
	{1, "edu.tt", 2, false},
	{1, "tv", 1, false},
	{1, "tw", 1, false},
	{1, "edu.tw", 2, false},
	{1, "gov.tw", 2, false},
	{1, "mil.tw", 2, false},
	{1, "com.tw", 2, false},
	{1, "net.tw", 2, false},
	{1, "org.tw", 2, false},
	{1, "idv.tw", 2, false},
	{1, "game.tw", 2, false},
	{1, "ebiz.tw", 2, false},
	{1, "club.tw", 2, false},
	{1, "xn--zf0ao64a.tw", 2, false},
	{1, "xn--uc0atv.tw", 2, false},
	{1, "xn--czrw28b.tw", 2, false},
	{1, "tz", 1, false},
	{1, "ac.tz", 2, false},
	{1, "co.tz", 2, false},
	{1, "go.tz", 2, false},
	{1, "hotel.tz", 2, false},
	{1, "info.tz", 2, false},
	{1, "me.tz", 2, false},
	{1, "mil.tz", 2, false},
	{1, "mobi.tz", 2, false},
	{1, "ne.tz", 2, false},
	{1, "or.tz", 2, false},
	{1, "sc.tz", 2, false},
	{1, "tv.tz", 2, false},
	{1, "ua", 1, false},
	{1, "com.ua", 2, false},
	{1, "edu.ua", 2, false},
	{1, "gov.ua", 2, false},
	{1, "in.ua", 2, false},
	{1, "net.ua", 2, false},
	{1, "org.ua", 2, false},
	{1, "cherkassy.ua", 2, false},
	{1, "cherkasy.ua", 2, false},
	{1, "chernigov.ua", 2, false},
	{1, "chernihiv.ua", 2, false},
	{1, "chernivtsi.ua", 2, false},
	{1, "chernovtsy.ua", 2, false},
	{1, "ck.ua", 2, false},
	{1, "cn.ua", 2, false},
	{1, "cr.ua", 2, false},
	{1, "crimea.ua", 2, false},
	{1, "cv.ua", 2, false},
	{1, "dn.ua", 2, false},
	{1, "dnepropetrovsk.ua", 2, false},
	{1, "dnipropetrovsk.ua", 2, false},
	{1, "donetsk.ua", 2, false},
	{1, "dp.ua", 2, false},
	{1, "if.ua", 2, false},
	{1, "ivano-frankivsk.ua", 2, false},
	{1, "kh.ua", 2, false},
	{1, "kharkiv.ua", 2, false},
	{1, "kharkov.ua", 2, false},
	{1, "kherson.ua", 2, false},
	{1, "khmelnitskiy.ua", 2, false},
	{1, "khmelnytskyi.ua", 2, false},
	{1, "kiev.ua", 2, false},
	{1, "kirovograd.ua", 2, false},
	{1, "km.ua", 2, false},
	{1, "kr.ua", 2, false},
	{1, "krym.ua", 2, false},
	{1, "ks.ua", 2, false},
	{1, "kv.ua", 2, false},
	{1, "kyiv.ua", 2, false},
	{1, "lg.ua", 2, false},
	{1, "lt.ua", 2, false},
	{1, "lugansk.ua", 2, false},
	{1, "lutsk.ua", 2, false},
	{1, "lv.ua", 2, false},
	{1, "lviv.ua", 2, false},
	{1, "mk.ua", 2, false},
	{1, "mykolaiv.ua", 2, false},
	{1, "nikolaev.ua", 2, false},
	{1, "od.ua", 2, false},
	{1, "odesa.ua", 2, false},
	{1, "odessa.ua", 2, false},
	{1, "pl.ua", 2, false},
	{1, "poltava.ua", 2, false},
	{1, "rivne.ua", 2, false},
	{1, "rovno.ua", 2, false},
	{1, "rv.ua", 2, false},
	{1, "sb.ua", 2, false},
	{1, "sebastopol.ua", 2, false},
	{1, "sevastopol.ua", 2, false},
	{1, "sm.ua", 2, false},
	{1, "sumy.ua", 2, false},
	{1, "te.ua", 2, false},
	{1, "ternopil.ua", 2, false},
	{1, "uz.ua", 2, false},
	{1, "uzhgorod.ua", 2, false},
	{1, "vinnica.ua", 2, false},
	{1, "vinnytsia.ua", 2, false},
	{1, "vn.ua", 2, false},
	{1, "volyn.ua", 2, false},
	{1, "yalta.ua", 2, false},
	{1, "zaporizhzhe.ua", 2, false},
	{1, "zaporizhzhia.ua", 2, false},
	{1, "zhitomir.ua", 2, false},
	{1, "zhytomyr.ua", 2, false},
	{1, "zp.ua", 2, false},
	{1, "zt.ua", 2, false},
	{1, "ug", 1, false},
	{1, "co.ug", 2, false},
	{1, "or.ug", 2, false},
	{1, "ac.ug", 2, false},
	{1, "sc.ug", 2, false},
	{1, "go.ug", 2, false},
	{1, "ne.ug", 2, false},
	{1, "com.ug", 2, false},
	{1, "org.ug", 2, false},
	{1, "uk", 1, false},
	{1, "ac.uk", 2, false},
	{1, "co.uk", 2, false},
	{1, "gov.uk", 2, false},
	{1, "ltd.uk", 2, false},
	{1, "me.uk", 2, false},
	{1, "net.uk", 2, false},
	{1, "nhs.uk", 2, false},
	{1, "org.uk", 2, false},
	{1, "plc.uk", 2, false},
	{1, "police.uk", 2, false},
	{2, "sch.uk", 3, false},
	{1, "us", 1, false},
	{1, "dni.us", 2, false},
	{1, "fed.us", 2, false},
	{1, "isa.us", 2, false},
	{1, "kids.us", 2, false},
	{1, "nsn.us", 2, false},
	{1, "ak.us", 2, false},
	{1, "al.us", 2, false},
	{1, "ar.us", 2, false},
	{1, "as.us", 2, false},
	{1, "az.us", 2, false},
	{1, "ca.us", 2, false},
	{1, "co.us", 2, false},
	{1, "ct.us", 2, false},
	{1, "dc.us", 2, false},
	{1, "de.us", 2, false},
	{1, "fl.us", 2, false},
	{1, "ga.us", 2, false},
	{1, "gu.us", 2, false},
	{1, "hi.us", 2, false},
	{1, "ia.us", 2, false},
	{1, "id.us", 2, false},
	{1, "il.us", 2, false},
	{1, "in.us", 2, false},
	{1, "ks.us", 2, false},
	{1, "ky.us", 2, false},
	{1, "la.us", 2, false},
	{1, "ma.us", 2, false},
	{1, "md.us", 2, false},
	{1, "me.us", 2, false},
	{1, "mi.us", 2, false},
	{1, "mn.us", 2, false},
	{1, "mo.us", 2, false},
	{1, "ms.us", 2, false},
	{1, "mt.us", 2, false},
	{1, "nc.us", 2, false},
	{1, "nd.us", 2, false},
	{1, "ne.us", 2, false},
	{1, "nh.us", 2, false},
	{1, "nj.us", 2, false},
	{1, "nm.us", 2, false},
	{1, "nv.us", 2, false},
	{1, "ny.us", 2, false},
	{1, "oh.us", 2, false},
	{1, "ok.us", 2, false},
	{1, "or.us", 2, false},
	{1, "pa.us", 2, false},
	{1, "pr.us", 2, false},
	{1, "ri.us", 2, false},
	{1, "sc.us", 2, false},
	{1, "sd.us", 2, false},
	{1, "tn.us", 2, false},
	{1, "tx.us", 2, false},
	{1, "ut.us", 2, false},
	{1, "vi.us", 2, false},
	{1, "vt.us", 2, false},
	{1, "va.us", 2, false},
	{1, "wa.us", 2, false},
	{1, "wi.us", 2, false},
	{1, "wv.us", 2, false},
	{1, "wy.us", 2, false},
	{1, "k12.ak.us", 3, false},
	{1, "k12.al.us", 3, false},
	{1, "k12.ar.us", 3, false},
	{1, "k12.as.us", 3, false},
	{1, "k12.az.us", 3, false},
	{1, "k12.ca.us", 3, false},
	{1, "k12.co.us", 3, false},
	{1, "k12.ct.us", 3, false},
	{1, "k12.dc.us", 3, false},
	{1, "k12.de.us", 3, false},
	{1, "k12.fl.us", 3, false},
	{1, "k12.ga.us", 3, false},
	{1, "k12.gu.us", 3, false},
	{1, "k12.ia.us", 3, false},
	{1, "k12.id.us", 3, false},
	{1, "k12.il.us", 3, false},
	{1, "k12.in.us", 3, false},
	{1, "k12.ks.us", 3, false},
	{1, "k12.ky.us", 3, false},
	{1, "k12.la.us", 3, false},
	{1, "k12.ma.us", 3, false},
	{1, "k12.md.us", 3, false},
	{1, "k12.me.us", 3, false},
	{1, "k12.mi.us", 3, false},
	{1, "k12.mn.us", 3, false},
	{1, "k12.mo.us", 3, false},
	{1, "k12.ms.us", 3, false},
	{1, "k12.mt.us", 3, false},
	{1, "k12.nc.us", 3, false},
	{1, "k12.ne.us", 3, false},
	{1, "k12.nh.us", 3, false},
	{1, "k12.nj.us", 3, false},
	{1, "k12.nm.us", 3, false},
	{1, "k12.nv.us", 3, false},
	{1, "k12.ny.us", 3, false},
	{1, "k12.oh.us", 3, false},
	{1, "k12.ok.us", 3, false},
	{1, "k12.or.us", 3, false},
	{1, "k12.pa.us", 3, false},
	{1, "k12.pr.us", 3, false},
	{1, "k12.sc.us", 3, false},
	{1, "k12.tn.us", 3, false},
	{1, "k12.tx.us", 3, false},
	{1, "k12.ut.us", 3, false},
	{1, "k12.vi.us", 3, false},
	{1, "k12.vt.us", 3, false},
	{1, "k12.va.us", 3, false},
	{1, "k12.wa.us", 3, false},
	{1, "k12.wi.us", 3, false},
	{1, "k12.wy.us", 3, false},
	{1, "cc.ak.us", 3, false},
	{1, "cc.al.us", 3, false},
	{1, "cc.ar.us", 3, false},
	{1, "cc.as.us", 3, false},
	{1, "cc.az.us", 3, false},
	{1, "cc.ca.us", 3, false},
	{1, "cc.co.us", 3, false},
	{1, "cc.ct.us", 3, false},
	{1, "cc.dc.us", 3, false},
	{1, "cc.de.us", 3, false},
	{1, "cc.fl.us", 3, false},
	{1, "cc.ga.us", 3, false},
	{1, "cc.gu.us", 3, false},
	{1, "cc.hi.us", 3, false},
	{1, "cc.ia.us", 3, false},
	{1, "cc.id.us", 3, false},
	{1, "cc.il.us", 3, false},
	{1, "cc.in.us", 3, false},
	{1, "cc.ks.us", 3, false},
	{1, "cc.ky.us", 3, false},
	{1, "cc.la.us", 3, false},
	{1, "cc.ma.us", 3, false},
	{1, "cc.md.us", 3, false},
	{1, "cc.me.us", 3, false},
	{1, "cc.mi.us", 3, false},
	{1, "cc.mn.us", 3, false},
	{1, "cc.mo.us", 3, false},
	{1, "cc.ms.us", 3, false},
	{1, "cc.mt.us", 3, false},
	{1, "cc.nc.us", 3, false},
	{1, "cc.nd.us", 3, false},
	{1, "cc.ne.us", 3, false},
	{1, "cc.nh.us", 3, false},
	{1, "cc.nj.us", 3, false},
	{1, "cc.nm.us", 3, false},
	{1, "cc.nv.us", 3, false},
	{1, "cc.ny.us", 3, false},
	{1, "cc.oh.us", 3, false},
	{1, "cc.ok.us", 3, false},
	{1, "cc.or.us", 3, false},
	{1, "cc.pa.us", 3, false},
	{1, "cc.pr.us", 3, false},
	{1, "cc.ri.us", 3, false},
	{1, "cc.sc.us", 3, false},
	{1, "cc.sd.us", 3, false},
	{1, "cc.tn.us", 3, false},
	{1, "cc.tx.us", 3, false},
	{1, "cc.ut.us", 3, false},
	{1, "cc.vi.us", 3, false},
	{1, "cc.vt.us", 3, false},
	{1, "cc.va.us", 3, false},
	{1, "cc.wa.us", 3, false},
	{1, "cc.wi.us", 3, false},
	{1, "cc.wv.us", 3, false},
	{1, "cc.wy.us", 3, false},
	{1, "lib.ak.us", 3, false},
	{1, "lib.al.us", 3, false},
	{1, "lib.ar.us", 3, false},
	{1, "lib.as.us", 3, false},
	{1, "lib.az.us", 3, false},
	{1, "lib.ca.us", 3, false},
	{1, "lib.co.us", 3, false},
	{1, "lib.ct.us", 3, false},
	{1, "lib.dc.us", 3, false},
	{1, "lib.fl.us", 3, false},
	{1, "lib.ga.us", 3, false},
	{1, "lib.gu.us", 3, false},
	{1, "lib.hi.us", 3, false},
	{1, "lib.ia.us", 3, false},
	{1, "lib.id.us", 3, false},
	{1, "lib.il.us", 3, false},
	{1, "lib.in.us", 3, false},
	{1, "lib.ks.us", 3, false},
	{1, "lib.ky.us", 3, false},
	{1, "lib.la.us", 3, false},
	{1, "lib.ma.us", 3, false},
	{1, "lib.md.us", 3, false},
	{1, "lib.me.us", 3, false},
	{1, "lib.mi.us", 3, false},
	{1, "lib.mn.us", 3, false},
	{1, "lib.mo.us", 3, false},
	{1, "lib.ms.us", 3, false},
	{1, "lib.mt.us", 3, false},
	{1, "lib.nc.us", 3, false},
	{1, "lib.nd.us", 3, false},
	{1, "lib.ne.us", 3, false},
	{1, "lib.nh.us", 3, false},
	{1, "lib.nj.us", 3, false},
	{1, "lib.nm.us", 3, false},
	{1, "lib.nv.us", 3, false},
	{1, "lib.ny.us", 3, false},
	{1, "lib.oh.us", 3, false},
	{1, "lib.ok.us", 3, false},
	{1, "lib.or.us", 3, false},
	{1, "lib.pa.us", 3, false},
	{1, "lib.pr.us", 3, false},
	{1, "lib.ri.us", 3, false},
	{1, "lib.sc.us", 3, false},
	{1, "lib.sd.us", 3, false},
	{1, "lib.tn.us", 3, false},
	{1, "lib.tx.us", 3, false},
	{1, "lib.ut.us", 3, false},
	{1, "lib.vi.us", 3, false},
	{1, "lib.vt.us", 3, false},
	{1, "lib.va.us", 3, false},
	{1, "lib.wa.us", 3, false},
	{1, "lib.wi.us", 3, false},
	{1, "lib.wy.us", 3, false},
	{1, "pvt.k12.ma.us", 4, false},
	{1, "chtr.k12.ma.us", 4, false},
	{1, "paroch.k12.ma.us", 4, false},
	{1, "ann-arbor.mi.us", 3, false},
	{1, "cog.mi.us", 3, false},
	{1, "dst.mi.us", 3, false},
	{1, "eaton.mi.us", 3, false},
	{1, "gen.mi.us", 3, false},
	{1, "mus.mi.us", 3, false},
	{1, "tec.mi.us", 3, false},
	{1, "washtenaw.mi.us", 3, false},
	{1, "uy", 1, false},
	{1, "com.uy", 2, false},
	{1, "edu.uy", 2, false},
	{1, "gub.uy", 2, false},
	{1, "mil.uy", 2, false},
	{1, "net.uy", 2, false},
	{1, "org.uy", 2, false},
	{1, "uz", 1, false},
	{1, "co.uz", 2, false},
	{1, "com.uz", 2, false},
	{1, "net.uz", 2, false},
	{1, "org.uz", 2, false},
	{1, "va", 1, false},
	{1, "vc", 1, false},
	{1, "com.vc", 2, false},
	{1, "net.vc", 2, false},
	{1, "org.vc", 2, false},
	{1, "gov.vc", 2, false},
	{1, "mil.vc", 2, false},
	{1, "edu.vc", 2, false},
	{1, "ve", 1, false},
	{1, "arts.ve", 2, false},
	{1, "co.ve", 2, false},
	{1, "com.ve", 2, false},
	{1, "e12.ve", 2, false},
	{1, "edu.ve", 2, false},
	{1, "firm.ve", 2, false},
	{1, "gob.ve", 2, false},
	{1, "gov.ve", 2, false},
	{1, "info.ve", 2, false},
	{1, "int.ve", 2, false},
	{1, "mil.ve", 2, false},
	{1, "net.ve", 2, false},
	{1, "org.ve", 2, false},
	{1, "rec.ve", 2, false},
	{1, "store.ve", 2, false},
	{1, "tec.ve", 2, false},
	{1, "web.ve", 2, false},
	{1, "vg", 1, false},
	{1, "vi", 1, false},
	{1, "co.vi", 2, false},
	{1, "com.vi", 2, false},
	{1, "k12.vi", 2, false},
	{1, "net.vi", 2, false},
	{1, "org.vi", 2, false},
	{1, "vn", 1, false},
	{1, "com.vn", 2, false},
	{1, "net.vn", 2, false},
	{1, "org.vn", 2, false},
	{1, "edu.vn", 2, false},
	{1, "gov.vn", 2, false},
	{1, "int.vn", 2, false},
	{1, "ac.vn", 2, false},
	{1, "biz.vn", 2, false},
	{1, "info.vn", 2, false},
	{1, "name.vn", 2, false},
	{1, "pro.vn", 2, false},
	{1, "health.vn", 2, false},
	{1, "vu", 1, false},
	{1, "com.vu", 2, false},
	{1, "edu.vu", 2, false},
	{1, "net.vu", 2, false},
	{1, "org.vu", 2, false},
	{1, "wf", 1, false},
	{1, "ws", 1, false},
	{1, "com.ws", 2, false},
	{1, "net.ws", 2, false},
	{1, "org.ws", 2, false},
	{1, "gov.ws", 2, false},
	{1, "edu.ws", 2, false},
	{1, "yt", 1, false},
	{1, "xn--mgbaam7a8h", 1, false},
	{1, "xn--y9a3aq", 1, false},
	{1, "xn--54b7fta0cc", 1, false},
	{1, "xn--90ae", 1, false},
	{1, "xn--mgbcpq6gpa1a", 1, false},
	{1, "xn--90ais", 1, false},
	{1, "xn--fiqs8s", 1, false},
	{1, "xn--fiqz9s", 1, false},
	{1, "xn--lgbbat1ad8j", 1, false},
	{1, "xn--wgbh1c", 1, false},
	{1, "xn--e1a4c", 1, false},
	{1, "xn--qxa6a", 1, false},
	{1, "xn--mgbah1a3hjkrd", 1, false},
	{1, "xn--node", 1, false},
	{1, "xn--qxam", 1, false},
	{1, "xn--j6w193g", 1, false},
	{1, "xn--55qx5d.xn--j6w193g", 2, false},
	{1, "xn--wcvs22d.xn--j6w193g", 2, false},
	{1, "xn--mxtq1m.xn--j6w193g", 2, false},
	{1, "xn--gmqw5a.xn--j6w193g", 2, false},
	{1, "xn--od0alg.xn--j6w193g", 2, false},
	{1, "xn--uc0atv.xn--j6w193g", 2, false},
	{1, "xn--2scrj9c", 1, false},
	{1, "xn--3hcrj9c", 1, false},
	{1, "xn--45br5cyl", 1, false},
	{1, "xn--h2breg3eve", 1, false},
	{1, "xn--h2brj9c8c", 1, false},
	{1, "xn--mgbgu82a", 1, false},
	{1, "xn--rvc1e0am3e", 1, false},
	{1, "xn--h2brj9c", 1, false},
	{1, "xn--mgbbh1a", 1, false},
	{1, "xn--mgbbh1a71e", 1, false},
	{1, "xn--fpcrj9c3d", 1, false},
	{1, "xn--gecrj9c", 1, false},
	{1, "xn--s9brj9c", 1, false},
	{1, "xn--45brj9c", 1, false},
	{1, "xn--xkc2dl3a5ee0h", 1, false},
	{1, "xn--mgba3a4f16a", 1, false},
	{1, "xn--mgba3a4fra", 1, false},
	{1, "xn--mgbtx2b", 1, false},
	{1, "xn--mgbayh7gpa", 1, false},
	{1, "xn--3e0b707e", 1, false},
	{1, "xn--80ao21a", 1, false},
	{1, "xn--q7ce6a", 1, false},
	{1, "xn--fzc2c9e2c", 1, false},
	{1, "xn--xkc2al3hye2a", 1, false},
	{1, "xn--mgbc0a9azcg", 1, false},
	{1, "xn--d1alf", 1, false},
	{1, "xn--l1acc", 1, false},
	{1, "xn--mix891f", 1, false},
	{1, "xn--mix082f", 1, false},
	{1, "xn--mgbx4cd0ab", 1, false},
	{1, "xn--mgb9awbf", 1, false},
	{1, "xn--mgbai9azgqp6j", 1, false},
	{1, "xn--mgbai9a5eva00b", 1, false},
	{1, "xn--ygbi2ammx", 1, false},
	{1, "xn--90a3ac", 1, false},
	{1, "xn--o1ac.xn--90a3ac", 2, false},
	{1, "xn--c1avg.xn--90a3ac", 2, false},
	{1, "xn--90azh.xn--90a3ac", 2, false},
	{1, "xn--d1at.xn--90a3ac", 2, false},
	{1, "xn--o1ach.xn--90a3ac", 2, false},
	{1, "xn--80au.xn--90a3ac", 2, false},
	{1, "xn--p1ai", 1, false},
	{1, "xn--wgbl6a", 1, false},
	{1, "xn--mgberp4a5d4ar", 1, false},
	{1, "xn--mgberp4a5d4a87g", 1, false},
	{1, "xn--mgbqly7c0a67fbc", 1, false},
	{1, "xn--mgbqly7cvafr", 1, false},
	{1, "xn--mgbpl2fh", 1, false},
	{1, "xn--yfro4i67o", 1, false},
	{1, "xn--clchc0ea0b2g2a9gcd", 1, false},
	{1, "xn--ogbpf8fl", 1, false},
	{1, "xn--mgbtf8fl", 1, false},
	{1, "xn--o3cw4h", 1, false},
	{1, "xn--12c1fe0br.xn--o3cw4h", 2, false},
	{1, "xn--12co0c3b4eva.xn--o3cw4h", 2, false},
	{1, "xn--h3cuzk1di.xn--o3cw4h", 2, false},
	{1, "xn--o3cyx2a.xn--o3cw4h", 2, false},
	{1, "xn--m3ch0j3a.xn--o3cw4h", 2, false},
	{1, "xn--12cfi8ixb8l.xn--o3cw4h", 2, false},
	{1, "xn--pgbs0dh", 1, false},
	{1, "xn--kpry57d", 1, false},
	{1, "xn--kprw13d", 1, false},
	{1, "xn--nnx388a", 1, false},
	{1, "xn--j1amh", 1, false},
	{1, "xn--mgb2ddes", 1, false},
	{1, "xxx", 1, false},
	{1, "ye", 1, false},
	{1, "com.ye", 2, false},
	{1, "edu.ye", 2, false},
	{1, "gov.ye", 2, false},
	{1, "net.ye", 2, false},
	{1, "mil.ye", 2, false},
	{1, "org.ye", 2, false},
	{1, "ac.za", 2, false},
	{1, "agric.za", 2, false},
	{1, "alt.za", 2, false},
	{1, "co.za", 2, false},
	{1, "edu.za", 2, false},
	{1, "gov.za", 2, false},
	{1, "grondar.za", 2, false},
	{1, "law.za", 2, false},
	{1, "mil.za", 2, false},
	{1, "net.za", 2, false},
	{1, "ngo.za", 2, false},
	{1, "nic.za", 2, false},
	{1, "nis.za", 2, false},
	{1, "nom.za", 2, false},
	{1, "org.za", 2, false},
	{1, "school.za", 2, false},
	{1, "tm.za", 2, false},
	{1, "web.za", 2, false},
	{1, "zm", 1, false},
	{1, "ac.zm", 2, false},
	{1, "biz.zm", 2, false},
	{1, "co.zm", 2, false},
	{1, "com.zm", 2, false},
	{1, "edu.zm", 2, false},
	{1, "gov.zm", 2, false},
	{1, "info.zm", 2, false},
	{1, "mil.zm", 2, false},
	{1, "net.zm", 2, false},
	{1, "org.zm", 2, false},
	{1, "sch.zm", 2, false},
	{1, "zw", 1, false},
	{1, "ac.zw", 2, false},
	{1, "co.zw", 2, false},
	{1, "gov.zw", 2, false},
	{1, "mil.zw", 2, false},
	{1, "org.zw", 2, false},
	{1, "aaa", 1, false},
	{1, "aarp", 1, false},
	{1, "abarth", 1, false},
	{1, "abb", 1, false},
	{1, "abbott", 1, false},
	{1, "abbvie", 1, false},
	{1, "abc", 1, false},
	{1, "able", 1, false},
	{1, "abogado", 1, false},
	{1, "abudhabi", 1, false},
	{1, "academy", 1, false},
	{1, "accenture", 1, false},
	{1, "accountant", 1, false},
	{1, "accountants", 1, false},
	{1, "aco", 1, false},
	{1, "actor", 1, false},
	{1, "adac", 1, false},
	{1, "ads", 1, false},
	{1, "adult", 1, false},
	{1, "aeg", 1, false},
	{1, "aetna", 1, false},
	{1, "afamilycompany", 1, false},
	{1, "afl", 1, false},
	{1, "africa", 1, false},
	{1, "agakhan", 1, false},
	{1, "agency", 1, false},
	{1, "aig", 1, false},
	{1, "airbus", 1, false},
	{1, "airforce", 1, false},
	{1, "airtel", 1, false},
	{1, "akdn", 1, false},
	{1, "alfaromeo", 1, false},
	{1, "alibaba", 1, false},
	{1, "alipay", 1, false},
	{1, "allfinanz", 1, false},
	{1, "allstate", 1, false},
	{1, "ally", 1, false},
	{1, "alsace", 1, false},
	{1, "alstom", 1, false},
	{1, "amazon", 1, false},
	{1, "americanexpress", 1, false},
	{1, "americanfamily", 1, false},
	{1, "amex", 1, false},
	{1, "amfam", 1, false},
	{1, "amica", 1, false},
	{1, "amsterdam", 1, false},
	{1, "analytics", 1, false},
	{1, "android", 1, false},
	{1, "anquan", 1, false},
	{1, "anz", 1, false},
	{1, "aol", 1, false},
	{1, "apartments", 1, false},
	{1, "app", 1, false},
	{1, "apple", 1, false},
	{1, "aquarelle", 1, false},
	{1, "arab", 1, false},
	{1, "aramco", 1, false},
	{1, "archi", 1, false},
	{1, "army", 1, false},
	{1, "art", 1, false},
	{1, "arte", 1, false},
	{1, "asda", 1, false},
	{1, "associates", 1, false},
	{1, "athleta", 1, false},
	{1, "attorney", 1, false},
	{1, "auction", 1, false},
	{1, "audi", 1, false},
	{1, "audible", 1, false},
	{1, "audio", 1, false},
	{1, "auspost", 1, false},
	{1, "author", 1, false},
	{1, "auto", 1, false},
	{1, "autos", 1, false},
	{1, "avianca", 1, false},
	{1, "aws", 1, false},
	{1, "axa", 1, false},
	{1, "azure", 1, false},
	{1, "baby", 1, false},
	{1, "baidu", 1, false},
	{1, "banamex", 1, false},
	{1, "bananarepublic", 1, false},
	{1, "band", 1, false},
	{1, "bank", 1, false},
	{1, "bar", 1, false},
	{1, "barcelona", 1, false},
	{1, "barclaycard", 1, false},
	{1, "barclays", 1, false},
	{1, "barefoot", 1, false},
	{1, "bargains", 1, false},
	{1, "baseball", 1, false},
	{1, "basketball", 1, false},
	{1, "bauhaus", 1, false},
	{1, "bayern", 1, false},
	{1, "bbc", 1, false},
	{1, "bbt", 1, false},
	{1, "bbva", 1, false},
	{1, "bcg", 1, false},
	{1, "bcn", 1, false},
	{1, "beats", 1, false},
	{1, "beauty", 1, false},
	{1, "beer", 1, false},
	{1, "bentley", 1, false},
	{1, "berlin", 1, false},
	{1, "best", 1, false},
	{1, "bestbuy", 1, false},
	{1, "bet", 1, false},
	{1, "bharti", 1, false},
	{1, "bible", 1, false},
	{1, "bid", 1, false},
	{1, "bike", 1, false},
	{1, "bing", 1, false},
	{1, "bingo", 1, false},
	{1, "bio", 1, false},
	{1, "black", 1, false},
	{1, "blackfriday", 1, false},
	{1, "blockbuster", 1, false},
	{1, "blog", 1, false},
	{1, "bloomberg", 1, false},
	{1, "blue", 1, false},
	{1, "bms", 1, false},
	{1, "bmw", 1, false},
	{1, "bnpparibas", 1, false},
	{1, "boats", 1, false},
	{1, "boehringer", 1, false},
	{1, "bofa", 1, false},
	{1, "bom", 1, false},
	{1, "bond", 1, false},
	{1, "boo", 1, false},
	{1, "book", 1, false},
	{1, "booking", 1, false},
	{1, "bosch", 1, false},
	{1, "bostik", 1, false},
	{1, "boston", 1, false},
	{1, "bot", 1, false},
	{1, "boutique", 1, false},
	{1, "box", 1, false},
	{1, "bradesco", 1, false},
	{1, "bridgestone", 1, false},
	{1, "broadway", 1, false},
	{1, "broker", 1, false},
	{1, "brother", 1, false},
	{1, "brussels", 1, false},
	{1, "budapest", 1, false},
	{1, "bugatti", 1, false},
	{1, "build", 1, false},
	{1, "builders", 1, false},
	{1, "business", 1, false},
	{1, "buy", 1, false},
	{1, "buzz", 1, false},
	{1, "bzh", 1, false},
	{1, "cab", 1, false},
	{1, "cafe", 1, false},
	{1, "cal", 1, false},
	{1, "call", 1, false},
	{1, "calvinklein", 1, false},
	{1, "cam", 1, false},
	{1, "camera", 1, false},
	{1, "camp", 1, false},
	{1, "cancerresearch", 1, false},
	{1, "canon", 1, false},
	{1, "capetown", 1, false},
	{1, "capital", 1, false},
	{1, "capitalone", 1, false},
	{1, "car", 1, false},
	{1, "caravan", 1, false},
	{1, "cards", 1, false},
	{1, "care", 1, false},
	{1, "career", 1, false},
	{1, "careers", 1, false},
	{1, "cars", 1, false},
	{1, "casa", 1, false},
	{1, "case", 1, false},
	{1, "cash", 1, false},
	{1, "casino", 1, false},
	{1, "catering", 1, false},
	{1, "catholic", 1, false},
	{1, "cba", 1, false},
	{1, "cbn", 1, false},
	{1, "cbre", 1, false},
	{1, "cbs", 1, false},
	{1, "center", 1, false},
	{1, "ceo", 1, false},
	{1, "cern", 1, false},
	{1, "cfa", 1, false},
	{1, "cfd", 1, false},
	{1, "chanel", 1, false},
	{1, "channel", 1, false},
	{1, "charity", 1, false},
	{1, "chase", 1, false},
	{1, "chat", 1, false},
	{1, "cheap", 1, false},
	{1, "chintai", 1, false},
	{1, "christmas", 1, false},
	{1, "chrome", 1, false},
	{1, "church", 1, false},
	{1, "cipriani", 1, false},
	{1, "circle", 1, false},
	{1, "cisco", 1, false},
	{1, "citadel", 1, false},
	{1, "citi", 1, false},
	{1, "citic", 1, false},
	{1, "city", 1, false},
	{1, "cityeats", 1, false},
	{1, "claims", 1, false},
	{1, "cleaning", 1, false},
	{1, "click", 1, false},
	{1, "clinic", 1, false},
	{1, "clinique", 1, false},
	{1, "clothing", 1, false},
	{1, "cloud", 1, false},
	{1, "club", 1, false},
	{1, "clubmed", 1, false},
	{1, "coach", 1, false},
	{1, "codes", 1, false},
	{1, "coffee", 1, false},
	{1, "college", 1, false},
	{1, "cologne", 1, false},
	{1, "comcast", 1, false},
	{1, "commbank", 1, false},
	{1, "community", 1, false},
	{1, "company", 1, false},
	{1, "compare", 1, false},
	{1, "computer", 1, false},
	{1, "comsec", 1, false},
	{1, "condos", 1, false},
	{1, "construction", 1, false},
	{1, "consulting", 1, false},
	{1, "contact", 1, false},
	{1, "contractors", 1, false},
	{1, "cooking", 1, false},
	{1, "cookingchannel", 1, false},
	{1, "cool", 1, false},
	{1, "corsica", 1, false},
	{1, "country", 1, false},
	{1, "coupon", 1, false},
	{1, "coupons", 1, false},
	{1, "courses", 1, false},
	{1, "cpa", 1, false},
	{1, "credit", 1, false},
	{1, "creditcard", 1, false},
	{1, "creditunion", 1, false},
	{1, "cricket", 1, false},
	{1, "crown", 1, false},
	{1, "crs", 1, false},
	{1, "cruise", 1, false},
	{1, "cruises", 1, false},
	{1, "csc", 1, false},
	{1, "cuisinella", 1, false},
	{1, "cymru", 1, false},
	{1, "cyou", 1, false},
	{1, "dabur", 1, false},
	{1, "dad", 1, false},
	{1, "dance", 1, false},
	{1, "data", 1, false},
	{1, "date", 1, false},
	{1, "dating", 1, false},
	{1, "datsun", 1, false},
	{1, "day", 1, false},
	{1, "dclk", 1, false},
	{1, "dds", 1, false},
	{1, "deal", 1, false},
	{1, "dealer", 1, false},
	{1, "deals", 1, false},
	{1, "degree", 1, false},
	{1, "delivery", 1, false},
	{1, "dell", 1, false},
	{1, "deloitte", 1, false},
	{1, "delta", 1, false},
	{1, "democrat", 1, false},
	{1, "dental", 1, false},
	{1, "dentist", 1, false},
	{1, "desi", 1, false},
	{1, "design", 1, false},
	{1, "dev", 1, false},
	{1, "dhl", 1, false},
	{1, "diamonds", 1, false},
	{1, "diet", 1, false},
	{1, "digital", 1, false},
	{1, "direct", 1, false},
	{1, "directory", 1, false},
	{1, "discount", 1, false},
	{1, "discover", 1, false},
	{1, "dish", 1, false},
	{1, "diy", 1, false},
	{1, "dnp", 1, false},
	{1, "docs", 1, false},
	{1, "doctor", 1, false},
	{1, "dog", 1, false},
	{1, "domains", 1, false},
	{1, "dot", 1, false},
	{1, "download", 1, false},
	{1, "drive", 1, false},
	{1, "dtv", 1, false},
	{1, "dubai", 1, false},
	{1, "duck", 1, false},
	{1, "dunlop", 1, false},
	{1, "dupont", 1, false},
	{1, "durban", 1, false},
	{1, "dvag", 1, false},
	{1, "dvr", 1, false},
	{1, "earth", 1, false},
	{1, "eat", 1, false},
	{1, "eco", 1, false},
	{1, "edeka", 1, false},
	{1, "education", 1, false},
	{1, "email", 1, false},
	{1, "emerck", 1, false},
	{1, "energy", 1, false},
	{1, "engineer", 1, false},
	{1, "engineering", 1, false},
	{1, "enterprises", 1, false},
	{1, "epson", 1, false},
	{1, "equipment", 1, false},
	{1, "ericsson", 1, false},
	{1, "erni", 1, false},
	{1, "esq", 1, false},
	{1, "estate", 1, false},
	{1, "etisalat", 1, false},
	{1, "eurovision", 1, false},
	{1, "eus", 1, false},
	{1, "events", 1, false},
	{1, "exchange", 1, false},
	{1, "expert", 1, false},
	{1, "exposed", 1, false},
	{1, "express", 1, false},
	{1, "extraspace", 1, false},
	{1, "fage", 1, false},
	{1, "fail", 1, false},
	{1, "fairwinds", 1, false},
	{1, "faith", 1, false},
	{1, "family", 1, false},
	{1, "fan", 1, false},
	{1, "fans", 1, false},
	{1, "farm", 1, false},
	{1, "farmers", 1, false},
	{1, "fashion", 1, false},
	{1, "fast", 1, false},
	{1, "fedex", 1, false},
	{1, "feedback", 1, false},
	{1, "ferrari", 1, false},
	{1, "ferrero", 1, false},
	{1, "fiat", 1, false},
	{1, "fidelity", 1, false},
	{1, "fido", 1, false},
	{1, "film", 1, false},
	{1, "final", 1, false},
	{1, "finance", 1, false},
	{1, "financial", 1, false},
	{1, "fire", 1, false},
	{1, "firestone", 1, false},
	{1, "firmdale", 1, false},
	{1, "fish", 1, false},
	{1, "fishing", 1, false},
	{1, "fit", 1, false},
	{1, "fitness", 1, false},
	{1, "flickr", 1, false},
	{1, "flights", 1, false},
	{1, "flir", 1, false},
	{1, "florist", 1, false},
	{1, "flowers", 1, false},
	{1, "fly", 1, false},
	{1, "foo", 1, false},
	{1, "food", 1, false},
	{1, "foodnetwork", 1, false},
	{1, "football", 1, false},
	{1, "ford", 1, false},
	{1, "forex", 1, false},
	{1, "forsale", 1, false},
	{1, "forum", 1, false},
	{1, "foundation", 1, false},
	{1, "fox", 1, false},
	{1, "free", 1, false},
	{1, "fresenius", 1, false},
	{1, "frl", 1, false},
	{1, "frogans", 1, false},
	{1, "frontdoor", 1, false},
	{1, "frontier", 1, false},
	{1, "ftr", 1, false},
	{1, "fujitsu", 1, false},
	{1, "fun", 1, false},
	{1, "fund", 1, false},
	{1, "furniture", 1, false},
	{1, "futbol", 1, false},
	{1, "fyi", 1, false},
	{1, "gal", 1, false},
	{1, "gallery", 1, false},
	{1, "gallo", 1, false},
	{1, "gallup", 1, false},
	{1, "game", 1, false},
	{1, "games", 1, false},
	{1, "gap", 1, false},
	{1, "garden", 1, false},
	{1, "gay", 1, false},
	{1, "gbiz", 1, false},
	{1, "gdn", 1, false},
	{1, "gea", 1, false},
	{1, "gent", 1, false},
	{1, "genting", 1, false},
	{1, "george", 1, false},
	{1, "ggee", 1, false},
	{1, "gift", 1, false},
	{1, "gifts", 1, false},
	{1, "gives", 1, false},
	{1, "giving", 1, false},
	{1, "glade", 1, false},
	{1, "glass", 1, false},
	{1, "gle", 1, false},
	{1, "global", 1, false},
	{1, "globo", 1, false},
	{1, "gmail", 1, false},
	{1, "gmbh", 1, false},
	{1, "gmo", 1, false},
	{1, "gmx", 1, false},
	{1, "godaddy", 1, false},
	{1, "gold", 1, false},
	{1, "goldpoint", 1, false},
	{1, "golf", 1, false},
	{1, "goo", 1, false},
	{1, "goodyear", 1, false},
	{1, "goog", 1, false},
	{1, "google", 1, false},
	{1, "gop", 1, false},
	{1, "got", 1, false},
	{1, "grainger", 1, false},
	{1, "graphics", 1, false},
	{1, "gratis", 1, false},
	{1, "green", 1, false},
	{1, "gripe", 1, false},
	{1, "grocery", 1, false},
	{1, "group", 1, false},
	{1, "guardian", 1, false},
	{1, "gucci", 1, false},
	{1, "guge", 1, false},
	{1, "guide", 1, false},
	{1, "guitars", 1, false},
	{1, "guru", 1, false},
	{1, "hair", 1, false},
	{1, "hamburg", 1, false},
	{1, "hangout", 1, false},
	{1, "haus", 1, false},
	{1, "hbo", 1, false},
	{1, "hdfc", 1, false},
	{1, "hdfcbank", 1, false},
	{1, "health", 1, false},
	{1, "healthcare", 1, false},
	{1, "help", 1, false},
	{1, "helsinki", 1, false},
	{1, "here", 1, false},
	{1, "hermes", 1, false},
	{1, "hgtv", 1, false},
	{1, "hiphop", 1, false},
	{1, "hisamitsu", 1, false},
	{1, "hitachi", 1, false},
	{1, "hiv", 1, false},
	{1, "hkt", 1, false},
	{1, "hockey", 1, false},
	{1, "holdings", 1, false},
	{1, "holiday", 1, false},
	{1, "homedepot", 1, false},
	{1, "homegoods", 1, false},
	{1, "homes", 1, false},
	{1, "homesense", 1, false},
	{1, "honda", 1, false},
	{1, "horse", 1, false},
	{1, "hospital", 1, false},
	{1, "host", 1, false},
	{1, "hosting", 1, false},
	{1, "hot", 1, false},
	{1, "hoteles", 1, false},
	{1, "hotels", 1, false},
	{1, "hotmail", 1, false},
	{1, "house", 1, false},
	{1, "how", 1, false},
	{1, "hsbc", 1, false},
	{1, "hughes", 1, false},
	{1, "hyatt", 1, false},
	{1, "hyundai", 1, false},
	{1, "ibm", 1, false},
	{1, "icbc", 1, false},
	{1, "ice", 1, false},
	{1, "icu", 1, false},
	{1, "ieee", 1, false},
	{1, "ifm", 1, false},
	{1, "ikano", 1, false},
	{1, "imamat", 1, false},
	{1, "imdb", 1, false},
	{1, "immo", 1, false},
	{1, "immobilien", 1, false},
	{1, "inc", 1, false},
	{1, "industries", 1, false},
	{1, "infiniti", 1, false},
	{1, "ing", 1, false},
	{1, "ink", 1, false},
	{1, "institute", 1, false},
	{1, "insurance", 1, false},
	{1, "insure", 1, false},
	{1, "international", 1, false},
	{1, "intuit", 1, false},
	{1, "investments", 1, false},
	{1, "ipiranga", 1, false},
	{1, "irish", 1, false},
	{1, "ismaili", 1, false},
	{1, "ist", 1, false},
	{1, "istanbul", 1, false},
	{1, "itau", 1, false},
	{1, "itv", 1, false},
	{1, "jaguar", 1, false},
	{1, "java", 1, false},
	{1, "jcb", 1, false},
	{1, "jeep", 1, false},
	{1, "jetzt", 1, false},
	{1, "jewelry", 1, false},
	{1, "jio", 1, false},
	{1, "jll", 1, false},
	{1, "jmp", 1, false},
	{1, "jnj", 1, false},
	{1, "joburg", 1, false},
	{1, "jot", 1, false},
	{1, "joy", 1, false},
	{1, "jpmorgan", 1, false},
	{1, "jprs", 1, false},
	{1, "juegos", 1, false},
	{1, "juniper", 1, false},
	{1, "kaufen", 1, false},
	{1, "kddi", 1, false},
	{1, "kerryhotels", 1, false},
	{1, "kerrylogistics", 1, false},
	{1, "kerryproperties", 1, false},
	{1, "kfh", 1, false},
	{1, "kia", 1, false},
	{1, "kim", 1, false},
	{1, "kinder", 1, false},
	{1, "kindle", 1, false},
	{1, "kitchen", 1, false},
	{1, "kiwi", 1, false},
	{1, "koeln", 1, false},
	{1, "komatsu", 1, false},
	{1, "kosher", 1, false},
	{1, "kpmg", 1, false},
	{1, "kpn", 1, false},
	{1, "krd", 1, false},
	{1, "kred", 1, false},
	{1, "kuokgroup", 1, false},
	{1, "kyoto", 1, false},
	{1, "lacaixa", 1, false},
	{1, "lamborghini", 1, false},
	{1, "lamer", 1, false},
	{1, "lancaster", 1, false},
	{1, "lancia", 1, false},
	{1, "land", 1, false},
	{1, "landrover", 1, false},
	{1, "lanxess", 1, false},
	{1, "lasalle", 1, false},
	{1, "lat", 1, false},
	{1, "latino", 1, false},
	{1, "latrobe", 1, false},
	{1, "law", 1, false},
	{1, "lawyer", 1, false},
	{1, "lds", 1, false},
	{1, "lease", 1, false},
	{1, "leclerc", 1, false},
	{1, "lefrak", 1, false},
	{1, "legal", 1, false},
	{1, "lego", 1, false},
	{1, "lexus", 1, false},
	{1, "lgbt", 1, false},
	{1, "lidl", 1, false},
	{1, "life", 1, false},
	{1, "lifeinsurance", 1, false},
	{1, "lifestyle", 1, false},
	{1, "lighting", 1, false},
	{1, "like", 1, false},
	{1, "lilly", 1, false},
	{1, "limited", 1, false},
	{1, "limo", 1, false},
	{1, "lincoln", 1, false},
	{1, "linde", 1, false},
	{1, "link", 1, false},
	{1, "lipsy", 1, false},
	{1, "live", 1, false},
	{1, "living", 1, false},
	{1, "lixil", 1, false},
	{1, "llc", 1, false},
	{1, "llp", 1, false},
	{1, "loan", 1, false},
	{1, "loans", 1, false},
	{1, "locker", 1, false},
	{1, "locus", 1, false},
	{1, "loft", 1, false},
	{1, "lol", 1, false},
	{1, "london", 1, false},
	{1, "lotte", 1, false},
	{1, "lotto", 1, false},
	{1, "love", 1, false},
	{1, "lpl", 1, false},
	{1, "lplfinancial", 1, false},
	{1, "ltd", 1, false},
	{1, "ltda", 1, false},
	{1, "lundbeck", 1, false},
	{1, "luxe", 1, false},
	{1, "luxury", 1, false},
	{1, "macys", 1, false},
	{1, "madrid", 1, false},
	{1, "maif", 1, false},
	{1, "maison", 1, false},
	{1, "makeup", 1, false},
	{1, "man", 1, false},
	{1, "management", 1, false},
	{1, "mango", 1, false},
	{1, "map", 1, false},
	{1, "market", 1, false},
	{1, "marketing", 1, false},
	{1, "markets", 1, false},
	{1, "marriott", 1, false},
	{1, "marshalls", 1, false},
	{1, "maserati", 1, false},
	{1, "mattel", 1, false},
	{1, "mba", 1, false},
	{1, "mckinsey", 1, false},
	{1, "med", 1, false},
	{1, "media", 1, false},
	{1, "meet", 1, false},
	{1, "melbourne", 1, false},
	{1, "meme", 1, false},
	{1, "memorial", 1, false},
	{1, "men", 1, false},
	{1, "menu", 1, false},
	{1, "merckmsd", 1, false},
	{1, "miami", 1, false},
	{1, "microsoft", 1, false},
	{1, "mini", 1, false},
	{1, "mint", 1, false},
	{1, "mit", 1, false},
	{1, "mitsubishi", 1, false},
	{1, "mlb", 1, false},
	{1, "mls", 1, false},
	{1, "mma", 1, false},
	{1, "mobile", 1, false},
	{1, "moda", 1, false},
	{1, "moe", 1, false},
	{1, "moi", 1, false},
	{1, "mom", 1, false},
	{1, "monash", 1, false},
	{1, "money", 1, false},
	{1, "monster", 1, false},
	{1, "mormon", 1, false},
	{1, "mortgage", 1, false},
	{1, "moscow", 1, false},
	{1, "moto", 1, false},
	{1, "motorcycles", 1, false},
	{1, "mov", 1, false},
	{1, "movie", 1, false},
	{1, "msd", 1, false},
	{1, "mtn", 1, false},
	{1, "mtr", 1, false},
	{1, "mutual", 1, false},
	{1, "nab", 1, false},
	{1, "nagoya", 1, false},
	{1, "natura", 1, false},
	{1, "navy", 1, false},
	{1, "nba", 1, false},
	{1, "nec", 1, false},
	{1, "netbank", 1, false},
	{1, "netflix", 1, false},
	{1, "network", 1, false},
	{1, "neustar", 1, false},
	{1, "new", 1, false},
	{1, "news", 1, false},
	{1, "next", 1, false},
	{1, "nextdirect", 1, false},
	{1, "nexus", 1, false},
	{1, "nfl", 1, false},
	{1, "ngo", 1, false},
	{1, "nhk", 1, false},
	{1, "nico", 1, false},
	{1, "nike", 1, false},
	{1, "nikon", 1, false},
	{1, "ninja", 1, false},
	{1, "nissan", 1, false},
	{1, "nissay", 1, false},
	{1, "nokia", 1, false},
	{1, "northwesternmutual", 1, false},
	{1, "norton", 1, false},
	{1, "now", 1, false},
	{1, "nowruz", 1, false},
	{1, "nowtv", 1, false},
	{1, "nra", 1, false},
	{1, "nrw", 1, false},
	{1, "ntt", 1, false},
	{1, "nyc", 1, false},
	{1, "obi", 1, false},
	{1, "observer", 1, false},
	{1, "off", 1, false},
	{1, "office", 1, false},
	{1, "okinawa", 1, false},
	{1, "olayan", 1, false},
	{1, "olayangroup", 1, false},
	{1, "oldnavy", 1, false},
	{1, "ollo", 1, false},
	{1, "omega", 1, false},
	{1, "one", 1, false},
	{1, "ong", 1, false},
	{1, "onl", 1, false},
	{1, "online", 1, false},
	{1, "ooo", 1, false},
	{1, "open", 1, false},
	{1, "oracle", 1, false},
	{1, "orange", 1, false},
	{1, "organic", 1, false},
	{1, "origins", 1, false},
	{1, "osaka", 1, false},
	{1, "otsuka", 1, false},
	{1, "ott", 1, false},
	{1, "ovh", 1, false},
	{1, "page", 1, false},
	{1, "panasonic", 1, false},
	{1, "paris", 1, false},
	{1, "pars", 1, false},
	{1, "partners", 1, false},
	{1, "parts", 1, false},
	{1, "party", 1, false},
	{1, "passagens", 1, false},
	{1, "pay", 1, false},
	{1, "pccw", 1, false},
	{1, "pet", 1, false},
	{1, "pfizer", 1, false},
	{1, "pharmacy", 1, false},
	{1, "phd", 1, false},
	{1, "philips", 1, false},
	{1, "phone", 1, false},
	{1, "photo", 1, false},
	{1, "photography", 1, false},
	{1, "photos", 1, false},
	{1, "physio", 1, false},
	{1, "pics", 1, false},
	{1, "pictet", 1, false},
	{1, "pictures", 1, false},
	{1, "pid", 1, false},
	{1, "pin", 1, false},
	{1, "ping", 1, false},
	{1, "pink", 1, false},
	{1, "pioneer", 1, false},
	{1, "pizza", 1, false},
	{1, "place", 1, false},
	{1, "play", 1, false},
	{1, "playstation", 1, false},
	{1, "plumbing", 1, false},
	{1, "plus", 1, false},
	{1, "pnc", 1, false},
	{1, "pohl", 1, false},
	{1, "poker", 1, false},
	{1, "politie", 1, false},
	{1, "porn", 1, false},
	{1, "pramerica", 1, false},
	{1, "praxi", 1, false},
	{1, "press", 1, false},
	{1, "prime", 1, false},
	{1, "prod", 1, false},
	{1, "productions", 1, false},
	{1, "prof", 1, false},
	{1, "progressive", 1, false},
	{1, "promo", 1, false},
	{1, "properties", 1, false},
	{1, "property", 1, false},
	{1, "protection", 1, false},
	{1, "pru", 1, false},
	{1, "prudential", 1, false},
	{1, "pub", 1, false},
	{1, "pwc", 1, false},
	{1, "qpon", 1, false},
	{1, "quebec", 1, false},
	{1, "quest", 1, false},
	{1, "qvc", 1, false},
	{1, "racing", 1, false},
	{1, "radio", 1, false},
	{1, "raid", 1, false},
	{1, "read", 1, false},
	{1, "realestate", 1, false},
	{1, "realtor", 1, false},
	{1, "realty", 1, false},
	{1, "recipes", 1, false},
	{1, "red", 1, false},
	{1, "redstone", 1, false},
	{1, "redumbrella", 1, false},
	{1, "rehab", 1, false},
	{1, "reise", 1, false},
	{1, "reisen", 1, false},
	{1, "reit", 1, false},
	{1, "reliance", 1, false},
	{1, "ren", 1, false},
	{1, "rent", 1, false},
	{1, "rentals", 1, false},
	{1, "repair", 1, false},
	{1, "report", 1, false},
	{1, "republican", 1, false},
	{1, "rest", 1, false},
	{1, "restaurant", 1, false},
	{1, "review", 1, false},
	{1, "reviews", 1, false},
	{1, "rexroth", 1, false},
	{1, "rich", 1, false},
	{1, "richardli", 1, false},
	{1, "ricoh", 1, false},
	{1, "ril", 1, false},
	{1, "rio", 1, false},
	{1, "rip", 1, false},
	{1, "rmit", 1, false},
	{1, "rocher", 1, false},
	{1, "rocks", 1, false},
	{1, "rodeo", 1, false},
	{1, "rogers", 1, false},
	{1, "room", 1, false},
	{1, "rsvp", 1, false},
	{1, "rugby", 1, false},
	{1, "ruhr", 1, false},
	{1, "run", 1, false},
	{1, "rwe", 1, false},
	{1, "ryukyu", 1, false},
	{1, "saarland", 1, false},
	{1, "safe", 1, false},
	{1, "safety", 1, false},
	{1, "sakura", 1, false},
	{1, "sale", 1, false},
	{1, "salon", 1, false},
	{1, "samsclub", 1, false},
	{1, "samsung", 1, false},
	{1, "sandvik", 1, false},
	{1, "sandvikcoromant", 1, false},
	{1, "sanofi", 1, false},
	{1, "sap", 1, false},
	{1, "sarl", 1, false},
	{1, "sas", 1, false},
	{1, "save", 1, false},
	{1, "saxo", 1, false},
	{1, "sbi", 1, false},
	{1, "sbs", 1, false},
	{1, "sca", 1, false},
	{1, "scb", 1, false},
	{1, "schaeffler", 1, false},
	{1, "schmidt", 1, false},
	{1, "scholarships", 1, false},
	{1, "school", 1, false},
	{1, "schule", 1, false},
	{1, "schwarz", 1, false},
	{1, "science", 1, false},
	{1, "scjohnson", 1, false},
	{1, "scot", 1, false},
	{1, "search", 1, false},
	{1, "seat", 1, false},
	{1, "secure", 1, false},
	{1, "security", 1, false},
	{1, "seek", 1, false},
	{1, "select", 1, false},
	{1, "sener", 1, false},
	{1, "services", 1, false},
	{1, "ses", 1, false},
	{1, "seven", 1, false},
	{1, "sew", 1, false},
	{1, "sex", 1, false},
	{1, "sexy", 1, false},
	{1, "sfr", 1, false},
	{1, "shangrila", 1, false},
	{1, "sharp", 1, false},
	{1, "shaw", 1, false},
	{1, "shell", 1, false},
	{1, "shia", 1, false},
	{1, "shiksha", 1, false},
	{1, "shoes", 1, false},
	{1, "shop", 1, false},
	{1, "shopping", 1, false},
	{1, "shouji", 1, false},
	{1, "show", 1, false},
	{1, "showtime", 1, false},
	{1, "silk", 1, false},
	{1, "sina", 1, false},
	{1, "singles", 1, false},
	{1, "site", 1, false},
	{1, "ski", 1, false},
	{1, "skin", 1, false},
	{1, "sky", 1, false},
	{1, "skype", 1, false},
	{1, "sling", 1, false},
	{1, "smart", 1, false},
	{1, "smile", 1, false},
	{1, "sncf", 1, false},
	{1, "soccer", 1, false},
	{1, "social", 1, false},
	{1, "softbank", 1, false},
	{1, "software", 1, false},
	{1, "sohu", 1, false},
	{1, "solar", 1, false},
	{1, "solutions", 1, false},
	{1, "song", 1, false},
	{1, "sony", 1, false},
	{1, "soy", 1, false},
	{1, "spa", 1, false},
	{1, "space", 1, false},
	{1, "sport", 1, false},
	{1, "spot", 1, false},
	{1, "srl", 1, false},
	{1, "stada", 1, false},
	{1, "staples", 1, false},
	{1, "star", 1, false},
	{1, "statebank", 1, false},
	{1, "statefarm", 1, false},
	{1, "stc", 1, false},
	{1, "stcgroup", 1, false},
	{1, "stockholm", 1, false},
	{1, "storage", 1, false},
	{1, "store", 1, false},
	{1, "stream", 1, false},
	{1, "studio", 1, false},
	{1, "study", 1, false},
	{1, "style", 1, false},
	{1, "sucks", 1, false},
	{1, "supplies", 1, false},
	{1, "supply", 1, false},
	{1, "support", 1, false},
	{1, "surf", 1, false},
	{1, "surgery", 1, false},
	{1, "suzuki", 1, false},
	{1, "swatch", 1, false},
	{1, "swiftcover", 1, false},
	{1, "swiss", 1, false},
	{1, "sydney", 1, false},
	{1, "systems", 1, false},
	{1, "tab", 1, false},
	{1, "taipei", 1, false},
	{1, "talk", 1, false},
	{1, "taobao", 1, false},
	{1, "target", 1, false},
	{1, "tatamotors", 1, false},
	{1, "tatar", 1, false},
	{1, "tattoo", 1, false},
	{1, "tax", 1, false},
	{1, "taxi", 1, false},
	{1, "tci", 1, false},
	{1, "tdk", 1, false},
	{1, "team", 1, false},
	{1, "tech", 1, false},
	{1, "technology", 1, false},
	{1, "temasek", 1, false},
	{1, "tennis", 1, false},
	{1, "teva", 1, false},
	{1, "thd", 1, false},
	{1, "theater", 1, false},
	{1, "theatre", 1, false},
	{1, "tiaa", 1, false},
	{1, "tickets", 1, false},
	{1, "tienda", 1, false},
	{1, "tiffany", 1, false},
	{1, "tips", 1, false},
	{1, "tires", 1, false},
	{1, "tirol", 1, false},
	{1, "tjmaxx", 1, false},
	{1, "tjx", 1, false},
	{1, "tkmaxx", 1, false},
	{1, "tmall", 1, false},
	{1, "today", 1, false},
	{1, "tokyo", 1, false},
	{1, "tools", 1, false},
	{1, "top", 1, false},
	{1, "toray", 1, false},
	{1, "toshiba", 1, false},
	{1, "total", 1, false},
	{1, "tours", 1, false},
	{1, "town", 1, false},
	{1, "toyota", 1, false},
	{1, "toys", 1, false},
	{1, "trade", 1, false},
	{1, "trading", 1, false},
	{1, "training", 1, false},
	{1, "travel", 1, false},
	{1, "travelchannel", 1, false},
	{1, "travelers", 1, false},
	{1, "travelersinsurance", 1, false},
	{1, "trust", 1, false},
	{1, "trv", 1, false},
	{1, "tube", 1, false},
	{1, "tui", 1, false},
	{1, "tunes", 1, false},
	{1, "tushu", 1, false},
	{1, "tvs", 1, false},
	{1, "ubank", 1, false},
	{1, "ubs", 1, false},
	{1, "unicom", 1, false},
	{1, "university", 1, false},
	{1, "uno", 1, false},
	{1, "uol", 1, false},
	{1, "ups", 1, false},
	{1, "vacations", 1, false},
	{1, "vana", 1, false},
	{1, "vanguard", 1, false},
	{1, "vegas", 1, false},
	{1, "ventures", 1, false},
	{1, "verisign", 1, false},
	{1, "versicherung", 1, false},
	{1, "vet", 1, false},
	{1, "viajes", 1, false},
	{1, "video", 1, false},
	{1, "vig", 1, false},
	{1, "viking", 1, false},
	{1, "villas", 1, false},
	{1, "vin", 1, false},
	{1, "vip", 1, false},
	{1, "virgin", 1, false},
	{1, "visa", 1, false},
	{1, "vision", 1, false},
	{1, "viva", 1, false},
	{1, "vivo", 1, false},
	{1, "vlaanderen", 1, false},
	{1, "vodka", 1, false},
	{1, "volkswagen", 1, false},
	{1, "volvo", 1, false},
	{1, "vote", 1, false},
	{1, "voting", 1, false},
	{1, "voto", 1, false},
	{1, "voyage", 1, false},
	{1, "vuelos", 1, false},
	{1, "wales", 1, false},
	{1, "walmart", 1, false},
	{1, "walter", 1, false},
	{1, "wang", 1, false},
	{1, "wanggou", 1, false},
	{1, "watch", 1, false},
	{1, "watches", 1, false},
	{1, "weather", 1, false},
	{1, "weatherchannel", 1, false},
	{1, "webcam", 1, false},
	{1, "weber", 1, false},
	{1, "website", 1, false},
	{1, "wedding", 1, false},
	{1, "weibo", 1, false},
	{1, "weir", 1, false},
	{1, "whoswho", 1, false},
	{1, "wien", 1, false},
	{1, "wiki", 1, false},
	{1, "williamhill", 1, false},
	{1, "win", 1, false},
	{1, "windows", 1, false},
	{1, "wine", 1, false},
	{1, "winners", 1, false},
	{1, "wme", 1, false},
	{1, "wolterskluwer", 1, false},
	{1, "woodside", 1, false},
	{1, "work", 1, false},
	{1, "works", 1, false},
	{1, "world", 1, false},
	{1, "wow", 1, false},
	{1, "wtc", 1, false},
	{1, "wtf", 1, false},
	{1, "xbox", 1, false},
	{1, "xerox", 1, false},
	{1, "xfinity", 1, false},
	{1, "xihuan", 1, false},
	{1, "xin", 1, false},
	{1, "xn--11b4c3d", 1, false},
	{1, "xn--1ck2e1b", 1, false},
	{1, "xn--1qqw23a", 1, false},
	{1, "xn--30rr7y", 1, false},
	{1, "xn--3bst00m", 1, false},
	{1, "xn--3ds443g", 1, false},
	{1, "xn--3oq18vl8pn36a", 1, false},
	{1, "xn--3pxu8k", 1, false},
	{1, "xn--42c2d9a", 1, false},
	{1, "xn--45q11c", 1, false},
	{1, "xn--4gbrim", 1, false},
	{1, "xn--55qw42g", 1, false},
	{1, "xn--55qx5d", 1, false},
	{1, "xn--5su34j936bgsg", 1, false},
	{1, "xn--5tzm5g", 1, false},
	{1, "xn--6frz82g", 1, false},
	{1, "xn--6qq986b3xl", 1, false},
	{1, "xn--80adxhks", 1, false},
	{1, "xn--80aqecdr1a", 1, false},
	{1, "xn--80asehdb", 1, false},
	{1, "xn--80aswg", 1, false},
	{1, "xn--8y0a063a", 1, false},
	{1, "xn--9dbq2a", 1, false},
	{1, "xn--9et52u", 1, false},
	{1, "xn--9krt00a", 1, false},
	{1, "xn--b4w605ferd", 1, false},
	{1, "xn--bck1b9a5dre4c", 1, false},
	{1, "xn--c1avg", 1, false},
	{1, "xn--c2br7g", 1, false},
	{1, "xn--cck2b3b", 1, false},
	{1, "xn--cckwcxetd", 1, false},
	{1, "xn--cg4bki", 1, false},
	{1, "xn--czr694b", 1, false},
	{1, "xn--czrs0t", 1, false},
	{1, "xn--czru2d", 1, false},
	{1, "xn--d1acj3b", 1, false},
	{1, "xn--eckvdtc9d", 1, false},
	{1, "xn--efvy88h", 1, false},
	{1, "xn--fct429k", 1, false},
	{1, "xn--fhbei", 1, false},
	{1, "xn--fiq228c5hs", 1, false},
	{1, "xn--fiq64b", 1, false},
	{1, "xn--fjq720a", 1, false},
	{1, "xn--flw351e", 1, false},
	{1, "xn--fzys8d69uvgm", 1, false},
	{1, "xn--g2xx48c", 1, false},
	{1, "xn--gckr3f0f", 1, false},
	{1, "xn--gk3at1e", 1, false},
	{1, "xn--hxt814e", 1, false},
	{1, "xn--i1b6b1a6a2e", 1, false},
	{1, "xn--imr513n", 1, false},
	{1, "xn--io0a7i", 1, false},
	{1, "xn--j1aef", 1, false},
	{1, "xn--jlq480n2rg", 1, false},
	{1, "xn--jlq61u9w7b", 1, false},
	{1, "xn--jvr189m", 1, false},
	{1, "xn--kcrx77d1x4a", 1, false},
	{1, "xn--kput3i", 1, false},
	{1, "xn--mgba3a3ejt", 1, false},
	{1, "xn--mgba7c0bbn0a", 1, false},
	{1, "xn--mgbaakc7dvf", 1, false},
	{1, "xn--mgbab2bd", 1, false},
	{1, "xn--mgbca7dzdo", 1, false},
	{1, "xn--mgbi4ecexp", 1, false},
	{1, "xn--mgbt3dhd", 1, false},
	{1, "xn--mk1bu44c", 1, false},
	{1, "xn--mxtq1m", 1, false},
	{1, "xn--ngbc5azd", 1, false},
	{1, "xn--ngbe9e0a", 1, false},
	{1, "xn--ngbrx", 1, false},
	{1, "xn--nqv7f", 1, false},
	{1, "xn--nqv7fs00ema", 1, false},
	{1, "xn--nyqy26a", 1, false},
	{1, "xn--otu796d", 1, false},
	{1, "xn--p1acf", 1, false},
	{1, "xn--pssy2u", 1, false},
	{1, "xn--q9jyb4c", 1, false},
	{1, "xn--qcka1pmc", 1, false},
	{1, "xn--rhqv96g", 1, false},
	{1, "xn--rovu88b", 1, false},
	{1, "xn--ses554g", 1, false},
	{1, "xn--t60b56a", 1, false},
	{1, "xn--tckwe", 1, false},
	{1, "xn--tiq49xqyj", 1, false},
	{1, "xn--unup4y", 1, false},
	{1, "xn--vermgensberater-ctb", 1, false},
	{1, "xn--vermgensberatung-pwb", 1, false},
	{1, "xn--vhquv", 1, false},
	{1, "xn--vuq861b", 1, false},
	{1, "xn--w4r85el8fhu5dnra", 1, false},
	{1, "xn--w4rs40l", 1, false},
	{1, "xn--xhq521b", 1, false},
	{1, "xn--zfr164b", 1, false},
	{1, "xyz", 1, false},
	{1, "yachts", 1, false},
	{1, "yahoo", 1, false},
	{1, "yamaxun", 1, false},
	{1, "yandex", 1, false},
	{1, "yodobashi", 1, false},
	{1, "yoga", 1, false},
	{1, "yokohama", 1, false},
	{1, "you", 1, false},
	{1, "youtube", 1, false},
	{1, "yun", 1, false},
	{1, "zappos", 1, false},
	{1, "zara", 1, false},
	{1, "zero", 1, false},
	{1, "zip", 1, false},
	{1, "zone", 1, false},
	{1, "zuerich", 1, false},
	{1, "cc.ua", 2, true},
	{1, "inf.ua", 2, true},
	{1, "ltd.ua", 2, true},
	{1, "611.to", 2, true},
	{1, "graphox.us", 2, true},
	{2, "devcdnaccesso.com", 3, true},
	{1, "adobeaemcloud.com", 2, true},
	{1, "adobeaemcloud.net", 2, true},
	{2, "dev.adobeaemcloud.com", 4, true},
	{1, "beep.pl", 2, true},
	{1, "barsy.ca", 2, true},
	{2, "compute.estate", 3, true},
	{2, "alces.network", 3, true},
	{1, "kasserver.com", 2, true},
	{1, "altervista.org", 2, true},
	{1, "alwaysdata.net", 2, true},
	{1, "cloudfront.net", 2, true},
	{2, "compute.amazonaws.com", 4, true},
	{2, "compute-1.amazonaws.com", 4, true},
	{2, "compute.amazonaws.com.cn", 5, true},
	{1, "us-east-1.amazonaws.com", 3, true},
	{1, "cn-north-1.eb.amazonaws.com.cn", 5, true},
	{1, "cn-northwest-1.eb.amazonaws.com.cn", 5, true},
	{1, "elasticbeanstalk.com", 2, true},
	{1, "ap-northeast-1.elasticbeanstalk.com", 3, true},
	{1, "ap-northeast-2.elasticbeanstalk.com", 3, true},
	{1, "ap-northeast-3.elasticbeanstalk.com", 3, true},
	{1, "ap-south-1.elasticbeanstalk.com", 3, true},
	{1, "ap-southeast-1.elasticbeanstalk.com", 3, true},
	{1, "ap-southeast-2.elasticbeanstalk.com", 3, true},
	{1, "ca-central-1.elasticbeanstalk.com", 3, true},
	{1, "eu-central-1.elasticbeanstalk.com", 3, true},
	{1, "eu-west-1.elasticbeanstalk.com", 3, true},
	{1, "eu-west-2.elasticbeanstalk.com", 3, true},
	{1, "eu-west-3.elasticbeanstalk.com", 3, true},
	{1, "sa-east-1.elasticbeanstalk.com", 3, true},
	{1, "us-east-1.elasticbeanstalk.com", 3, true},
	{1, "us-east-2.elasticbeanstalk.com", 3, true},
	{1, "us-gov-west-1.elasticbeanstalk.com", 3, true},
	{1, "us-west-1.elasticbeanstalk.com", 3, true},
	{1, "us-west-2.elasticbeanstalk.com", 3, true},
	{2, "elb.amazonaws.com", 4, true},
	{2, "elb.amazonaws.com.cn", 5, true},
	{1, "awsglobalaccelerator.com", 2, true},
	{1, "s3.amazonaws.com", 3, true},
	{1, "s3-ap-northeast-1.amazonaws.com", 3, true},
	{1, "s3-ap-northeast-2.amazonaws.com", 3, true},
	{1, "s3-ap-south-1.amazonaws.com", 3, true},
	{1, "s3-ap-southeast-1.amazonaws.com", 3, true},
	{1, "s3-ap-southeast-2.amazonaws.com", 3, true},
	{1, "s3-ca-central-1.amazonaws.com", 3, true},
	{1, "s3-eu-central-1.amazonaws.com", 3, true},
	{1, "s3-eu-west-1.amazonaws.com", 3, true},
	{1, "s3-eu-west-2.amazonaws.com", 3, true},
	{1, "s3-eu-west-3.amazonaws.com", 3, true},
	{1, "s3-external-1.amazonaws.com", 3, true},
	{1, "s3-fips-us-gov-west-1.amazonaws.com", 3, true},
	{1, "s3-sa-east-1.amazonaws.com", 3, true},
	{1, "s3-us-gov-west-1.amazonaws.com", 3, true},
	{1, "s3-us-east-2.amazonaws.com", 3, true},
	{1, "s3-us-west-1.amazonaws.com", 3, true},
	{1, "s3-us-west-2.amazonaws.com", 3, true},
	{1, "s3.ap-northeast-2.amazonaws.com", 4, true},
	{1, "s3.ap-south-1.amazonaws.com", 4, true},
	{1, "s3.cn-north-1.amazonaws.com.cn", 5, true},
	{1, "s3.ca-central-1.amazonaws.com", 4, true},
	{1, "s3.eu-central-1.amazonaws.com", 4, true},
	{1, "s3.eu-west-2.amazonaws.com", 4, true},
	{1, "s3.eu-west-3.amazonaws.com", 4, true},
	{1, "s3.us-east-2.amazonaws.com", 4, true},
	{1, "s3.dualstack.ap-northeast-1.amazonaws.com", 5, true},
	{1, "s3.dualstack.ap-northeast-2.amazonaws.com", 5, true},
	{1, "s3.dualstack.ap-south-1.amazonaws.com", 5, true},
	{1, "s3.dualstack.ap-southeast-1.amazonaws.com", 5, true},
	{1, "s3.dualstack.ap-southeast-2.amazonaws.com", 5, true},
	{1, "s3.dualstack.ca-central-1.amazonaws.com", 5, true},
	{1, "s3.dualstack.eu-central-1.amazonaws.com", 5, true},
	{1, "s3.dualstack.eu-west-1.amazonaws.com", 5, true},
	{1, "s3.dualstack.eu-west-2.amazonaws.com", 5, true},
	{1, "s3.dualstack.eu-west-3.amazonaws.com", 5, true},
	{1, "s3.dualstack.sa-east-1.amazonaws.com", 5, true},
	{1, "s3.dualstack.us-east-1.amazonaws.com", 5, true},
	{1, "s3.dualstack.us-east-2.amazonaws.com", 5, true},
	{1, "s3-website-us-east-1.amazonaws.com", 3, true},
	{1, "s3-website-us-west-1.amazonaws.com", 3, true},
	{1, "s3-website-us-west-2.amazonaws.com", 3, true},
	{1, "s3-website-ap-northeast-1.amazonaws.com", 3, true},
	{1, "s3-website-ap-southeast-1.amazonaws.com", 3, true},
	{1, "s3-website-ap-southeast-2.amazonaws.com", 3, true},
	{1, "s3-website-eu-west-1.amazonaws.com", 3, true},
	{1, "s3-website-sa-east-1.amazonaws.com", 3, true},
	{1, "s3-website.ap-northeast-2.amazonaws.com", 4, true},
	{1, "s3-website.ap-south-1.amazonaws.com", 4, true},
	{1, "s3-website.ca-central-1.amazonaws.com", 4, true},
	{1, "s3-website.eu-central-1.amazonaws.com", 4, true},
	{1, "s3-website.eu-west-2.amazonaws.com", 4, true},
	{1, "s3-website.eu-west-3.amazonaws.com", 4, true},
	{1, "s3-website.us-east-2.amazonaws.com", 4, true},
	{1, "amsw.nl", 2, true},
	{1, "t3l3p0rt.net", 2, true},
	{1, "tele.amune.org", 3, true},
	{1, "apigee.io", 2, true},
	{1, "appspacehosted.com", 2, true},
	{1, "appspaceusercontent.com", 2, true},
	{1, "on-aptible.com", 2, true},
	{1, "user.aseinet.ne.jp", 4, true},
	{1, "gv.vc", 2, true},
	{1, "d.gv.vc", 3, true},
	{1, "user.party.eus", 3, true},
	{1, "pimienta.org", 2, true},
	{1, "poivron.org", 2, true},
	{1, "potager.org", 2, true},
	{1, "sweetpepper.org", 2, true},
	{1, "myasustor.com", 2, true},
	{1, "myfritz.net", 2, true},
	{2, "awdev.ca", 3, true},
	{2, "advisor.ws", 3, true},
	{1, "b-data.io", 2, true},
	{1, "backplaneapp.io", 2, true},
	{1, "balena-devices.com", 2, true},
	{2, "banzai.cloud", 3, true},
	{1, "app.banzaicloud.io", 3, true},
	{2, "backyards.banzaicloud.io", 4, true},
	{1, "betainabox.com", 2, true},
	{1, "bnr.la", 2, true},
	{1, "blackbaudcdn.net", 2, true},
	{1, "of.je", 2, true},
	{1, "boomla.net", 2, true},
	{1, "boutir.com", 2, true},
	{1, "boxfuse.io", 2, true},
	{1, "square7.ch", 2, true},
	{1, "bplaced.com", 2, true},
	{1, "bplaced.de", 2, true},
	{1, "square7.de", 2, true},
	{1, "bplaced.net", 2, true},
	{1, "square7.net", 2, true},
	{1, "browsersafetymark.io", 2, true},
	{1, "uk0.bigv.io", 3, true},
	{1, "dh.bytemark.co.uk", 4, true},
	{1, "vm.bytemark.co.uk", 4, true},
	{1, "cafjs.com", 2, true},
	{1, "mycd.eu", 2, true},
	{1, "carrd.co", 2, true},
	{1, "crd.co", 2, true},
	{1, "uwu.ai", 2, true},
	{1, "ae.org", 2, true},
	{1, "br.com", 2, true},
	{1, "cn.com", 2, true},
	{1, "com.de", 2, true},
	{1, "com.se", 2, true},
	{1, "de.com", 2, true},
	{1, "eu.com", 2, true},
	{1, "gb.net", 2, true},
	{1, "hu.net", 2, true},
	{1, "jp.net", 2, true},
	{1, "jpn.com", 2, true},
	{1, "mex.com", 2, true},
	{1, "ru.com", 2, true},
	{1, "sa.com", 2, true},
	{1, "se.net", 2, true},
	{1, "uk.com", 2, true},
	{1, "uk.net", 2, true},
	{1, "us.com", 2, true},
	{1, "za.bz", 2, true},
	{1, "za.com", 2, true},
	{1, "ar.com", 2, true},
	{1, "gb.com", 2, true},
	{1, "hu.com", 2, true},
	{1, "kr.com", 2, true},
	{1, "no.com", 2, true},
	{1, "qc.com", 2, true},
	{1, "uy.com", 2, true},
	{1, "africa.com", 2, true},
	{1, "gr.com", 2, true},
	{1, "in.net", 2, true},
	{1, "web.in", 2, true},
	{1, "us.org", 2, true},
	{1, "co.com", 2, true},
	{1, "aus.basketball", 2, true},
	{1, "nz.basketball", 2, true},
	{1, "radio.am", 2, true},
	{1, "radio.fm", 2, true},
	{1, "c.la", 2, true},
	{1, "certmgr.org", 2, true},
	{1, "cx.ua", 2, true},
	{1, "discourse.group", 2, true},
	{1, "discourse.team", 2, true},
	{1, "virtueeldomein.nl", 2, true},
	{1, "cleverapps.io", 2, true},
	{2, "lcl.dev", 3, true},
	{2, "lclstage.dev", 3, true},
	{2, "stg.dev", 3, true},
	{2, "stgstage.dev", 3, true},
	{1, "clic2000.net", 2, true},
	{1, "clickrising.net", 2, true},
	{1, "c66.me", 2, true},
	{1, "cloud66.ws", 2, true},
	{1, "cloud66.zone", 2, true},
	{1, "jdevcloud.com", 2, true},
	{1, "wpdevcloud.com", 2, true},
	{1, "cloudaccess.host", 2, true},
	{1, "freesite.host", 2, true},
	{1, "cloudaccess.net", 2, true},
	{1, "cloudcontrolled.com", 2, true},
	{1, "cloudcontrolapp.com", 2, true},
	{1, "cloudera.site", 2, true},
	{1, "pages.dev", 2, true},
	{1, "trycloudflare.com", 2, true},
	{1, "workers.dev", 2, true},
	{1, "wnext.app", 2, true},
	{1, "co.ca", 2, true},
	{2, "otap.co", 3, true},
	{1, "co.cz", 2, true},
	{1, "c.cdn77.org", 3, true},
	{1, "cdn77-ssl.net", 2, true},
	{1, "r.cdn77.net", 3, true},
	{1, "rsc.cdn77.org", 3, true},
	{1, "ssl.origin.cdn77-secure.org", 4, true},
	{1, "cloudns.asia", 2, true},
	{1, "cloudns.biz", 2, true},
	{1, "cloudns.club", 2, true},
	{1, "cloudns.cc", 2, true},
	{1, "cloudns.eu", 2, true},
	{1, "cloudns.in", 2, true},
	{1, "cloudns.info", 2, true},
	{1, "cloudns.org", 2, true},
	{1, "cloudns.pro", 2, true},
	{1, "cloudns.pw", 2, true},
	{1, "cloudns.us", 2, true},
	{1, "cnpy.gdn", 2, true},
	{1, "co.nl", 2, true},
	{1, "co.no", 2, true},
	{1, "webhosting.be", 2, true},
	{1, "hosting-cluster.nl", 2, true},
	{1, "ac.ru", 2, true},
	{1, "edu.ru", 2, true},
	{1, "gov.ru", 2, true},
	{1, "int.ru", 2, true},
	{1, "mil.ru", 2, true},
	{1, "test.ru", 2, true},
	{1, "dyn.cosidns.de", 3, true},
	{1, "dynamisches-dns.de", 2, true},
	{1, "dnsupdater.de", 2, true},
	{1, "internet-dns.de", 2, true},
	{1, "l-o-g-i-n.de", 2, true},
	{1, "dynamic-dns.info", 2, true},
	{1, "feste-ip.net", 2, true},
	{1, "knx-server.net", 2, true},
	{1, "static-access.net", 2, true},
	{1, "realm.cz", 2, true},
	{2, "cryptonomic.net", 3, true},
	{1, "cupcake.is", 2, true},
	{1, "curv.dev", 2, true},
	{1, "multibaas.app", 2, true},
	{1, "multibaas.com", 2, true},
	{2, "customer-oci.com", 3, true},
	{2, "oci.customer-oci.com", 4, true},
	{2, "ocp.customer-oci.com", 4, true},
	{2, "ocs.customer-oci.com", 4, true},
	{1, "cyon.link", 2, true},
	{1, "cyon.site", 2, true},
	{1, "fnwk.site", 2, true},
	{1, "folionetwork.site", 2, true},
	{1, "platform0.app", 2, true},
	{1, "daplie.me", 2, true},
	{1, "localhost.daplie.me", 3, true},
	{1, "dattolocal.com", 2, true},
	{1, "dattorelay.com", 2, true},
	{1, "dattoweb.com", 2, true},
	{1, "mydatto.com", 2, true},
	{1, "dattolocal.net", 2, true},
	{1, "mydatto.net", 2, true},
	{1, "biz.dk", 2, true},
	{1, "co.dk", 2, true},
	{1, "firm.dk", 2, true},
	{1, "reg.dk", 2, true},
	{1, "store.dk", 2, true},
	{1, "dyndns.dappnode.io", 3, true},
	{2, "dapps.earth", 3, true},
	{2, "bzz.dapps.earth", 4, true},
	{1, "builtwithdark.com", 2, true},
	{1, "edgestack.me", 2, true},
	{1, "debian.net", 2, true},
	{1, "deno.dev", 2, true},
	{1, "deno-staging.dev", 2, true},
	{1, "dedyn.io", 2, true},
	{1, "jozi.biz", 2, true},
	{1, "dnshome.de", 2, true},
	{1, "online.th", 2, true},
	{1, "shop.th", 2, true},
	{1, "drayddns.com", 2, true},
	{1, "shoparena.pl", 2, true},
	{1, "dreamhosters.com", 2, true},
	{1, "mydrobo.com", 2, true},
	{1, "drud.io", 2, true},
	{1, "drud.us", 2, true},
	{1, "duckdns.org", 2, true},
	{1, "bip.sh", 2, true},
	{1, "bitbridge.net", 2, true},
	{1, "dy.fi", 2, true},
	{1, "tunk.org", 2, true},
	{1, "dyndns-at-home.com", 2, true},
	{1, "dyndns-at-work.com", 2, true},
	{1, "dyndns-blog.com", 2, true},
	{1, "dyndns-free.com", 2, true},
	{1, "dyndns-home.com", 2, true},
	{1, "dyndns-ip.com", 2, true},
	{1, "dyndns-mail.com", 2, true},
	{1, "dyndns-office.com", 2, true},
	{1, "dyndns-pics.com", 2, true},
	{1, "dyndns-remote.com", 2, true},
	{1, "dyndns-server.com", 2, true},
	{1, "dyndns-web.com", 2, true},
	{1, "dyndns-wiki.com", 2, true},
	{1, "dyndns-work.com", 2, true},
	{1, "dyndns.biz", 2, true},
	{1, "dyndns.info", 2, true},
	{1, "dyndns.org", 2, true},
	{1, "dyndns.tv", 2, true},
	{1, "at-band-camp.net", 2, true},
	{1, "ath.cx", 2, true},
	{1, "barrel-of-knowledge.info", 2, true},
	{1, "barrell-of-knowledge.info", 2, true},
	{1, "better-than.tv", 2, true},
	{1, "blogdns.com", 2, true},
	{1, "blogdns.net", 2, true},
	{1, "blogdns.org", 2, true},
	{1, "blogsite.org", 2, true},
	{1, "boldlygoingnowhere.org", 2, true},
	{1, "broke-it.net", 2, true},
	{1, "buyshouses.net", 2, true},
	{1, "cechire.com", 2, true},
	{1, "dnsalias.com", 2, true},
	{1, "dnsalias.net", 2, true},
	{1, "dnsalias.org", 2, true},
	{1, "dnsdojo.com", 2, true},
	{1, "dnsdojo.net", 2, true},
	{1, "dnsdojo.org", 2, true},
	{1, "does-it.net", 2, true},
	{1, "doesntexist.com", 2, true},
	{1, "doesntexist.org", 2, true},
	{1, "dontexist.com", 2, true},
	{1, "dontexist.net", 2, true},
	{1, "dontexist.org", 2, true},
	{1, "doomdns.com", 2, true},
	{1, "doomdns.org", 2, true},
	{1, "dvrdns.org", 2, true},
	{1, "dyn-o-saur.com", 2, true},
	{1, "dynalias.com", 2, true},
	{1, "dynalias.net", 2, true},
	{1, "dynalias.org", 2, true},
	{1, "dynathome.net", 2, true},
	{1, "dyndns.ws", 2, true},
	{1, "endofinternet.net", 2, true},
	{1, "endofinternet.org", 2, true},
	{1, "endoftheinternet.org", 2, true},
	{1, "est-a-la-maison.com", 2, true},
	{1, "est-a-la-masion.com", 2, true},
	{1, "est-le-patron.com", 2, true},
	{1, "est-mon-blogueur.com", 2, true},
	{1, "for-better.biz", 2, true},
	{1, "for-more.biz", 2, true},
	{1, "for-our.info", 2, true},
	{1, "for-some.biz", 2, true},
	{1, "for-the.biz", 2, true},
	{1, "forgot.her.name", 3, true},
	{1, "forgot.his.name", 3, true},
	{1, "from-ak.com", 2, true},
	{1, "from-al.com", 2, true},
	{1, "from-ar.com", 2, true},
	{1, "from-az.net", 2, true},
	{1, "from-ca.com", 2, true},
	{1, "from-co.net", 2, true},
	{1, "from-ct.com", 2, true},
	{1, "from-dc.com", 2, true},
	{1, "from-de.com", 2, true},
	{1, "from-fl.com", 2, true},
	{1, "from-ga.com", 2, true},
	{1, "from-hi.com", 2, true},
	{1, "from-ia.com", 2, true},
	{1, "from-id.com", 2, true},
	{1, "from-il.com", 2, true},
	{1, "from-in.com", 2, true},
	{1, "from-ks.com", 2, true},
	{1, "from-ky.com", 2, true},
	{1, "from-la.net", 2, true},
	{1, "from-ma.com", 2, true},
	{1, "from-md.com", 2, true},
	{1, "from-me.org", 2, true},
	{1, "from-mi.com", 2, true},
	{1, "from-mn.com", 2, true},
	{1, "from-mo.com", 2, true},
	{1, "from-ms.com", 2, true},
	{1, "from-mt.com", 2, true},
	{1, "from-nc.com", 2, true},
	{1, "from-nd.com", 2, true},
	{1, "from-ne.com", 2, true},
	{1, "from-nh.com", 2, true},
	{1, "from-nj.com", 2, true},
	{1, "from-nm.com", 2, true},
	{1, "from-nv.com", 2, true},
	{1, "from-ny.net", 2, true},
	{1, "from-oh.com", 2, true},
	{1, "from-ok.com", 2, true},
	{1, "from-or.com", 2, true},
	{1, "from-pa.com", 2, true},
	{1, "from-pr.com", 2, true},
	{1, "from-ri.com", 2, true},
	{1, "from-sc.com", 2, true},
	{1, "from-sd.com", 2, true},
	{1, "from-tn.com", 2, true},
	{1, "from-tx.com", 2, true},
	{1, "from-ut.com", 2, true},
	{1, "from-va.com", 2, true},
	{1, "from-vt.com", 2, true},
	{1, "from-wa.com", 2, true},
	{1, "from-wi.com", 2, true},
	{1, "from-wv.com", 2, true},
	{1, "from-wy.com", 2, true},
	{1, "ftpaccess.cc", 2, true},
	{1, "fuettertdasnetz.de", 2, true},
	{1, "game-host.org", 2, true},
	{1, "game-server.cc", 2, true},
	{1, "getmyip.com", 2, true},
	{1, "gets-it.net", 2, true},
	{1, "go.dyndns.org", 3, true},
	{1, "gotdns.com", 2, true},
	{1, "gotdns.org", 2, true},
	{1, "groks-the.info", 2, true},
	{1, "groks-this.info", 2, true},
	{1, "ham-radio-op.net", 2, true},
	{1, "here-for-more.info", 2, true},
	{1, "hobby-site.com", 2, true},
	{1, "hobby-site.org", 2, true},
	{1, "home.dyndns.org", 3, true},
	{1, "homedns.org", 2, true},
	{1, "homeftp.net", 2, true},
	{1, "homeftp.org", 2, true},
	{1, "homeip.net", 2, true},
	{1, "homelinux.com", 2, true},
	{1, "homelinux.net", 2, true},
	{1, "homelinux.org", 2, true},
	{1, "homeunix.com", 2, true},
	{1, "homeunix.net", 2, true},
	{1, "homeunix.org", 2, true},
	{1, "iamallama.com", 2, true},
	{1, "in-the-band.net", 2, true},
	{1, "is-a-anarchist.com", 2, true},
	{1, "is-a-blogger.com", 2, true},
	{1, "is-a-bookkeeper.com", 2, true},
	{1, "is-a-bruinsfan.org", 2, true},
	{1, "is-a-bulls-fan.com", 2, true},
	{1, "is-a-candidate.org", 2, true},
	{1, "is-a-caterer.com", 2, true},
	{1, "is-a-celticsfan.org", 2, true},
	{1, "is-a-chef.com", 2, true},
	{1, "is-a-chef.net", 2, true},
	{1, "is-a-chef.org", 2, true},
	{1, "is-a-conservative.com", 2, true},
	{1, "is-a-cpa.com", 2, true},
	{1, "is-a-cubicle-slave.com", 2, true},
	{1, "is-a-democrat.com", 2, true},
	{1, "is-a-designer.com", 2, true},
	{1, "is-a-doctor.com", 2, true},
	{1, "is-a-financialadvisor.com", 2, true},
	{1, "is-a-geek.com", 2, true},
	{1, "is-a-geek.net", 2, true},
	{1, "is-a-geek.org", 2, true},
	{1, "is-a-green.com", 2, true},
	{1, "is-a-guru.com", 2, true},
	{1, "is-a-hard-worker.com", 2, true},
	{1, "is-a-hunter.com", 2, true},
	{1, "is-a-knight.org", 2, true},
	{1, "is-a-landscaper.com", 2, true},
	{1, "is-a-lawyer.com", 2, true},
	{1, "is-a-liberal.com", 2, true},
	{1, "is-a-libertarian.com", 2, true},
	{1, "is-a-linux-user.org", 2, true},
	{1, "is-a-llama.com", 2, true},
	{1, "is-a-musician.com", 2, true},
	{1, "is-a-nascarfan.com", 2, true},
	{1, "is-a-nurse.com", 2, true},
	{1, "is-a-painter.com", 2, true},
	{1, "is-a-patsfan.org", 2, true},
	{1, "is-a-personaltrainer.com", 2, true},
	{1, "is-a-photographer.com", 2, true},
	{1, "is-a-player.com", 2, true},
	{1, "is-a-republican.com", 2, true},
	{1, "is-a-rockstar.com", 2, true},
	{1, "is-a-socialist.com", 2, true},
	{1, "is-a-soxfan.org", 2, true},
	{1, "is-a-student.com", 2, true},
	{1, "is-a-teacher.com", 2, true},
	{1, "is-a-techie.com", 2, true},
	{1, "is-a-therapist.com", 2, true},
	{1, "is-an-accountant.com", 2, true},
	{1, "is-an-actor.com", 2, true},
	{1, "is-an-actress.com", 2, true},
	{1, "is-an-anarchist.com", 2, true},
	{1, "is-an-artist.com", 2, true},
	{1, "is-an-engineer.com", 2, true},
	{1, "is-an-entertainer.com", 2, true},
	{1, "is-by.us", 2, true},
	{1, "is-certified.com", 2, true},
	{1, "is-found.org", 2, true},
	{1, "is-gone.com", 2, true},
	{1, "is-into-anime.com", 2, true},
	{1, "is-into-cars.com", 2, true},
	{1, "is-into-cartoons.com", 2, true},
	{1, "is-into-games.com", 2, true},
	{1, "is-leet.com", 2, true},
	{1, "is-lost.org", 2, true},
	{1, "is-not-certified.com", 2, true},
	{1, "is-saved.org", 2, true},
	{1, "is-slick.com", 2, true},
	{1, "is-uberleet.com", 2, true},
	{1, "is-very-bad.org", 2, true},
	{1, "is-very-evil.org", 2, true},
	{1, "is-very-good.org", 2, true},
	{1, "is-very-nice.org", 2, true},
	{1, "is-very-sweet.org", 2, true},
	{1, "is-with-theband.com", 2, true},
	{1, "isa-geek.com", 2, true},
	{1, "isa-geek.net", 2, true},
	{1, "isa-geek.org", 2, true},
	{1, "isa-hockeynut.com", 2, true},
	{1, "issmarterthanyou.com", 2, true},
	{1, "isteingeek.de", 2, true},
	{1, "istmein.de", 2, true},
	{1, "kicks-ass.net", 2, true},
	{1, "kicks-ass.org", 2, true},
	{1, "knowsitall.info", 2, true},
	{1, "land-4-sale.us", 2, true},
	{1, "lebtimnetz.de", 2, true},
	{1, "leitungsen.de", 2, true},
	{1, "likes-pie.com", 2, true},
	{1, "likescandy.com", 2, true},
	{1, "merseine.nu", 2, true},
	{1, "mine.nu", 2, true},
	{1, "misconfused.org", 2, true},
	{1, "mypets.ws", 2, true},
	{1, "myphotos.cc", 2, true},
	{1, "neat-url.com", 2, true},
	{1, "office-on-the.net", 2, true},
	{1, "on-the-web.tv", 2, true},
	{1, "podzone.net", 2, true},
	{1, "podzone.org", 2, true},
	{1, "readmyblog.org", 2, true},
	{1, "saves-the-whales.com", 2, true},
	{1, "scrapper-site.net", 2, true},
	{1, "scrapping.cc", 2, true},
	{1, "selfip.biz", 2, true},
	{1, "selfip.com", 2, true},
	{1, "selfip.info", 2, true},
	{1, "selfip.net", 2, true},
	{1, "selfip.org", 2, true},
	{1, "sells-for-less.com", 2, true},
	{1, "sells-for-u.com", 2, true},
	{1, "sells-it.net", 2, true},
	{1, "sellsyourhome.org", 2, true},
	{1, "servebbs.com", 2, true},
	{1, "servebbs.net", 2, true},
	{1, "servebbs.org", 2, true},
	{1, "serveftp.net", 2, true},
	{1, "serveftp.org", 2, true},
	{1, "servegame.org", 2, true},
	{1, "shacknet.nu", 2, true},
	{1, "simple-url.com", 2, true},
	{1, "space-to-rent.com", 2, true},
	{1, "stuff-4-sale.org", 2, true},
	{1, "stuff-4-sale.us", 2, true},
	{1, "teaches-yoga.com", 2, true},
	{1, "thruhere.net", 2, true},
	{1, "traeumtgerade.de", 2, true},
	{1, "webhop.biz", 2, true},
	{1, "webhop.info", 2, true},
	{1, "webhop.net", 2, true},
	{1, "webhop.org", 2, true},
	{1, "worse-than.tv", 2, true},
	{1, "writesthisblog.com", 2, true},
	{1, "ddnss.de", 2, true},
	{1, "dyn.ddnss.de", 3, true},
	{1, "dyndns.ddnss.de", 3, true},
	{1, "dyndns1.de", 2, true},
	{1, "dyn-ip24.de", 2, true},
	{1, "home-webserver.de", 2, true},
	{1, "dyn.home-webserver.de", 3, true},
	{1, "myhome-server.de", 2, true},
	{1, "ddnss.org", 2, true},
	{1, "definima.net", 2, true},
	{1, "definima.io", 2, true},
	{1, "ondigitalocean.app", 2, true},
	{1, "bci.dnstrace.pro", 3, true},
	{1, "ddnsfree.com", 2, true},
	{1, "ddnsgeek.com", 2, true},
	{1, "giize.com", 2, true},
	{1, "gleeze.com", 2, true},
	{1, "kozow.com", 2, true},
	{1, "loseyourip.com", 2, true},
	{1, "ooguy.com", 2, true},
	{1, "theworkpc.com", 2, true},
	{1, "casacam.net", 2, true},
	{1, "dynu.net", 2, true},
	{1, "accesscam.org", 2, true},
	{1, "camdvr.org", 2, true},
	{1, "freeddns.org", 2, true},
	{1, "mywire.org", 2, true},
	{1, "webredirect.org", 2, true},
	{1, "myddns.rocks", 2, true},
	{1, "blogsite.xyz", 2, true},
	{1, "dynv6.net", 2, true},
	{1, "e4.cz", 2, true},
	{1, "en-root.fr", 2, true},
	{1, "mytuleap.com", 2, true},
	{1, "onred.one", 2, true},
	{1, "staging.onred.one", 3, true},
	{1, "service.one", 2, true},
	{1, "enonic.io", 2, true},
	{1, "customer.enonic.io", 3, true},
	{1, "eu.org", 2, true},
	{1, "al.eu.org", 3, true},
	{1, "asso.eu.org", 3, true},
	{1, "at.eu.org", 3, true},
	{1, "au.eu.org", 3, true},
	{1, "be.eu.org", 3, true},
	{1, "bg.eu.org", 3, true},
	{1, "ca.eu.org", 3, true},
	{1, "cd.eu.org", 3, true},
	{1, "ch.eu.org", 3, true},
	{1, "cn.eu.org", 3, true},
	{1, "cy.eu.org", 3, true},
	{1, "cz.eu.org", 3, true},
	{1, "de.eu.org", 3, true},
	{1, "dk.eu.org", 3, true},
	{1, "edu.eu.org", 3, true},
	{1, "ee.eu.org", 3, true},
	{1, "es.eu.org", 3, true},
	{1, "fi.eu.org", 3, true},
	{1, "fr.eu.org", 3, true},
	{1, "gr.eu.org", 3, true},
	{1, "hr.eu.org", 3, true},
	{1, "hu.eu.org", 3, true},
	{1, "ie.eu.org", 3, true},
	{1, "il.eu.org", 3, true},
	{1, "in.eu.org", 3, true},
	{1, "int.eu.org", 3, true},
	{1, "is.eu.org", 3, true},
	{1, "it.eu.org", 3, true},
	{1, "jp.eu.org", 3, true},
	{1, "kr.eu.org", 3, true},
	{1, "lt.eu.org", 3, true},
	{1, "lu.eu.org", 3, true},
	{1, "lv.eu.org", 3, true},
	{1, "mc.eu.org", 3, true},
	{1, "me.eu.org", 3, true},
	{1, "mk.eu.org", 3, true},
	{1, "mt.eu.org", 3, true},
	{1, "my.eu.org", 3, true},
	{1, "net.eu.org", 3, true},
	{1, "ng.eu.org", 3, true},
	{1, "nl.eu.org", 3, true},
	{1, "no.eu.org", 3, true},
	{1, "nz.eu.org", 3, true},
	{1, "paris.eu.org", 3, true},
	{1, "pl.eu.org", 3, true},
	{1, "pt.eu.org", 3, true},
	{1, "q-a.eu.org", 3, true},
	{1, "ro.eu.org", 3, true},
	{1, "ru.eu.org", 3, true},
	{1, "se.eu.org", 3, true},
	{1, "si.eu.org", 3, true},
	{1, "sk.eu.org", 3, true},
	{1, "tr.eu.org", 3, true},
	{1, "uk.eu.org", 3, true},
	{1, "us.eu.org", 3, true},
	{1, "eurodir.ru", 2, true},
	{1, "eu-1.evennode.com", 3, true},
	{1, "eu-2.evennode.com", 3, true},
	{1, "eu-3.evennode.com", 3, true},
	{1, "eu-4.evennode.com", 3, true},
	{1, "us-1.evennode.com", 3, true},
	{1, "us-2.evennode.com", 3, true},
	{1, "us-3.evennode.com", 3, true},
	{1, "us-4.evennode.com", 3, true},
	{1, "twmail.cc", 2, true},
	{1, "twmail.net", 2, true},
	{1, "twmail.org", 2, true},
	{1, "mymailer.com.tw", 3, true},
	{1, "url.tw", 2, true},
	{1, "onfabrica.com", 2, true},
	{1, "apps.fbsbx.com", 3, true},
	{1, "ru.net", 2, true},
	{1, "adygeya.ru", 2, true},
	{1, "bashkiria.ru", 2, true},
	{1, "bir.ru", 2, true},
	{1, "cbg.ru", 2, true},
	{1, "com.ru", 2, true},
	{1, "dagestan.ru", 2, true},
	{1, "grozny.ru", 2, true},
	{1, "kalmykia.ru", 2, true},
	{1, "kustanai.ru", 2, true},
	{1, "marine.ru", 2, true},
	{1, "mordovia.ru", 2, true},
	{1, "msk.ru", 2, true},
	{1, "mytis.ru", 2, true},
	{1, "nalchik.ru", 2, true},
	{1, "nov.ru", 2, true},
	{1, "pyatigorsk.ru", 2, true},
	{1, "spb.ru", 2, true},
	{1, "vladikavkaz.ru", 2, true},
	{1, "vladimir.ru", 2, true},
	{1, "abkhazia.su", 2, true},
	{1, "adygeya.su", 2, true},
	{1, "aktyubinsk.su", 2, true},
	{1, "arkhangelsk.su", 2, true},
	{1, "armenia.su", 2, true},
	{1, "ashgabad.su", 2, true},
	{1, "azerbaijan.su", 2, true},
	{1, "balashov.su", 2, true},
	{1, "bashkiria.su", 2, true},
	{1, "bryansk.su", 2, true},
	{1, "bukhara.su", 2, true},
	{1, "chimkent.su", 2, true},
	{1, "dagestan.su", 2, true},
	{1, "east-kazakhstan.su", 2, true},
	{1, "exnet.su", 2, true},
	{1, "georgia.su", 2, true},
	{1, "grozny.su", 2, true},
	{1, "ivanovo.su", 2, true},
	{1, "jambyl.su", 2, true},
	{1, "kalmykia.su", 2, true},
	{1, "kaluga.su", 2, true},
	{1, "karacol.su", 2, true},
	{1, "karaganda.su", 2, true},
	{1, "karelia.su", 2, true},
	{1, "khakassia.su", 2, true},
	{1, "krasnodar.su", 2, true},
	{1, "kurgan.su", 2, true},
	{1, "kustanai.su", 2, true},
	{1, "lenug.su", 2, true},
	{1, "mangyshlak.su", 2, true},
	{1, "mordovia.su", 2, true},
	{1, "msk.su", 2, true},
	{1, "murmansk.su", 2, true},
	{1, "nalchik.su", 2, true},
	{1, "navoi.su", 2, true},
	{1, "north-kazakhstan.su", 2, true},
	{1, "nov.su", 2, true},
	{1, "obninsk.su", 2, true},
	{1, "penza.su", 2, true},
	{1, "pokrovsk.su", 2, true},
	{1, "sochi.su", 2, true},
	{1, "spb.su", 2, true},
	{1, "tashkent.su", 2, true},
	{1, "termez.su", 2, true},
	{1, "togliatti.su", 2, true},
	{1, "troitsk.su", 2, true},
	{1, "tselinograd.su", 2, true},
	{1, "tula.su", 2, true},
	{1, "tuva.su", 2, true},
	{1, "vladikavkaz.su", 2, true},
	{1, "vladimir.su", 2, true},
	{1, "vologda.su", 2, true},
	{1, "channelsdvr.net", 2, true},
	{1, "u.channelsdvr.net", 3, true},
	{1, "edgecompute.app", 2, true},
	{1, "fastly-terrarium.com", 2, true},
	{1, "fastlylb.net", 2, true},
	{1, "map.fastlylb.net", 3, true},
	{1, "freetls.fastly.net", 3, true},
	{1, "map.fastly.net", 3, true},
	{1, "a.prod.fastly.net", 4, true},
	{1, "global.prod.fastly.net", 4, true},
	{1, "a.ssl.fastly.net", 4, true},
	{1, "b.ssl.fastly.net", 4, true},
	{1, "global.ssl.fastly.net", 4, true},
	{1, "fastvps-server.com", 2, true},
	{1, "fastvps.host", 2, true},
	{1, "myfast.host", 2, true},
	{1, "fastvps.site", 2, true},
	{1, "myfast.space", 2, true},
	{1, "fedorainfracloud.org", 2, true},
	{1, "fedorapeople.org", 2, true},
	{1, "cloud.fedoraproject.org", 3, true},
	{1, "app.os.fedoraproject.org", 4, true},
	{1, "app.os.stg.fedoraproject.org", 5, true},
	{1, "couk.me", 2, true},
	{1, "ukco.me", 2, true},
	{1, "conn.uk", 2, true},
	{1, "copro.uk", 2, true},
	{1, "hosp.uk", 2, true},
	{1, "mydobiss.com", 2, true},
	{1, "fh-muenster.io", 2, true},
	{1, "filegear.me", 2, true},
	{1, "filegear-au.me", 2, true},
	{1, "filegear-de.me", 2, true},
	{1, "filegear-gb.me", 2, true},
	{1, "filegear-ie.me", 2, true},
	{1, "filegear-jp.me", 2, true},
	{1, "filegear-sg.me", 2, true},
	{1, "firebaseapp.com", 2, true},
	{1, "fireweb.app", 2, true},
	{1, "flap.id", 2, true},
	{1, "fly.dev", 2, true},
	{1, "edgeapp.net", 2, true},
	{1, "shw.io", 2, true},
	{1, "flynnhosting.net", 2, true},
	{1, "forgeblocks.com", 2, true},
	{2, "id.forgerock.io", 4, true},
	{1, "framer.app", 2, true},
	{1, "framercanvas.com", 2, true},
	{1, "ravpage.co.il", 3, true},
	{1, "0e.vc", 2, true},
	{1, "freebox-os.com", 2, true},
	{1, "freeboxos.com", 2, true},
	{1, "fbx-os.fr", 2, true},
	{1, "fbxos.fr", 2, true},
	{1, "freebox-os.fr", 2, true},
	{1, "freeboxos.fr", 2, true},
	{1, "freedesktop.org", 2, true},
	{1, "freemyip.com", 2, true},
	{1, "wien.funkfeuer.at", 3, true},
	{2, "futurecms.at", 3, true},
	{2, "ex.futurecms.at", 4, true},
	{2, "in.futurecms.at", 4, true},
	{1, "futurehosting.at", 2, true},
	{1, "futuremailing.at", 2, true},
	{2, "ex.ortsinfo.at", 4, true},
	{2, "kunden.ortsinfo.at", 4, true},
	{2, "statics.cloud", 3, true},
	{1, "service.gov.uk", 3, true},
	{1, "gehirn.ne.jp", 3, true},
	{1, "usercontent.jp", 2, true},
	{1, "gentapps.com", 2, true},
	{1, "gentlentapis.com", 2, true},
	{1, "lab.ms", 2, true},
	{1, "cdn-edges.net", 2, true},
	{1, "ghost.io", 2, true},
	{1, "gsj.bz", 2, true},
	{1, "githubusercontent.com", 2, true},
	{1, "github.dev", 2, true},
	{1, "githubpreview.dev", 2, true},
	{1, "github.io", 2, true},
	{1, "gitlab.io", 2, true},
	{1, "gitapp.si", 2, true},
	{1, "gitpage.si", 2, true},
	{1, "glitch.me", 2, true},
	{1, "co.ro", 2, true},
	{1, "shop.ro", 2, true},
	{1, "lolipop.io", 2, true},
	{1, "cloudapps.digital", 2, true},
	{1, "london.cloudapps.digital", 3, true},
	{1, "pymnt.uk", 2, true},
	{1, "homeoffice.gov.uk", 3, true},
	{1, "ro.im", 2, true},
	{1, "goip.de", 2, true},
	{1, "run.app", 2, true},
	{1, "a.run.app", 3, true},
	{1, "web.app", 2, true},
	{2, "0emm.com", 3, true},
	{1, "appspot.com", 2, true},
	{2, "r.appspot.com", 4, true},
	{1, "codespot.com", 2, true},
	{1, "googleapis.com", 2, true},
	{1, "googlecode.com", 2, true},
	{1, "pagespeedmobilizer.com", 2, true},
	{1, "publishproxy.com", 2, true},
	{1, "withgoogle.com", 2, true},
	{1, "withyoutube.com", 2, true},
	{2, "gateway.dev", 3, true},
	{1, "cloud.goog", 2, true},
	{1, "translate.goog", 2, true},
	{1, "cloudfunctions.net", 2, true},
	{1, "blogspot.ae", 2, true},
	{1, "blogspot.al", 2, true},
	{1, "blogspot.am", 2, true},
	{1, "blogspot.ba", 2, true},
	{1, "blogspot.be", 2, true},
	{1, "blogspot.bg", 2, true},
	{1, "blogspot.bj", 2, true},
	{1, "blogspot.ca", 2, true},
	{1, "blogspot.cf", 2, true},
	{1, "blogspot.ch", 2, true},
	{1, "blogspot.cl", 2, true},
	{1, "blogspot.co.at", 3, true},
	{1, "blogspot.co.id", 3, true},
	{1, "blogspot.co.il", 3, true},
	{1, "blogspot.co.ke", 3, true},
	{1, "blogspot.co.nz", 3, true},
	{1, "blogspot.co.uk", 3, true},
	{1, "blogspot.co.za", 3, true},
	{1, "blogspot.com", 2, true},
	{1, "blogspot.com.ar", 3, true},
	{1, "blogspot.com.au", 3, true},
	{1, "blogspot.com.br", 3, true},
	{1, "blogspot.com.by", 3, true},
	{1, "blogspot.com.co", 3, true},
	{1, "blogspot.com.cy", 3, true},
	{1, "blogspot.com.ee", 3, true},
	{1, "blogspot.com.eg", 3, true},
	{1, "blogspot.com.es", 3, true},
	{1, "blogspot.com.mt", 3, true},
	{1, "blogspot.com.ng", 3, true},
	{1, "blogspot.com.tr", 3, true},
	{1, "blogspot.com.uy", 3, true},
	{1, "blogspot.cv", 2, true},
	{1, "blogspot.cz", 2, true},
	{1, "blogspot.de", 2, true},
	{1, "blogspot.dk", 2, true},
	{1, "blogspot.fi", 2, true},
	{1, "blogspot.fr", 2, true},
	{1, "blogspot.gr", 2, true},
	{1, "blogspot.hk", 2, true},
	{1, "blogspot.hr", 2, true},
	{1, "blogspot.hu", 2, true},
	{1, "blogspot.ie", 2, true},
	{1, "blogspot.in", 2, true},
	{1, "blogspot.is", 2, true},
	{1, "blogspot.it", 2, true},
	{1, "blogspot.jp", 2, true},
	{1, "blogspot.kr", 2, true},
	{1, "blogspot.li", 2, true},
	{1, "blogspot.lt", 2, true},
	{1, "blogspot.lu", 2, true},
	{1, "blogspot.md", 2, true},
	{1, "blogspot.mk", 2, true},
	{1, "blogspot.mr", 2, true},
	{1, "blogspot.mx", 2, true},
	{1, "blogspot.my", 2, true},
	{1, "blogspot.nl", 2, true},
	{1, "blogspot.no", 2, true},
	{1, "blogspot.pe", 2, true},
	{1, "blogspot.pt", 2, true},
	{1, "blogspot.qa", 2, true},
	{1, "blogspot.re", 2, true},
	{1, "blogspot.ro", 2, true},
	{1, "blogspot.rs", 2, true},
	{1, "blogspot.ru", 2, true},
	{1, "blogspot.se", 2, true},
	{1, "blogspot.sg", 2, true},
	{1, "blogspot.si", 2, true},
	{1, "blogspot.sk", 2, true},
	{1, "blogspot.sn", 2, true},
	{1, "blogspot.td", 2, true},
	{1, "blogspot.tw", 2, true},
	{1, "blogspot.ug", 2, true},
	{1, "blogspot.vn", 2, true},
	{1, "goupile.fr", 2, true},
	{1, "awsmppl.com", 2, true},
	{1, "xn--gnstigbestellen-zvb.de", 2, true},
	{1, "xn--gnstigliefern-wob.de", 2, true},
	{1, "fin.ci", 2, true},
	{1, "free.hr", 2, true},
	{1, "caa.li", 2, true},
	{1, "ua.rs", 2, true},
	{1, "conf.se", 2, true},
	{1, "hs.zone", 2, true},
	{1, "hs.run", 2, true},
	{1, "hashbang.sh", 2, true},
	{1, "hasura.app", 2, true},
	{1, "hasura-app.io", 2, true},
	{1, "hepforge.org", 2, true},
	{1, "herokuapp.com", 2, true},
	{1, "herokussl.com", 2, true},
	{1, "myravendb.com", 2, true},
	{1, "ravendb.community", 2, true},
	{1, "ravendb.me", 2, true},
	{1, "development.run", 2, true},
	{1, "ravendb.run", 2, true},
	{1, "secaas.hk", 2, true},
	{1, "orx.biz", 2, true},
	{1, "biz.gl", 2, true},
	{1, "col.ng", 2, true},
	{1, "firm.ng", 2, true},
	{1, "gen.ng", 2, true},
	{1, "ltd.ng", 2, true},
	{1, "ngo.ng", 2, true},
	{1, "edu.scot", 2, true},
	{1, "sch.so", 2, true},
	{1, "org.yt", 2, true},
	{1, "hostyhosting.io", 2, true},
	{1, "xn--hkkinen-5wa.fi", 2, true},
	{2, "moonscale.io", 3, true},
	{1, "moonscale.net", 2, true},
	{1, "iki.fi", 2, true},
	{1, "impertrixcdn.com", 2, true},
	{1, "impertrix.com", 2, true},
	{1, "smushcdn.com", 2, true},
	{1, "wphostedmail.com", 2, true},
	{1, "wpmucdn.com", 2, true},
	{1, "tempurl.host", 2, true},
	{1, "wpmudev.host", 2, true},
	{1, "dyn-berlin.de", 2, true},
	{1, "in-berlin.de", 2, true},
	{1, "in-brb.de", 2, true},
	{1, "in-butter.de", 2, true},
	{1, "in-dsl.de", 2, true},
	{1, "in-dsl.net", 2, true},
	{1, "in-dsl.org", 2, true},
	{1, "in-vpn.de", 2, true},
	{1, "in-vpn.net", 2, true},
	{1, "in-vpn.org", 2, true},
	{1, "biz.at", 2, true},
	{1, "info.at", 2, true},
	{1, "info.cx", 2, true},
	{1, "ac.leg.br", 3, true},
	{1, "al.leg.br", 3, true},
	{1, "am.leg.br", 3, true},
	{1, "ap.leg.br", 3, true},
	{1, "ba.leg.br", 3, true},
	{1, "ce.leg.br", 3, true},
	{1, "df.leg.br", 3, true},
	{1, "es.leg.br", 3, true},
	{1, "go.leg.br", 3, true},
	{1, "ma.leg.br", 3, true},
	{1, "mg.leg.br", 3, true},
	{1, "ms.leg.br", 3, true},
	{1, "mt.leg.br", 3, true},
	{1, "pa.leg.br", 3, true},
	{1, "pb.leg.br", 3, true},
	{1, "pe.leg.br", 3, true},
	{1, "pi.leg.br", 3, true},
	{1, "pr.leg.br", 3, true},
	{1, "rj.leg.br", 3, true},
	{1, "rn.leg.br", 3, true},
	{1, "ro.leg.br", 3, true},
	{1, "rr.leg.br", 3, true},
	{1, "rs.leg.br", 3, true},
	{1, "sc.leg.br", 3, true},
	{1, "se.leg.br", 3, true},
	{1, "sp.leg.br", 3, true},
	{1, "to.leg.br", 3, true},
	{1, "pixolino.com", 2, true},
	{1, "na4u.ru", 2, true},
	{1, "iopsys.se", 2, true},
	{1, "ipifony.net", 2, true},
	{1, "mein-iserv.de", 2, true},
	{1, "schulserver.de", 2, true},
	{1, "test-iserv.de", 2, true},
	{1, "iserv.dev", 2, true},
	{1, "iobb.net", 2, true},
	{1, "mel.cloudlets.com.au", 4, true},
	{1, "cloud.interhostsolutions.be", 3, true},
	{1, "users.scale.virtualcloud.com.br", 5, true},
	{1, "mycloud.by", 2, true},
	{1, "alp1.ae.flow.ch", 4, true},
	{1, "appengine.flow.ch", 3, true},
	{1, "es-1.axarnet.cloud", 3, true},
	{1, "diadem.cloud", 2, true},
	{1, "vip.jelastic.cloud", 3, true},
	{1, "jele.cloud", 2, true},
	{1, "it1.eur.aruba.jenv-aruba.cloud", 5, true},
	{1, "it1.jenv-aruba.cloud", 3, true},
	{1, "keliweb.cloud", 2, true},
	{1, "cs.keliweb.cloud", 3, true},
	{1, "oxa.cloud", 2, true},
	{1, "tn.oxa.cloud", 3, true},
	{1, "uk.oxa.cloud", 3, true},
	{1, "primetel.cloud", 2, true},
	{1, "uk.primetel.cloud", 3, true},
	{1, "ca.reclaim.cloud", 3, true},
	{1, "uk.reclaim.cloud", 3, true},
	{1, "us.reclaim.cloud", 3, true},
	{1, "ch.trendhosting.cloud", 3, true},
	{1, "de.trendhosting.cloud", 3, true},
	{1, "jele.club", 2, true},
	{1, "amscompute.com", 2, true},
	{1, "clicketcloud.com", 2, true},
	{1, "dopaas.com", 2, true},
	{1, "hidora.com", 2, true},
	{1, "paas.hosted-by-previder.com", 3, true},
	{1, "rag-cloud.hosteur.com", 3, true},
	{1, "rag-cloud-ch.hosteur.com", 3, true},
	{1, "jcloud.ik-server.com", 3, true},
	{1, "jcloud-ver-jpc.ik-server.com", 3, true},
	{1, "demo.jelastic.com", 3, true},
	{1, "kilatiron.com", 2, true},
	{1, "paas.massivegrid.com", 3, true},
	{1, "jed.wafaicloud.com", 3, true},
	{1, "lon.wafaicloud.com", 3, true},
	{1, "ryd.wafaicloud.com", 3, true},
	{1, "j.scaleforce.com.cy", 4, true},
	{1, "jelastic.dogado.eu", 3, true},
	{1, "fi.cloudplatform.fi", 3, true},
	{1, "demo.datacenter.fi", 3, true},
	{1, "paas.datacenter.fi", 3, true},
	{1, "jele.host", 2, true},
	{1, "mircloud.host", 2, true},
	{1, "paas.beebyte.io", 3, true},
	{1, "sekd1.beebyteapp.io", 3, true},
	{1, "jele.io", 2, true},
	{1, "cloud-fr1.unispace.io", 3, true},
	{1, "jc.neen.it", 3, true},
	{1, "cloud.jelastic.open.tim.it", 5, true},
	{1, "jcloud.kz", 2, true},
	{1, "upaas.kazteleport.kz", 3, true},
	{1, "cloudjiffy.net", 2, true},
	{1, "fra1-de.cloudjiffy.net", 3, true},
	{1, "west1-us.cloudjiffy.net", 3, true},
	{1, "jls-sto1.elastx.net", 3, true},
	{1, "jls-sto2.elastx.net", 3, true},
	{1, "jls-sto3.elastx.net", 3, true},
	{1, "faststacks.net", 2, true},
	{1, "fr-1.paas.massivegrid.net", 4, true},
	{1, "lon-1.paas.massivegrid.net", 4, true},
	{1, "lon-2.paas.massivegrid.net", 4, true},
	{1, "ny-1.paas.massivegrid.net", 4, true},
	{1, "ny-2.paas.massivegrid.net", 4, true},
	{1, "sg-1.paas.massivegrid.net", 4, true},
	{1, "jelastic.saveincloud.net", 3, true},
	{1, "nordeste-idc.saveincloud.net", 3, true},
	{1, "j.scaleforce.net", 3, true},
	{1, "jelastic.tsukaeru.net", 3, true},
	{1, "sdscloud.pl", 2, true},
	{1, "unicloud.pl", 2, true},
	{1, "mircloud.ru", 2, true},
	{1, "jelastic.regruhosting.ru", 3, true},
	{1, "enscaled.sg", 2, true},
	{1, "jele.site", 2, true},
	{1, "jelastic.team", 2, true},
	{1, "orangecloud.tn", 2, true},
	{1, "j.layershift.co.uk", 4, true},
	{1, "phx.enscaled.us", 3, true},
	{1, "mircloud.us", 2, true},
	{1, "myjino.ru", 2, true},
	{2, "hosting.myjino.ru", 4, true},
	{2, "landing.myjino.ru", 4, true},
	{2, "spectrum.myjino.ru", 4, true},
	{2, "vps.myjino.ru", 4, true},
	{2, "triton.zone", 3, true},
	{2, "cns.joyent.com", 4, true},
	{1, "js.org", 2, true},
	{1, "kaas.gg", 2, true},
	{1, "khplay.nl", 2, true},
	{1, "keymachine.de", 2, true},
	{1, "kinghost.net", 2, true},
	{1, "uni5.net", 2, true},
	{1, "knightpoint.systems", 2, true},
	{1, "oya.to", 2, true},
	{1, "kuleuven.cloud", 2, true},
	{1, "ezproxy.kuleuven.be", 3, true},
	{1, "co.krd", 2, true},
	{1, "edu.krd", 2, true},
	{1, "krellian.net", 2, true},
	{1, "webthings.io", 2, true},
	{1, "git-repos.de", 2, true},
	{1, "lcube-server.de", 2, true},
	{1, "svn-repos.de", 2, true},
	{1, "leadpages.co", 2, true},
	{1, "lpages.co", 2, true},
	{1, "lpusercontent.com", 2, true},
	{1, "lelux.site", 2, true},
	{1, "co.business", 2, true},
	{1, "co.education", 2, true},
	{1, "co.events", 2, true},
	{1, "co.financial", 2, true},
	{1, "co.network", 2, true},
	{1, "co.place", 2, true},
	{1, "co.technology", 2, true},
	{1, "app.lmpm.com", 3, true},
	{1, "linkyard.cloud", 2, true},
	{1, "linkyard-cloud.ch", 2, true},
	{1, "members.linode.com", 3, true},
	{2, "nodebalancer.linode.com", 4, true},
	{2, "linodeobjects.com", 3, true},
	{1, "we.bs", 2, true},
	{1, "localzone.xyz", 2, true},
	{1, "loginline.app", 2, true},
	{1, "loginline.dev", 2, true},
	{1, "loginline.io", 2, true},
	{1, "loginline.services", 2, true},
	{1, "loginline.site", 2, true},
	{1, "lohmus.me", 2, true},
	{1, "krasnik.pl", 2, true},
	{1, "leczna.pl", 2, true},
	{1, "lubartow.pl", 2, true},
	{1, "lublin.pl", 2, true},
	{1, "poniatowa.pl", 2, true},
	{1, "swidnik.pl", 2, true},
	{1, "glug.org.uk", 3, true},
	{1, "lug.org.uk", 3, true},
	{1, "lugs.org.uk", 3, true},
	{1, "barsy.bg", 2, true},
	{1, "barsy.co.uk", 3, true},
	{1, "barsyonline.co.uk", 3, true},
	{1, "barsycenter.com", 2, true},
	{1, "barsyonline.com", 2, true},
	{1, "barsy.club", 2, true},
	{1, "barsy.de", 2, true},
	{1, "barsy.eu", 2, true},
	{1, "barsy.in", 2, true},
	{1, "barsy.info", 2, true},
	{1, "barsy.io", 2, true},
	{1, "barsy.me", 2, true},
	{1, "barsy.menu", 2, true},
	{1, "barsy.mobi", 2, true},
	{1, "barsy.net", 2, true},
	{1, "barsy.online", 2, true},
	{1, "barsy.org", 2, true},
	{1, "barsy.pro", 2, true},
	{1, "barsy.pub", 2, true},
	{1, "barsy.shop", 2, true},
	{1, "barsy.site", 2, true},
	{1, "barsy.support", 2, true},
	{1, "barsy.uk", 2, true},
	{2, "magentosite.cloud", 3, true},
	{1, "mayfirst.info", 2, true},
	{1, "mayfirst.org", 2, true},
	{1, "hb.cldmail.ru", 3, true},
	{1, "cn.vu", 2, true},
	{1, "mazeplay.com", 2, true},
	{1, "mcpe.me", 2, true},
	{1, "mcdir.me", 2, true},
	{1, "mcdir.ru", 2, true},
	{1, "mcpre.ru", 2, true},
	{1, "vps.mcdir.ru", 3, true},
	{1, "hra.health", 2, true},
	{1, "miniserver.com", 2, true},
	{1, "memset.net", 2, true},
	{2, "cloud.metacentrum.cz", 4, true},
	{1, "custom.metacentrum.cz", 3, true},
	{1, "flt.cloud.muni.cz", 4, true},
	{1, "usr.cloud.muni.cz", 4, true},
	{1, "meteorapp.com", 2, true},
	{1, "eu.meteorapp.com", 3, true},
	{1, "co.pl", 2, true},
	{2, "azurecontainer.io", 3, true},
	{1, "azurewebsites.net", 2, true},
	{1, "azure-mobile.net", 2, true},
	{1, "cloudapp.net", 2, true},
	{1, "azurestaticapps.net", 2, true},
	{1, "centralus.azurestaticapps.net", 3, true},
	{1, "eastasia.azurestaticapps.net", 3, true},
	{1, "eastus2.azurestaticapps.net", 3, true},
	{1, "westeurope.azurestaticapps.net", 3, true},
	{1, "westus2.azurestaticapps.net", 3, true},
	{1, "csx.cc", 2, true},
	{1, "mintere.site", 2, true},
	{1, "forte.id", 2, true},
	{1, "mozilla-iot.org", 2, true},
	{1, "bmoattachments.org", 2, true},
	{1, "net.ru", 2, true},
	{1, "org.ru", 2, true},
	{1, "pp.ru", 2, true},
	{1, "hostedpi.com", 2, true},
	{1, "customer.mythic-beasts.com", 3, true},
	{1, "caracal.mythic-beasts.com", 3, true},
	{1, "fentiger.mythic-beasts.com", 3, true},
	{1, "lynx.mythic-beasts.com", 3, true},
	{1, "ocelot.mythic-beasts.com", 3, true},
	{1, "oncilla.mythic-beasts.com", 3, true},
	{1, "onza.mythic-beasts.com", 3, true},
	{1, "sphinx.mythic-beasts.com", 3, true},
	{1, "vs.mythic-beasts.com", 3, true},
	{1, "x.mythic-beasts.com", 3, true},
	{1, "yali.mythic-beasts.com", 3, true},
	{1, "cust.retrosnub.co.uk", 4, true},
	{1, "ui.nabu.casa", 3, true},
	{1, "pony.club", 2, true},
	{1, "of.fashion", 2, true},
	{1, "in.london", 2, true},
	{1, "of.london", 2, true},
	{1, "from.marketing", 2, true},
	{1, "with.marketing", 2, true},
	{1, "for.men", 2, true},
	{1, "repair.men", 2, true},
	{1, "and.mom", 2, true},
	{1, "for.mom", 2, true},
	{1, "for.one", 2, true},
	{1, "under.one", 2, true},
	{1, "for.sale", 2, true},
	{1, "that.win", 2, true},
	{1, "from.work", 2, true},
	{1, "to.work", 2, true},
	{1, "nctu.me", 2, true},
	{1, "netlify.app", 2, true},
	{1, "4u.com", 2, true},
	{1, "ngrok.io", 2, true},
	{1, "nh-serv.co.uk", 3, true},
	{1, "nfshost.com", 2, true},
	{2, "developer.app", 3, true},
	{1, "noop.app", 2, true},
	{2, "northflank.app", 3, true},
	{2, "code.run", 3, true},
	{1, "noticeable.news", 2, true},
	{1, "dnsking.ch", 2, true},
	{1, "mypi.co", 2, true},
	{1, "n4t.co", 2, true},
	{1, "001www.com", 2, true},
	{1, "ddnslive.com", 2, true},
	{1, "myiphost.com", 2, true},
	{1, "forumz.info", 2, true},
	{1, "16-b.it", 2, true},
	{1, "32-b.it", 2, true},
	{1, "64-b.it", 2, true},
	{1, "soundcast.me", 2, true},
	{1, "tcp4.me", 2, true},
	{1, "dnsup.net", 2, true},
	{1, "hicam.net", 2, true},
	{1, "now-dns.net", 2, true},
	{1, "ownip.net", 2, true},
	{1, "vpndns.net", 2, true},
	{1, "dynserv.org", 2, true},
	{1, "now-dns.org", 2, true},
	{1, "x443.pw", 2, true},
	{1, "now-dns.top", 2, true},
	{1, "ntdll.top", 2, true},
	{1, "freeddns.us", 2, true},
	{1, "crafting.xyz", 2, true},
	{1, "zapto.xyz", 2, true},
	{1, "nsupdate.info", 2, true},
	{1, "nerdpol.ovh", 2, true},
	{1, "blogsyte.com", 2, true},
	{1, "brasilia.me", 2, true},
	{1, "cable-modem.org", 2, true},
	{1, "ciscofreak.com", 2, true},
	{1, "collegefan.org", 2, true},
	{1, "couchpotatofries.org", 2, true},
	{1, "damnserver.com", 2, true},
	{1, "ddns.me", 2, true},
	{1, "ditchyourip.com", 2, true},
	{1, "dnsfor.me", 2, true},
	{1, "dnsiskinky.com", 2, true},
	{1, "dvrcam.info", 2, true},
	{1, "dynns.com", 2, true},
	{1, "eating-organic.net", 2, true},
	{1, "fantasyleague.cc", 2, true},
	{1, "geekgalaxy.com", 2, true},
	{1, "golffan.us", 2, true},
	{1, "health-carereform.com", 2, true},
	{1, "homesecuritymac.com", 2, true},
	{1, "homesecuritypc.com", 2, true},
	{1, "hopto.me", 2, true},
	{1, "ilovecollege.info", 2, true},
	{1, "loginto.me", 2, true},
	{1, "mlbfan.org", 2, true},
	{1, "mmafan.biz", 2, true},
	{1, "myactivedirectory.com", 2, true},
	{1, "mydissent.net", 2, true},
	{1, "myeffect.net", 2, true},
	{1, "mymediapc.net", 2, true},
	{1, "mypsx.net", 2, true},
	{1, "mysecuritycamera.com", 2, true},
	{1, "mysecuritycamera.net", 2, true},
	{1, "mysecuritycamera.org", 2, true},
	{1, "net-freaks.com", 2, true},
	{1, "nflfan.org", 2, true},
	{1, "nhlfan.net", 2, true},
	{1, "no-ip.ca", 2, true},
	{1, "no-ip.co.uk", 3, true},
	{1, "no-ip.net", 2, true},
	{1, "noip.us", 2, true},
	{1, "onthewifi.com", 2, true},
	{1, "pgafan.net", 2, true},
	{1, "point2this.com", 2, true},
	{1, "pointto.us", 2, true},
	{1, "privatizehealthinsurance.net", 2, true},
	{1, "quicksytes.com", 2, true},
	{1, "read-books.org", 2, true},
	{1, "securitytactics.com", 2, true},
	{1, "serveexchange.com", 2, true},
	{1, "servehumour.com", 2, true},
	{1, "servep2p.com", 2, true},
	{1, "servesarcasm.com", 2, true},
	{1, "stufftoread.com", 2, true},
	{1, "ufcfan.org", 2, true},
	{1, "unusualperson.com", 2, true},
	{1, "workisboring.com", 2, true},
	{1, "3utilities.com", 2, true},
	{1, "bounceme.net", 2, true},
	{1, "ddns.net", 2, true},
	{1, "ddnsking.com", 2, true},
	{1, "gotdns.ch", 2, true},
	{1, "hopto.org", 2, true},
	{1, "myftp.biz", 2, true},
	{1, "myftp.org", 2, true},
	{1, "myvnc.com", 2, true},
	{1, "no-ip.biz", 2, true},
	{1, "no-ip.info", 2, true},
	{1, "no-ip.org", 2, true},
	{1, "noip.me", 2, true},
	{1, "redirectme.net", 2, true},
	{1, "servebeer.com", 2, true},
	{1, "serveblog.net", 2, true},
	{1, "servecounterstrike.com", 2, true},
	{1, "serveftp.com", 2, true},
	{1, "servegame.com", 2, true},
	{1, "servehalflife.com", 2, true},
	{1, "servehttp.com", 2, true},
	{1, "serveirc.com", 2, true},
	{1, "serveminecraft.net", 2, true},
	{1, "servemp3.com", 2, true},
	{1, "servepics.com", 2, true},
	{1, "servequake.com", 2, true},
	{1, "sytes.net", 2, true},
	{1, "webhop.me", 2, true},
	{1, "zapto.org", 2, true},
	{1, "stage.nodeart.io", 3, true},
	{1, "nodum.co", 2, true},
	{1, "nodum.io", 2, true},
	{1, "pcloud.host", 2, true},
	{1, "nyc.mn", 2, true},
	{1, "nom.ae", 2, true},
	{1, "nom.af", 2, true},
	{1, "nom.ai", 2, true},
	{1, "nom.al", 2, true},
	{1, "nym.by", 2, true},
	{1, "nom.bz", 2, true},
	{1, "nym.bz", 2, true},
	{1, "nom.cl", 2, true},
	{1, "nym.ec", 2, true},
	{1, "nom.gd", 2, true},
	{1, "nom.ge", 2, true},
	{1, "nom.gl", 2, true},
	{1, "nym.gr", 2, true},
	{1, "nom.gt", 2, true},
	{1, "nym.gy", 2, true},
	{1, "nym.hk", 2, true},
	{1, "nom.hn", 2, true},
	{1, "nym.ie", 2, true},
	{1, "nom.im", 2, true},
	{1, "nom.ke", 2, true},
	{1, "nym.kz", 2, true},
	{1, "nym.la", 2, true},
	{1, "nym.lc", 2, true},
	{1, "nom.li", 2, true},
	{1, "nym.li", 2, true},
	{1, "nym.lt", 2, true},
	{1, "nym.lu", 2, true},
	{1, "nom.lv", 2, true},
	{1, "nym.me", 2, true},
	{1, "nom.mk", 2, true},
	{1, "nym.mn", 2, true},
	{1, "nym.mx", 2, true},
	{1, "nom.nu", 2, true},
	{1, "nym.nz", 2, true},
	{1, "nym.pe", 2, true},
	{1, "nym.pt", 2, true},
	{1, "nom.pw", 2, true},
	{1, "nom.qa", 2, true},
	{1, "nym.ro", 2, true},
	{1, "nom.rs", 2, true},
	{1, "nom.si", 2, true},
	{1, "nym.sk", 2, true},
	{1, "nom.st", 2, true},
	{1, "nym.su", 2, true},
	{1, "nym.sx", 2, true},
	{1, "nom.tj", 2, true},
	{1, "nym.tw", 2, true},
	{1, "nom.ug", 2, true},
	{1, "nom.uy", 2, true},
	{1, "nom.vc", 2, true},
	{1, "nom.vg", 2, true},
	{1, "static.observableusercontent.com", 3, true},
	{1, "cya.gg", 2, true},
	{1, "omg.lol", 2, true},
	{1, "cloudycluster.net", 2, true},
	{1, "omniwe.site", 2, true},
	{1, "nid.io", 2, true},
	{1, "opensocial.site", 2, true},
	{1, "opencraft.hosting", 2, true},
	{1, "orsites.com", 2, true},
	{1, "operaunite.com", 2, true},
	{1, "authgear-staging.com", 2, true},
	{1, "authgearapps.com", 2, true},
	{1, "skygearapp.com", 2, true},
	{1, "outsystemscloud.com", 2, true},
	{2, "webpaas.ovh.net", 4, true},
	{2, "hosting.ovh.net", 4, true},
	{1, "ownprovider.com", 2, true},
	{1, "own.pm", 2, true},
	{2, "owo.codes", 3, true},
	{1, "ox.rs", 2, true},
	{1, "oy.lc", 2, true},
	{1, "pgfog.com", 2, true},
	{1, "pagefrontapp.com", 2, true},
	{1, "pagexl.com", 2, true},
	{2, "paywhirl.com", 3, true},
	{1, "bar0.net", 2, true},
	{1, "bar1.net", 2, true},
	{1, "bar2.net", 2, true},
	{1, "rdv.to", 2, true},
	{1, "art.pl", 2, true},
	{1, "gliwice.pl", 2, true},
	{1, "krakow.pl", 2, true},
	{1, "poznan.pl", 2, true},
	{1, "wroc.pl", 2, true},
	{1, "zakopane.pl", 2, true},
	{1, "pantheonsite.io", 2, true},
	{1, "gotpantheon.com", 2, true},
	{1, "mypep.link", 2, true},
	{1, "perspecta.cloud", 2, true},
	{1, "lk3.ru", 2, true},
	{1, "ra-ru.ru", 2, true},
	{1, "zsew.ru", 2, true},
	{1, "on-web.fr", 2, true},
	{1, "bc.platform.sh", 3, true},
	{1, "ent.platform.sh", 3, true},
	{1, "eu.platform.sh", 3, true},
	{1, "us.platform.sh", 3, true},
	{2, "platformsh.site", 3, true},
	{2, "tst.site", 3, true},
	{1, "platter-app.com", 2, true},
	{1, "platter-app.dev", 2, true},
	{1, "platterp.us", 2, true},
	{1, "pdns.page", 2, true},
	{1, "plesk.page", 2, true},
	{1, "pleskns.com", 2, true},
	{1, "dyn53.io", 2, true},
	{1, "co.bn", 2, true},
	{1, "xen.prgmr.com", 3, true},
	{1, "priv.at", 2, true},
	{1, "prvcy.page", 2, true},
	{2, "dweb.link", 3, true},
	{1, "protonet.io", 2, true},
	{1, "chirurgiens-dentistes-en-france.fr", 2, true},
	{1, "byen.site", 2, true},
	{1, "pubtls.org", 2, true},
	{1, "pythonanywhere.com", 2, true},
	{1, "eu.pythonanywhere.com", 3, true},
	{1, "qoto.io", 2, true},
	{1, "qualifioapp.com", 2, true},
	{1, "qbuser.com", 2, true},
	{1, "cloudsite.builders", 2, true},
	{1, "instantcloud.cn", 2, true},
	{1, "ras.ru", 2, true},
	{1, "qa2.com", 2, true},
	{1, "qcx.io", 2, true},
	{2, "sys.qcx.io", 4, true},
	{1, "dev-myqnapcloud.com", 2, true},
	{1, "alpha-myqnapcloud.com", 2, true},
	{1, "myqnapcloud.com", 2, true},
	{2, "quipelements.com", 3, true},
	{1, "vapor.cloud", 2, true},
	{1, "vaporcloud.io", 2, true},
	{1, "rackmaze.com", 2, true},
	{1, "rackmaze.net", 2, true},
	{1, "g.vbrplsbx.io", 3, true},
	{2, "on-k3s.io", 3, true},
	{2, "on-rancher.cloud", 3, true},
	{2, "on-rio.io", 3, true},
	{1, "readthedocs.io", 2, true},
	{1, "rhcloud.com", 2, true},
	{1, "app.render.com", 3, true},
	{1, "onrender.com", 2, true},
	{1, "repl.co", 2, true},
	{1, "id.repl.co", 3, true},
	{1, "repl.run", 2, true},
	{1, "resindevice.io", 2, true},
	{1, "devices.resinstaging.io", 3, true},
	{1, "hzc.io", 2, true},
	{1, "wellbeingzone.eu", 2, true},
	{1, "wellbeingzone.co.uk", 3, true},
	{1, "git-pages.rit.edu", 3, true},
	{1, "xn--90amc.xn--p1acf", 2, true},
	{1, "xn--j1aef.xn--p1acf", 2, true},
	{1, "xn--j1ael8b.xn--p1acf", 2, true},
	{1, "xn--h1ahn.xn--p1acf", 2, true},
	{1, "xn--j1adp.xn--p1acf", 2, true},
	{1, "xn--c1avg.xn--p1acf", 2, true},
	{1, "xn--80aaa0cvac.xn--p1acf", 2, true},
	{1, "xn--h1aliz.xn--p1acf", 2, true},
	{1, "xn--90a1af.xn--p1acf", 2, true},
	{1, "xn--41a.xn--p1acf", 2, true},
	{1, "sandcats.io", 2, true},
	{1, "logoip.de", 2, true},
	{1, "logoip.com", 2, true},
	{1, "schokokeks.net", 2, true},
	{1, "gov.scot", 2, true},
	{1, "service.gov.scot", 3, true},
	{1, "scrysec.com", 2, true},
	{1, "firewall-gateway.com", 2, true},
	{1, "firewall-gateway.de", 2, true},
	{1, "my-gateway.de", 2, true},
	{1, "my-router.de", 2, true},
	{1, "spdns.de", 2, true},
	{1, "spdns.eu", 2, true},
	{1, "firewall-gateway.net", 2, true},
	{1, "my-firewall.org", 2, true},
	{1, "myfirewall.org", 2, true},
	{1, "spdns.org", 2, true},
	{1, "seidat.net", 2, true},
	{1, "senseering.net", 2, true},
	{1, "magnet.page", 2, true},
	{1, "biz.ua", 2, true},
	{1, "co.ua", 2, true},
	{1, "pp.ua", 2, true},
	{1, "shiftcrypto.dev", 2, true},
	{1, "shiftcrypto.io", 2, true},
	{1, "shiftedit.io", 2, true},
	{1, "myshopblocks.com", 2, true},
	{1, "myshopify.com", 2, true},
	{1, "shopitsite.com", 2, true},
	{1, "shopware.store", 2, true},
	{1, "mo-siemens.io", 2, true},
	{1, "1kapp.com", 2, true},
	{1, "appchizi.com", 2, true},
	{1, "applinzi.com", 2, true},
	{1, "sinaapp.com", 2, true},
	{1, "vipsinaapp.com", 2, true},
	{1, "siteleaf.net", 2, true},
	{1, "bounty-full.com", 2, true},
	{1, "alpha.bounty-full.com", 3, true},
	{1, "beta.bounty-full.com", 3, true},
	{1, "small-web.org", 2, true},
	{1, "try-snowplow.com", 2, true},
	{1, "srht.site", 2, true},
	{1, "stackhero-network.com", 2, true},
	{1, "static.land", 2, true},
	{1, "dev.static.land", 3, true},
	{1, "sites.static.land", 3, true},
	{1, "storebase.store", 2, true},
	{1, "vps-host.net", 2, true},
	{1, "atl.jelastic.vps-host.net", 4, true},
	{1, "njs.jelastic.vps-host.net", 4, true},
	{1, "ric.jelastic.vps-host.net", 4, true},
	{1, "playstation-cloud.com", 2, true},
	{1, "apps.lair.io", 3, true},
	{2, "stolos.io", 3, true},
	{1, "spacekit.io", 2, true},
	{1, "customer.speedpartner.de", 3, true},
	{1, "api.stdlib.com", 3, true},
	{1, "storj.farm", 2, true},
	{1, "utwente.io", 2, true},
	{1, "soc.srcf.net", 3, true},
	{1, "user.srcf.net", 3, true},
	{1, "temp-dns.com", 2, true},
	{2, "s5y.io", 3, true},
	{2, "sensiosite.cloud", 3, true},
	{1, "syncloud.it", 2, true},
	{1, "diskstation.me", 2, true},
	{1, "dscloud.biz", 2, true},
	{1, "dscloud.me", 2, true},
	{1, "dscloud.mobi", 2, true},
	{1, "dsmynas.com", 2, true},
	{1, "dsmynas.net", 2, true},
	{1, "dsmynas.org", 2, true},
	{1, "familyds.com", 2, true},
	{1, "familyds.net", 2, true},
	{1, "familyds.org", 2, true},
	{1, "i234.me", 2, true},
	{1, "myds.me", 2, true},
	{1, "synology.me", 2, true},
	{1, "vpnplus.to", 2, true},
	{1, "direct.quickconnect.to", 3, true},
	{1, "taifun-dns.de", 2, true},
	{1, "gda.pl", 2, true},
	{1, "gdansk.pl", 2, true},
	{1, "gdynia.pl", 2, true},
	{1, "med.pl", 2, true},
	{1, "sopot.pl", 2, true},
	{1, "edugit.org", 2, true},
	{1, "telebit.app", 2, true},
	{1, "telebit.io", 2, true},
	{2, "telebit.xyz", 3, true},
	{1, "gwiddle.co.uk", 3, true},
	{1, "thingdustdata.com", 2, true},
	{1, "cust.dev.thingdust.io", 4, true},
	{1, "cust.disrec.thingdust.io", 4, true},
	{1, "cust.prod.thingdust.io", 4, true},
	{1, "cust.testing.thingdust.io", 4, true},
	{2, "firenet.ch", 3, true},
	{2, "svc.firenet.ch", 4, true},
	{1, "arvo.network", 2, true},
	{1, "azimuth.network", 2, true},
	{1, "tlon.network", 2, true},
	{1, "torproject.net", 2, true},
	{1, "pages.torproject.net", 3, true},
	{1, "bloxcms.com", 2, true},
	{1, "townnews-staging.com", 2, true},
	{1, "tbits.me", 2, true},
	{1, "12hp.at", 2, true},
	{1, "2ix.at", 2, true},
	{1, "4lima.at", 2, true},
	{1, "lima-city.at", 2, true},
	{1, "12hp.ch", 2, true},
	{1, "2ix.ch", 2, true},
	{1, "4lima.ch", 2, true},
	{1, "lima-city.ch", 2, true},
	{1, "trafficplex.cloud", 2, true},
	{1, "de.cool", 2, true},
	{1, "12hp.de", 2, true},
	{1, "2ix.de", 2, true},
	{1, "4lima.de", 2, true},
	{1, "lima-city.de", 2, true},
	{1, "1337.pictures", 2, true},
	{1, "clan.rip", 2, true},
	{1, "lima-city.rocks", 2, true},
	{1, "webspace.rocks", 2, true},
	{1, "lima.zone", 2, true},
	{2, "transurl.be", 3, true},
	{2, "transurl.eu", 3, true},
	{2, "transurl.nl", 3, true},
	{1, "tuxfamily.org", 2, true},
	{1, "dd-dns.de", 2, true},
	{1, "diskstation.eu", 2, true},
	{1, "diskstation.org", 2, true},
	{1, "dray-dns.de", 2, true},
	{1, "draydns.de", 2, true},
	{1, "dyn-vpn.de", 2, true},
	{1, "dynvpn.de", 2, true},
	{1, "mein-vigor.de", 2, true},
	{1, "my-vigor.de", 2, true},
	{1, "my-wan.de", 2, true},
	{1, "syno-ds.de", 2, true},
	{1, "synology-diskstation.de", 2, true},
	{1, "synology-ds.de", 2, true},
	{1, "uber.space", 2, true},
	{2, "uberspace.de", 3, true},
	{1, "hk.com", 2, true},
	{1, "hk.org", 2, true},
	{1, "ltd.hk", 2, true},
	{1, "inc.hk", 2, true},
	{1, "virtualuser.de", 2, true},
	{1, "virtual-user.de", 2, true},
	{1, "urown.cloud", 2, true},
	{1, "dnsupdate.info", 2, true},
	{1, "lib.de.us", 3, true},
	{1, "2038.io", 2, true},
	{1, "vercel.app", 2, true},
	{1, "vercel.dev", 2, true},
	{1, "now.sh", 2, true},
	{1, "router.management", 2, true},
	{1, "v-info.info", 2, true},
	{1, "voorloper.cloud", 2, true},
	{1, "neko.am", 2, true},
	{1, "nyaa.am", 2, true},
	{1, "be.ax", 2, true},
	{1, "cat.ax", 2, true},
	{1, "es.ax", 2, true},
	{1, "eu.ax", 2, true},
	{1, "gg.ax", 2, true},
	{1, "mc.ax", 2, true},
	{1, "us.ax", 2, true},
	{1, "xy.ax", 2, true},
	{1, "nl.ci", 2, true},
	{1, "xx.gl", 2, true},
	{1, "app.gp", 2, true},
	{1, "blog.gt", 2, true},
	{1, "de.gt", 2, true},
	{1, "to.gt", 2, true},
	{1, "be.gy", 2, true},
	{1, "cc.hn", 2, true},
	{1, "blog.kg", 2, true},
	{1, "io.kg", 2, true},
	{1, "jp.kg", 2, true},
	{1, "tv.kg", 2, true},
	{1, "uk.kg", 2, true},
	{1, "us.kg", 2, true},
	{1, "de.ls", 2, true},
	{1, "at.md", 2, true},
	{1, "de.md", 2, true},
	{1, "jp.md", 2, true},
	{1, "to.md", 2, true},
	{1, "uwu.nu", 2, true},
	{1, "indie.porn", 2, true},
	{1, "vxl.sh", 2, true},
	{1, "ch.tc", 2, true},
	{1, "me.tc", 2, true},
	{1, "we.tc", 2, true},
	{1, "nyan.to", 2, true},
	{1, "at.vg", 2, true},
	{1, "blog.vu", 2, true},
	{1, "dev.vu", 2, true},
	{1, "me.vu", 2, true},
	{1, "v.ua", 2, true},
	{1, "wafflecell.com", 2, true},
	{1, "idnblogger.com", 2, true},
	{1, "indowapblog.com", 2, true},
	{1, "bloger.id", 2, true},
	{1, "wblog.id", 2, true},
	{1, "wbq.me", 2, true},
	{1, "fastblog.net", 2, true},
	{2, "webhare.dev", 3, true},
	{1, "reserve-online.net", 2, true},
	{1, "reserve-online.com", 2, true},
	{1, "bookonline.app", 2, true},
	{1, "hotelwithflight.com", 2, true},
	{1, "wedeploy.io", 2, true},
	{1, "wedeploy.me", 2, true},
	{1, "wedeploy.sh", 2, true},
	{1, "remotewd.com", 2, true},
	{1, "pages.wiardweb.com", 3, true},
	{1, "wmflabs.org", 2, true},
	{1, "toolforge.org", 2, true},
	{1, "wmcloud.org", 2, true},
	{1, "panel.gg", 2, true},
	{1, "daemon.panel.gg", 3, true},
	{1, "woltlab-demo.com", 2, true},
	{1, "myforum.community", 2, true},
	{1, "community-pro.de", 2, true},
	{1, "diskussionsbereich.de", 2, true},
	{1, "community-pro.net", 2, true},
	{1, "meinforum.net", 2, true},
	{1, "wpenginepowered.com", 2, true},
	{1, "js.wpenginepowered.com", 3, true},
	{1, "wixsite.com", 2, true},
	{1, "editorx.io", 2, true},
	{1, "half.host", 2, true},
	{1, "xnbay.com", 2, true},
	{1, "u2.xnbay.com", 3, true},
	{1, "u2-local.xnbay.com", 3, true},
	{1, "cistron.nl", 2, true},
	{1, "demon.nl", 2, true},
	{1, "xs4all.space", 2, true},
	{1, "yandexcloud.net", 2, true},
	{1, "storage.yandexcloud.net", 3, true},
	{1, "website.yandexcloud.net", 3, true},
	{1, "official.academy", 2, true},
	{1, "yolasite.com", 2, true},
	{1, "ybo.faith", 2, true},
	{1, "yombo.me", 2, true},
	{1, "homelink.one", 2, true},
	{1, "ybo.party", 2, true},
	{1, "ybo.review", 2, true},
	{1, "ybo.science", 2, true},
	{1, "ybo.trade", 2, true},
	{1, "nohost.me", 2, true},
	{1, "noho.st", 2, true},
	{1, "za.net", 2, true},
	{1, "za.org", 2, true},
	{1, "bss.design", 2, true},
	{1, "basicserver.io", 2, true},
	{1, "virtualserver.io", 2, true},
	{1, "enterprisecloud.nu", 2, true},
}

func init() {
	for i := range r {
		DefaultList.AddRule(&r[i])
	}
}
